/************************************************************************/
/*

    Analysis class.  Sorts a block of floating point numbers and puts
    itself on the done queue.

*/
/************************************************************************/

#include <stdio.h>
#include "rnd.h"

int n_block[4];
float in_block[EVENT_SIZE];
float out_block[OUT_SIZE];

main() {

    int i;

    int p1,p2,p3,p4;

    int nready = 0;

    int quicksort();

    acp_init();

    p1 = EVENT_SIZE * sizeof(float);
    p2 = IN_BLOCK;
    acp_declare_block(in_block,&p1,&p2);

    p1 = OUT_SIZE * sizeof(float);
    p2 = OUT_BLOCK;
    acp_declare_block(out_block,&p1,&p2);

    p1 = 4 * sizeof(int);
    p2 = N_BLOCK;
    acp_declare_block(n_block,&p1,&p2);

    p1 = READY_QUEUE;
    p2 = 0;
    acp_declare_queue(&p1,&p2);

    p1 = DONE_QUEUE;
    p2 = 0;
    acp_declare_queue(&p1,&p2);

    p1 = ACP_THIS_PROCESS;
    p2 = READY_QUEUE;
    acp_queue_process(&p1,&p2);

    for (i=0; i<4; i++) n_block[i] = 0;

    p1 = ACP_ALL_PROCESSES;
    p2 = BEGIN_SYNC;
    acp_sync(&p1,&p2);

    loop:

/* Wait for an event */

    for (i=0; i<4; i++) {
	while (n_block[i] != -1) sleep(1);
    }

/* Do the sorting */

    quicksort_float(in_block,0,EVENT_SIZE-1);
    for (i=0; i<OUT_SIZE; i++) out_block[i] = in_block[i];

/* Clear the notify block */

    for (i=0; i<4; i++) n_block[i] = 0;

/* Tell how many bytes are available */

    nready++;
    if (nready > OUT_SIZE) nready = 1;
    p1 = OUT_BLOCK;
    p2 = 400000;
    acp_data_ready(&p1,&p2);

/* Put ourselves on the done queue */

    p1 = ACP_THIS_PROCESS;
    p2 = DONE_QUEUE;
    acp_queue_process(&p1,&p2);

    goto loop;
}

/************************************************************************/
/*									*/
/*	QUICKSORT (recursive, floating point version)			*/
/*									*/
/************************************************************************/

quicksort_float(x,f,l)
float x[];
int f,l;
{
    float y;	/* temporary storage */

    int i,j;

    if (f >= l) return;

/* if the array is just two elements, just swap if necessary */

    if (f == l-1) {
	if (x[f] > x[l]) {
	    y = x[f];
	    x[f] = x[l];
	    x[l] = y;
	    return;
	}
    }

/* partition the table */

    i = f+1;
    while (x[i] < x[f]) {
	i++;
	if (i > l) {
	    y = x[l];
	    x[l] = x[f];
	    x[f] = y;
	    quicksort_float(x,f,l-1);
	    return;
	}
    }

    j = l;
    while (x[j] > x[f]) {
	j--;
	if (j < f) {
	    quicksort_float(x,f+1,l);
	    return;
	}
    }

    while (i < j) {

	y = x[i];
	x[i] = x[j];
	x[j] = y;

	i++;
	while (x[i] < x[f]) i++;

	j--;
	while (x[j] > x[f]) j--;

    }

    y = x[f];
    x[f] = x[j];
    x[j] = y;

    quicksort_float(x,f,j-1);
    quicksort_float(x,j+1,l);

}
