/*

 Name  :        fgtcol.c
 Descr :        Output program for cps low level asynchronous mode test job

 Author:	R. C. Yeager
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <cps_user.h>
#include "fgtched.h"

main() 

    {

    void gtimfunc();                    /* declare formatted time function   */

    int ofile;                          /* output file id                    */
    int otbyte = 0;                     /* byte count from writes            */
    int byte_count;                     /* cps data xfer byte count          */
    int block_no;                       /* cps block id (queue number)       */
    int queue_num;                      /* cps declare queue id              */
    int which_proc;                     /* cps process number                */
    int act_type;                       /* cps requested action type         */
    int msg_inc;                        /* block increment from getenv       */
    int num_of_msgs;                    /* number of blocks from getenv      */
    int max_omsg_size;                  /* output block size from getenv     */
    int nbuf;                           /* number of buffers from getenv     */
    int bsize;                          /* stored block size (in block)      */
    int list;                           /* cps_get_next_tape list id         */ 
    int pos;                            /* cps_get_next_tape list position   */
    int status;                         /* cps_get_next_tape return status   */
    int *flag;                          /* data buffer status array pointer  */
    int *process;                       /* process id array status pointer   */
    int *bcount;                        /* byte count array pointer          */
    int eoq = 0;                        /* end of queue flag                 */
    int eoj = 0;                        /* end of job flag                   */
    int bp;                             /* buffer pointer subscript          */
    int eventno;                        /* record counter                    */ 
    int nblocks;                        /* cps_wait_for_get number of blocks */
    int blocklist[1];                   /* cps_wait_for_get block id array   */

    HEADR *bufr,**buffin;               /* data buff ptrs cast w/HEADR struct*/
    char *getenv(char *);               /* declare char form of getenv       */
    char odev[8];                       /* output device type from getenv    */
    char fpout[80];                     /* output file name from getenv      */
    char cpsfpout[80];                  /* formatted cps tape driver name    */
    char errmsg[80];                    /* cps_get_next_tape error message   */
    char obuf[8];                       /* return variable from time function*/
    char device[80];                    /* cps_get_next_tape returned device */
    char vsn[12];                       /* cps_get_next_tape returned volser */
    char errormsg[80];                  /* formatted device error message    */
    char *sep = " ";                    /* strtok seperator character        */
	int start = 255, end = 256;

    cps_init();

    gtimfunc(obuf);
    printf("Starting  : Time %.8s\n",obuf);
    fflush(stdout);

    strcpy(odev,ODEV);
    strcpy(fpout,OFILE);
    num_of_msgs = MSGNUM;
    msg_inc = MSGINC;
    max_omsg_size = MSGOSIZE;
    nbuf = NBUF;

    if ((strncmp(odev,"tape",4) == 0) &&
        (strncmp(fpout,"cpsbatch",8) == 0)) 
        {
        list = 1;
        cps_get_next_tape(&list,device,vsn,&pos,&status,errmsg,80,12,80);
        if (status == CPS_OK) 
            {
            (strtok(device,sep));
            sprintf(cpsfpout,"/usr/cpsdev/%s_nrnsv",device);
            }
        else
            {
            printf("%s\n",errmsg);
            fflush(stdout);
            cps_stop_job();
            }
        }
    else
        {
        strcpy(cpsfpout,fpout);
        } 

#ifdef AIX
    sighold(SIGUSR1);
#endif
    if (strncmp(odev,"disk",4) == 0) 
        {
        if (0 > (ofile = open(cpsfpout,O_WRONLY|O_CREAT|O_TRUNC,0744))) 
            {
            perror(cpsfpout);
            fflush(stderr);
#ifdef AIX
            sigrelse(SIGUSR1);
#endif
            cps_stop_job();
            }
        }
    else if (strncmp(odev,"tape",4) == 0) 
        {
        if (0 > (ofile = open(cpsfpout,O_WRONLY,0))) 
            {
            perror(cpsfpout);
            fflush(stderr);
#ifdef AIX
            sigrelse(SIGUSR1);
#endif
            cps_stop_job();
            }
        }
#ifdef AIX
    sigrelse(SIGUSR1);
#endif

    sprintf(errormsg,"I/O error during write on %s",cpsfpout);
    bufr = (HEADR *) malloc(nbuf*max_omsg_size);
    buffin = (HEADR **) malloc(nbuf*sizeof(HEADR *));
    flag = (int *) malloc(nbuf*sizeof(int));
    process = (int *) malloc(nbuf*sizeof(int));
    bcount = (int *) malloc(nbuf*sizeof(int));

    for (bp = 0; bp < nbuf; bp++)
        {
        buffin[bp] = (HEADR *) ((char *)bufr+(bp*max_omsg_size));
        flag[bp] = 0;
        process[bp] = 0;
        bcount[bp] = 0;
        }

    which_proc = CPS_ALL_PROCESSES;
    act_type = BEGIN_SYNC;
    cps_sync(&which_proc,&act_type);

    gtimfunc(obuf);
    printf("Synchd    : Time %.8s\n",obuf);
    fflush(stdout);

    eventno = 1;
    bp = 0;

    while (!eoj) 

        {

        if (!eoq)
            {
			cps_hook( &start );
            for (bp = 0; bp < nbuf; bp++)
                {
                if (flag[bp] == 0)
                    {
                    byte_count = max_omsg_size;
                    block_no = OUT_BLOCK;
                    flag[bp] = 1;
                    cps_queue_xfer(buffin[bp],&byte_count,&block_no, 
                        &flag[bp],&bcount[bp]);
                    queue_num = DONE_QUEUE;
                    cps_dq_get(&queue_num,&process[bp]);
                    }
                }
			cps_hook( &end );
            nblocks = 1;
            blocklist[0] = OUT_BLOCK;
            cps_wait_for_get(&nblocks,blocklist); 
            }

        for (bp = 0; bp < nbuf; bp++)
            {
            if (flag[bp] == -1)
                {
                if (process[bp] != CPS_END_OF_QUEUE)
                    {
                    gtimfunc(obuf);
                    sprintf(buffin[bp]->header_oti, 
                        "From Event: Time %.8s", 
                        obuf);
                    byte_count = max_omsg_size;

                    if (strncmp(odev,"none",4) != 0) 
                        {
#ifdef AIX
                        sighold(SIGUSR1);
#endif
                        if (0 > (otbyte = write(ofile,buffin[bp],byte_count))) 
                            {
                            perror(errormsg);
                            fflush(stderr);
#ifdef AIX
                            sigrelse(SIGUSR1);
#endif
                            cps_stop_job();
                            }
#ifdef AIX
                        sigrelse(SIGUSR1);
#endif
                        }

                    bsize = buffin[bp]->block[0];

                    gtimfunc(obuf);
                    sprintf(buffin[bp]->header_oto, 
                       "To Output : Time %.8s Block %d Buf %d a %d r %d b %d", 
                        obuf,eventno,bp,otbyte,byte_count,bsize);
                    if (((eventno % msg_inc) == 0) || (eventno == num_of_msgs)) 
                        {
                        printf("%s \n",buffin[bp]->header_ini);
                        printf("%s \n",buffin[bp]->header_ino);
                        printf("%s \n",buffin[bp]->header_ani);
                        printf("%s \n",buffin[bp]->header_ano);
                        printf("%s \n",buffin[bp]->header_oti);
                        printf("%s \n",buffin[bp]->header_oto);
                        fflush(stdout);
                        }

                    eventno++;

                    }
                 else
                    {
                    eoq = -1;
                    }
                flag[bp] = 0;
                queue_num = READY_QUEUE;
                cps_queue_process(&process[bp],&queue_num);
                }
            }

        if (eoq) 
            {
            for (bp = 0; bp < nbuf; bp++)
                {
                if (flag[bp] == 1)
                    {
                    cps_sleep(1);
                    break;
                    }
                }
            if (bp == nbuf)
                {
                eoj = -1;
                }
            }
        }                                   /* end of while (!eoj) */


        gtimfunc(obuf);
        printf("Stopping  : Time %.8s\n",obuf);
        fflush(stdout);

        if (strncmp(odev,"none",4) != 0) 
            {
#ifdef AIX
            sighold(SIGUSR1);
#endif
            close(ofile);
#ifdef AIX
            sigrelse(SIGUSR1);
#endif
            }

        cps_stop_job();

    }                                   /* end of main block                 */

    /*  called function to format and return the time in format HH:MM:SS     */

    void gtimfunc(obuf) 

        {
        
        long tloc;                      /* return variable from time function*/
        char buf[30];                   /* formatted time variable           */
 
        time(&tloc);                    /* get current time                  */
        strcpy(buf,ctime(&tloc));       /* format current time into buf      */
        strncpy(obuf,&buf[11],8);       /* move time value to obuf           */
        return;                         /* return to caller                  */

        }

