/*

 Name  :        fgtcot.c
 Descr :        Output program for cps test job

 Author:	R. C. Yeager
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <acp_user.h>
#include "fgtched.h"

    int ofile;                          /* output file id                    */
    int otbyte = 0;                     /* byte count from writes            */
    int byte_count;                     /* cps data xfer byte count          */
    int block_no;                       /* cps block id (queue number)       */
    int msg_inc;                        /* block increment from getenv       */
    int num_of_msgs;                    /* number of blocks from getenv      */
    int max_omsg_size;                  /* output block size from getenv     */
    int eventno;                        /* record counter                    */ 
    int bsize;                          /* stored block size (in block)      */

    char obuf[8];                       /* return variable from time function*/
    char odev[8];                       /* output device type from getenv    */
    char errormsg[80];                  /* formatted device error message    */

main() 

    {

    void gtimfunc();                    /* declare formatted time function   */
    void write_data(int *buf, int *nbytes); /* declare output routine        */
    int pos;                            /* acp_get_next_tape list position   */
    int status;                         /* acp_get_next_tape return status   */
    int nbuf;                           /* number of buffers from getenv     */
    int which_proc;                     /* cps process number                */
    int act_type;                       /* cps requested action type         */
    int list;                           /* acp_get_next_tape list id         */ 
    int key[1];                         /* cps buffer pool key arg           */
    int queue_num;                      /* cps declare queue id              */
    int wait_num;                       /* cps declare queue id              */

    char *getenv(char *);               /* declare char form of getenv       */
    char fpout[80];                     /* output file name from getenv      */
    char cpsfpout[80];                  /* formatted cps tape driver name    */
    char device[80];                    /* acp_get_next_tape returned device */
    char vsn[12];                       /* acp_get_next_tape returned volser */
    char errmsg[80];                    /* acp_get_next_tape error message   */
    char *sep = " ";                    /* strtok seperator character        */

    acp_init();

    gtimfunc(obuf);
    printf("Starting  : Time %.8s\n",obuf);
    fflush(stdout);

    strcpy(odev,ODEV);
    strcpy(fpout,OFILE);
    num_of_msgs = MSGNUM;
    msg_inc = MSGINC;
    max_omsg_size = MSGOSIZE;
    nbuf = NBUF;

    if ((strncmp(odev,"tape",4) == 0) &&
        (strncmp(fpout,"cpsbatch",8) == 0)) 
        {
        list = 1;
        acp_get_next_tape(&list,device,vsn,&pos,&status,errmsg,80,12,80);
        if (status == ACP_OK) 
            {
            strtok(device,sep);
            sprintf(cpsfpout,"/usr/cpsdev/%s_nrnsv",device);
            }
        else
            {
            printf("%s\n",errmsg);
            fflush(stdout);
            acp_stop_job();
            }
        }
    else
        {
        strcpy(cpsfpout,fpout);
        } 

#ifdef AIX
    sighold(SIGUSR1);
#endif
    if (strncmp(odev,"disk",4) == 0) 
        {
        if (0 > (ofile = open(cpsfpout,O_WRONLY|O_CREAT|O_TRUNC,0744))) 
            {
            perror(cpsfpout);
            fflush(stderr);
#ifdef AIX
            sigrelse(SIGUSR1);
#endif 
            acp_stop_job();
            }
        }
    else if (strncmp(odev,"tape",4) == 0) 
        {
        if (0 > (ofile = open(cpsfpout,O_WRONLY,0))) 
            {
            perror(cpsfpout);
            fflush(stderr);
#ifdef AIX
            sigrelse(SIGUSR1);
#endif 
            acp_stop_job();
            }
        }
#ifdef AIX
    sigrelse(SIGUSR1);
#endif 

    sprintf(errormsg,"I/O error during write on %s",cpsfpout);

    which_proc = ACP_ALL_PROCESSES;
    act_type = BEGIN_SYNC;
    acp_sync(&which_proc,&act_type);

    gtimfunc(obuf);
    printf("Synchd    : Time %.8s\n",obuf);
    fflush(stdout);

    eventno = 1;
    wait_num = 1;

    while (1) 

        {

        byte_count = max_omsg_size;
        block_no = OUT_BLOCK;
        queue_num = DONE_QUEUE;
        acp_start_gets(key,&byte_count,&nbuf,&block_no,&queue_num);
   
        queue_num = DONE_QUEUE;
        acp_buffer_set_wait(key,&queue_num);

        queue_num = DONE_QUEUE;
        if (acp_end_of_data(key,&queue_num))
            {
            gtimfunc(obuf);
            printf("Stopping  : Time %.8s \n",obuf);
            fflush(stdout);
            if (strncmp(odev,"none",4) != 0) 
                {
#ifdef AIX
                sighold(SIGUSR1);
#endif 
                close(ofile);
#ifdef AIX
                sigrelse(SIGUSR1);
#endif 
                }
            acp_stop_job();
            }

        queue_num = READY_QUEUE;
        acp_handle_data(key,write_data,&queue_num);
 /*
            printf("Iteration %d \n",wait_num);
            fflush(stdout);
            wait_num++;
 */
        }

    }                                   /* end of main block                 */

    void write_data(int *ptr1,int *ptr2)

        {

        int nbytes;

        HEADR *bufr;

        nbytes = *ptr2;

        bufr = (HEADR *) ptr1;
        
        gtimfunc(obuf);
        sprintf(bufr->header_oti, 
            "From Event: Time %.8s ", 
            obuf);

        byte_count = max_omsg_size;
        bsize = bufr->block[0];

        if (strncmp(odev,"none",4) != 0) 
            {
#ifdef AIX
            sighold(SIGUSR1);
#endif 
            if (0 > (otbyte = write(ofile,bufr,byte_count))) 
                {
                perror(errormsg);
                fflush(stderr);
#ifdef AIX
                sigrelse(SIGUSR1);
#endif 
                acp_stop_job();
                }
#ifdef AIX
            sigrelse(SIGUSR1);
#endif 
            }

        gtimfunc(obuf);
        sprintf(bufr->header_oto, 
            "To Output : Time %.8s Block %d a %d r %d bl %d bf %d", 
            obuf,eventno,otbyte,byte_count,bsize,nbytes);

        if (((eventno % msg_inc) == 0) || (eventno == num_of_msgs)) 
            {
            printf("%s \n",bufr->header_ini);
            printf("%s \n",bufr->header_ino);
            printf("%s \n",bufr->header_ani);
            printf("%s \n",bufr->header_ano);
            printf("%s \n",bufr->header_oti);
            printf("%s \n",bufr->header_oto);
            fflush(stdout);
            }

        eventno++;

        return;

        }

    /*  called function to format and return the time in format HH:MM:SS     */

    void gtimfunc(obuf) 

        {
        
        long tloc;                      /* return variable from time function*/
        char buf[30];                   /* formatted time variable           */
 
        time(&tloc);                    /* get current time                  */
        strcpy(buf,ctime(&tloc));       /* format current time into buf      */
        strncpy(obuf,&buf[11],8);       /* move time value to obuf           */
        return;                         /* return to caller                  */

        }

