/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



#include <stdio.h>
#include "rnd.h"

float in_block[EVENT_SIZE];
float out_block[OUT_SIZE];

main() {

    int p1,p2,p3,p4;

    int quicksort();

    acp_init();

    p1 = EVENT_SIZE * sizeof(float);
    p2 = IN_BLOCK;
    acp_declare_block(in_block,&p1,&p2);

    p1 = OUT_SIZE * sizeof(float);
    p2 = OUT_BLOCK;
    acp_declare_block(out_block,&p1,&p2);

    p2 = SORT_REMSUB;
    p3 = 0;
    acp_declare_subroutine(quicksort,&p2,&p3);

    p1 = READY_QUEUE;
    p2 = 0;
    acp_declare_queue(&p1,&p2);

    p1 = DONE_QUEUE;
    p2 = 0;
    acp_declare_queue(&p1,&p2);

    p1 = ACP_THIS_PROCESS;
    p2 = READY_QUEUE;
    acp_queue_process(&p1,&p2);

    p1 = ACP_ALL_PROCESSES;
    p2 = BEGIN_SYNC;
    acp_sync(&p1,&p2);

    acp_service_calls();

}


quicksort() {

    int i;

    quicksort_float(in_block,0,EVENT_SIZE-1);
    for (i=0; i<OUT_SIZE; i++) out_block[i] = in_block[i];

}


/************************************************************************/
/*									*/
/*	QUICKSORT (recursive, floating point version)			*/
/*									*/
/************************************************************************/

quicksort_float(x,f,l)
float x[];
int f,l;
{
    float y;	/* temporary storage */

    int i,j;

    if (f >= l) return;

/* if the array is just two elements, just swap if necessary */

    if (f == l-1) {
	if (x[f] > x[l]) {
	    y = x[f];
	    x[f] = x[l];
	    x[l] = y;
	    return;
	}
    }

/* partition the table */

    i = f+1;
    while (x[i] < x[f]) {
	i++;
	if (i > l) {
	    y = x[l];
	    x[l] = x[f];
	    x[f] = y;
	    quicksort_float(x,f,l-1);
	    return;
	}
    }

    j = l;
    while (x[j] > x[f]) {
	j--;
	if (j < f) {
	    quicksort_float(x,f+1,l);
	    return;
	}
    }

    while (i < j) {

	y = x[i];
	x[i] = x[j];
	x[j] = y;

	i++;
	while (x[i] < x[f]) i++;

	j--;
	while (x[j] > x[f]) j--;

    }

    y = x[f];
    x[f] = x[j];
    x[j] = y;

    quicksort_float(x,f,j-1);
    quicksort_float(x,j+1,l);

}
