/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



/* Logical Process Numbering:

	0: Job Manager
	1-32: User Processes
	33-64: Shared Memory Processes
	65: PM
*/

struct shmem {

    long type;
    long src_lpn;
    long dst_lpn;
    long size;		/* #bytes in msg */
    char msg[MAX_MESSAGE_SIZE];

};

typedef struct shmem cps_message;

struct shmem_s {	/* Small-size message */

    long type;
    long src_lpn;
    long dst_lpn;
    long size;		/* #bytes in msg */
    char msg[SMALL_MSG_SIZE];

};

typedef struct shmem_s cps_msg;

struct qelem {

    int size;		/* 0 = small, 1 = large */
    int index;		/* index of this queue element */
    int next;		/* Index into array of qelem */
    int buf;		/* Index into array of shmem */

};

struct shmmap {		/* To be used in an array where 
			   max(0,(lpn - FIRST_SHM_LPN + 1)) is index.
			   (This allows the JM to be handled uniformly) */

    char hostname[64];
    char hostaddr[24];
    int portno;
    long sockaddr;
    int pid;

    int data_msg_count;
    int data_ack_count;

    int semnum;

    int sock;		/* Socket descriptor */
    int status; 	/* -1 = connected, 0 = unconnected */

    int sock2;		/* Possible dual connection */
    int status2;
};

struct procmap {	/* To be used in an array where index is lpn - 1 */

    int local; 		/* -1 if process is on same CPU as this one,
			   otherwise 0 */
    int pid;
    int chpid;		/* Child pid (for pairs of processes) */

    int shmid;		/* shm process this process is associated with -
			   index into the shmmap */

    int semnum;		/* semaphore number associated with this process */

    int status; 	/* -1 = alive, 0 = dead, 1 = finished */
};
