
%token CLASS TAPE CR NOTHING
%token TRACK EXABYTE NUMBER WORD LPAREN RPAREN COMMA
%token SHARE PROG CLASSES ALL NONE SPECIFIC CPU TYPE EQ 
%token NUMBR OF PER PROCESS MIN DBG AS MANY POSS TIMEOUT CPSDIR
%token TOK_SGI TOK_MIPS TOK_R3000 TOK_VMS TOK_ULTRIX TOK_AIX TOK_SUN TOK_SONY
%token TOK_HPUX TOK_UNIX
%token EXABYTE_8200 EXABYTE_8500
%token TOK_IO TOK_COMPUTE TOK_ANY TASK
%token ON LOGICAL

%{

extern char word[];
extern char yytext[];

#include "all.h"

static int cclass = 1;		/* Current class */

static int integers[32];
static int num_int = 0;

static int shclass = 0;

%}

%%

lines	: line
		| lines line
		;

line	: CLASS EQ NUMBER CR
			{
				cclass = $3;
				if (!class[cclass].valid) initialize_class(cclass);
			}
/* Old SDF format */
		| CPU EQ cpu_types {mark_word();} WORD
			{add_sdf_cpu(yytext,$3,TASK_ANY);} CR
/* New SDF format */
		| NUMBER ':' WORD ':' cpu_types ':' CR
			{
				create_logical_cpu( $1 - 1 );
				add_sdf_cpu(word,$5,TASK_ANY);
			}
		| NUMBER ':' WORD ':' cpu_types ':' task_types CR
			{
				create_logical_cpu( $1 - 1 );
				add_sdf_cpu(word,$5,$7);
			}
/* CPS directory */
		| CPSDIR EQ {mark_word();} WORD {add_cpsdir(yytext);} CR
/* Tape Types */
		| NUMBER TRACK TAPE EQ NUMBER CR
		| EXABYTE TAPE EQ NUMBER CR
		| EXABYTE_8200 TAPE EQ NUMBER CR
		| EXABYTE_8500 TAPE EQ NUMBER CR
/* JDF Process Info */
		| NUMBR OF PROCESS EQ num_proc_opt CR
			{class[cclass].nproc = $5;}
		| NUMBR OF CPU EQ NUMBER CR
			{class[cclass].ncpu = $5;}
		| PROCESS PER CPU EQ n_or_all CR
			{class[cclass].per_cpu = $5;}
        | PROCESS PER CPU EQ NUMBER ON CPU NUMBER CR
			{assign_processes_to_cpu(cclass,$5,$8,$8);}
		| PROCESS PER CPU EQ NUMBER ON CPU NUMBER '-' NUMBER CR
			{assign_processes_to_cpu(cclass,$5,$8,$10);}
		| MIN PROCESS EQ n_or_all CR
			{class[cclass].min_proc = $4;}
/* Obsolete */
		| DBG EQ NUMBER CR
			{class[cclass].debug = $3;}
		| TIMEOUT EQ NUMBER CR
			{class[cclass].timeout = $3;}
/* JDF CPU Type and Task */
		| CPU TYPE EQ cpu_types CR
			{class[cclass].cpu_type = $4;}
		| TASK TYPE EQ task_types CR
			{class[cclass].task = $4;}
/* Share Classes */
		| SHARE CLASSES EQ share_opt CR
		{
		    int i,j;

		    if ($4 == 0)
			{
				for (i=1; i<=MAX_CLASSES; i++)
				{
				    set_clear(class[i].share,MAX_CLASSES);
				    add_to_set(i,class[i].share);
				}
				shclass = -1;
		    }
		    else if ($4 == 1)
			{
				if (shclass == 0)
				{
					for (i=1; i<=MAX_CLASSES; i++)
					{
			        	set_clear(class[i].share,MAX_CLASSES);
						add_to_set(i,class[i].share);
				    }
				    shclass = -1;
				}
				for (i=0; i<num_int; i++)
				{
				    set_clear(class[integers[i]].share,MAX_CLASSES);
				    for (j=0; j<num_int; j++)
					{
						add_to_set(integers[j],
						class[integers[i]].share);
			    	}
				}
		    }
		    num_int = 0;
		}
/* Specific CPU */
		| SPECIFIC CPU EQ {mark_word();} cpulist CR
/* Logical CPU */
		| LOGICAL CPU EQ {num_int = 0;} logical_cpulist CR
/* Program name */
		| PROG EQ {mark_word();} proglist CR
/* Error */
		| error CR {add_syntax_error();}
		| CR
		;

cpu_types	: TOK_MIPS {$$ = CPU_MIPS;}
		| TOK_SGI {$$ = CPU_SGI;}
		| TOK_R3000 {$$ = CPU_R3000;}
		| TOK_VMS {$$ = CPU_VMS;}
		| TOK_ULTRIX {$$ = CPU_DEC;}
		| TOK_AIX {$$ = CPU_AIX;}
		| TOK_SUN {$$ = CPU_SUN;}
		| TOK_SONY {$$ = CPU_SONY;}
		| TOK_HPUX {$$ = CPU_HPUX;}
		| TOK_UNIX {$$ = CPU_SGI;}
		;

task_types	: TOK_IO {$$ = TASK_IO;}
		| TOK_COMPUTE {$$ = TASK_COMPUTE;}
		| TOK_ANY {$$ = TASK_ANY;}
		;

share_opt	: ALL	{$$ = -1;}
		| NONE		{$$ = 0;}
		| numlist	{$$ = 1; }
		;

n_or_all	: NUMBER {$$ = $1;}
		| ALL {$$ = -1;}
		;

num_proc_opt	: NUMBER {$$ = $1;}
		| AS MANY AS POSS {$$ = -1;}
		;

numlist		: numlist COMMA NUMBER {integers[num_int++] = $3;}
		| NUMBER {integers[num_int++] = $1;}
		;

cpulist		: cpulist WORD {add_cpu_to_class(cclass,yytext);}
		| WORD  {add_cpu_to_class(cclass,yytext);}
		;

logical_cpulist		: numlist
		{
			int i;
			for( i = 0; i < num_int; i++ )
				add_logical_cpu_to_class(cclass,integers[i]);
		}
		;

proglist	: proglist WORD {add_word_to_prog(cclass,yytext);}
		| WORD {add_word_to_prog(cclass,yytext);}
		;

%%

yyerror(s)
char *s;
{
}

yywrap()
{
return(1);
}
