/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of
 their employees, makes any warrenty, express or implied, or assumes any
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				JM_MON.C												*/
/*																		*/
/*																		*/
/************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include "all.h"

/******************************************************************/
/*
	FUNCTION NAME:

	  init_mon_shm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

init_mon_shm() {

	int addr;

	int i,p,c;

	char fname[40];

	FILE *fp,*fopen();

	int size;

/* Create the shared memory segment and attach to it */

	mon_shmid = shm_create(sizeof(struct cps_job_stats));
	if (mon_shmid == -1)
		return;
	addr = (int) shm_attach(mon_shmid);
	if (addr == -1)
		return;
	mon_shm = (struct cps_job_stats *) addr;

/* Initialize the structure */

	mon_shm->job_done = 0;
	mon_shm->num_proc = num_usr_procs;
	for (i=0; i<num_usr_procs; i++) {
		p = i+FIRST_USER_LPN;
		c = usr_map[i].class;
		mon_shm->class[i] = c;
		if (set_member(p,class[c].alive))
			mon_shm->p_state[i].state = PSTATE_RUNNING;
		else
			mon_shm->p_state[i].state = PSTATE_DEAD;
	}
	size = (MAX_PROCESSES*(12+MAX_USER_STATUS) + MAX_QUEUES*2)*sizeof(int);
	bzero(mon_shm->call_cnt,size);

	for (i=0; i<4; i++) {
		mon_shm->nsends[i] = 0;
		mon_shm->sndavg[i] = 0;
		mon_shm->sndcur[i] = 0;
		mon_shm->ngets[i] = 0;
		mon_shm->getavg[i] = 0;
		mon_shm->getcur[i] = 0;
	}

/* Write the shared memory id to a file */

	sprintf(fname,"%s.mdf",job_name);
	fp = fopen(fname,"w");
	fprintf(fp,"%d",mon_shmid);
	fclose(fp);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  monitor_update

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int getindx(s)
int s;
{
	if (s < 2048)
		return(0);
	else if (s < 8192)
		return(1);
	else if (s < 32768)
		return(2);
	else
		return(3);
}

monitor_update(opcode,lpn,datum,array)
int opcode,lpn,datum,*array;
{
	static int get_bcount = 0;
	static int send_bcount = 0;
	int bps;
	int i;
	int sum;

	switch (opcode) {
	case MOP_USER_STATUS :
		break;
	case MOP_INC_CALL :
		mon_shm->call_cnt[lpn-FIRST_USER_LPN][datum]++;
		if (datum == INC_SEND) {
			i = getindx(send_bcount);
			sum = mon_shm->nsends[i] * mon_shm->sndavg[i] + *array;
			mon_shm->nsends[i]++;
			mon_shm->sndcur[i] = *array;
			mon_shm->sndavg[i] = sum/mon_shm->nsends[i];
		}
		else if (datum == INC_GET) {
			i = getindx(get_bcount);
			sum = mon_shm->ngets[i] * mon_shm->getavg[i] + *array;
			mon_shm->ngets[i]++;
			mon_shm->getcur[i] = *array;
			mon_shm->getavg[i] = sum/mon_shm->ngets[i];
		}
		break;
	case MOP_STATE_CHANGE :
		switch (datum) {
		case PSTATE_GET_DONE :
			get_bcount = *array;
			datum = PSTATE_RUNNING;
			break;
		case PSTATE_SEND_DONE :
			send_bcount = *array;
			datum = PSTATE_RUNNING;
			break;
		case PSTATE_WAIT_GET :
		case PSTATE_WAIT_SEND :
		case PSTATE_WAIT_CALL :
			mon_shm->p_state[lpn-FIRST_USER_LPN].attrib = *array;
			break;
		}
		mon_shm->p_state[lpn-FIRST_USER_LPN].state = datum;
		break;
	case MOP_Q_PROC :
		break;
	case MOP_DQ_PROC :
		break;
	}

}
