/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				JM_PARENT.C												*/
/*																		*/
/*																		*/
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include <signal.h>
#include "../inc/types.h"
#include "../inc/time.h"
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <errno.h>

/********************/
/* Global Variables */
/********************/

static fd_set fdset;
static int nfds;
static struct timeval poll_t;

static struct {

	int sock;
	int status;		/* 1 = active, 0 = closed */
	char buf[256];
	int bufpt;

} rterm[MAX_PROCESSES];

/************************/
/* External Definitions */
/************************/



/******************************************************************/
/*
	FUNCTION NAME:

	  init_rterm
	  open_rterm
	  close_rterm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

init_rterm() {

	int i;

	poll_t.tv_sec = 0;
	poll_t.tv_usec = 0;
	nfds = 0;
	FD_ZERO(&fdset);
	for (i=0; i<MAX_PROCESSES; i++)
	{
		rterm[i].status = 0;
		rterm[i].bufpt = 0;
	}
}

open_rterm(p,sock)
int p,sock;
{
	nfds++;
	FD_SET(sock,&fdset);
	rterm[p].sock = sock;
	rterm[p].status = 1;
#if defined(R3000) || defined(MIPS)
	fcntl(sock,F_SETFL,O_NDELAY);	/* Make socket non-blocking */
#else
	fcntl(sock,F_SETFL,O_NONBLOCK);	/* Make socket non-blocking */
#endif
}

close_rterm(p)
int p;
{
	nfds--;
	FD_CLR(rterm[p].sock,&fdset);
	rterm[p].status = 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  read_rterm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int read_rterm(p) 
int p;
{
	char temp[1024];

	char *strrchr(),*strchr(),*sl,*sf,*s;

	int nleft;

	int n;

	rdmsg:
	n = recv(rterm[p].sock,temp,1023,0);
	if (n == 0)
	{
		close_rterm(p);
		return;
	}
	if (n == -1)
	{
		if (errno == EINTR) goto rdmsg;
#ifdef HPUX
		if (errno == EAGAIN) return;
#else
		if (errno == EWOULDBLOCK) return;
#endif
		printerr(FACILITY,ERROR_MSG,NULL,NULL,22);
		close_rterm(p);
		return;
	}
	temp[n] = '\0';

/* Find first & last newline character */

	sl = strrchr(temp,'\n');
	sf = strchr(temp,'\n');

/* Print up to first newline character */

	if (sf != NULL)
	{
		*sf = '\0';
		if (rterm[p].bufpt > 0)
			print2_usr_log(p+1,rterm[p].buf,temp);
		else
			print_usr_log(p+1,temp);
		rterm[p].bufpt = 0;
	}
	else
	{	/* Received no newlines (store or print, and return) */
		nleft = 255 - rterm[p].bufpt;
		if (n > nleft)
		{
			if (rterm[p].bufpt > 0)
				print2_usr_log(p+1,rterm[p].buf,temp);
			else
				print_usr_log(p+1,temp);
			rterm[p].bufpt = 0;
		}
		else
		{
			strcpy(&(rterm[p].buf[rterm[p].bufpt]),temp);
			rterm[p].bufpt = strlen(rterm[p].buf);
		}
		return;
	}

/* Print to last newline character */

	while (sf != sl)
	{
		s = sf + 1;
		sf = strchr(s,'\n');
		*sf = 0;	
		print_usr_log(p+1,s);
	}

/* Store everything beyond last newline (or print if too much) */

	s = sl + 1;
	nleft = strlen(s);
	if (nleft > 255)
	{
		print_usr_log(p+1,s);
		rterm[p].bufpt = 0;
	}
	else
	{
		strcpy(rterm[p].buf,s);
		rterm[p].bufpt = nleft;
	}

	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  wait_rterm()

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int wait_rterm() {

    int nfound;

    fd_set fds;

    int i;

    int nopen;

    struct timeval t;

    t.tv_sec = 60; t.tv_usec = 0;

    sel:
    fds = fdset;
    nfound = select(FD_SETSIZE,&fds,(fd_set *) 0,(fd_set *) 0,&t);

    if (nfound == 0) {		/* Time limit expired */

/* Check all the sockets to see if they're still connected */

    	for (i=0; i<MAX_PROCESSES; i++) {
		if (rterm[i].status == 1) {
			if (FD_ISSET(rterm[i].sock,&fds)) {
				if (!socket_connected(rterm[i].sock)) {
					close_rterm(i);
				}
			}
		}
	}

	goto check_nopen;

    }
    else if (nfound == -1) {
	if (errno == EINTR) goto sel;
	printerr(FACILITY,ERROR_MSG,NULL,NULL,23);
	exit(0);
    }

    for (i=0; i<MAX_PROCESSES; i++) {
	if (rterm[i].status == 1) {
	    if (FD_ISSET(rterm[i].sock,&fds)) {
		read_rterm(i);
	    }
	}
    }

    check_nopen:

    nopen = 0;
    for (i=0; i<MAX_PROCESSES; i++) {
	if (rterm[i].status == 1) {
	    nopen++;
	    break;
	}
    }

    if (nopen > 0) {
        return(0);
    }
    else {
	exit(0);
    }
}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_child_died

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

																  */
/******************************************************************/

void jm_child_died(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
/*
	exit(0);
*/
}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_parent

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

																  */
/******************************************************************/

jm_parent()
{
	int i;

	init_rterm();

	for (i=0; i<num_usr_procs; i++)
		if (usr_map[i].status)
			open_rterm(i,usr_map[i].sock);

	while (1) wait_rterm();

}

