/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				JM_STRUCT.C												*/
/*																		*/
/*																		*/
/************************************************************************/

static int last_cpu_index = 0;

/*****************/
/* Include Files */
/*****************/

#include "all.h"

#include "../inc/types.h"
#include "../inc/socket.h"

display_queue(fp,q)
FILE *fp;
int q;
{
	int i;

	struct cps_queue_element *e;

	fprintf(fp,"Queue #%d:",q);
	fprintf(fp,"  #args = %d,  total byte count = %d\n",
	jmq[q].nargs,jmq[q].bcount);
	fprintf(fp,"  ");
	for (i=0; i<jmq[q].nargs; i++) {
		fprintf(fp,"argcnt[%d]: %d\t",i,jmq[q].argcnt[i]);
		if (i % 4) 
			fprintf(fp,"\n  ");
		else if (i == jmq[q].nargs-1)
			fprintf(fp,"\n  ");
	}

	fprintf(fp,"Set of declared processes: ");
	print_set(fp,jmq[q].declared,MAX_PROCESSES);
	fprintf(fp,"\n");

	fprintf(fp,"  Set of processes waiting for full: ");
	print_set(fp,jmq[q].wait_full,MAX_PROCESSES);
	fprintf(fp,"\n");

	fprintf(fp,"  Set of processes waiting for empty: ");
	print_set(fp,jmq[q].wait_empty,MAX_PROCESSES);
	fprintf(fp,"\n");

	fprintf(fp,"  Set of processes queued: ");
	print_set(fp,jmq[q].q_set,MAX_PROCESSES);
	fprintf(fp,"\n");

	fprintf(fp,"  Order of queued processes: ");
	e = jmq[q].queued->front;
	while (e != NULL) {
		fprintf(fp,"%d ",e->lpn);
		e = e->next;
	}
	fprintf(fp,"\n");

	fprintf(fp,"  Set of processes waiting for dequeue: ");
	print_set(fp,jmq[q].dq_set,MAX_PROCESSES);
	fprintf(fp,"\n");

	fprintf(fp,"  Order of waiting processes: ");
	e = jmq[q].dequeue->front;
	while (e != NULL) {
		fprintf(fp,"%d ",e->lpn);
		e = e->next;
	}
	fprintf(fp,"\n");

	if (jmq[q].eoq)
		fprintf(fp,"  End of Queue has been queued\n");
	else
		fprintf(fp,"  Queue still open\n");

}

/******************************************************************/
/*
	FUNCTION NAME:

	  add_word_to_prog

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

add_word_to_prog(c,word)
int c;
char *word;
{
	if (class[c].program == NULL) {
		class[c].program = (char *) hmalloc(512);
		class[c].program[0] = '\0';
	}

	strcat(class[c].program,word);
	strcat(class[c].program," ");
}

/******************************************************************/
/*
	FUNCTION NAME:

	  add_cpu_to_class

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

add_cpu_to_class(c,name)
int c;
char *name;
{
	int i;
	char errmsg[80];

/* Find the CPU */

	for (i=0; i<num_sdf_cpus; i++)
	{
		if (cpu_name_compare(name,sdf_cpu[i].cpu_name) == 0)
		{
			add_to_set(c,sdf_cpu[i].class);
			sdf_cpu[i].nprocs += class[c].per_cpu;
			if (class[c].spec_cpu == NULL)
				class[c].spec_cpu = (int *) hmalloc(MAX_SDF_CPUS*sizeof(int));
			class[c].spec_cpu[class[c].nspec] = i;
			class[c].nspec++;
			add_to_set( i + 1, class[c].cpu);
			return;
		}
	}
	sprintf(errmsg,"%s not found in cps_sdf",name);
	printerr(FACILITY,FATAL_MSG,errmsg,NULL,55);
	exit(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  add_sdf_cpu

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

add_sdf_cpu(name,type,task)
char *name;
int type, task;
{
	int i;

	static char *envstr = NULL;
	static char cpsdir[80];
	char *getenv();

/* If CPU is already in the list, do not add it again */

	for (i=0; i<num_sdf_cpus; i++)
	{
		if (sdf_cpu[i].cpu_type == type)
		{
			if (cpu_name_compare(sdf_cpu[i].cpu_name,name) == 0)
			{
				last_cpu_index = i;
				return 0;
			}
		}
	}

/* Add the CPU to the list */

	i = num_sdf_cpus;
	num_sdf_cpus++;
	sdf_cpu[i].cpu_type = type;
	if( !strcmp( name, "localhost" ) )
	{
		char hostname[64];
		struct hostent *hp;
		gethostname( hostname, 64 );
		if( ( hp = gethostbyname( hostname ) ) != NULL )
			strcpy( hostname, hp->h_name );
		sdf_cpu[i].cpu_name = (char *) hmalloc(strlen(hostname)+1);
		strcpy(sdf_cpu[i].cpu_name,hostname);
	}
	else
	{
		sdf_cpu[i].cpu_name = (char *) hmalloc(strlen(name)+1);
		strcpy(sdf_cpu[i].cpu_name,name);
	}
	sdf_cpu[i].task = task;
	sdf_cpu[i].nprocs = 0;
	sdf_cpu[i].procs = create_empty_set(MAX_PROCESSES);
	sdf_cpu[i].class = create_empty_set(MAX_CLASSES);
	sdf_cpu[i].login_name[0] = '\0';
	sdf_cpu[i].password[0] = '\0';

	if (envstr == NULL)
	{
		envstr = getenv("CPSBIN");
		if (envstr == NULL)
		{
			envstr = getenv("CPS_DIR");
			if (envstr == NULL)
				envstr = "/usr/local/bin";
		}
		strcpy(cpsdir,envstr);
		envstr = cpsdir;
	}
	strcpy(sdf_cpu[i].cpsdir,cpsdir);
	last_cpu_index = i;
}

add_cpsdir(name)
char *name;
{
	strcpy(sdf_cpu[last_cpu_index].cpsdir,name);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  initialize_class

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

initialize_class(c)
int c;
{
	int i;
	class[c].valid = -1;
	class[c].sop = create_empty_set(MAX_PROCESSES);
	class[c].program = NULL;
	class[c].cpu_type = CPU_SGI;
	class[c].nspec = 0;
	class[c].spec_cpu = NULL;
	class[c].nproc = 0;
	class[c].per_cpu = 0;		/* Undefined */
	class[c].ncpu = 0;
	class[c].task = TASK_ANY;
	class[c].min_proc = 1;
	class[c].debug = 0;
	class[c].timeout = -1;		/* infinity */
	class[c].cpu = create_empty_set(MAX_SDF_CPUS);
	class[c].fini = create_empty_set(MAX_PROCESSES);
	class[c].alive = create_empty_set(MAX_PROCESSES);
	class[c].dead = create_empty_set(MAX_PROCESSES);
	for( i = 0; i < MAX_SDF_CPUS; i++ )
		class[c].logical_processes[i] = 0;
	class[c].logical_assignment = 0;
}
