/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
%token CPS_MON
%token ALIAS
%token ALL
%token AT
%token BANGBANG
%token BANGNUMBER
%token BANGSTRING
%token CAT
%token CD
%token PROMPT1
%token PROMPT2
%token COMMANDS
%token CONNECT
%token CONT
%token COREDIR
%token CR
%token DBX
%token DBXCOPY
%token DBXRUN
%token DBGDIR
%token DEBUG
%token DELETE
%token DISCONNECT
%token DISPLAY
%token EXIT
%token FILENAME
%token HELP
%token HISTORY
%token HOOK
%token HOOKS
%token HOSTS
%token IMPORT
%token JM
%token JML
%token JMINFO
%token JOB
%token KILL
%token MONITOR
%token NUMBER
%token OBJDIR
%token ON
%token OFF
%token OUTPUT
%token PAGER
%token PHOOKS
%token PING
%token PROCESS
%token PWD
%token QUESTION
%token QUEUE
%token QUEUES
%token QUIT
%token QUITFORCE
%token RANGE
%token REFRESH
%token REXEC
%token SET
%token SH
%token SHOW
%token START
%token STATE
%token STATUS
%token STEP
%token STOP
%token SYNC
%token SYNCS
%token TAIL
%token TRACE
%token UNALIAS
%token UNDEFINED
%token UNSET
%token UPDATE
%token UPSHOT
%token VIEW
%token WHERE
%token WORD

%{
#include "jmdb.h"

#define STORE store_command();

extern char		word[128];
extern char		*token[4];
extern int		num_of_toks;
extern int		n1, n2;
extern unsigned long *process_set;
extern char		yytext[];
extern char		txt[256];

extern int		find_bang();

extern ClientInfo	client;
extern JmInfo		jm;
extern int			num_procs;

%}

%%

input		:	init command
					{ return(0); }
			|	BANGBANG CR
					{ handle_bang( -1 ); return(0); }
			|	BANGNUMBER CR
					{ handle_bang( yylval ); return(0); }
			|	BANGSTRING CR
					{ handle_bang( find_bang( word ) ); return(0); }
			|	CR
					{ return(0); }
			|	error
					{ handle_error( yytext ); } CR { return(0); }
			;

command 	:	CPS_MON options CR
					{ STORE handle_cps_mon( txt ); }
			|	ALIAS opt_token opt_token CR
					{ STORE handle_alias( num_of_toks, token[0], token[1] ); }
			|	CD CR
					{ STORE handle_cd( getenv("HOME") ); }
			|	CD WORD CR
					{ STORE handle_cd( word ); }
			|	COMMANDS CR
					{ STORE handle_commands(); }
			|	CONNECT server NUMBER CR
					{ STORE connect_to_server( jm.jobname, word, $3, 1 ); }
			|	CONT CR
					{ STORE handle_cont(); }
			|	DBX dbx CR
					{ STORE handle_dbx( $2, NULL ); }
			|	DBX dbx WORD CR
					{ STORE handle_dbx( $2, word ); }
			|	DEBUG CR
					{ STORE handle_debug(); }
			|	DELETE delete CR
					{ STORE handle_delete( n1, n2 ); }
			|	DISCONNECT CR
					{ STORE handle_disconnect(); }
			|	EXIT CR
					{ STORE return( handle_exit() ); }
			|	HELP CR
					{ STORE handle_help(); }
			|	HISTORY CR
					{ STORE handle_history(); }
			|	HOOKS on_off CR
					{ STORE handle_hooks_switch( $2 ); }
			|	IMPORT CR
					{ STORE handle_import(); }
			|	JMINFO server CR
					{ STORE handle_jminfo( word, 0 ); }
			|	JMINFO ALL CR
					{ STORE handle_jminfo_all(); }
			|	JML display CR
					{ STORE handle_jml( $2 ); }
			|	JOB CR
					{ STORE handle_job(); }
			|	KILL kill CR
					{ STORE handle_kill( $2 ); }
			|	MONITOR monitor CR
					{ STORE handle_monitor( $2 ); }
			|	PING CR
					{ STORE handle_ping(); }
			|	PWD CR
					{ STORE handle_pwd(); }
			|	PROCESS CR
					{ STORE handle_process(); }
			|	QUEUES CR
					{ STORE handle_queues(); }
			|	QUIT CR
					{ return( handle_quit( 0 ) ); }
			|	QUITFORCE CR
					{ return( handle_quit( 1 ) ); }
			|	REFRESH CR
					{ STORE qmon_repaint(); qmon_update( 0 ); }
			|	REXEC CR
					{ STORE handle_rexec(); }
			|	SET CR
					{ STORE handle_set( VAR_LIST, n2, word ); }
			|	SET flag_var on_off CR
					{ STORE handle_set( n1, $3, word ); }
			|	SET flag_var CR
					{ STORE handle_set( n1, 1, word ); }
			|	SET string_var token CR
					{ STORE handle_set( n1, n2, token[0] ); }
			|	SH options CR
					{ STORE handle_sh( txt ); }
			|	STATE CR
					{ STORE handle_state( 0 ); }
			|	STATE JM CR
					{ STORE handle_state( -1 ); }
			|	STATUS CR
					{ STORE handle_status(); }
			|	STEP step CR
					{ STORE handle_step( $2, STEP_MODE, NULL, process_set ); }
			/**/
			|	STOP CR
					{ STORE handle_stop(); }
			|	STOP AT brkpt NUMBER CR
					{ STORE handle_set_breakpoint( $3, $4, $4 ); }
			|	STOP AT ALL brkpts CR
					{ STORE handle_set_breakpoint( $4, 0, -1 ); }
			|	STOP AT brkpts RANGE CR
					{ STORE handle_set_breakpoint( $3, n1, n2 ); }
			|	STOP AT procedure CR
					{ STORE handle_set_breakpoint( FUNC_HOOK_POINT, $3, $3 ); }
			/**/
			|	SYNCS CR
					{ STORE handle_syncs(); }
			|	TRACE trace CR
					{ STORE handle_trace( TRACE_MODE, process_set ); }
			|	UNALIAS token CR
					{ STORE handle_unalias( num_of_toks, token[0] ); }
			|	UNSET flag_var CR
					{ STORE handle_set( n1, 0, word ); }
			|	UNSET string_var CR
					{ STORE handle_set( n1, n2, "(null)" ); }
			|	UPSHOT on_off CR
					{ STORE handle_upshot_switch( $2 ); }
			|	UPSHOT options CR
					{ STORE handle_upshot_show( txt ); }
			|	VIEW trace CR
					{ STORE handle_trace( VIEW_MODE, process_set ); }
			|	WHERE CR
					{ STORE handle_where(); }
			;

init		:	/* empty */	{ init_lex_and_yacc(); }
			/*
			|	NULL_TOKEN	{ init_lex_and_yacc(); }
			*/
			;

opt_token	:	{ gettok(); }
			;	

token		:	{
					if( !gettok() )
					{
						handle_error( "" );
						return( 0 );
					}
				}
			;

on_off		:	ON  { $$ = 1; }
			|	OFF { $$ = 0; }
			;

options		:	/* empty */
			|	options WORD
			;

brkpts		:	HOOKS	{ $$ = USER_HOOK_POINT; }
			|	QUEUES	{ $$ = QUEUE_POINT; }
			|	SYNCS	{ $$ = SYNC_POINT; }
			;

brkpt		:	HOOK	{ $$ = USER_HOOK_POINT; }
			|	QUEUE	{ $$ = QUEUE_POINT; }
			|	SYNC	{ $$ = SYNC_POINT; }
			;

monitor		:	/* empty */	{ $$ = 30; }
			|	NUMBER		{ $$ = $1; }
			;

procedure	:	WORD { $$ = get_procedure_number( word ); }
			;

dbx			:	JM		{ $$ = 0; }
			|	NUMBER	{ $$ = $1; }
			;

delete		:	NUMBER	{ n1 = $1; n2 = $1; }
			|	ALL		{ n1 = 0;  n2 = -1; }
			|	RANGE
			;

display		:	/* empty */	{ $$ = 0; }
			|	CAT		{ $$ = 0; }
			|	TAIL	{ $$ = 1; }
			;

kill		:	ALL		{ $$ = -1; }
			|	NUMBER	{ $$ = $1; }
			;

server		:	/* empty */	{ strcpy( word, client.hostname ); }
			|	WORD
			;

string_var	:	COREDIR			{ n1 = VAR_COREDIR; }
			|	DBGDIR			{ n1 = VAR_DBGDIR; }
			|	DBX				{ n1 = VAR_DBX; }
			|	DBXRUN			{ n1 = VAR_DBXRUN; }
			|	OBJDIR			{ n1 = VAR_OBJDIR; }
			|	OUTPUT			{ n1 = VAR_OUTPUT; }
			|	PAGER			{ n1 = VAR_PAGER; }
			|	HOSTS			{ n1 = VAR_HOSTS; }
			|	PROMPT1			{ n1 = VAR_PROMPT1; }
			|	PROMPT2			{ n1 = VAR_PROMPT2; }
			;

flag_var	:	DBXCOPY			{ n1 = VAR_DBXCOPY; }
			|	DISPLAY			{ n1 = VAR_DISPLAY; }
			|	QUEUES			{ n1 = VAR_QUEUES; }
			|	SYNCS			{ n1 = VAR_SYNCS; }
			;

step		:	/* empty */	{ $$ = 1; add_process_to_set( 1, 0, num_procs ); }
			|	NUMBER		{ $$ = $1; add_process_to_set( 1, 0, num_procs ); }
			;

trace		:	/* empty */	{ add_process_to_set( 1, 0, num_procs ); } 
			|	RANGE		{ add_process_to_set( 1, n1, n2 ); }
			|	process_list
			;

process_list:	process_list NUMBER	{ add_process_to_set( 0, $2, $2 ); }
			|	NUMBER				{ add_process_to_set( 1, $1, $1 ); }
			;

%%

yyerror( s )
char *s;
{
}

yywrap()
{
}

add_process_to_set( clear, p1, p2 )
int clear;
int p1, p2;
{
	int index;

	if( process_set == NULL )
		process_set = (unsigned long *)create_empty_set( MAX_PROCESSES + 1 );

	if( clear )
		set_clear( process_set, MAX_PROCESSES + 1 );

	for( index = p1; index <= p2; index++ )
		add_to_set( index + 1, process_set );
}

init_lex_and_yacc()
{
	num_of_toks = 0;
}
