/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#define GLOBALS 1
#include "jmdb.h"
#include "version.h"

char Interrupt = 1;

extern JmInfo jm;
extern ClientInfo client;
extern VariableInfo var;

/* options to main program */
extern int	optind;
extern char	*optarg;

/******************************************************************/
/*
	FUNCTION NAME:

		sigint_handler

	DESCRIPTION:

		interrupt handler for jmdb

	FORMAL PARAMETERS:

		standard handler parameters
		calls siglongjmp to jump to the last place
		that sigsetjmp was called

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void sigint_handler( sig, code, sc )
int sig, code;
struct sigcontext *sc;
{
	monitor_exit( MONITOR_INTERRUPTED );

	if( Interrupt == 0 )
		Interrupt = 1;
	else
	{
		fprintf( stdout, STR_INTERRUPT );
		siglongjmp( env, 1 );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		sig_establish_handler

	DESCRIPTION:

		establishes signal handler

	FORMAL PARAMETERS:

		standard parameters

	FUNCTION VALUE:

		none

*/
/******************************************************************/
int sig_establish_handler( sig, handler )
int sig;
void (*handler)();
{
	struct sigaction act;

	act.sa_handler = handler;
	bzero( &( act.sa_mask ), sizeof( sigset_t ) );
	act.sa_flags = 0;

	if( sigaction( sig, &act, 0 ) == -1 )
	{
		perror( "sigaction" );
		return( 1 );
	}
	return( 0 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		initialize_structures

	DESCRIPTION:

		initialize all structures

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
initialize_structures()
{
	client.prognum	= 0;
	client.version	= 1;
	client.cl		= NULL;
	var.coredir		= NULL;
	var.objdir		= NULL;
	var.dbgdir		= NULL;
	var.output		= NULL;
	var.queues		= FALSE;
	var.syncs		= FALSE;
	var.display		= FALSE;
	var.dbxcopy		= FALSE;
	var.hosts		= NULL;
	var.pager		= NULL;
}

/******************************************************************/
/*
	PROGRAM NAME:

		jmdb main routine

	DESCRIPTION:

		JM Debugger ( jmdb )

	FORMAL PARAMETERS:

		standard parameters

	FUNCTION VALUE:

		none

*/
/******************************************************************/
main( argc, argv )
int argc;
char **argv;
{
	int opt;
	char host[128];
	u_long prog;
	u_long vers;
	int connect_now = 0;

	if( argc == 2 && !strcmp( argv[1], "-help" ) )
	{
		fprintf( stdout, STR_USAGE );
		exit(0);
	}

	create_default_aliases();
	initialize_structures();

	/* Get program options */
	while( ( opt = getopt( argc, argv, "a:f:" ) ) != -1 )
	{
		switch( (char) opt )
		{
		case 'a':
			strcpy( host, optarg );
			prog = atoi( argv[optind++] );
			vers = 1;
			connect_now = 1;
			break;
		case 'f':
			strcpy( jm.jobname, optarg );
			prog = 0;
			vers = 0;
			connect_now = 1;
			break;
		}
	}

	/* Establish interrupt handler */
	if( sig_establish_handler( SIGINT, sigint_handler ) )
	{
		fprintf( stdout, STR_SIGNAL_HANDLER_ERR, "SIGINT" );
		exit( 1 );
	}

	/* Print out jmdb version and opening message */
	fprintf( stdout, STR_OPENING, jmdb_version, jmdb_date );

	/* Handle interrupt while trying to connect to server at startup */
	if( connect_now )
	{
		if( sigsetjmp( env, 1 ) == 0 )
			connect_to_server( jm.jobname, host, prog, vers );
	}

	/* Handle interrupt while reading .jmdbinit */
	if( sigsetjmp( env, 1 ) == 0 )
		jmdb_init();

	/* All other interrupts return here */
	sigsetjmp( env, 1 );

	/* Close pipes and .jmdbinit */
	close_tmp_files();

	/* Get input from user */
	while( !get_input() );

	/* Exit jmdb */
	jmdb_exit();
}
