/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

#include "jmdb.h"

int declared_queue[ MAX_QUEUES ];
int q_num[ MAX_QUEUES ];
int num_queues;
short queue_data[ MAX_QUEUES ][ MAX_PROCESSES ];

extern ClientInfo	client;
extern int			num_procs;

/******************************************************************/
/*
	FUNCTION NAME:

		qmon_display

	DESCRIPTION:

		graphical display of queues

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
qmon_update( mode )
int mode;
{
	int q;
	int queue_set;
	int i, j;

	if( !connected( client.cl ) ) return;

	for( q = 0; q < MAX_QUEUES; q++ )
	{
		if( declared_queue[ q ] )
		{
			queue_set = *rpc_get_q_set_1( &q, client.cl );
			for( i = 0; i < num_procs; i++ )
			{
				if( ( queue_set>>i ) & 1 )
				{
					if( !mode | !queue_data[ q ][ i ] )
					{
						fill_rectangle( i*50+50, q_num[ q ]*50+50, 25, 25, 3 );
						queue_data[ q ][ i ] = 1;
					}
				}
				else
				{
					if( !mode | queue_data[ q ][ i ] )
					{
						fill_rectangle( i*50+50, q_num[ q ]*50+50, 25, 25, 2 );
						queue_data[ q ][ i ] = 0;
					}
				}
			} /* end of for loop */
		}
	} /* end of for loop */

	update_display();
}

/******************************************************************/
/*
	FUNCTION NAME:

		qmon_init

	DESCRIPTION:

		initialize graphical display of queues

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
int qmon_init()
{
	int i, j, q;

	if( create_window() )
		return( 1 );
	else
	{
		num_queues = 0;

		for( q = 0; q < MAX_QUEUES; q++ )
			declared_queue[ q ] = 0;

		for( q = 0; q < MAX_QUEUES; q++ )
		{
			if( *rpc_get_q_declared_1( &q, client.cl ) )
			{
				declared_queue[ q ] = 1;
				q_num[ q ] = num_queues++;
			}
		}
		for( i = 0; i < MAX_QUEUES; i++ )
			for( j = 0; j < MAX_PROCESSES; j++ )
				queue_data[ i ][ j ] = 0;

		qmon_repaint();

		return( 0 );
	}
}


qmon_repaint()
{
	int i, q;
	char string[4];

	window_resize( num_queues * 50 + 50 );

	for( i = 0; i < num_procs; i++ )
	{
		sprintf( string, "P%d", i + 1 );
		draw_string( 50*i+50, 25, string, 1 );
	}

	for( q = 0; q < MAX_QUEUES; q++ )
	{
		if( declared_queue[ q ] )
		{
			sprintf( string, "Q%d", q );
			draw_string( 0, 50*q_num[ q ]+70, string, 1 );
			for( i = 0; i < num_procs; i++ )
			{
				draw_box( i*50+50, q_num[ q ]*50+50, 25, 25, 1, 0 );
				fill_rectangle( i*50+50, q_num[ q ]*50+50, 25, 25, 2 );
			}
		}
	}

	update_display();
}

qmon_newq( queue )
int queue;
{
	char string[4];
	int i;

	if( !declared_queue[ queue ] )
	{
		declared_queue[ queue ] = 1;
		q_num[ queue ] = num_queues++;
		qmon_repaint();
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		init_declared_queues

	DESCRIPTION:

		initialize declared_queues array

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
init_declared_queues()
{
	int i;

	for( i = 0; i < MAX_QUEUES; i++ )
		declared_queue[ i ] = 0;
}
