
#include "all.h"
#include "../inc/types.h"
#include "../inc/time.h"
#include <stdio.h>
#include <sys/socket.h>
#include <rpc/rpc.h>

/******************************************************************/
/*
	FUNCTION NAME:

		gettransient

	DESCRIPTION:

		Generates a transient program number for use by server.

	FORMAL PARAMETERS:

		Input is prototype ( TCP or UDP ), program version

	FUNCTION VALUE:

		returns program number

                                  */
/******************************************************************/
int
gettransient( proto, vers, sockp )
int proto, vers, *sockp;
{
	static int prognum = 0x40000000;
	int s, len, socktype;
	struct sockaddr_in addr;
	static struct timeval TIMEOUT = { 25, 0 };
	CLIENT *client;

	switch( proto )
	{
		case IPPROTO_UDP:
			socktype = SOCK_DGRAM;
			break;
		case IPPROTO_TCP:
			socktype = SOCK_STREAM;
			break;
		default:
			return(0);
	}
	if( *sockp == RPC_ANYSOCK )
	{
		if( ( s = socket( AF_INET, socktype, 0 ) ) <  0 )
		{
			perror( "socket" );
			return(0);
		}
		*sockp = s;
	}
	else
		s = *sockp;

	addr.sin_addr.s_addr = 0;
	addr.sin_family = AF_INET;
	addr.sin_port = 0;
	len = sizeof( addr );
	bind( s, &addr, len );

	if( getsockname( s, &addr, &len ) < 0 )
	{
		perror( "getsockname" );
		return(0);
	}

	while( !pmap_set( prognum++, vers, proto, ntohs( addr.sin_port ) ) )
		continue;

	return( prognum - 1 );
}
