/*

 Author:	Roberto Ullfig
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL:ULLFIG
		Internet: ullfig@fnrobo.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.	    
	   All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/******************************************************************/
/*
		RPC File Information for Server and Client

			jm_rpc.x		RPCL Protocol Definition File

				generates the following four files. These files
				should not be edited as they are recreated whenever
				jm_rpc.x is modified.

			jm_rpc.h		RPC Header File
			jm_rpcclnt.c	Client-Side Stubs
			jm_rpcsvc.c		Server-Side Stubs
			jm_rpcxdr.c		XDR Declarations

		Server Code
			jm_rpc.c		Initialization and Termination
			jm_rpcsub.c		Subroutines

		Client Code
			jm_rpccreate.c	Initialization

                                  */
/******************************************************************/
/*
	Include Files
*/
#include "all.h"
#include "../inc/types.h"
#include <stdio.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#include "jm_rpc.h"

/*
	External Variables
*/
extern void jmprog_1();
static int prognum = 0;
int rpc_call;

/******************************************************************/
/*
	FUNCTION NAME:

		jm_rpc_init

	DESCRIPTION:

		Creates an RPC service and writes hostname, prognnum, and version
		to a ".prog" file so that clients can later call the server
		Program numbers are generated dynamically so as to allow more than
		one server per machine

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

                                  */
/******************************************************************/
void
jm_rpc_init()
{
	SVCXPRT *transp;
	char fname[256];
	FILE *fp;
	int sock;
	char errmsg[132];

	sprintf(fname,"%s.prog",job_name);
	fp = fopen(fname,"w");
	if( fp == NULL )
	{
		sprintf(errmsg, "Failed to open %s",fname);
		printerr("RPC",WARNING_MSG,errmsg,NULL,2);
		sprintf(errmsg, "RPC service not started");
		printerr("RPC",WARNING_MSG,errmsg,NULL,1);
		return;
	}

	sock = RPC_ANYSOCK;

	prognum = gettransient(IPPROTO_TCP, JMVERS, &sock);

	fprintf(fp,"%s %d %d\n",jm_hostname,prognum,JMVERS);
	fclose(fp);

	transp = svctcp_create(sock, 1024, 1024);
	if( transp == NULL )
	{
		sprintf(errmsg,"Could not create tcp service\n");
		printerr("RPC",WARNING_MSG,errmsg,NULL,3);
		sprintf(errmsg, "RPC service not started");
		printerr("RPC",WARNING_MSG,errmsg,NULL,1);
		pmap_unset( prognum, JMVERS );
		close( sock );
		return;
	}
	if (!svc_register(transp, prognum, JMVERS, jmprog_1, 0))
	{
		sprintf(errmsg,"Could not register tcp service\n");
		printerr("RPC",WARNING_MSG,errmsg,NULL,4);
		sprintf(errmsg, "RPC service not started");
		printerr("RPC",WARNING_MSG,errmsg,NULL,1);
		pmap_unset( prognum, JMVERS );
		close( sock );
		svc_destroy( transp );
		return;
	}
	printf("jm[%s] RPC service started on %s\n",jm_hostname,jm_hostname);
	fprintf(jml,"jm[%s] RPC service started on %s\n",jm_hostname,jm_hostname);
}

/******************************************************************/
/*
	FUNCTION NAME:

		jm_rpc_exit

	DESCRIPTION:

		Terminates RPC service

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

                                  */
/******************************************************************/
void
jm_rpc_exit()
{
	callback_endjob();
	if( prognum )
		pmap_unset( prognum, JMVERS );
}

jm_rpc_manage( fdset, fds, rpcset )
fd_set *fdset, *fds, *rpcset;
{
	*rpcset = svc_fdset;

	if( jm_pause )
	{
		*fds = *rpcset;
	}
	else
	{
		*fds = *fdset;
		set_union( fds, rpcset, FD_SETSIZE );
		if( jm_step )
		{
			jm_pause = TRUE;
			jm_step  = FALSE;
		}
	}
	rpc_call = FALSE;
}

void
jm_rpc_wait_for_request()
{
	fd_set rpcset;
	int nfound;

	jm_pause = TRUE;
	while( jm_pause )
	{
		rpcset = svc_fdset;
		nfound = select( FD_SETSIZE, &rpcset, NULL, NULL, NULL );
		svc_getreqset( &rpcset );
	}
	rpc_call = TRUE;
}

jm_rpc_breakpoint( break_point, value, location )
int break_point;
int value;
int *location;
{
	if( break_point )
	{
		*location = value;
		jm_rpc_wait_for_request();
	}
}

jm_rpc_get_request( set )
fd_set *set;
{
	if( !rpc_call )
		svc_getreqset( set );
}
