/*

 Author:	Roberto Ullfig
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL:ULLFIG
		Internet: ullfig@fnrobo.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.	    
	   All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#include "../inc/types.h"
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <rpc/rpc.h>
#include "jm_rpc.h"

/******************************************************************/
/*
	FUNCTION NAME:

		create_rpc_client

	DESCRIPTION:

		Subroutine called by client to inquire program number of
		JM and connect to it.

	FORMAL PARAMETERS:

		Input is name of file containing program number
		or hostname, prognum, and version
		Output is pointer to CLIENT data type

	FUNCTION VALUE:

		none

                                  */
/******************************************************************/
create_rpc_client( jobname, hostname, prognum, version, client )
char *jobname;
char *hostname;
u_long *prognum, *version;
CLIENT **client;
{
	struct sockaddr_in addr;
	struct hostent *hp;
	int sock = RPC_ANYSOCK;

	hp = gethostbyname( hostname );
	if( hp == NULL )
	{
		fprintf( stderr, "%s:%s\n", hostname, strerror( errno ) );
		return( 0 );
	}
	bcopy( hp->h_addr, &addr.sin_addr, hp->h_length );
	addr.sin_family = AF_INET;
	addr.sin_port = 0;

	/* *client = clnt_create( hostname, *prognum, *version, "tcp" ); */
	*client = clnttcp_create( &addr, *prognum, *version, &sock, 0, 0 );
	if( *client == NULL )
	{
		clnt_pcreateerror( hostname );
		return( 0 );
	}
	(*client)->cl_auth = authunix_create_default();

	strcpy( jobname, *rpc_get_job_name_1( NULL, *client ) );

	return( 1 );
}

int
read_prog_file( jobname, hostname, prognum, version )
char *jobname;
char *hostname;
int *prognum;
int *version;
{
	char fname[128];
	FILE *fp;

	sprintf(fname,"%s.prog",jobname);
	fp = fopen(fname,"r");
	if( fp == NULL )
	{
		fprintf(stderr,"Could not open file: %s\n",fname);
		return( 0 );
	}
	
	fscanf(fp,"%s %d %d",hostname,prognum,version);
	fclose(fp);

	return( 1 );
}

destroy_rpc_client( client )
CLIENT *client;
{
	if( client != NULL )
	{
		auth_destroy( client->cl_auth );
		clnt_destroy( client );
	}
}
