/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													

 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/****************************************************************************/
/*																			*/
/*																			*/
/*						CPS Shared Memory Manager							*/
/*																			*/
/*																			*/
/****************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include <stdio.h>
#include <errno.h>
#include "../inc/types.h"
#include "../inc/time.h"

#ifdef SGI
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef HPUX
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef AIX
#include <sys/select.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef MIPS
#include <sys/socket.h>
#include <bsd/netinet/in.h>
#include <netdb.h>
#endif

#ifdef R3000
#include <bsd/sys/socket.h>
#include <bsd/netinet/in.h>
#include <bsd/netdb.h>
#endif

#ifdef SUNOS
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef ULTRIX
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif


#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>

#include "../inc/perr.h"
#include "../inc/param.h"
#include "../inc/shm.h"
#include "../inc/cps_msgs.h"
#include "../inc/xfunc.h"


/**********/
/* Macros */
/**********/

#include "../inc/cps_msg_rt.h"

#define FACILITY "SHM"

/**********************************/
/* External Non-Integer Functions */
/**********************************/

FILE *fopen();

/********************/
/* Global Variables */
/********************/

FILE *logfp = NULL;

fd_set fdset;

int shm_lpn;			/* LPN of this shared memory process */
int shm_index;			/* Index of this process in smap */
int shm_sem;			/* Semaphore number of this process */
int num_semaphores = 0;

int num_large_buffers,num_small_buffers;

int shmid;			/* Shared memory identifier */
extern int q_semid;

int shm_exist = 0;
int sem_exist = 0;

int num_user_proc;		/* Number of user processes */
int num_local_proc;		/* Number of local user processes */
int num_shm_proc;		/* Number of shm processes */

int (*p_error)();

int current_socket;
int current_pid;
int current_lpn;

int passive_socket;

extern char local_hostaddr[24];
extern char local_hostname[64];
extern int assigned_port_number;

extern int suppress_errors;

int num_fresh_dead = 0;
int fresh_dead[ MAX_PROCESSES ];

struct data_xfer {

	int index;			/* Index in array of this record */

	int type; 			/* 0 = acp_send, 1 = acp_get */
	int src_lpn;
	int dst_lpn;
	int not_lpn;		/* Notify lpn */
	int blockno;
	int offset;
	int size;
	int vaddr;			/* Virtual address */
	int nleft;			/* Number of packets left to be sent */
	int outstand;		/* Number packets sent but still waiting for acks */

	struct data_xfer *prev;	/* Previous xfer in circular list */
	struct data_xfer *next;	/* Next data xfer in list */
};

#define NBUF_SAVED 3
#define MAX_OUTSTAND 3

#define MAX_DX 128

struct data_xfer dx[MAX_DX];	/* Data transfers in progress */
struct data_xfer *adx = NULL;	/* First active data xfer in circular list */
struct data_xfer *idx = NULL;	/* First inactive xfer record in list */
struct data_xfer *cdx = NULL;	/* Data xfers completed */
int num_cdx = 0;

int total_outstand = 0;

/******************************/
/* List of active TCP sockets */
/******************************/

struct active_socket {

	int index;
	int sock;
	int socknum;
	struct active_socket *next;

};

struct active_socket *asock = NULL;

/****************************************************/
/* Private stash of small and large message buffers */
/****************************************************/

#define MAX_SMALL_STASH 20
#define MAX_LARGE_STASH 10

struct private_stash {
	struct qelem *q;
	struct private_stash *next;
};

struct private_stash *fstash = NULL;	/* Free list */
struct private_stash *sstash = NULL;
struct private_stash *lstash = NULL;
int sstash_size = 0;
int lstash_size = 0;

/****************************/
/* Shared Memory Structures */
/****************************/

int *nprocs;
int *semids;
struct procmap *pmap;
struct shmmap *smap;
cps_message *shmem;
cps_msg *sshmem;
extern struct qelem *qelem;
extern int *q_head;	
extern int *q_tail;
extern int *q_block;

/**********************/
/* External Variables */
/**********************/

extern int hm_report;
extern int hm_mem_interval;

/************************************************************************/
/*

	Function: get_inactive_xfer, activate_xfer, xfer_complete, initialize_dx,
		  calculate_nleft

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

dump_xfers(i)
int i;
{
	struct data_xfer *x;

	fprintf(logfp,"current xfer = %d\n",i);
	fprintf(logfp,"total_outstand = %d\n",total_outstand);
	fprintf(logfp,"Inactive List:\n");

	x = idx;
	while (x != NULL) {
		fprintf(logfp,"%d ",x->index);
		x = x->next;
	}
	fprintf(logfp,"\n");

	fprintf(logfp,"Active List:\n");
	if (adx != NULL) {
		x = adx;
		do {
			fprintf(logfp,"%d: outstand=%d, nleft=%d, size=%d\n",
				x->index,x->outstand,x->nleft,x->size);
			x = x->next;
		}  while (x != adx);
	}

	fprintf(logfp,"Completed List:\n");
	x = cdx;
	while (x != NULL) {
		fprintf(logfp,"%d ",x->index);
		x = x->next;
	}
	fprintf(logfp,"\n");
	fflush(logfp);
}

int get_inactive_xfer()		/* Causes dx structure to be activated */
{
	struct data_xfer *x;

/* If there are too many ongoing data transfers, abort the job */

	if (idx == NULL) shm_shutdown(8);

/* Grab the first inactive record */

	x = idx;
	idx = idx->next;

/* Return the index of the inactive record obtained */

	return(x->index);
}

activate_xfer(i)
int i;
{
	if (adx == NULL) {		/* This is the first active xfer */
		adx = &dx[i];
		adx->next = adx;
		adx->prev = adx;
	}
	else {			/* Add xfer to circular list */
		dx[i].prev = adx;
		dx[i].next = adx->next;
		adx->next = &dx[i];
		dx[i].next->prev = &dx[i];
	}
}

xfer_complete(i)
int i;
{

/* Remove the record from the circular list of active transfers */

	if (dx[i].next == (&dx[i])) {	/* No more active xfers */
		adx = NULL;
	}
	else {
		if (adx == &dx[i]) {
			adx = dx[i].prev;
		}
		dx[i].next->prev = dx[i].prev;
		dx[i].prev->next = dx[i].next;
	}

/* Put the record on the list of completed transfers */

	dx[i].prev = NULL;
	dx[i].next = cdx;
	cdx = &dx[i];
	num_cdx++;
}

#define MAX_CHUNK (MAX_MESSAGE_SIZE - 5*sizeof(long))

#define calculate_nleft(index)				\
	dx[index].nleft = dx[index].size/MAX_CHUNK; 	\
	if ((dx[index].nleft*MAX_CHUNK) != dx[index].size) dx[index].nleft++;

initialize_dx()
{
	int i;

	idx = &dx[0];
	for (i=0; i<MAX_DX-1; i++) {
		dx[i].index = i;
		dx[i].prev = NULL;
		dx[i].next = &dx[i+1];
	}
	dx[MAX_DX-1].index = MAX_DX-1;
	dx[MAX_DX-1].prev = NULL;
	dx[MAX_DX-1].next = NULL;
}

/************************************************************************/
/*

	Function:  activate_socket

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int activate_socket(index,sock,socknum)
int index,sock,socknum;
{
	struct active_socket *s;

	s = (struct active_socket *) hmalloc(sizeof(struct active_socket));
	s->index = index;
	s->sock = sock;
	s->socknum = socknum;
	if (asock == NULL) {
		s->next = s;
		asock = s;
	}
	else {
		s->next = asock->next;
		asock->next = s;
	}

	return 0;
}

/************************************************************************/
/*

	Function: Private stash push and pop functions.

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

struct qelem *sstash_pop()
{
	struct qelem *q;
	struct private_stash *s;

	if (sstash != NULL) {
		s = sstash;
		q = s->q;		/* Get the buffer */
		sstash = s->next;	/* Put list element on the free list */
		s->next = fstash;
		fstash = s;
		sstash_size--;
	}
	else {
		q = NULL;
	}
	return(q);
}

int sstash_pop_n(qp,max_qp)
struct qelem *qp[];
int max_qp;
{
	struct private_stash *s;
	int n,nqp;
	int i;
	int j;

/* Get as many buffers as possible from our private stash */

	i = 0;
	while ((sstash != NULL) && (i < max_qp)) {
		s = sstash;
		qp[i] = s->q;		/* Get the buffer */
		sstash = s->next;	/* Put list element on the free list */
		s->next = fstash;
		fstash = s;
		sstash_size--;
		i++;
	}
	n = max_qp - i;
	if (n == 0) return i;	/* Got all the buffers requested */

/* Get the remaining buffers by popping from the free queue */

/*
	for( j = 0; j < n; j++ )
	{
		qp[ i ] = q_pop_retry( num_semaphores-1 );
		i++;
	}
*/

	nqp = q_pop_no_wait_n(num_semaphores-1,&qp[i],n);
	i += nqp;

	return i;
}

int sstash_pop_retry_n(qp,max_qp)
struct qelem *qp[];
int max_qp;
{
	int nqp, n, j;

	nqp = sstash_pop_n(qp,max_qp);

	if( nqp < max_qp )
	{
		n = max_qp - nqp;
		for( j = 0; j < n; j++ )
		{
			qp[ nqp ] = q_pop_retry( num_semaphores-1 );
			nqp++;
		}
	}
	return nqp;
}

struct qelem *lstash_pop()
{
	struct qelem *q;
	struct private_stash *s;

	if (lstash != NULL) {
		s = lstash;
		q = s->q;		/* Get the buffer */
		lstash = s->next;	/* Put list element on the free list */
		s->next = fstash;
		fstash = s;
		lstash_size--;
	}
	else {
		q = NULL;
	}
	return(q);
}

int lstash_pop_n(qp,max_qp)
struct qelem *qp[];
int max_qp;
{
	struct private_stash *s;
	int n,nqp;
	int i;

/* Get as many buffers as possible from our private stash */

	i = 0;
	while ((lstash != NULL) && (i < max_qp)) {
		s = lstash;
		qp[i] = s->q;		/* Get the buffer */
		lstash = s->next;	/* Put list element on the free list */
		s->next = fstash;
		fstash = s;
		lstash_size--;
		i++;
	}
	n = max_qp - i;
	if (n == 0) return i;	/* Got all the buffers requested */

/* Get the remaining buffers by popping from the free queue */

	nqp = q_pop_no_wait_n(0,&qp[i],n);
	i += nqp;

	return i;
}

int sstash_push(q)
struct qelem *q;
{
	struct private_stash *s;

	if (q->size != 0) {
		lstash_push(q);
		return 0;
	}

	if (sstash_size >= MAX_SMALL_STASH) {
		q_push_retry(num_semaphores-1,q);
	}
	else {
		if (fstash != NULL) {	/* Get a list element */
			s = fstash;
			fstash = s->next;
		}
		else {
			s = (struct private_stash *) hmalloc(sizeof(struct private_stash));
		}
		s->q = q;
		s->next = sstash;
		sstash = s;
		sstash_size++;
	}
	return 0;
}

int lstash_push(q)
struct qelem *q;
{
	struct private_stash *s;

	if (q->size != 1) {
		sstash_push(q);
		return 0;
	}

	if (lstash_size >= MAX_LARGE_STASH) {
		q_push_retry(0,q);
	}
	else {
		if (fstash != NULL) {	/* Get a list element */
			s = fstash;
			fstash = s->next;
		}
		else {
			s = (struct private_stash *) hmalloc(sizeof(struct private_stash));
		}
		s->q = q;
		s->next = lstash;
		lstash = s;
		lstash_size++;
	}
	return 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  dump_shm

  DESCRIPTION:

	  [@tbs@]

								  */
/******************************************************************/

dump_queues(fp)
FILE *fp;
{
	int i;

	struct qelem *q;

	fprintf(fp,"\nQueues (%d)\n",num_semaphores);
	fprintf(fp,  "------\n");
	fflush(fp);
	for (i=0; i<num_semaphores; i++) {
		if (q_head[i] == -1)
			q = NULL;	
		else
			q = &qelem[q_head[i]];
		fprintf(fp,"%d: q = ",i);
		fflush(fp);
		while (q != NULL) {
			fprintf(fp,"%d",q->buf);
			if (q->size == 0)
				fprintf(fp,"s ");
			else
			fprintf(fp,"l ");
				fflush(fp);
			if (q->next == -1)
				q = NULL;
			else
				q = &(qelem[q->next]);
		}
		fprintf(fp,"\n");
		fflush(fp);
	}
	fflush(fp);
}

dump_shm()
{
	FILE *fp;

	char f[80];

	int i;

	struct qelem *q;

	fp = logfp;

	fprintf(fp,"Number of user processes = %d\n",nprocs[0]);
	fprintf(fp,"Number of shm processes = %d\n",nprocs[1]);
	fprintf(fp,"Number of local processes = %d\n",nprocs[2]);
	fprintf(fp,"Queue semaphore id = %d\n",semids[0]);
	fprintf(fp,"\nShared Memory Processes\n");
	fprintf(fp,  "-----------------------\n");
	for (i=0; i<=num_shm_proc; i++) {
		fprintf(fp,"(%s,%s,%d) pid=%d semnum=%d sock=%d status=%d\n",
			smap[i].hostname,smap[i].hostaddr,smap[i].portno,
			smap[i].pid,smap[i].semnum,smap[i].sock,smap[i].status);
	}

	fprintf(fp,"\nUser Processes\n");
	fprintf(fp,  "---------------\n");
	for (i=0; i<num_user_proc; i++) {
		fprintf(fp,"local=%d pid=%d shmid=%d semnum=%d\n",
			pmap[i].local,pmap[i].pid,pmap[i].shmid,pmap[i].semnum);
	}

	dump_queues(fp);
	fflush(fp);
}

/******************************************************************/
/*
	FUNCTION NAME:

	 shm_exit

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

shm_exit(x) 
int x;
{
#ifdef DIAG
	fprintf(logfp,"shm_exit(%d) called\n",x);
	fclose(logfp);
#endif
	if (shm_exist) shmctl(shmid,IPC_RMID,NULL);
	if (sem_exist) semctl(q_semid,0,IPC_RMID);
	exit(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	 send_startup_msg

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_startup_msg(msg) 
char *msg;
{
	if (send_counted(1,msg) < 0) shm_exit(1);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  log_shm_error

  DESCRIPTION:

	  Sends an error message to the JM.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

log_shm_error(msg)
char *msg;
{
	cps_msg m;

	if (logfp != NULL) {
		fprintf(logfp,"%s\n",msg);
		fflush(logfp);
		return;
	}

	if (smap[0].status == 0) return;

/* Build the message */

	cps_bld_shm_error(&m,shm_lpn,0,msg);

/* Send it to the JM */

	send_to_jm(&m);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  display_error

  DESCRIPTION:

	  Sends an error message to the Job Manager.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

display_error(str)
char *str;
{
	(*p_error)(str);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  shm_disconnect

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

shm_disconnect(i)
int i;
{

	struct active_socket *s0,*s1;

/* Remove the socket from the set of file descriptors, and close the socket */

	if (smap[i].status) {
		smap[i].status = 0;
		FD_CLR(smap[i].sock,&fdset);
		socket_close(smap[i].sock);
	}
	if (smap[i].status2) {
		smap[i].status2 = 0;
		FD_CLR(smap[i].sock2,&fdset);
		socket_close(smap[i].sock2);
	}

/* Remove the socket from the circular list of active sockets */

	s0 = asock;
	asock = NULL;
	s1 = s0;
	do {
		if (s1->index != i) {
			activate_socket(s1->index,s1->sock,s1->socknum);
		}
		if (s1->next == s0) s1->next = NULL;
		s1 = s1->next;
	}
	while (s1 != NULL);

/* Free the old circular list */

	while (s0 != NULL) {
		s1 = s0->next;
		hfree(s0);
		s0 = s1;
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	  shm_shutdown

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

shm_shutdown(x)
int x;
{
	cps_msg msg;

	int i;

	char logmsg[80];

/* Report why we are shutting down */

	if (x == 1)
		sprintf(logmsg,
			"Shutting down; no more processes alive on this system");
	else if (x ==2)
		sprintf(logmsg,"Shutting down; failed to send message to JM");
	else if (x == 3)
		sprintf(logmsg,"Shutting down; error polling sockets");
	else if (x == 4)
		sprintf(logmsg,"Shutting down; failed to connect to SHM");
	else if (x == 5)
		sprintf(logmsg,
			"Shutting down; failed to get message from JM");
	else if (x == 6)
		sprintf(logmsg,"Shutting down; lost connection with SHM");
	else if (x ==7)
		sprintf(logmsg,"Shutting down; received shutdown message from JM");
	else if (x == 8)
		sprintf(logmsg,
			"Shutting down; maximum number of active transfers exceeded");
	else if (x == 10)
		sprintf(logmsg,"Shutting down; bad message destination");
	else if (x == 13)
		sprintf(logmsg,"Shutting down; error waiting on sockets");
	else
		sprintf(logmsg,"Shutting down, reason=%d",x);
	if( x != 2 && x != 5 )
		log_shm_error(logmsg);

/* Tell the JM we are shutting down (if possible) */

	if (smap[0].status) {
		cps_bld_shm_shutdown(&msg,shm_lpn,0);
		suppress_errors = -1;
		send_binary(smap[0].sock,&msg,4*sizeof(long)+msg.size);
		suppress_errors = 0;
	}

/* Kill all the local processes */

	for (i=0; i<num_user_proc; i++) {
		if (pmap[i].local && (pmap[i].status == -1)) {
			sprintf(logmsg,"Killing %d",pmap[i].pid);
			if( x != 2 && x != 5 )
				log_shm_error(logmsg);
			kill(pmap[i].pid,SIGTERM);
		}
	}

	shm_exit(2);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  signal_user

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

signal_user(lpn)
int lpn;
{
	int i;
 
	i = lpn - FIRST_USER_LPN;

	if (pmap[i].status == 0) return;	/* Process is dead */
	if (pmap[i].status == 1) return;	/* Process is finished */

	retry:
	if (kill(pmap[i].pid,SIGUSR1) == -1) {
		if (errno == EINTR) goto retry;
		usr_proc_dead(lpn,"SHM failed to signal process",-1);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  usr_proc_dead

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

usr_proc_dead(lpn,text,tell_jm)
int lpn,tell_jm;
char *text;
{
	cps_msg m;

	struct qelem *q;

	int i,j;

	int s;

	char logmsg[256];

	if (tell_jm) {
		sprintf(logmsg,"User process %d dead",lpn);
		log_shm_error(logmsg);
	}

	i = lpn - FIRST_USER_LPN;
	if ((i < 0) || (i >= num_user_proc)) {
		printerr(FACILITY,WARNING_MSG,"LPN out of range",NULL,1);
		return;
	}

/* If the process is local and not already dead, kill it */
/* Mark it as dead in the shared memory */

	if (pmap[i].local && (pmap[i].status == -1)) kill(pmap[i].pid,SIGTERM);
	pmap[i].status = 0;

/* Inform the JM of the dead process */

	if (tell_jm) {
		cps_bld_usr_proc_dead(&m,shm_lpn,0,lpn,text);
		send_to_jm(&m);
	}

/* Clear its queues */



/* If there are no more local processes that are alive, shutdown */

	j = 0;
	for (i=0; i<num_user_proc; i++) {
		if (pmap[i].local && (pmap[i].status == -1)) {
			j = 1;
			break;
		}
	}
	if (j == 0) shm_shutdown(1);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_to_jm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_to_jm(msg)
cps_message *msg;
{
	hold_signals();
	suppress_errors = -1;
	if (send_binary(smap[0].sock,msg,4*sizeof(long)+msg->size) == -1) {
		suppress_errors = 0;
		shm_shutdown(2);
	}
	suppress_errors = 0;
	release_signals();
	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_to_shm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_to_shm(i,msg)
int i;
cps_message *msg;
{
	hold_signals();
	if (smap[i].status == 0) {			/* Not yet connected */
		if (connect_to_shm(i) < 0) shm_shutdown(4);
	}

	suppress_errors = -1;
	if (send_binary(smap[i].sock,msg,4*sizeof(long)+msg->size) == -1) {
		suppress_errors = 0;
		shm_shutdown(6);
	}
	suppress_errors = 0;
	release_signals();
	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_to_user

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_to_user(lpn,q,m)
int lpn;
struct qelem *q;
cps_message *m;
{
	int j;

	j = lpn - FIRST_USER_LPN;

/* Put message on local user's queue */

	if (pmap[j].local) {	
		q_push_retry(pmap[j].semnum,q);
		signal_user(lpn);
	}

/* Send message to user's shared memory process */

	else {	
		j = pmap[j].shmid;
		send_to_shm(j,m);
		lstash_push(q);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  shm_init

  DESCRIPTION:

	  Allocates memory for the shared memory and initializes pointers.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int shm_init() {

	int size;

	char *base_addr,*addr;

	char text[80];

/* Calculate the size of the shared memory segment */

	size = 0;
	size += 3 * sizeof(int);
	size += 2 * sizeof(int);
	size += num_user_proc * sizeof(struct procmap);
	size += (num_shm_proc + 1) * sizeof(struct shmmap);
	num_large_buffers = SHM_FUNC(num_semaphores);
	num_small_buffers = S_SHM_FUNC(num_semaphores);
	size += num_large_buffers * sizeof(cps_message);
	size += num_small_buffers * sizeof(cps_msg);
	size += (num_large_buffers + num_small_buffers) * sizeof(struct qelem);
	size += (num_semaphores) * sizeof(struct qelem *);
	size += (num_semaphores) * sizeof(struct qelem *);
	size += (num_semaphores) * sizeof(int);

/* Create the shared memory and attach */

	if ((shmid = shm_create(size)) == -1) return(-1);
	shm_exist = -1;
	if ((int)(base_addr = shm_attach(shmid)) == -1) return(-1);

/* Initialize pointers */

	addr = base_addr;

	nprocs = (int *) addr;
	addr += 3 * sizeof(int);

	semids = (int *) addr;
	addr += 2 * sizeof(int);

	pmap = (struct procmap *) addr;
	addr += num_user_proc * sizeof(struct procmap);

	smap = (struct shmmap *) addr;
	addr += (num_shm_proc + 1) * sizeof(struct shmmap);

	shmem = (cps_message *) addr;
	addr += SHM_FUNC(num_semaphores) * sizeof(cps_message);

	sshmem = (cps_msg *) addr;
	addr += S_SHM_FUNC(num_semaphores) * sizeof(cps_msg);

	qelem = (struct qelem *) addr;
	addr += (SHM_FUNC(num_semaphores)+S_SHM_FUNC(num_semaphores))
		* sizeof(struct qelem);

	q_head = (int *) addr;
	addr += (num_semaphores) * sizeof(int *);

	q_tail = (int *) addr;
	addr += (num_semaphores) * sizeof(int *);

	q_block = (int *) addr;

	return(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  setup_shm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int setup_shm()
{
	int i,j;

	int index;

	char msg[20];

	char text[80];

	struct hostent *hp,*gethostbyname();

	int n;

	nprocs[0] = num_user_proc;
	nprocs[1] = num_shm_proc;
	nprocs[2] = num_local_proc;

	semids[0] = 0;
	semids[1] = 0;

/* Initialize process map */

	for (i=0; i<num_user_proc; i++) {
		pmap[i].local = 0;  	/* Will be changed to -1 by local processes */
		pmap[i].pid = 0;		/* Will be set by local processes */
		pmap[i].shmid = 0;
		pmap[i].semnum = -1;
		pmap[i].status = -1;	/* Assume all are alive to begin with */
	}

/* Zero out shm map */

	for (i=0; i<=num_shm_proc; i++) {
		smap[i].hostname[0] = 0;
		smap[i].hostaddr[0] = 0;
		smap[i].portno = 0;
		smap[i].pid = 0;
		smap[i].semnum = -1;
		smap[i].sock = 0;
		smap[i].status = 0;
		smap[i].sock2 = 0;
		smap[i].status2 = 0;
		smap[i].data_ack_count = 0;
		smap[i].data_msg_count = 0;
	}

/* Get location of JM & other shm processes */

	for (i=0; i<=num_shm_proc; i++) {
		if (read_counted(0,text,80) < 0) {
		}
		strcpy(smap[i].hostname,text);
		if (read_counted(0,text,80) < 0) {
		}
		strcpy(smap[i].hostaddr,text);
		if (read_counted(0,text,80) < 0) {
		}
		smap[i].portno = atoi(text);

		hp = gethostbyname(smap[i].hostname);
		if (hp != NULL) {
			if (hp->h_length > 4) {
			}
			bcopy(hp->h_addr,(char *)&(smap[i].sockaddr),4);
		}

	}

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(0,msg,20) < 0) {
	}

	if (strcmp(msg,"suicide") == 0) shm_exit(3);

/* Write our pid in the shared memory */

	smap[shm_index].pid = getpid();

/* Record the JM's current socket & status in the shared memory */

	smap[0].sock = 1;		/* standard output */
	smap[0].status = -1;	/* connected */

/* Get user-->shm mappings */

	i = 0;
	while (1) {
		if (read_counted(0,msg,20) < 0) {
		}
		j = atoi(msg);

		if (j == -1) {
			if (i < num_user_proc) {
				printerr(FACILITY,FATAL_MSG,"Failed to get user to shm map",NULL,2);
				shm_exit(4);
			}
			break;
		}
		if (i < num_user_proc) pmap[i].shmid = j;
		i++;
		if (i > MAX_PROCESSES) {
			printerr(FACILITY,FATAL_MSG,"Failed to get user to shm map",NULL,3);
			shm_exit(5);
		}
	}

/* Tell the JM that we are ok */

	send_startup_msg("ok");

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(0,msg,20) < 0) {
	}
	if (strcmp(msg,"suicide") == 0) shm_exit(6);

/* Assign semaphores to each process accessing shared memory 
  (0 is for free queue) */

	index = shm_lpn - FIRST_SHM_LPN + 1;

	j = 1;
	for (i=0; i<num_user_proc; i++) {
		if (pmap[i].shmid == index) {
			pmap[i].semnum = j;
			j++;
		}
	}

	smap[index].semnum = j;
}

/******************************************************************/
/*
	FUNCTION NAME:

	 hold_signals

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int hold_signals() {

	sig_block_one(SIGUSR1);
	sig_block_one(SIGALRM);

}

/******************************************************************/
/*
	FUNCTION NAME:

	 release_signals

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int release_signals() {

	sig_pop_mask();
	sig_pop_mask();

}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_fresh_dead

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int handle_fresh_dead() {

	int i,j;

/* For each process we couldn't signal, check to see if we still believe the
   process is alive, if so, declare it dead. */

	hold_signals();
	for (i=0; i<num_fresh_dead; i++) {
		j = fresh_dead[i];
		if (pmap[j].status == -1) {
			usr_proc_dead(j+FIRST_USER_LPN,"Unable to signal process",-1);
		}
	}
	num_fresh_dead = 0;
	release_signals();

}

/******************************************************************/
/*
	FUNCTION NAME:

	  connect_to_shm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int connect_to_shm(i)
int i;
{
	int s;

	if (smap[i].status != 0) abort();

	s = create_active_socket();
	if (s == -1) return(-1);

	if (initiate_connection(s,smap[i].hostname,smap[i].portno) == -1) return -1;

/* Add socket to set of descriptors */

	smap[i].sock = s;
	smap[i].status = -1;
	FD_SET(smap[i].sock,&fdset);
	activate_socket(i,s,1);

	return(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  accept_shm_connect

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int accept_shm_connect()
{
	int s;

	int i;

	struct sockaddr_in peer;
	int peerlen;

/* Accept the connection */

	if ((s = accept_connection(passive_socket)) == -1) return(-1);

/* Get the address of the partner shm */

	peerlen = sizeof(struct sockaddr_in);
	if (getpeername(s,&peer,&peerlen) < 0) {
		socket_close(s);
		return(-1);
	}

/* Find the shm process with that address */

	for (i=1; i<=num_shm_proc; i++) {
		if (peer.sin_addr.s_addr == smap[i].sockaddr) break;
	}
	if (i > num_shm_proc) {
		return(-1);
	}

	if (smap[i].status) {
		smap[i].sock2 = s;
		smap[i].status2 = -1;
		FD_SET(smap[i].sock2,&fdset);
		activate_socket(i,s,2);
	}
	else {
		smap[i].sock = s;
		smap[i].status = -1;
		FD_SET(smap[i].sock,&fdset);
		activate_socket(i,s,1);
	}
	return(i);
}

/************************************************************************/
/*

	Function: handle_data_ack

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int handle_data_ack(i)
int i;
{
	dx[i].outstand--;
	total_outstand--;
	if ((dx[i].outstand == 0) && (dx[i].size <= 0) && (dx[i].nleft == 0)) {
		xfer_complete(i);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_data_ack

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_data_ack(i,s_index)
int i,s_index;
{
	struct qelem *q;
	cps_message *m;

	if (s_index == shm_index) {		/* Data ack sent to ourself */
		handle_data_ack(i);
	}
	else {				/* Send data ack to another shmproc */
		q = sstash_pop();
		if (q == NULL) q = q_pop_retry(num_semaphores-1);
		m = (cps_message *) &sshmem[q->buf];
		cps_bld_data_ack(m,shm_lpn,s_index+FIRST_SHM_LPN-1,i);
		send_to_shm(s_index,m);
		sstash_push(q);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  get_tcp_msg

  DESCRIPTION:

	  Read a message from the JM or another SHM directly into a shared
	  memory buffer and process it.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int get_tcp_msg(i,x)
int i;
int x;
{
	struct qelem *q;

	int qn;

	int blocked;

	int s;

	if (x == 1)
		s = smap[i].sock;
	else
		s = smap[i].sock2;

/* Get a large buffer from the free queue */

	q = lstash_pop();
	if (q == NULL) q = q_pop_no_wait(0);
	if (q == NULL) return(1);

/* Read the message into the large buffer */

	if (s == 1) s = 0;
	if (read_binary(s,
		(char *) &(shmem[q->buf]), 4*sizeof(long)+MAX_MESSAGE_SIZE) == -1) {
		return(-1);
	}

/* Process the message */

	process_message(q);

/* Successfull return */

	return(0);	
}

/******************************************************************/
/*
	FUNCTION NAME:

	  process_message

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  IMPLICIT OUTPUTS:

  FUNCTION VALUE:

								  */
/******************************************************************/

process_message(q)
struct qelem *q;
{
	cps_message *m;

	int i,j,k;

	int type,src_lpn,dst_lpn,size;
	char *msg;

	char text[80];

	int dead_lpn;

	int vaddr,blockno,offset,xfer_size,xfer_type,xfer_src_lpn,xfer_dst_lpn;

	struct qelem *qp[ MAX_DX*2 ];
	int iqp, nqp;
	cps_message *mp[ MAX_DX*2 ];
	long *x, *y;

/* Determine whether its a large or small buffer */

	i = q->buf;
	if (q->size ==  0)
		m = (cps_message *) &sshmem[i];		/* Small buffers */
	else 
		m = &shmem[i];				/* Large buffers */

/* Get the message information */

	type = (int) m->type;
	size = (int) m->size;
	src_lpn = (int) m->src_lpn;
	dst_lpn = (int) m->dst_lpn;
	msg = m->msg;
#ifdef DIAG
	log_cps_msg(logfp,m);
#endif
/* Handle any messages directed at us */

	if (dst_lpn == shm_lpn) {

		switch (type) {

		case MSG_DATA_ACK :

		cps_get_data_ack(m,&i);
		handle_data_ack(i);
		break;

		case MSG_XFER_REQUEST :

		i = get_inactive_xfer();
		cps_get_xfer_request(m,&xfer_type,&xfer_src_lpn,
			&xfer_dst_lpn,&vaddr,&blockno,&offset,&xfer_size);
		dx[i].type = xfer_type;
		dx[i].src_lpn = xfer_src_lpn;
		dx[i].dst_lpn = xfer_dst_lpn;
		if (xfer_type == 0) {
			dx[i].not_lpn = xfer_src_lpn;
		}
		else if (xfer_type == 1) {
			dx[i].not_lpn = xfer_dst_lpn;
		}
		else if (xfer_type > 0) {	/* Asynchronous sends */
			dx[i].not_lpn = xfer_src_lpn;
		}
		else {				/* Asynchronous gets */
			dx[i].not_lpn = xfer_dst_lpn;
		}
		dx[i].vaddr = vaddr;
		dx[i].blockno = blockno;
		dx[i].offset = offset;
		dx[i].size = xfer_size;
		dx[i].outstand = 0;
		calculate_nleft(i);
		activate_xfer(i);

		break;

		case MSG_USR_PROC_DEAD :

		cps_get_usr_proc_dead(m,&dead_lpn,&msg);

/* If the message did not come from the JM, then tell the JM */

		if (src_lpn == 0)
			usr_proc_dead(dead_lpn,msg,0);
		else
			usr_proc_dead(dead_lpn,msg,-1);

		break;

		case MSG_SHM_SHUTDOWN :	 /* JM told us to shutdown */

		shm_disconnect(0);	/* Do not tell JM we are shutting down */
		shm_shutdown(7);
		break;

		default :
		break;

		}
		lstash_push(q);
		return;
	}

/* Handle messages going to the Job Manager and other SHMs */

	if ((dst_lpn == 0) || 
		((dst_lpn >= FIRST_SHM_LPN) && 
		 (dst_lpn < FIRST_SHM_LPN + MAX_SHM_PROCESSES))) {

/* Send the message over TCP */

		j = dst_lpn - FIRST_SHM_LPN + 1;
		if (j < 0) j = 0;
		if (j == 0)
			send_to_jm(m);
		else
			send_to_shm(j,m);

		lstash_push(q);
		return;
	}

/* Handle messages going to user processes */

	if ((dst_lpn >= FIRST_USER_LPN) && (dst_lpn < FIRST_SHM_LPN)) {

		if (type == MSG_DATA_XFER) {

			j = dst_lpn - FIRST_USER_LPN;

/* Send a data ack if necessary */

			if (pmap[j].local) {	
				j = pmap[src_lpn - FIRST_USER_LPN].shmid;
				cps_get_data_ack(m,&i);
				send_data_ack(i,j);
			}
		}
		send_to_user(dst_lpn,q,m);

	}
/* Handle a broadcast from the JM */
	else if( dst_lpn == -1 )
	{
		nqp = sstash_pop_retry_n( &qp[1], num_local_proc - 1 );
		iqp = 0;
		for( j = 0; j < num_user_proc; j++ )
		{
			if( pmap[j].local )
			{
				if( iqp == 0 )
				{
					qp[iqp] = q;
					mp[iqp] = (cps_message *) &shmem[ qp[iqp] ->buf ];
				}
				else
					mp[iqp] = (cps_message *) &sshmem[ qp[iqp]->buf ];

				mp[iqp]->type = type;
				mp[iqp]->size = size;
				mp[iqp]->src_lpn = src_lpn;
				mp[iqp]->dst_lpn = dst_lpn;
				x = (long *) mp[iqp]->msg;
				y = (long *) m->msg;
				*x = *y;
				x++;y++;
				*x = *y;

				send_to_user( j + FIRST_USER_LPN, qp[iqp], mp[iqp] );
				iqp++;
			}
		}
	}

/* Should never execute these lines */

	else {	

		sprintf(text,"Bad message destination, type=%d, src=%d, dst=%d",
			type,src_lpn,dst_lpn);
		printerr(FACILITY,ERROR_MSG,text,NULL,17);
		shm_shutdown(10);

	}

}

/************************************************************************/
/*

	Function: completed_xfers

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int completed_xfers()
{
	struct qelem *qp[MAX_DX*2],*q;
	int nqp;

	int i,j,n;

	struct data_xfer *x;

	cps_message *m;

/* Try to get as many small buffers as possible */

	nqp = sstash_pop_n(qp,2*num_cdx);
	if (nqp == 0) return 0;
	if ((nqp % 2) != 0) {	/* Got an odd number of buffers */
		sstash_push(qp[nqp-1]);	/* Return one buffer to the free queue */
		nqp--;
		if (nqp == 0) return 0;
	}
	n = nqp/2;

/* Send xfer_complete messages & block_arrived messages */

	for (i=0; i<n; i++) {
		j = 2*i;
		x = cdx;

		m = (cps_message *) &sshmem[qp[j]->buf];
		cps_bld_xfer_complete(m,shm_lpn,x->not_lpn,x->type,x->blockno);
		send_to_user(x->not_lpn,qp[j],m);

		m = (cps_message *) &sshmem[qp[j+1]->buf];
		cps_bld_block_arrived(m,shm_lpn,x->dst_lpn,x->blockno);
		send_to_user(x->dst_lpn,qp[j+1],m);

/* Move record from completed list to inactive list */

		cdx = x->next;
		x->next = idx;
		idx = x;
	}
	num_cdx -= n;
	return 0;
}

/************************************************************************/
/*

	Function: active_xfers

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int active_xfers()
{

	int n,t,p;

	struct qelem *qp[MAX_DX*MAX_OUTSTAND],*q;
	int nqp;

	int i;

	cps_message *m;
	int chunk;

	struct data_xfer *x;

	x = adx;
	n = 0;
	do {
		if ((x->outstand < MAX_OUTSTAND) && (x->nleft > 0)) {
			t = x->nleft > MAX_OUTSTAND ? MAX_OUTSTAND : x->nleft;
			if (t > (MAX_OUTSTAND - x->outstand)) {
				n += MAX_OUTSTAND - x->outstand;
			}
			else {
				n += t;
			}
		}
		x = x->next;
	} while (x != adx);

/* Calculate the number of large buffers we should allocate */

	t =  n + total_outstand;
	p = num_large_buffers - NBUF_SAVED;
	if (t > p) n = p - total_outstand;
	if (n == 0) return 0;

/* n contains the number of large buffers to request */

/* Allocate the buffers */

	nqp = lstash_pop_n(qp,n);
	if (nqp == 0) return 0;		/* Couldn't get any buffers */

/* Sequence through the circular list of active data xfers until we use
   all the buffers */

	for (i=0; i<nqp; i++) {	/* For each message buffer */

		m = &shmem[qp[i]->buf];

/* Find an active data xfer where we can send more data */

		while ((adx->outstand == MAX_OUTSTAND) || (adx->nleft == 0)) 
			adx = adx->next;

/* Send one buffer-full */

		chunk = adx->nleft > 1 ? MAX_CHUNK : adx->size;
		cps_bld_fill_buffer(m,adx->dst_lpn,adx->src_lpn,adx->index,
			adx->type,adx->vaddr,adx->blockno,adx->offset,chunk);
		send_to_user(adx->src_lpn,qp[i],m);

/* Do the bookkeeping */

		adx->outstand++;
		adx->nleft--;
		adx->offset += MAX_CHUNK;
		adx->vaddr += MAX_CHUNK;
		adx->size -= MAX_CHUNK;
		total_outstand++;
	}

	return 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  manage_shared_memory

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

manage_shared_memory() 
{
	fd_set fds;
	int n;

	int i;

	int status;

	struct timeval poll_t;

	struct qelem *q,*qp[32];
	int nqp,iqp;

	struct active_socket *s;
	int sock;

	poll_t.tv_sec = 0;
	poll_t.tv_usec = 0;

/* Initialize set of file descriptors */

	FD_ZERO(&fdset);
	FD_SET(smap[0].sock,&fdset);	/* Job Manager's socket */
	activate_socket(0,smap[0].sock,1);
	FD_SET(passive_socket,&fdset);  /* Socket for accepting connections */

	start:

/* Poll the sockets */

	poll_sock:

	fds = fdset;
	n = select(FD_SETSIZE,&fds,(fd_set *) 0, (fd_set *) 0,&poll_t);
	if (n < 0) {
		if (errno == EINTR) goto poll_sock;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,7);
		shm_shutdown(3);
	}
	if (n > 0) goto read_sock;		/* Data is waiting */

/* Check for freshly killed processes */

	check_dead:

	if (num_fresh_dead > 0) handle_fresh_dead();

/* Check the queue, process all messages found there */

	check_queue:

	nqp = q_pop_no_wait_n(shm_sem,qp,32);	/* Get all pending msgs */
	while (nqp != 0) {
		for (iqp = 0; iqp < nqp; iqp++) {
			process_message(qp[iqp]);		/* Process each message */
		}
		nqp = q_pop_no_wait_n(shm_sem,qp,32);	/* Get new pending msgs */
	}

/* Check for pending data transfers */

	check_pend:

	if (cdx != NULL) completed_xfers();		/* Check for completed xfers */
	if (adx != NULL) active_xfers();		/* Check active xfers */

	if (num_cdx != 0) goto start;
	if (q_head[shm_sem] != -1) goto start;

/* Wait on sockets */

	wait_on_sock:

	fds = fdset;
	n = select(FD_SETSIZE,&fds,(fd_set *) 0, (fd_set *) 0,0);
	if (n < 0) {
		if (errno == EINTR) goto check_dead;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,7);
		shm_shutdown(13);
	}
	if (n == 0) {
#ifdef AIX
		goto check_dead;
#else
		printerr(FACILITY,ERROR_MSG,"Select returned status of zero",
			NULL,25);
		shm_shutdown(8);
#endif
	}

/* Check what's coming in from over the network, put messages onto the queue */

	read_sock:

/* Check for new connection requests */

	if (FD_ISSET(passive_socket,&fds)) {
		hold_signals();
		accept_shm_connect();
		release_signals();
		goto start;
	}

/* Check JM and all SHMs, get maximum of one message from each */

	s = asock;
	do {
		i = asock->index;
		sock = asock->socknum == 1 ? smap[i].sock : smap[i].sock2;

		if (FD_ISSET(sock,&fds)) {
			status = get_tcp_msg(i,asock->socknum);
			if (status == -1) {
				shm_disconnect(i);
				if (i == 0) {
					shm_shutdown(5);
				}
				else {
					goto start;	/* The circular list of active sockets has
				   been modified, so exit the loop */
				}
			}
			else if (status == 1) {
				goto start;
			}
		}
		asock = asock->next;

	} while (asock != s);

	goto start;

}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_sigusr1

  DESCRIPTION:

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

void handle_sigusr1(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{

/* This empty routine is necessary so that when a message arrives on the 
   shm's queue, if the shm is waiting on a select, it will be interrupted
   and the message on the queue will be seen.
*/

}

/******************************************************************/
/*
	FUNCTION NAME:

	  check_usr_procs

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

void check_usr_procs(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	int i,s;

	static int counter = 0;

/* Send a SIGUSR1 every 10 seconds */

	counter++;
	s = 0;
	if (counter == 5) {
		counter = 0;
		s = SIGUSR1;
	}

	for (i=0; i<num_user_proc; i++) {
		if (pmap[i].local && (pmap[i].status == -1)) {
			if (kill(pmap[i].pid,s) == -1) {
				if (errno == ESRCH) {		/* Process does not exist */
					fresh_dead[num_fresh_dead] = i;
					num_fresh_dead++;
				}
			}
		}
	}
	alarm(2);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  main

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

main() {

	char msg[80];
	char host[64];
	struct hostent *hp;

/* Turn on periodic memory allocation reporting */

	hm_mem_interval = 1000000;
	hm_report = -1;

	gethostname(host,64);
	if( ( hp = gethostbyname( host ) ) != NULL )
			strcpy( host, hp->h_name );
	sprintf(msg,"Shared Memory Manager starting on %s",host);
	write(1,msg,strlen(msg)+1);

	sprintf(msg,"Cooperative Processes Software V2.2\n");
	write(1,msg,strlen(msg)+1);

/* Initialize the inactive data xfer list */

	initialize_dx();

/* Redirect stderr to /dev/null */

	freopen("/dev/null","w",stderr);

/* Catch SIGUSR1 */

	sig_establish_handler(SIGUSR1,handle_sigusr1);

/* Ignore SIGPIPE */

	sig_establish_handler(SIGPIPE,SIG_IGN);

/* Print error messages using display_startup_error */

	p_error = send_startup_msg;

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(0,msg,80) < 0) {
	}
	if (strcmp(msg,"suicide") == 0) shm_exit(7);

/* Get lpn, number of processes & number of shared memory managers */

	if (read_counted(0,msg,80) < 0) {
	}
	shm_lpn = atoi(msg);
	if (read_counted(0,msg,80) < 0) {
	}
	num_user_proc = atoi(msg);
	if (read_counted(0,msg,80) < 0) {
	}
	num_shm_proc = atoi(msg);
	if (read_counted(0,msg,80) < 0) {
	}
	num_local_proc = atoi(msg);
	shm_index = shm_lpn - FIRST_SHM_LPN + 1;
	num_semaphores = num_local_proc+3;

/* Open a log file for debugging */
#ifdef DIAG
	sprintf(msg,"shm_%d.log",shm_lpn);
	logfp = fopen(msg,"w");
#endif

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(0,msg,80) < 0) {
	}
	if (strcmp(msg,"suicide") == 0) shm_exit(8);

/* Create the shared memory */

	if (shm_init() == -1) shm_exit(9);

/* Create a passive socket for accepting connection */

	if ((passive_socket = create_passive_socket(0)) == -1) shm_exit(10);

/* Tell the JM we have reached this point without any errors */

	send_startup_msg("ok");

/* Send the shared memory id */

	sprintf(msg,"%d",shmid);
	send_startup_msg(msg);

/* Send the location of this process */

	send_startup_msg(local_hostname);
	send_startup_msg(local_hostaddr);
	sprintf(msg,"%d",assigned_port_number);
	send_startup_msg(msg);

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(0,msg,80) < 0) {
	}
	if (strcmp(msg,"suicide") == 0) shm_exit(11);

/* Initialize shared memory */

	setup_shm();

/* Initialize queues & Create semaphores */

	q_init(num_semaphores);
	if (sem_init(num_semaphores) == -1) shm_exit(12);
	shm_sem = smap[shm_index].semnum;
	sem_exist = -1;

/* Store semaphore ids in the shared memory */

	semids[0] = q_semid;

/* Tell JM that we are ok */

	send_startup_msg("ok");

/* Send semaphore id to the JM */

	sprintf(msg,"%d",q_semid);
	send_startup_msg(msg);

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(0,msg,80) < 0) {
	}
	if (strcmp(msg,"suicide") == 0) shm_exit(13);

/* Change the error logging routine */

	p_error = log_shm_error;

/* Set up an alarm which checks to see what processes are alive every 
   other second */

	sig_establish_handler(SIGALRM,check_usr_procs);

	alarm(2);

/* Direct messages through the shared memory */

	manage_shared_memory();

}
