/* XCPS_SYSMON.C V2.0						7 Jan, 1993

 Author:		Bill Meyer
				Fermi National Accelerator Laboratory - Mail Station 369
				P.O Box 500
				Batavia, IL 60510
				Internet: meyer@fnfaus.fnal.gov

 Copyright (c) 1993 Universities Research Association, Inc.
	   All Rights Reserved


 This material resulted from work developed under a Government Contract and
 is subject to the following license:  The Government retains a paid-up,
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative
 works, perform publicly and display publicly by or for the Government,
 including the right to distribute to other Government contractors.  Neither
 the United States nor the United States Department of Energy, nor any of
 their employees, makes any warrenty, express or implied, or assumes any
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or
 represents that its use would not infringe privately owned rights.
*/

#include "../inc/types.h"

#ifdef AIX
#include <sys/select.h>
#endif
#include <sys/param.h>
#include <sys/socket.h>
#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include <rpcsvc/rstat.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#define DB_GLOBALS 1
#define STATS index->stats_info
#define PREV  index->prev_v

#include "../inc/database.h"
#include "../inc/db_functions.h"
#include "../inc/perr.h"

#include "xcps_sysmon.h"

/* Xm include files */
#include "xm.h"

/* x include files */
#include "xdefines.h"
#include "xstructs.h"
#include "xexterns.h"

#define TITLE "xcps_sysmon 2.0"
#define ICON  "xcps_sysmon"

#define ACTIVE_THRESHOLD 0.10


#define USAGE "\
Usage: xcps_sysmon [options]\n\
\n\
OPTION           X-RESOURCE     DESCRIPTION\n\
for graphic mode:\n\
-f <filename>    *cpuFile       monitor hosts listed in file\n\
-p <prod_system> *prodsys       monitor CPS batch production system\n\
-u <seconds>     *update        polling interval\n\
-r <seconds>     *retry         retry interval\n\
-t <seconds>     *timeout       timeout interval\n\
-m <type>        *monitor       type is either: cpu, page, swap, intr, disk,\n\
                                cntxt, load, pkts, ipkts, opkts, colls, errs\n\
-bs              *boxSize       Graph Box Size\n\
-bfg             *boxForeground Graph Box Foreground Color\n\
-bbg             *boxBackground Graph Box Background Color\n\
-npolls          *npolls        Number of times to poll before exiting\n\
-n               *numRows       Number of rows of Graph Boxes\n\
for non-graphic mode:\n\
-nox                            non-graphic mode option. Must be the first\n\
                                option if others are following.\n\
these options are accept in the non-graphic mode:\n\
-l <filename>                   output file name.\n\
-f <filename>, -u <seconds>, -r <seconds>, -t <timeout>, -m <type>, -npools.\n"

Colormap cmap;

Dimension drawing_width, drawing_height;
Widget CurrentMonitorWidget;
Widget PollDialog, PollTextField;
Widget TimeWidget;
Pixel BackgroundColor;
Pixel ForegroundColor;

Boolean xversion;

int npolls = 0;

Dimension TextHeight;

int total_time = 0;
FILE *log;
long wastetime = 0;


typedef struct
{
	char *boxForeground;
	char *boxBackground;
	int  boxSize;
	int  numRows;
	char *cpuFile;
	char *prodsys;
	int  update;
	int  retry;
	long timeout;
	char *monitor;
	char *domain;
	char *logfile;
	int  npolls;
} AppData;

AppData app_data;

XtResource resources[] =
{
	{
		"boxForeground",
		"BoxForeground",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,boxForeground),
		XmRImmediate,
		"red"
	},
	{
		"boxBackground",
		"BoxColorBackground",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,boxBackground),
		XmRImmediate,
		"MediumBlue"
	},
	{
		"boxSize",
		"BoxSize",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,boxSize),
		XmRImmediate,
		(XtPointer) 125
	},
	{
		"numRows",
		"numRows",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,numRows),
		XmRImmediate,
		(XtPointer) 0
	},
	{
		"cpuFile",
		"CpuFile",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,cpuFile),
		XmRImmediate,
		(XtPointer) NULL
	},
	{
		"update",
		"Update",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,update),
		XmRImmediate,
		(XtPointer) 5
	},
	{
		"retry",
		"Retry",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,retry),
		XmRImmediate,
		(XtPointer) 60
	},
	{
		"timeout",
		"Timeout",
		XmRInt,
		sizeof(long),
		XtOffsetOf(AppData,timeout),
		XmRImmediate,
		(XtPointer) 10
	},
	{
		"prodsys",
		"Prodsys",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,prodsys),
		XmRImmediate,
		(XtPointer) NULL
	},
	{
		"monitor",
		"Monitor",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,monitor),
		XmRImmediate,
		"cpu"
	},
	{
		"domain",
		"Domain",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,domain),
		XmRImmediate,
		"localdomain"
	},
	{
		"logfile",
		"Logfile",
		XmRString,
		sizeof(char*),
		XtOffsetOf(AppData,logfile),
		XmRImmediate,
		(XtPointer) NULL
	},
	{
		"npolls",
		"Npolls",
		XmRInt,
		sizeof(int),
		XtOffsetOf(AppData,npolls),
		XmRImmediate,
		(XtPointer) -1
	},
};

XrmOptionDescRec options[] =
{
	{"-boxSize",			"*boxSize",				XrmoptionSepArg, NULL },
	{"-bs",					"*boxSize",				XrmoptionSepArg, NULL },
	{"-boxForeground",		"*boxForeground",		XrmoptionSepArg, NULL },
	{"-bfg",				"*boxForeground",		XrmoptionSepArg, NULL },
	{"-boxBackground",		"*boxBackground",		XrmoptionSepArg, NULL },
	{"-bbg",				"*boxBackground",		XrmoptionSepArg, NULL },
	{"-numRows",			"*numRows",				XrmoptionSepArg, NULL },
	{"-n",					"*numRows",				XrmoptionSepArg, NULL },
	{"-fn",					"*fontList",			XrmoptionSepArg, NULL },
	{"-font",				"*fontList",			XrmoptionSepArg, NULL },
	{"-f",					"*cpuFile",				XrmoptionSepArg, NULL },
	{"-u",					"*update",				XrmoptionSepArg, NULL },
	{"-r",					"*retry",				XrmoptionSepArg, NULL },
	{"-t",					"*timeout",				XrmoptionSepArg, NULL },
	{"-p",					"*prodsys",				XrmoptionSepArg, NULL },
	{"-monitor",			"*monitor",				XrmoptionSepArg, NULL },
	{"-m",					"*monitor",				XrmoptionSepArg, NULL },
	{"-d",					"*domain",				XrmoptionSepArg, NULL },
	{"-l",					"*logfile",				XrmoptionSepArg, NULL },
	{"-npolls",				"*npolls",				XrmoptionSepArg, NULL },
};

String fallbacks[] =
{
	"xcps_sysmon*fontList: -adobe-helvetica-medium-r-normal--14-140-75-75-p-77-iso8859-1",
	NULL
};

void PollDialogCallback( widget, data )
Widget widget;
int data;
{
	char *string;
	switch( data )
	{
		case 0:
			string = XmTextFieldGetString( PollTextField );
			update = atoi( string );
			if( update < 1 )
				update = 1;
			old_update = update;
			XtFree( string );
			XtUnmanageChild( PollDialog );
			break;
		case 1:
			update = old_update;
			XtUnmanageChild( PollDialog );
			break;
		default:
			break;
	}
}

void PollDialogMask( widget, data, cbs )
Widget widget;
XtPointer data;
XmTextVerifyCallbackStruct *cbs;
{
	int len;

	if( cbs->startPos < cbs->currInsert ) /* backspace */
		return;

	for( len = 0; len < cbs->text->length; len++ )
	{
		if( !isdigit( cbs->text->ptr[ len ] ) )
		{
			cbs->doit = False;
			return;
		}
	}
}

void XcpsmonPollDialog( widget )
Widget widget;
{
	Widget dialog, action_area, control_area, separator;
	Widget label, text_field;
	XmString xm_string;
	char string[16];
	static ActionAreaItem action_items[] =
	{
		{ "OK",	 PollDialogCallback, NULL },
		{ "Cancel", PollDialogCallback, NULL }
	};

	if( PollDialog )
	{
		sprintf( string, "%d", update );
		XmTextFieldSetString( PollTextField, string );
		XtManageChild( PollDialog );
		return;
	}
	dialog = XmCreateFormDialog( widget, "dialog", NULL, 0 );
	PollDialog = dialog;

	action_items[ 0 ].data = (caddr_t)0;
	action_items[ 1 ].data = (caddr_t)1;
	
	action_area = CreateActionArea( dialog, action_items,
		XtNumber( action_items ), 0 );

	separator = CreateActionSeparator( dialog, action_area );

	control_area = CreateControlArea( dialog, separator );

	xm_string = XmStringCreateSimple( "Polling Interval:" );
	label = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass,	 control_area,
		XmNlabelString,		 xm_string,
		XmNtopAttachment,	   XmATTACH_FORM,
		XmNleftAttachment,	  XmATTACH_FORM,
		NULL );
	XmStringFree( xm_string );

	text_field = XtVaCreateManagedWidget( "text-field",
		xmTextFieldWidgetClass, control_area,
		XmNtopAttachment,	   XmATTACH_WIDGET,
		XmNtopWidget,		   label,
		XmNleftAttachment,	  XmATTACH_FORM,
		XmNwidth,			   50,
		NULL );
	sprintf( string, "%d", update );
	XmTextFieldSetString( text_field, string );
	XtAddCallback( text_field, XmNactivateCallback, PollDialogCallback, 
		NULL );
	XtAddCallback( text_field, XmNmodifyVerifyCallback, PollDialogMask, 
		NULL );
	PollTextField = text_field;

	xm_string = XmStringCreateSimple( "seconds" );
	label = XtVaCreateManagedWidget( "label",
		xmLabelWidgetClass,	 control_area,
		XmNlabelString,		 xm_string,
		XmNbottomAttachment,	XmATTACH_OPPOSITE_WIDGET,
		XmNbottomWidget,		text_field,
		XmNbottomOffset,		5,
		XmNleftAttachment,	  XmATTACH_WIDGET,
		XmNleftWidget,		  text_field,
		XmNleftOffset,		  5,
		NULL );
	XmStringFree( xm_string );

	XtManageChild( dialog );
}

void XcpssysmonInit(argc,argv)
int *argc;
char **argv[];
{
	AppShell = XtVaAppInitialize( &context, "XCps_sysmon",
		options, XtNumber( options ),
		argc, argv, fallbacks, NULL );
	XtVaSetValues( AppShell, XtNiconName, ICON, NULL );

	XtGetApplicationResources( AppShell,
		&app_data, resources, XtNumber( resources ), NULL, 0 );

}

XcpssysmonCommandLine()
{
	if( strcmp( app_data.domain, "localdomain" ) )
	{
		char env_string[128];
		sprintf( env_string, "%s=%s", "BATCH_DOMAIN", app_data.domain );
		putenv( env_string );
	}

	if( app_data.prodsys != NULL )
	{
		input_option = FROM_PM;
		get_cpulist_from_pm(app_data.prodsys);
	}
	else if( app_data.cpuFile != NULL )
	{
		input_option = FROM_FILE;
		get_cpulist_from_file(app_data.cpuFile);
	}

	if( !strcmp( app_data.monitor, "cpu" ) )
		option = P_CPU;
	else if( !strcmp( app_data.monitor, "page" ) )
		option = P_PAGE;
	else if( !strcmp( app_data.monitor, "swap" ) )
		option = P_SWAP;
	else if( !strcmp( app_data.monitor, "intr" ) )
		option = P_INTR;
	else if( !strcmp( app_data.monitor, "disk" ) )
		option = P_DISK;
	else if( !strcmp( app_data.monitor, "cntxt" ) )
		option = P_CNTXT;
	else if( !strcmp( app_data.monitor, "load" ) )
		option = P_LOAD;
	else if( !strcmp( app_data.monitor, "pkts" ) )
		option = P_PKTS;
	else if( !strcmp( app_data.monitor, "ipkts" ) )
		option = P_IPKTS;
	else if( !strcmp( app_data.monitor, "opkts" ) )
		option = P_OPKTS;
	else if( !strcmp( app_data.monitor, "colls" ) )
		option = P_COLLS;
	else if( !strcmp( app_data.monitor, "errs" ) )
		option = P_ERRS;

	update = app_data.update;
	if( update < 1 )
		update = 1;
	old_update = update;

}

main (int argc, char **argv)
{
	struct CPULIST *index;
	char buffer[255];
	extern char *optarg;
	extern int optind;
	int i, n;
	XColor xcolor;
    time_t starttime;

	if( ( argc == 2) && !strcmp( argv[1], "-help" ) )
	{
		printf(USAGE);
		exit(0);
	}

	if( argc >= 2 && !strcmp( argv[1], "-nox" ) )
	{
		xversion = False;
		app_data.timeout = 10;
		app_data.retry = 60;
		app_data.update = 5;
		app_data.npolls = -1;
		if (argc > 2) 
		{
			for (i = 2; i < argc; i++) 
			{
			    if (!strcmp(argv[i], "-f")) {
		            input_option = FROM_FILE;
                    app_data.cpuFile = argv[++i];
					get_cpulist_from_file(app_data.cpuFile);
                }
			    else if (!strcmp(argv[i], "-npolls"))
						app_data.npolls = atoi(argv[++i]);
                else if (!strcmp(argv[i], "-t"))
						app_data.timeout = atol(argv[++i]);
                else if (!strcmp(argv[i], "-r"))
						app_data.retry = atoi(argv[++i]);
                else if (!strcmp(argv[i], "-u"))
						update = app_data.update = atoi(argv[++i]);
                else if (!strcmp(argv[i], "-l"))
						app_data.logfile = argv[++i];
			    else if (!strcmp(argv[i], "-m"))
						app_data.monitor = argv[++i];
			}
			fprintf(stderr, " -- Parameters --\ncpufile: %s\ttimeout: %d\nretry: %d\tupdate: %d\nnpolls: %d\n", app_data.cpuFile, app_data.timeout, app_data.retry, app_data.update, app_data.npolls);
        }
	}
	else
		xversion = True;
    
	if( xversion )
	{
		XcpssysmonInit( &argc, argv );	
		XcpssysmonCommandLine();
		if( argc > 1 )
		{
			printf(USAGE);
			exit(0);
		}
	}

	if(input_option == UNDEFINED)
	{
		input_option = FROM_FILE;
		get_cpulist_from_file("default");
	}

	if(!total_cpus)
	{
		fprintf(stderr, "No cpus found.\n");
		exit(-1);
	}

	establish_remote_connections();
	signal(SIGINT, sigint_handler);

	time( &starttime );

	if ( app_data.logfile != NULL) {
		if ((log = fopen( app_data.logfile, "w+" )) == NULL)
	    {
		    perror( app_data.logfile );
		    fprintf(stderr, "could not open log file \"%s\"\n", app_data.logfile);
			exit(-1);
	    }
    }
	else  
        if (!xversion)	
		    log = stdout;

    if (log != NULL) {
        fprintf( log, "# xcps_sysmon log file \"%s\"\n", 
    			app_data.logfile );
        fprintf( log, "# Created %s", ctime(&starttime));
        fprintf( log, "# Polling Interval %d seconds\n", update );
        fflush( log );
	}	 

	if( xversion )
	{
		MainWindow = CreatemainWindow(AppShell);

		XtAddCallback(MainWindow, XmNdestroyCallback, sigint_handler, 0);
		XtManageChild(MainWindow);

/* Show it */
	
		XtPopup(AppShell, XtGrabNone);

/* Initialize maximum values and menu options */

		processOptions(MainWindow, menuOptions[option], NULL);

/*  Get the first set of data and start the update timer */

		get_data(0);

/*  Enter the X Window message loop */

		XtAppMainLoop (context);
	}
	else
	{
		while(1)
		{
			if (wastetime < update) 
			    sleep(update - wastetime);
			wastetime = 0;
			get_data(1);
		}
	}
}

Widget CreateMenus( mainWindow )
Widget mainWindow;
{
	int index;
	Widget MenuBar, widget;
	XmString xm_string;
	Widget file_menu, monitor_menu, options_menu;

	MenuBar = XmCreateMenuBar( mainWindow, "MenuBar", NULL, 0 );
	XtVaSetValues( MenuBar,
		XmNleftAttachment,  XmATTACH_FORM,
		XmNrightAttachment, XmATTACH_FORM,
		NULL );

	file_menu = XmCreatePulldownMenu( MenuBar, "file_menu", NULL, 0 );
	monitor_menu = XmCreatePulldownMenu( MenuBar, "monitor_menu", NULL, 0 );
	options_menu = XmCreatePulldownMenu( MenuBar, "options_menu", NULL, 0 );

	xm_string = XmStringCreateSimple( "File" );
	XtVaCreateManagedWidget( "File",
		xmCascadeButtonWidgetClass, MenuBar,
		XmNlabelString,			 xm_string,
		XmNmnemonic,				'F',
		XmNsubMenuId,			   file_menu,
		NULL );
	XmStringFree( xm_string );

	xm_string = XmStringCreateSimple( "Ctrl-H" );
	widget = XtVaCreateManagedWidget( "Help",
		xmPushButtonWidgetClass, file_menu,
		XmNmnemonic,             'H',
		XmNaccelerator,          "Ctrl<Key>H",
		XmNacceleratorText,      xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget,
		XmNactivateCallback, processOptions, menuOptions[12] );

	xm_string = XmStringCreateSimple( "Ctrl-X" );
	widget = XtVaCreateManagedWidget( "Exit",
		xmPushButtonWidgetClass, file_menu,
		XmNmnemonic,             'E',
		XmNaccelerator,          "Ctrl<Key>X",
		XmNacceleratorText,      xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget,
		XmNactivateCallback, processOptions, menuOptions[13] );

	xm_string = XmStringCreateSimple( "Monitor" );
	XtVaCreateManagedWidget( "Monitor",
		xmCascadeButtonWidgetClass, MenuBar,
		XmNlabelString,			 xm_string,
		XmNmnemonic,				'M',
		XmNsubMenuId,			   monitor_menu,
		NULL );
	XmStringFree( xm_string );

	for(index = 0; index < 12; index++)
	{
		widget = XtVaCreateManagedWidget( menuOptions[index],
			xmToggleButtonWidgetClass, monitor_menu,
			XmNindicatorType,		  XmONE_OF_MANY,
			NULL );
		XtAddCallback( widget,
			XmNvalueChangedCallback, processOptions, menuOptions[index] );
		if( index == 0 )
		{
			XtVaSetValues( widget, XmNset, True, NULL );
			CurrentMonitorWidget = widget;
		}
	}

	xm_string = XmStringCreateSimple( "Options" );
	XtVaCreateManagedWidget( "Options",
		xmCascadeButtonWidgetClass, MenuBar,
		XmNlabelString,			 xm_string,
		XmNmnemonic,				'O',
		XmNsubMenuId,			   options_menu,
		NULL );
	XmStringFree( xm_string );

	xm_string = XmStringCreateSimple( "Ctrl-P" );
	widget = XtVaCreateManagedWidget( "Polling Interval",
		xmPushButtonWidgetClass, options_menu,
		XmNmnemonic,             'P',
		XmNaccelerator,          "Ctrl<Key>P",
		XmNacceleratorText,      xm_string,
		NULL );
	XmStringFree( xm_string );
	XtAddCallback( widget,
		XmNactivateCallback, XcpsmonPollDialog, NULL );

	TimeWidget = XtVaCreateManagedWidget( "00:00:00",
		xmCascadeButtonWidgetClass, MenuBar,
		NULL );

	XtVaSetValues( MenuBar,
		XmNmenuHelpWidget, TimeWidget,
		NULL );

	XtManageChild( MenuBar );

	return( MenuBar );
}

DrawAxisLabels( widget, gc )
Widget widget;
GC	 gc;
{
	char string[16];
	if(option == P_LOAD)
		sprintf(string, "%-3.2f", maxvalue[option]);
	else
		sprintf(string, "%-3.0f", maxvalue[option]);

	XSetForeground( display, gc, BackgroundColor );

	XDrawString( display, XtWindow(widget), gc,
		2, TextHeight, string, strlen(string) );
	XDrawString( display, XtWindow(widget), gc,
		2, drawing_height-2, "0", 1 );
	/*
	XDrawLine( display, XtWindow(widget), gc,
		2, drawing_height/2,
		2 + ( drawing_width / 10 ), drawing_height/2 );
	XDrawLine( display, XtWindow(widget), gc,
		2, 2,
		2 + ( drawing_width / 5 ), 2 );
	*/
}

/******************************************************************/
/*
	FUNCTION NAME:

	CreatemainWindow()

  DESCRIPTION:

	Creates the main application window, including all widgets.

  FORMAL PARAMETERS:

	Widget parent - Pointer to the calling widget

																  */
/******************************************************************/

Widget CreatemainWindow(Widget parent)
{
	char hostnm[MAXHOSTNAMELEN];
	Boolean ok;
	struct CPULIST *index;
	int x, y, m, n;
	XmString xmstr;
	Widget mainWindow;
	Widget main_bulletinboard;
	Widget scrolledWindow;
	Widget bulletinboard_frame;
	Widget titleblock_frame[6];
	Widget menuBar;
	Widget menuShell;
	Widget pulldownMenu;
	Widget pushButton;
	Widget cascadeButton;
	Widget rowcol;
	char buffer[255];
	int num_rowcol_elements = 0;

	char            firsttime = True;
	Font            font_id;

/* Create a Main Window widget */

	n = 0;
	mainWindow = XtCreateWidget("mainWindow", xmFormWidgetClass,
		parent, args, n);

	display = XtDisplay( mainWindow );

	screen = XtScreen( mainWindow );

	menuBar = CreateMenus( mainWindow );

/* Create a Scrolled Window below the menu bar */

	n = 0;
	XtSetArg(args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
	XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
	XtSetArg(args[n], XmNtopWidget, menuBar); n++;
	XtSetArg(args[n], XmNtopOffset, 0); n++;
	XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNleftOffset, 0); n++;
	XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNrightOffset, 0); n++;
	XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
	XtSetArg(args[n], XmNbottomOffset, 0); n++;
	scrolledWindow = XtCreateWidget("scrolledWindow",
		xmScrolledWindowWidgetClass, mainWindow, args, n);
	XtManageChild(scrolledWindow);

	XtVaGetValues( scrolledWindow, XmNbackground, &BackgroundColor, NULL );
	XtVaGetValues( scrolledWindow, XmNforeground, &ForegroundColor, NULL );

/* Place the Bulletin Board inside the Frame */

	n = 0;
	main_bulletinboard = XtCreateWidget("main_bulletinboard",
		xmFormWidgetClass, scrolledWindow, args, n);
	XtManageChild(main_bulletinboard);

/* Create the Time titleblock */

	if(input_option == FROM_PM)
	{
		char title[ 128 ];
		sprintf( title, "%s - %s:%s - %s", TITLE,
			p->jm_cpu->cpu_name, p->ps_name, menuOptions[option] );
		XtVaSetValues( AppShell, XtNtitle, title, NULL );
	}
	else
	{
		char title[ 128 ];
		sprintf( title, "%s - %s", TITLE, menuOptions[option] );
		XtVaSetValues( AppShell, XtNtitle, title, NULL );
	}

/* Create a RowColumn widget where each individual node will be placed */

	n = 0;
	XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
	XtSetArg(args[n], XmNspacing, 5); n++;
	XtSetArg(args[n], XmNpacking, XmPACK_COLUMN); n++;
	XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM ); n++;
	XtSetArg(args[n], XmNleftOffset, 5 ); n++;
	XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM ); n++;
	XtSetArg(args[n], XmNrightOffset, 5 ); n++;
	XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM ); n++;
	XtSetArg(args[n], XmNbottomOffset, 5 ); n++;
	XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET ); n++;
	XtSetArg(args[n], XmNtopWidget, titleblock[UPDATE] ); n++;
	XtSetArg(args[n], XmNtopOffset, 5 ); n++;
	XtSetArg(args[n], XmNnumColumns, app_data.numRows ); n++;
	rowcol = XtCreateWidget("rowcol", xmRowColumnWidgetClass,
		main_bulletinboard, args, n);
	XtManageChild(rowcol);

/* Add each node to the RowColumn widget */

	index = head;
	while(index != NULL) {

/* Use a bulletin board as the foundation */

		n = 0;
		index->bulletinboard = XtCreateWidget("bulletinboard",
			xmFormWidgetClass, rowcol, args, n);
		XtManageChild(index->bulletinboard);

/* Place a Frame for the label */

		n = 0;
		XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
		XtSetArg(args[n], XmNtopOffset, 0); n++;
		XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
		XtSetArg(args[n], XmNleftOffset, 0); n++;
		XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
		XtSetArg(args[n], XmNrightOffset, 0); n++;
		index->label_frame = XtCreateWidget("label_frame",
			xmFrameWidgetClass, index->bulletinboard, args, n);
		XtManageChild(index->label_frame);

/* Place a label inside the frame that contains the node name */

		n = 0;
		XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++;
		XtSetArg(args[n], XmNlabelString, (xmstr = XmStringCreateLtoR(
			index->name, XmSTRING_DEFAULT_CHARSET))); n++;
		index->label = XtCreateWidget("label",
			xmLabelWidgetClass, index->label_frame, args, n);
		XmStringFree(xmstr);
		XtManageChild(index->label);

		if( firsttime )
		{
			XmFontList      font_list;
			XmFontContext   font_context;
			XmStringCharSet font_charset;
			XFontStruct     *font;
			XmString        xm_string;

			font_id = -1;
			XtVaGetValues( index->label, XmNfontList, &font_list, NULL );

			xm_string = XmStringCreateLtoR( "0", XmSTRING_DEFAULT_CHARSET );
			TextHeight = XmStringHeight( font_list, xm_string );
			XmStringFree( xm_string );

			if( XmFontListInitFontContext( &font_context, font_list ) )
			{
				if( XmFontListGetNextFont( font_context, &font_charset,
					&font ) )
				{
					font_id = (Font)font->fid;
					XmFontListFreeFontContext( font_context );
					free( font_charset );
				}
			}
			firsttime = False;
		}

/* Create a Frame for the Drawing Area */

		n = 0;
		XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
		XtSetArg(args[n], XmNtopWidget, index->label_frame); n++;
		XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
		XtSetArg(args[n], XmNleftOffset, 0); n++;
		XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
		XtSetArg(args[n], XmNrightOffset, 0); n++;
		XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
		XtSetArg(args[n], XmNbottomOffset, 0); n++;
		index->drawingarea_frame = XtCreateWidget("drawingarea_frame",
			xmFrameWidgetClass, index->bulletinboard, args, n);
		XtManageChild(index->drawingarea_frame);

/* Create the Drawing Area */

		n = 0;
		XtSetArg(args[n], XmNwidth, app_data.boxSize); n++;
		XtSetArg(args[n], XmNheight, app_data.boxSize); n++;
		index->drawingarea = XtCreateWidget("box",
			xmDrawingAreaWidgetClass, index->drawingarea_frame, args, n);
		XtManageChild(index->drawingarea);

/* Add the EXPOSE callback so that when a window which was covering ours is
   removed, the screen gets repainted immediately, w/o having to wait for
   the update timer to expire */

		XtAddCallback(index->drawingarea, XmNexposeCallback, redraw, NULL);

/* Create a pixmap for each Drawing Area */

		XtVaGetValues( index->drawingarea,
			XmNwidth, &drawing_width,
			XmNheight, &drawing_height,
			NULL );

		if(!(index->pixmap = XCreatePixmap(XtDisplay(index->drawingarea),
			/*
			RootWindowOfScreen(XtScreen(index->drawingarea)),
			*/
			RootWindowOfScreen(screen),
			drawing_width+5, drawing_height,
			/*
			DefaultDepthOfScreen(XtScreen(index->drawingarea)))))
			*/
			DefaultDepthOfScreen(screen))))
				XtError("Can't Create Pixmap");

		if(!(index->gc = XCreateGC(display, index->pixmap, NULL, 0)))
			XtError("Can't create GC");

		if( font_id != -1 )
			XSetFont( display, index->gc, font_id );

/* Do the next node */

		index = index->next;
		num_rowcol_elements++;
	}

	return(mainWindow);
}

/******************************************************************/
/*
	FUNCTION NAME:

	get_data()

  DESCRIPTION:

	Connects to each node, gets information, and calls display_data()

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void get_data(int disp)
{
	struct CPULIST *index;
	struct timeval time_out;
    time_t onetime, twotime;
	static long retryconnect = 0;
	static long printtime = 0;    

	time_out.tv_sec = app_data.timeout;
	time_out.tv_usec = 0;

    total_update += update;
	retryconnect += update;
	printtime += update;

	index = head;
	while(index != NULL)
	{
		/* try reaching dead nodes every -app_data.retry- seconds */
		if( ( total_update >= app_data.retry ) &&
			( index->status != RPC_SUCCESS ) ) {

            if (printtime >= 3600) {
                time(&onetime);
                fprintf(stderr, "%s", ctime(&onetime));
				fflush(stderr);
				printtime = 0;
            }
				
            if (index->connect == FALSE ) { 
			 if (retryconnect >= 4*app_data.retry) {
				 retryconnect = 0;
			     time( &onetime );
			     fprintf(stderr, "Attempting reconnection with %s...", 
				   		 index->name );
			     fflush(stderr);
                 if((index->client = clntudp_create(&index->server_addr, 
				   RSTATPROG, RSTATVERS_TIME, time_out, &index->sock)) == NULL) 
                 {
                     fprintf(stderr, "Failed\n" );
                     fprintf(stderr, "sysmon: Cannot establish connection%s",
					        clnt_spcreateerror("")); 
				     fflush(stderr);			 
		         }
		         else { 
		        /* Established connection, authenticate it */
		             fprintf(stderr, "OK\n");
					 fflush(stderr);
			         index->connect = TRUE;
			         index->client->cl_auth = authunix_create_default();
			         index->prev_v.cp_time[CP_USER] = 0.0;
			         index->prev_v.cp_time[CP_NICE] = 0.0;
			         index->prev_v.cp_time[CP_SYS] = 0.0;
			         index->prev_v.cp_time[CP_IDLE] = 0.0;

			         index->status = clnt_call(index->client, RSTATPROC_STATS,
				                               xdr_void, 0, xdr_statstime, 
					   						  &index->stats_info, time_out);

			         if( index->status == RPC_SUCCESS )
			         {
				         if( xversion )
				   	        XtVaSetValues( index->label, XmNbackground, 
				   						  BackgroundColor, NULL );
			         }
			         else 
				         fprintf(stderr, "client call failed.\n" );
			         fflush( stderr );
		         }
				 time( &twotime);
          		 wastetime = twotime - onetime;
              }
		    }	
			else {
			    fprintf(stderr, "Attempting reconnection with %s...", 
						index->name );
			    fflush(stderr);
			
			    index->status = clnt_call(index->client, RSTATPROC_STATS,
			                              xdr_void, 0, xdr_statstime, 
										  &index->stats_info, time_out);

			    if( index->status == RPC_SUCCESS )
			    {
			        index->prev_v.cp_time[CP_USER] = 0.0;
			        index->prev_v.cp_time[CP_NICE] = 0.0;
			        index->prev_v.cp_time[CP_SYS] = 0.0;
			        index->prev_v.cp_time[CP_IDLE] = 0.0;

			        fprintf(stderr, "OK\n");
			        if( xversion )
				        XtVaSetValues( index->label, XmNbackground, 
									  BackgroundColor, NULL );
			    }
			    else
				    fprintf(stderr, "Failed\n" );
			    fflush(stderr);
            }
        }			 
		/* try reaching only nodes that have responded previously */
		else {
			if( index->status == RPC_SUCCESS )
			{
				index->status = clnt_call(index->client, RSTATPROC_STATS,
					xdr_void, 0, xdr_statstime, &index->stats_info, time_out);

				if( index->status != RPC_SUCCESS )
				{
					fprintf(stderr, "Lost connection with %s\n", index->name );
					fflush( stderr );

					if( xversion )
						XtVaSetValues( index->label, XmNbackground,
							          (Pixel) SetColor( display, 
									   DefaultScreen( display ),
								       app_data.boxForeground, index->gc ),
									   NULL );
				}
			}
        }		
		index = index->next;
	}
	if(disp)
		display_data();

	if( total_update >= app_data.retry )
		total_update = 0;

	if( npolls == app_data.npolls )
		exit( 0 );

/* Refresh the Update Timer */

	if( xversion )
	    update_id = XtAppAddTimeOut(context, update*1000L, get_data, 
									(caddr_t)1);
}

/******************************************************************/
/*
	FUNCTION NAME:

	display_data()

  DESCRIPTION:

	Displays data for each node.
	Converts RSTATD data into displayable data,
		scales it, scrolls the window, and draws it onto the screen(),
		then updates the clock.

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void display_data(void)
{
	long t;
	int n;
	long temp, total;
	float resize = 0.0;
	char buffer[255];
	struct CPULIST *index;
	struct tm *curt;

	npolls++;

	total_time += update;
	index = head;
	while(index != NULL)
	{
		if(index->status == RPC_SUCCESS)
		{

/* Update the previous data */

			for(n = 0; n < MAXITEMS; n++)
				index->pdata[n] = index->cdata[n];

/* Update the new data */

			/* cpu utilization */
			if( log != NULL )
			{
				temp = STATS.cp_time[CP_USER] - PREV.cp_time[CP_USER] +
					STATS.cp_time[CP_NICE] - PREV.cp_time[CP_NICE] +
					STATS.cp_time[CP_SYS]  - PREV.cp_time[CP_SYS];

				total = temp + STATS.cp_time[CP_IDLE] - PREV.cp_time[CP_IDLE];

                if (total == 0 || (total < 0 && temp > 0)) 
					index->cdata[P_CPU] = 0.0;
                else
                    index->cdata[P_CPU]=temp*100.0/total;

				/* changed - lauro.
				index->cdata[P_CPU]=(total > 0 ? (temp*100.0/total) : 0.0);
				index->cdata[P_CPU]=(temp > 0 ? (temp*100.0/total) : 0.0);
				*/

				/*
				fprintf(log,"total=%d temp=%d\n", total, temp);
				*/
				fprintf(log,"%s\t%2.0f\n",index->name,index->cdata[P_CPU]);
				fflush(log);
			}

			switch(option)
			{
			case P_CPU:

			if( log == NULL )
			{
				temp = STATS.cp_time[CP_USER] - PREV.cp_time[CP_USER] +
					STATS.cp_time[CP_NICE] - PREV.cp_time[CP_NICE] +
					STATS.cp_time[CP_SYS]  - PREV.cp_time[CP_SYS];
				total = temp + STATS.cp_time[CP_IDLE] - PREV.cp_time[CP_IDLE];
			}

			/* changed - lauro.
			index->cdata[P_CPU]=(total > 0 ? (temp*100.0/total) : 0.0);
			index->cdata[P_CPU]=(temp > 0 ? (temp*100.0/total) : 0.0);
			*/
			index->cdata[P_CPU]=temp*100.0/total;
			break;

			case P_PAGE:

			index->cdata[P_PAGE] = STATS.v_pgpgin  - PREV.v_pgpgin +
				STATS.v_pgpgout - PREV.v_pgpgout;
			if(index->cdata[P_PAGE] > maxvalue[P_PAGE])
				resize = max(index->cdata[P_PAGE], resize);
			break;

			case P_SWAP:

			index->cdata[P_SWAP] = STATS.v_pswpin  - PREV.v_pswpin +
				STATS.v_pswpout - PREV.v_pswpout;
			if(index->cdata[P_SWAP] > maxvalue[P_SWAP])
				resize = max(index->cdata[P_SWAP], resize);
			break;

			case P_INTR:

			index->cdata[P_INTR] = STATS.v_intr - PREV.v_intr;
			if(index->cdata[P_INTR] > maxvalue[P_INTR])
				resize = max(index->cdata[P_INTR], resize);
			break;

			case P_DISK:

			index->cdata[P_DISK] = STATS.dk_xfer[0] - PREV.dk_xfer[0] +
				STATS.dk_xfer[1] - PREV.dk_xfer[1] +
				STATS.dk_xfer[2] - PREV.dk_xfer[2] +
				STATS.dk_xfer[3] - PREV.dk_xfer[3];
			if(index->cdata[P_DISK] > maxvalue[P_DISK])
				resize = max(index->cdata[P_DISK], resize);
			break;

			case P_CNTXT:

			index->cdata[P_CNTXT] = STATS.v_swtch - PREV.v_swtch;
			if(index->cdata[P_CNTXT] > maxvalue[P_CNTXT])
				resize = max(index->cdata[P_CNTXT], resize);
			break;

			case P_LOAD:

			index->cdata[P_LOAD] = (float)(STATS.avenrun[0]/256.0);
			if(index->cdata[P_LOAD] > maxvalue[P_LOAD])
				resize = max(index->cdata[P_LOAD], resize);
			break;

			case P_PKTS:

			index->cdata[P_PKTS]= STATS.if_ipackets - PREV.if_ipackets +
				STATS.if_opackets - PREV.if_opackets;
			if(index->cdata[P_PKTS] > maxvalue[P_PKTS])
				resize = max(index->cdata[P_PKTS], resize);
			break;
			
			case P_IPKTS:

			index->cdata[P_IPKTS]= STATS.if_ipackets - PREV.if_ipackets;
			if(index->cdata[P_IPKTS] > maxvalue[P_IPKTS])
				resize = max(index->cdata[P_IPKTS], resize);
			break;

			case P_OPKTS:

			index->cdata[P_OPKTS]= STATS.if_opackets - PREV.if_opackets;
			if(index->cdata[P_OPKTS] > maxvalue[P_OPKTS])
				resize = max(index->cdata[P_OPKTS], resize);
			break;

			case P_COLLS:

			index->cdata[P_COLLS] = STATS.if_collisions -
				PREV.if_collisions;
			if(index->cdata[P_COLLS] > maxvalue[P_COLLS])
				resize = max(index->cdata[P_COLLS], resize);
			break;

			case P_ERRS:

			index->cdata[P_ERRS] = STATS.if_ierrors - PREV.if_ierrors +
				STATS.if_oerrors - PREV.if_oerrors;
			if(index->cdata[P_ERRS] > maxvalue[P_ERRS])
				resize = max(index->cdata[P_ERRS], resize);
			break;
			}

			for(temp = 0; temp < MAXITEMS; temp++)
			{
				if(index->cdata[temp] < 0.0)
					index->cdata[temp] = 0.0;
				if(temp != P_LOAD && temp != P_CPU)
					index->cdata[temp] /= update;
			}

			if( xversion )
			{

#define scale(a) \
	((int)((a)*((drawing_height-4)/maxvalue[option])))

/*
 * Scroll the pixmap to the left
 * The pixmap is the source and the destination,
 * with about 5 pixels inbetween
 */
				XCopyArea(display, index->pixmap, index->pixmap, index->gc,
					0, 0, drawing_width+5, drawing_height, -5, 0);

/* Set the graphics context to display the peak value */

				SetColor( display, DefaultScreen( display ),
					app_data.boxForeground, index->gc );

/* Draw the line to the pixmap */

				XDrawLine(display, index->pixmap, index->gc,
					drawing_width-6,
					drawing_height-2-scale(index->pdata[option]),
					drawing_width-1,
					drawing_height-2-scale(index->cdata[option]));

				{
				XPoint points[4];
				points[0].x = drawing_width-6;
				points[0].y = drawing_height;
				points[1].x = drawing_width-6;
				points[1].y = drawing_height-2-scale(index->pdata[option]);

				points[2].x = drawing_width;
				points[2].y = drawing_height-2-scale(index->cdata[option]);
				points[3].x = drawing_width;
				points[3].y = drawing_height;

				XFillPolygon( display, index->pixmap, index->gc,
					points, 4, Convex, CoordModeOrigin );
				}

/* Copy the pixmap to the display */

				XCopyArea(display, index->pixmap, XtWindow(index->drawingarea),
					index->gc, 0, 0, drawing_width, drawing_height, 0, 0);

				DrawAxisLabels( index->drawingarea, index->gc );

			}

		}
		else
		{
			if( log != NULL )
			{
				fprintf(log,"%s\t0\n",index->name);
				fflush(log);
			}

			if( xversion )
			{
				XCopyArea( display, index->pixmap, index->pixmap, index->gc,
					0, 0, drawing_width+5, drawing_height, -5, 0 );
				SetColor( display, DefaultScreen( display ),
					app_data.boxForeground, index->gc );
				XCopyArea(display, index->pixmap, XtWindow(index->drawingarea),
					index->gc, 0, 0, drawing_width, drawing_height, 0, 0);
				DrawAxisLabels( index->drawingarea, index->gc );
			}

			for( n = 0; n < MAXITEMS; n++)
				index->cdata[n] = index->pdata[n];
		}

/* Copy the new RPC data to the old RPC data */

		index->prev_v = index->stats_info;  

/* Do the next node */

		index = index->next;
	}

/* A new value went above the maximum */

	if(resize)
	{

/* Keep doubling in size until max is greater than the value */

		while(maxvalue[option] < resize)
			maxvalue[option] += maxvalue[option];

		resize = 0.0;
		reset();
	}

/* Update the clock display */

	if( xversion )
	{
		t = time((long *)0);
		curt = localtime(&t);
		sprintf(buffer, "%.2d:%.2d:%.2d",
			curt->tm_hour, curt->tm_min, curt->tm_sec);
		XtVaSetValues( TimeWidget,
			XtVaTypedArg, XmNlabelString,
			XmRString, buffer, strlen(buffer), NULL );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	void redraw(Widget, XtPointer, XmDrawingAreaCallbackStruct *)

  DESCRIPTION:

	Redraws the screen by copying the pixmaps
	onto the drawingareas.

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void redraw(Widget parent, XtPointer unused, XmDrawingAreaCallbackStruct *cbs)
{
	struct CPULIST *index;

/* There are 3 reasons why we might be here, we only want XmCR_EXPOSE */

	if(cbs->reason == XmCR_EXPOSE)
	{
		index = head;
		while(index != NULL)
		{
			if(index->drawingarea == parent)
			{
				XCopyArea(display, index->pixmap,
					XtWindow(index->drawingarea),
					index->gc, 0, 0,
					drawing_width, drawing_height, 0, 0);
				DrawAxisLabels( index->drawingarea, index->gc );
				return;
			}
			index = index->next;
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	void reset(void)

  DESCRIPTION:

	Resets certain titleblocks, and clears the
	pixmap for each node.

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void reset(void)
{
	struct CPULIST *index;
	char buffer[255];
	int n;

	if(input_option == FROM_PM)
	{
		char title[ 128 ];
		sprintf( title, "%s -  %s:%s - %s", TITLE,
			p->jm_cpu->cpu_name, p->ps_name, menuOptions[option] );
		XtVaSetValues( AppShell, XtNtitle, title, NULL );
	}
	else
	{
		char title[ 128 ];
		sprintf( title, "%s - %s", TITLE, menuOptions[option] );
		XtVaSetValues( AppShell, XtNtitle, title, NULL );
	}

	index = head;
	while(index != NULL) {

/* The background color */

		SetColor( display, DefaultScreen( display ),
			app_data.boxBackground, index->gc );

/* Clear each pixmap by filling it with the background color */

		XFillRectangle(display, index->pixmap, index->gc,
			0, 0, drawing_width+5, drawing_height);

/* Set the data for this option back to zero */

		for(n = 0; n < MAXITEMS; n++)
			index->cdata[n] = index->pdata[n] = 0;

/* Do the next node */

		index = index->next;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	void get_cpulist_from_file(char *file)

  DESCRIPTION:

	Reads a list of node name from a file and stores it in the
	CPULIST.

  FORMAL PARAMETERS:

	char *file:	Name of the file to load the cpu list from.

																  */
/******************************************************************/

void get_cpulist_from_file(char *file)
{
	FILE *fp;
	char filename[81], name[81];
	struct CPULIST *index, *pointers[300];
	int i, n = 0;

/* Get the users home path */

	if(!strcmp(file, "default"))
		sprintf(filename, "%s/%s", getenv("HOME"), ".sysmon");
	else 
		strcpy(filename, file);

/* Try to open the file */
	if((fp = fopen(filename, "r")) == NULL)
	{
		fprintf(stderr, "Cannot locate the %s file.\n\n", filename);
		exit(-1);
	}

/* Read each name and add it to the CPULIST */

	while(fscanf(fp, "%s", name) != EOF) {
		acpu(name);
    }

	fclose(fp);

	/* new code 9/19/94 */
	if( app_data.numRows == 0 )
		app_data.numRows = ( total_cpus + 4 ) / 5;
	/* */

/* Create an array of pointers to each element so that they may be sorted */

	index = head;
	while(index != NULL)
	{
		pointers[n] = index;
		n++;
		index = index->next;
	}
	pointers[n] = NULL;

/* Sort the array */

	/*
	qsort(pointers, n, sizeof(struct CPULIST *), compare);
	*/

/* Re-link the list */

	head = pointers[0];
	for(i = 0; i < n; i++)
		pointers[i]->next = pointers[i+1];
}

/******************************************************************/
/*
	FUNCTION NAME:

	void get_cpulist_from_pm(void)

  DESCRIPTION:

	Get the list of nodes which belong to the Production System
	from the Job Manager.

  FORMAL PARAMETERS:

	name - The name of the Production System

																  */
/******************************************************************/

void get_cpulist_from_pm(char *name)
{
	struct CPULIST *index, *pointers[300];
	struct prod_cpu_list *pc;
	int i, n;

/* Connect to the PM */

	fprintf(stderr,"Connecting to PM...\n");
	if(pm_connect() == -1) {
		printerr(FACILITY, FATAL_MSG, "Failed to connect to a PM", NULL, 1);
		exit(0);
	}

/* Ask for a PS by name */

	p = find_prodsys(db.ps, name);
	if(p == NULL) {
		printerr(FACILITY, FATAL_MSG, "Production system not found", NULL, 2);
		exit(0);
	}

/* Add the JM node first */

	acpu(p->jm_cpu->cpu_name);

/* Add the rest, but make sure they're not the JM's node */

	pc = p->cpus;
	while(pc != NULL)
	{
		if(strcmp(pc->cpu->cpu_name, p->jm_cpu->cpu_name))
		acpu(pc->cpu->cpu_name);
		pc = pc->next;
	}

/* Begin the sort routine */
/* A separate set of sorting routines is required for 
   get_cpulist_from_pm and get_cpulist_from_file because
   here, we want the JM node to always be first */

/* Build an array of pointers */

	n = 0;
	index = head->next;
	while(index != NULL)
	{
		pointers[n] = index;
		n++;
		index = index->next;
	}
	pointers[n] = NULL;

/* Sort the node names */

	/*
	qsort(pointers, n, sizeof(struct CPULIST *), compare);
	*/

/* Re-link the list */

	head->next = pointers[0];
	for(i = 0; i < n; i++)
		pointers[i]->next = pointers[i+1];

	pm_disconnect();
	fprintf(stderr, "Disconnected from PM\n");
}

static int compare(struct CPULIST **x, struct CPULIST **y)
{
	return(strcmp((*x)->name, (*y)->name));
}

/******************************************************************/
/*
	FUNCTION NAME:

	void acpu(char *name)

  DESCRIPTION:

	Create a link in the CPULIST list and add the node to it.
	Also setup preliminary networking stuff.

  FORMAL PARAMETERS:

	name - The name of the node

																  */
/******************************************************************/

void acpu(char *name)
{
	struct hostent *hp;
	struct CPULIST *new, *index;

/* Get information about the remote node */

	if((hp = gethostbyname(name)) == NULL)
	{
		fprintf(stderr, "Host not found: %.6s\n", name);
		return;
	}

/* Create a new link */

	new = (struct CPULIST *)malloc(sizeof(struct CPULIST));

/* Determine where to add it */

	if(head == NULL)
		head = index = new;
	else
	{
		index = head;
		while(index->next != NULL)
			index = index->next;
		index->next = new;
		index = new;
	}
	index->active = 0;
	total_cpus++;

/* Truncate the name after the first '.' */

	/*
	truncate(name);
	*/

/* Copy it to the CPULIST list */

	strcpy(index->name, name);

/* Setup some RPC necessities */

	bcopy(hp->h_addr, (caddr_t)&index->server_addr.sin_addr, hp->h_length);
	index->server_addr.sin_family = AF_INET;
	index->server_addr.sin_port = 0;
	index->status = RPC_SUCCESS;
	index->sock = RPC_ANYSOCK;
	index->next = NULL;
}

/******************************************************************/
/*
	FUNCTION NAME:

	void rcpu(char *name)

  DESCRIPTION:

	Parse the CPULIST list and remove 'name' from the list.

  FORMAL PARAMETERS:

	name - The name of the node

																  */
/******************************************************************/

void rcpu(char *name)
{
	struct CPULIST *index;

/* Start from the beginning */

	index = head;
	if(!strcmp(index->name, name))
	{
		free(index);
		head = index = index->next;
	}
	else
	{
		while(strcmp(index->next->name, name))
			index = index->next;
		free(index->next);
		index->next = index->next->next;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	void truncate(char *name)

  DESCRIPTION:

	Truncate an Internet name after the first '.'

	EX: fnsfa.fnal.gov --> fnsfa

  FORMAL PARAMETERS:

	name - The name of the node

																  */
/******************************************************************/

void truncate(char *name)
{
	register int n;

/* Brute force, but necessary */

	for(n = 0; n <= strlen(name); n++)
		if(name[n] == '.')
			break;
	name[n] = '\0';
}

/******************************************************************/
/*
	FUNCTION NAME:

	void destroy_list(void)

  DESCRIPTION:

	Destroys all links in the CPULIST, and frees all allocated memory

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void destroy_list(void)
{
	struct CPULIST *index;

	index = head;
	while(index != NULL)
	{
		free(index);
		index = index->next;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	void establish_remote_connections(void)

  DESCRIPTION:

	Establishes RPC remote connections with each node
	in the CPULIST list.

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void establish_remote_connections(void)
{
	struct CPULIST *index;
	struct timeval time_out;

/* Number of seconds RPC will WAIT before attempting another connection.
   NOT the number of seconds before it FAILs a connection */

	/*
	time_out.tv_sec = 5;
	*/
	time_out.tv_sec = app_data.timeout;
	time_out.tv_usec = 0;

/* Create UDP connections with the RSTATD daemon running on the remote nodes */
	/*
	pmap_settimeouts( time_out, time_out );
	*/

	index = head;
	while(index != NULL)
	{
		fprintf(stderr,"Attempting connection with %s...", index->name);
		fflush(stderr);

		if((index->client = clntudp_create(&index->server_addr, RSTATPROG,
			RSTATVERS_TIME, time_out, &index->sock)) == NULL)
		{
				fprintf(stderr, "Failed\n" );
				fprintf(stderr,
					"sysmon: Cannot establish connection%s",
					clnt_spcreateerror(""));
                fflush(stderr);

/* Failed connection, remove node from the list */
		
		/* changed for the following lines - lauro. 
				rcpu(index->name);
				total_cpus--;
         */
			    index->status = RPC_FAILED;
			    index->connect = FALSE;
		}
		else
		{

/* Established connection, authenticate it */

			fprintf(stderr, "OK\n");
			fflush(stderr);
			index->connect = TRUE;
			index->client->cl_auth = authunix_create_default();
			index->prev_v.cp_time[CP_USER] = 0.0;
			index->prev_v.cp_time[CP_NICE] = 0.0;
			index->prev_v.cp_time[CP_SYS] = 0.0;
			index->prev_v.cp_time[CP_IDLE] = 0.0;
		}
		index = index->next;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	void processOptions(Widget w, XtPointer client_data,
						XtPointer call_data)

  DESCRIPTION:

	Process the menu options.

  FORMAL PARAMETERS:

	w		- Pointer to the calling widget
	client_data - Pointer to a string passed by the client
	unused		- Not used

																  */
/******************************************************************/

void processOptions(Widget w, XtPointer client_data, XtPointer unused)
{
	int n;

/* Process the Exit option */

	if(!strcmp(client_data, "Exit"))
		sigint_handler();

/* Process the Help option */

	if(!strcmp(client_data, "Help"))
	{
		help_handler(XtParent(w));
		return;
	}

	if( w != MainWindow )
	{
		if( w != CurrentMonitorWidget ) 
		{
			XtVaSetValues( CurrentMonitorWidget, XmNset, False ,NULL );
			CurrentMonitorWidget = w;
		}
		else
			XtVaSetValues( CurrentMonitorWidget, XmNset, True ,NULL );
	}

/* Parse through the list of menu options */

	option = P_CPU;
	while(strcmp(menuOptions[option], client_data))
		option++;

/* Reset the maximum value for the option */

	if(option == P_CPU)
		maxvalue[option] = 100.0;	/* CPU max is always 100% */
	else if(option == P_LOAD)
		maxvalue[option] = 0.001;	/* LOAD is relatively small */
	else
		maxvalue[option] = 1.0;		/* Reset max on all others */

/* Reset the display to reflect the changes */

	reset();
}

/******************************************************************/
/*
	FUNCTION NAME:

	void help_handler(Widget parent)

  DESCRIPTION:

	Create a window containing help.

  FORMAL PARAMETERS:

	None

																  */
/******************************************************************/

void help_handler(Widget parent)
{
	int n;
	XmString xmstr;
	Boolean ok;
	Widget bulletinb, scrolledw, pushb, frame, label;

	n = 0;
	PopupShell = XtCreatePopupShell("xcps_sysmon v2.0   -   HELP",
		transientShellWidgetClass, MainWindow, args, n);

	n = 0;
	XtSetArg(args[n], XmNresizePolicy, XmRESIZE_GROW); n++;
	XtSetArg(args[n], XmNx, 0); n++;
	XtSetArg(args[n], XmNy, 0); n++;
	XtSetArg(args[n], XmNwidth, 665); n++;
	XtSetArg(args[n], XmNheight, 360); n++;
	bulletinb = XtCreateWidget("bulletinb",
		xmBulletinBoardWidgetClass, PopupShell, args, n);
	XtManageChild(bulletinb);

	n = 0;
	XtSetArg(args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
	XtSetArg(args[n], XmNx, 10); n++;
	XtSetArg(args[n], XmNy, 10); n++;
	XtSetArg(args[n], XmNwidth, 655); n++;
	XtSetArg(args[n], XmNheight, 300); n++;
	scrolledw = XtCreateWidget("scrolledw",
		xmScrolledWindowWidgetClass, bulletinb, args, n);
	XtManageChild(scrolledw);

	n = 0;
	XtSetArg(args[n], XmNfontList, CONVERT(parent,
	"-b&h-lucidatypewriter-bold-r-normal-sans-14-*",
		"FontList", &ok)); if(ok) n++;
	XtSetArg(args[n], XmNlabelString, (xmstr=XmStringCreateLtoR(
	"Dismiss", XmSTRING_DEFAULT_CHARSET))); n++;
	XtSetArg(args[n], XmNrecomputeSize, False); n++;
	XtSetArg(args[n], XmNx, 10); n++;
	XtSetArg(args[n], XmNy, 320); n++;
	XtSetArg(args[n], XmNwidth, 629); n++;
	XtSetArg(args[n], XmNheight, 30); n++;
	pushb = XtCreateWidget("pushb",
		xmPushButtonWidgetClass, bulletinb, args, n);
	XmStringFree(xmstr);
	XtManageChild(pushb);
	XtAddCallback(pushb, XmNarmCallback, DestroyPopup, NULL);

	n = 0;
	XtSetArg(args[n], XmNshadowType, XmSHADOW_IN); n++;
	XtSetArg(args[n], XmNborderWidth, 0); n++;
	XtSetArg(args[n], XmNx, 0); n++;
	XtSetArg(args[n], XmNy, 0); n++;
	frame = XtCreateWidget("frame", xmFrameWidgetClass,
		scrolledw, args, n);
	XtManageChild(frame);

	n = 0;
	XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++;
	XtSetArg(args[n], XmNstringDirection, XmSTRING_DIRECTION_L_TO_R); n++;
	XtSetArg(args[n], XmNfontList,
		CONVERT(parent, "-b&h-lucidatypewriter-bold-r-normal-sans-14-*",
		"FontList", &ok)); if(ok) n++;
	XtSetArg(args[n], XmNrecomputeSize, False); n++;
	XtSetArg(args[n], XmNlabelString, (xmstr=XmStringCreateLtoR(
	HelpText, XmSTRING_DEFAULT_CHARSET))); n++;
	XtSetArg(args[n], XmNx, 2); n++;
	XtSetArg(args[n], XmNy, 2); n++;
	XtSetArg(args[n], XmNwidth, 620); n++;
	XtSetArg(args[n], XmNheight, 1100); n++;
	label = XtCreateWidget("label", xmLabelWidgetClass, frame, args, n);
	XtManageChild(label);

	XtPopup(PopupShell, XtGrabNone);
}

void DestroyPopup(void)
{
	XtPopdown(PopupShell);
	XtDestroyWidget(PopupShell);
}

/******************************************************************/
/*
	FUNCTION NAME:

	void sigint_handler(int sig, int code, struct sigcontext *sc)

  DESCRIPTION:

	Handle all interrupt signals so that we can exit gracefully.
	Close all socket connections, free all memory allocated to the
	pixmaps, graphic contexts, and CPULIST.

  FORMAL PARAMETERS:

	sig	- Caught signal
	code	- Unused for this signal
	sc	- The processors context

																  */
/******************************************************************/

void sigint_handler(int sig, int code, struct sigcontext *sc)
{
	struct CPULIST *index;

	index = head;
	while(index != NULL && index->connect == TRUE)
	{
		clnt_destroy(index->client);
		close(index->sock);
		if( xversion )
		{
			XFreePixmap(display, index->pixmap);
			XFreeGC(display, index->gc);
		}
		free(index);
		index = index->next;
	}
	exit(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	static caddr_t CONVERT(Widget w, char *from_string,
				char *to_type, Boolean *success)

  DESCRIPTION:

	An adaptation of a routine generated by Builders Accessory.
	Removed misc garbage and streamlined the code.

	Converts one format to another for use by the X windows interface.

  FORMAL PARAMETERS:

	w		- Pointer to the calling widget
	from_string - String to be converted
	to_type		- The type to be converted to (XString, Pixel, etc...)

																  */
/******************************************************************/

static caddr_t CONVERT(w, from_string, to_type, success)
Widget w;
char *from_string;
char *to_type;
Boolean *success;
{
	XrmValue			fromVal, toVal; /* resource holders			 */
	Boolean			 convResult;	 /* return value				 */
	unsigned char	   oByte;		  /* one byte result			  */
	unsigned short	  tByte;		  /* two byte result			  */
	caddr_t			 fByte;		  /* four byte result			 */
	caddr_t			 aByte;		  /* allocated result			 */
	int				 to_size;
	fByte = aByte = NULL;
	*success = False;

	if(!strcmp(XmRString, to_type))
	{
		*success = True;
		return( from_string );
	}

	to_size = strlen(from_string);
	fromVal.size = strlen(from_string) + 1;
	fromVal.addr = from_string;

	switch( to_size )
	{
		case 1:
			toVal.size = sizeof(unsigned char);
			toVal.addr = (caddr_t)&oByte;
			break;
		case 2:
			toVal.size = sizeof(unsigned short);
			toVal.addr = (caddr_t)&tByte;
			break;
		default:
			toVal.size = sizeof(caddr_t);
			toVal.addr = (caddr_t)&fByte;
			break;
	}

	convResult = XtConvertAndStore(w, XmRString, &fromVal, to_type, &toVal);

	if( convResult )
	{
		switch( to_size )
		{
			case 1:
				fByte = (caddr_t)oByte;
				break;
			case 2:
				fByte = (caddr_t)tByte;
				break;
			default:
				break;
		}
	}

	if( !convResult && toVal.size != to_size )
	{
		toVal.addr = (caddr_t)XtMalloc(toVal.size);
		fByte = aByte = toVal.addr;
		convResult = XtConvertAndStore(w, XmRString, &fromVal, to_type, &toVal);
	}

	if( !convResult )
	{
		XtFree(aByte);
		aByte = NULL;
	}

	*success = convResult;
	return( fByte );
}

display_error(msg)
char *msg;
{
	fprintf(stderr, "%s\n",msg);
	fflush(stdout);
}

