/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/************************************************************************/
/*																		*/
/*																		*/
/*	Background message handler for CPS user processes					*/
/*																		*/
/*																		*/
/************************************************************************/

#include "all.h"
#include <signal.h>
#include "../inc/cps_msgs.h"

#if defined(MIPS) || defined(R3000)
#include <signal.h>
extern int in_sig_handler[];
#endif

static char warning[132];
static char *b_fntype,*b_func;
static int b_sop,b_status = CPS_OK;
static char b_errmsg[132];
static int inside_handler = 0;

static int save_error_info()
{
	b_fntype = acp_fntype;
	b_func = acp_func;
	acp_fntype = "background_";
	acp_func = "handler";
	if (acp_return_status != CPS_OK) {
		strcpy(b_errmsg,acp_error_info);
		b_status = acp_return_status;
		b_sop = acp_set_of_procs;
	}
}
static int restore_error_info()
{
	acp_fntype = b_fntype;
	acp_func = b_func;
	if (b_status != CPS_OK) {
		strcpy(acp_error_info,b_errmsg);
		acp_return_status = b_status;
		acp_set_of_procs = b_sop;
	}
}
static int log_bg_error(msg)
char *msg;
{
	sprintf(warning,"%s%s: %s",acp_fntype,acp_func,msg);
	acp_log_msg(warning);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_declare_queue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_declare_queue(m)
cps_message *m;
{
	int i;
	int q,nargs,*bcount;

	int s;

/* Extract information out of the message */

	cps_get_declare_queue(m,&q,&nargs,&bcount);

/* Check the information */

	s = acp_check_qnum(q);
	if (s != CPS_OK) {
		log_bg_error(acp_error_info);
		return;
	}

	s = acp_check_narg(q,nargs);
	if (s != CPS_OK) {
		log_bg_error(acp_error_info);
		return;
	}

	s = acp_check_qargs(q,nargs,bcount);
	if (s != CPS_OK) {
		log_bg_error(acp_error_info);
		return;
	}

/* Update global structures */

	acp_queue_declared[q] = nargs;
	if (nargs > 0) {
		acp_queue_bcount[q] = (int *) hmalloc(nargs*sizeof(int));
		bcopy(bcount,acp_queue_bcount[q],nargs*sizeof(int));
	}
	for (i=0; i<nargs; i++) acp_queue_bsum[q] += bcount[i];

	acp_request_complete++;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_sub_decl

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_sub_decl(m)
cps_message *m;
{
	struct acp_remote_sub *rsub,*r;

	int i;

/* Extract the information */

	cps_get_sub_decl(m,&rsub);

/* Add it to the list */

	r = (struct acp_remote_sub *) hmalloc(sizeof(struct acp_remote_sub));
	*r = *rsub;

	r->num_active = 0;
	r->inst = (struct acp_sub_instance **) 
	hmalloc(MAX_SUB_INSTANCES*sizeof(struct acp_sub_instance *));
	for (i=0; i<MAX_SUB_INSTANCES; i++) r->inst[i] = NULL;

	r->next = acp_subs;
	acp_subs = r;

	acp_rsub_complete++;

}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_remote_call

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_remote_call(m,q)
cps_message *m;
struct qelem *q;
{

/* Push the subroutine call onto the pending queue */

	if (acp_rms_tail == NULL) {
		q->next = -1;
		acp_rms_tail = q;
		acp_rms_head = q;
	}
	else {
		acp_rms_tail->next = q->index;
		q->next = -1;
		acp_rms_tail = q;
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_explicit_msg

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_explicit_msg(m)
cps_message *m;
{
	struct acp_msg_block *block;
	int size;
	long *x;

	x = (long *) m->msg;
	acp_xpl_ready[ *x ]++;
	size = 4*sizeof(long) + m->size;
	block = acp_checkout_block( size );
	memcpy( block->buf, (char *) m, size );
	acp_queue_block(block,&acp_xpl);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_remote_return

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_remote_return(m)
cps_message *m;
{

	int id,inst,action;
	char *args;

	int i;

	struct acp_remote_sub *rsub;


/* Extract the information (same as for CPS_REMOTE_CALL) */

	cps_get_rem_call(m,&id,&inst,&action,&args);

/* Find the subroutine declaration */

	rsub = acp_find_remsub(id);
	if (rsub == NULL) {
		log_bg_error("Failed to obtain remote subroutine declaration from JM");
		return;
	}

/* Copy arguments back (for WRITE_ONLY or READ_WRITE) */

	for (i=0; i<rsub->nargs; i++) {
		if (rsub->argtyp[i] != CPS_READ_ONLY)
			bcopy(args,rsub->inst[inst]->baddr[i],rsub->bcount[i]);
		args += rsub->bcount[i];
	}

/* Mark instance complete */

	rsub->inst[inst]->complete = -1;

/* If action is CPS_NO_WAIT, mark instance inactive */

	if (action == CPS_NO_WAIT) {
		rsub->inst[inst]->active = 0;
		rsub->num_active--;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_dequeue_reply

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_dequeue_reply(m)
cps_message *m;
{

	int lpn,q;
	char *data;

/* Extract the information (same as for queue process) */

	cps_get_queue_process(m,&lpn,&q,&data);

	if (q == acp_queue_number) {
		acp_qreq_complete++;
		acp_process_dequeued = lpn;
		if (lpn != CPS_EMPTY) 
			bcopy(data,acp_queue_data,acp_queue_bsum[q]);
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	  setup_asynch_xfer

								  */
/******************************************************************/

static int setup_asynch_xfer(aqx)
struct active_qxfer *aqx;
{
	struct queued_xfer *qx;

	int j,k;

	struct acp_data_block *dblock;
	int *bcount;

	qx = aqx->head;
	if (qx == NULL) {		/* This shouldn't happen */
		abort();
	}

	bcount = qx->bcount;

	if (aqx->id > 0) {			/* Asynchronous send */

		*bcount = qx->nbytes;

		j = aqx->lpn;

/* Fill in the information in the direct transfer record */

		if (acp_dxfer_nleft[j] >= MAX_L_XFERS) {
		}

		k = acp_dxfer_nleft[j] + acp_dxfer_ptr[j];

		acp_l_dxfer[j][k].type = XFER_SEND;
		acp_l_dxfer[j][k].need_to_clear = 0;
		acp_l_dxfer[j][k].state = XFER_NEED_CONNECT;
		acp_l_dxfer[j][k].op = 2;
		acp_l_dxfer[j][k].abort = qx->abort;	/* v2.9 */
		acp_l_dxfer[j][k].done = qx->flag;
		acp_l_dxfer[j][k].buf = qx->addr;
		acp_l_dxfer[j][k].blockno = qx->blockid;
		acp_l_dxfer[j][k].offset = 0;
		acp_l_dxfer[j][k].nbytes = qx->nbytes;
		if (qx->next == NULL) {
			acp_l_dxfer[j][k].lflag = -1;
		}
		else {
			acp_l_dxfer[j][k].lflag = 0;
		}
		acp_l_dxfer[j][k].aflag = -1;

		acp_dxfer_nleft[j]++;
	}
	else {				/* Asynchronous get */

/* Determine how many bytes to get */

		j = aqx->lpn-FIRST_USER_LPN;
		dblock = acp_find_block(acp_remote_blocks[j],qx->blockid);
		if (dblock != NULL) {
			if (dblock->bcount > qx->nbytes) {	/* Request too large */
			}

			else if (dblock->bcount > 0) {	/* Get a partial block */
				qx->nbytes = dblock->bcount;
			}
		}
		*bcount = qx->nbytes;

/* Fill in the information in the direct transfer record */

		j = aqx->lpn;

		if (acp_dxfer_nleft[j] >= MAX_L_XFERS) {
		}

		k = acp_dxfer_nleft[j] + acp_dxfer_ptr[j];

		acp_l_dxfer[j][k].type = XFER_GET;
		acp_l_dxfer[j][k].need_to_clear = 0;
		acp_l_dxfer[j][k].state = XFER_NEED_CONNECT;
		acp_l_dxfer[j][k].op = 2;
		acp_l_dxfer[j][k].abort = qx->abort;	/* v2.9 */
		acp_l_dxfer[j][k].done = qx->flag;
		acp_l_dxfer[j][k].buf = qx->addr;
		acp_l_dxfer[j][k].blockno = qx->blockid;
		acp_l_dxfer[j][k].offset = 0;
		acp_l_dxfer[j][k].nbytes = qx->nbytes;
		if (qx->next == NULL) {
			acp_l_dxfer[j][k].lflag = -1;
		}
		else {
			acp_l_dxfer[j][k].lflag = 0;
		}
		acp_l_dxfer[j][k].aflag = -1;

		acp_dxfer_nleft[j]++;
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_asynch_dq_reply

								  */
/******************************************************************/

static int handle_asynch_dq_reply(m)
cps_message *m;
{
	int lpn,qnum,id;

	struct active_qxfer *aqx,*aqx1;

	struct queued_xfer *qx,*qx1;

	struct qelem *q;
	cps_message *m1;

/* Extract the information */

	cps_get_asynch_dq_reply(m,&lpn,&qnum,&id);

/* Find the asynchronous qxfer record */

	aqx1 = NULL;
	aqx = acp_aqx;	
	while (aqx != NULL) {
		if (aqx->id == id) break;
		aqx1 = aqx;
		aqx = aqx->next;
	}
	if (aqx == NULL) {	/* Something went wrong */
		abort();
	}

	aqx->lpn = lpn;
	*(aqx->lpn_addr) = lpn;

/* Do the data transfers (if a process was dequeued) */

	if (lpn != CPS_END_OF_QUEUE) {

		while (1) {

/* Check to see if this is the last data transfer */

			if (aqx->head == NULL) {
				if (aqx1 == NULL) {
					acp_aqx = aqx->next;
				}
				else {
					aqx1->next = aqx->next;
				}
				hfree(aqx);

/* Tell the remote process to look for a transfer */

				q = acp_q_pop1(S_FREE_Q);
				m1 = (cps_message *) SMALL_MSG(q);
				cps_bld_direct_xfer(m1,acp_local_lpn,lpn,acp_local_lpn);
				acp_q_push1(acp_shm_qnum,q);

/* Tell ourselves to do the transfer */

				q = acp_q_pop1(S_FREE_Q);
				m1 = (cps_message *) SMALL_MSG(q);
				cps_bld_direct_xfer(m1,acp_local_lpn,acp_local_lpn,lpn);
				acp_q_push1(acp_shm_qnum,q);
				acp_signal_shm();

				return;
			}

/* Store parameters for the next data transfer */

			else {

				setup_asynch_xfer(aqx);

			}

/* Remove the transfer from the list. */

			qx = aqx->head;
			aqx->head = qx->next;
			hfree(qx);

		}
	}
	else {	/* We got the end of queue */

/* Mark all transfers complete and deallocate */
/* Decrement number of gets pending for each queued transfer */

		qx = aqx->head;
		while (qx != NULL) {
			*(qx->flag) = -1;
			qx1 = qx;
/* Fix on 2/17/94 MRF */
			if (aqx->id < 0) acp_zero_pget(qx->blockid);
			qx = qx->next;
			hfree(qx1);
		}

/* Deallocate the aqx record */

		if (aqx1 == NULL) {
			acp_aqx = aqx->next;
		}
		else {
			aqx1->next = aqx->next;
		}
		hfree(aqx);

	}

}

/************************************************************************/
/*

	Function:   acp_local_block_arrived
		remote_block_arrived

*/
/************************************************************************/

int acp_local_block_arrived(blockno)
int blockno;
{
	struct acp_data_block *block;

	block = acp_find_block(acp_blocks,blockno);
	if  (block != NULL) block->arrived = -1;

}

int remote_block_arrived(lpn,blockno)
int lpn,blockno;
{
	struct acp_data_block *block;
	int i;

	i = lpn - FIRST_USER_LPN;
	block = acp_find_block(acp_remote_blocks[i],blockno);
	if  (block != NULL) block->arrived = -1;

}

int acp_local_block_arrived1(blockno,bcount)
int blockno,bcount;
{
	struct acp_data_block *block;

	block = acp_find_block(acp_blocks,blockno);
	if  (block != NULL) {
		block->arrived = -1;
		block->bcount = bcount;
	}
}

int remote_block_arrived1(lpn,blockno,bcount)
int lpn,blockno,bcount;
{
	struct acp_data_block *block;
	int i;

	i = lpn - FIRST_USER_LPN;
	block = acp_find_block(acp_remote_blocks[i],blockno);
	if  (block != NULL) {
		block->arrived = -1;
		block->bcount = bcount;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_declare_block

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_declare_block(m)
cps_message *m;
{
	int lpn,blockno;

	struct acp_data_block *dblock;

	int i;

	cps_get_declare_block(m,&lpn,&blockno);

	i = lpn-FIRST_USER_LPN;

	if ((dblock = acp_find_block(acp_remote_blocks[i],blockno)) != NULL) {
		return;
	}
	else {
		dblock = (struct acp_data_block *) 
		hmalloc(sizeof(struct acp_data_block));
	}

	dblock->bcount = -1;
	dblock->blockno = blockno;
	dblock->next = acp_remote_blocks[i];
	acp_remote_blocks[i] = dblock;

}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_data_ready

								  */
/******************************************************************/

static int handle_data_ready(m)
cps_message *m;
{
	int lpn,blockno,bcount;

	struct acp_data_block *dblock;

	int i;

	cps_get_data_ready(m,&lpn,&blockno,&bcount);

	i = lpn-FIRST_USER_LPN;

	if ((dblock = acp_find_block(acp_remote_blocks[i],blockno)) == NULL) {
		dblock = (struct acp_data_block *) 
		hmalloc(sizeof(struct acp_data_block));
		dblock->blockno = blockno;
		dblock->next = acp_remote_blocks[i];
		acp_remote_blocks[i] = dblock;
	}

	dblock->bcount = bcount;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_vbbc_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_vbbc_info(m)
cps_message *m;
{
	int lpn,blockno,n;
	unsigned long *paddr;
	int crate,vme_addr,offset;

	struct acp_data_block *dblock;
	struct acp_vbbc_info *vbbc;
	unsigned long *x;

	int i;

	cps_get_vbbc_info(m,&lpn,&blockno,&crate,&vme_addr,&offset,&n,&paddr);

	i = lpn-FIRST_USER_LPN;

	if ((dblock = acp_find_block(acp_remote_blocks[i],blockno)) == NULL) {
		return 0;
	}

	vbbc = (struct acp_vbbc_info *) hmalloc(sizeof(struct acp_vbbc_info));
	x = (unsigned long *) hmalloc(n*sizeof(unsigned long));
	vbbc->crate = crate;
	vbbc->vme_addr = vme_addr;
	vbbc->npaddr = n;
	vbbc->paddr = x;
	vbbc->offset = offset;
	bcopy(paddr,x,n*sizeof(long));

	dblock->vbbc = vbbc;

	return 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_usr_proc_dead

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static handle_usr_proc_dead(m)
cps_message *m;
{
	int lpn,i;
	char *msg;

	cps_get_usr_proc_dead(m,&lpn,&msg);

	for (i=1; i<=MAX_CLASSES; i++)
	{
		if (acp_class_proc[i] != NULL)
		{
			if (set_member(lpn,acp_class_proc[i]))
			{
				remove_from_set(lpn,acp_class_proc[i]);
				break;
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_fill_buffer

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_fill_buffer(m,q)
cps_message *m;
struct qelem *q;
{
	int vaddr,index,xfer_type,blockno,offset,size;
	char *buf;

	struct acp_data_block *block;

	int src_lpn,dst_lpn;

	src_lpn = m->src_lpn;
	dst_lpn = m->dst_lpn;
	cps_get_fill_buffer(m,&index,&xfer_type,&vaddr,&blockno,&offset,&size);
	if ((xfer_type == 0) || (xfer_type > 1)) {	/* acp_send */

/* Get the data from the virtual address */

/* Using the buffer provided, build a CPS_DATA_XFER message with the source
   and destination lpns reversed */

		buf = (char *) vaddr;
		cps_bld_data_xfer(m,dst_lpn,src_lpn,index,xfer_type,blockno,offset,
			size,buf);
		acp_q_push1(acp_shm_qnum,q);
		acp_signal_shm();

	}
	else {			/* acp_get */

/* Get the data from the block & offset */

		block = acp_find_block(acp_blocks,blockno);
		if  (block == NULL) {

/* Do some error handling */

			acp_q_push1(0,q);
			return;
		}
		buf = block->buf + offset;
		cps_bld_data_xfer(m,dst_lpn,src_lpn,index,xfer_type,vaddr,0,size,buf);
		acp_q_push1(acp_shm_qnum,q);
		acp_signal_shm();

	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_data_xfer

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int handle_data_xfer(m)
cps_message *m;
{
	int index,xfer_type,addr,offset,size;
	char *srcbuf,*dstbuf;
	int *s,*d;

	struct acp_data_block *block;

	int i;

	cps_get_data_xfer(m,&index,&xfer_type,&addr,&offset,&size,&srcbuf);

	if ((xfer_type == 0) || (xfer_type > 1)) {	/* acp_send */

/* Store the data according to blockno & offset */

		block = acp_find_block(acp_blocks,addr);
		if  (block == NULL) {

/* Do error handling here */

			return;
		}
		dstbuf = block->buf + offset;
		memcpy(dstbuf,srcbuf,size);

	}
	else {			/* acp_get */

/* Store the data in the virtual address */

		d = (int *) addr;
		s = (int *) srcbuf;
		size /= sizeof(int);
		for (i=0; i<size; i++) {
			*d += *s;
			d++; s++;
		}

	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  handle_small_msg

								  */
/******************************************************************/

static int handle_small_msg(m)
cps_message *m;
{
	int blockno,offset,size;
	char *srcbuf,*dstbuf;

	struct acp_data_block *block;

	cps_get_small_msg(m,&blockno,&offset,&size,&srcbuf);

/* Store the data according to blockno & offset */

	block = acp_find_block(acp_blocks,blockno);
	if  (block == NULL) return;

	dstbuf = block->buf + offset;
	memcpy(dstbuf,srcbuf,size);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_handle_sigusr1

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_handle_sigusr1(sig,code,sc)
int sig,code;
#if defined(R3000)
char *sc;
#else
void *sc;
#endif
{

	int i,blockno;
	struct qelem *q;
	cps_message *m;

	int qn;

	int msgtyp;

	if (inside_handler) {
		printerr(FACILITY,ERROR_MSG,"Re-entered signal handler",NULL,14);
		abort();
	}

	inside_handler = -1;
#if defined(MIPS) || defined(R3000)
	in_sig_handler[SIGUSR1] = -1;
#endif
	save_error_info();

	while (q_head[acp_local_qnum] != -1) {

/* Get a message */

		q = acp_q_pop1(acp_local_qnum);
		if (q->size) {
			m = LARGE_MSG(q);
		}
		else {
			m = (cps_message *) SMALL_MSG(q);
		}
		msgtyp = (int) m->type;
#ifdef DIAG
		log_cps_msg(acp_fp,m);
#endif
		switch (msgtyp) {

			case MSG_SMALL_MSG :
				handle_small_msg(m);
				break;
			case MSG_DATA_XFER :
				handle_data_xfer(m);
				break;
			case MSG_DIRECT_XFER :
				if (acp_handle_direct_xfer(m,q) == -1) {
					m = NULL;
				}
				break;
			case MSG_FILL_BUFFER :
				handle_fill_buffer(m,q);
				m = NULL;
				break;
			case MSG_REQ_COMPLETE :
				acp_request_complete++;
				break;
			case MSG_SYNC_COMPLETE :
				acp_sync_complete++;
				break;
			case MSG_HOOK_COMPLETE :
				acp_hook_complete++;
				break;
			case MSG_HOOK_SWITCH :
				proc_hooks = !proc_hooks;
				break;
			case MSG_WAIT_COMPLETE :
				acp_wait_complete++;
				break;
			case MSG_RSUB_COMPLETE :
				acp_rsub_complete++;
				break;
			case MSG_XFER_COMPLETE :
				cps_get_xfer_complete(m,&i,&blockno);
				if (i == 0) {
					acp_send_complete++;
				}
				else if (i == 1) {
					acp_get_complete++;
				}
				break;
			case MSG_BLOCK_ARRIVED :
				cps_get_block_arrived(m,&blockno);
				acp_local_block_arrived(blockno);
				break;
			case MSG_DECLARE_QUEUE :
				handle_declare_queue(m);
				break;
			case MSG_END_OF_QUEUE :
				cps_get_end_of_queue(m,&qn);
				acp_queue_declared[qn] = -2;
				break;
			case MSG_SUB_DECL :
				handle_sub_decl(m);
				break;
			case MSG_REMOTE_CALL :
				handle_remote_call(m,q);
				m = NULL;
				break;
			case MSG_REMOTE_RETURN :
				handle_remote_return(m);
				break;
			case MSG_DEQUEUE_REPLY :
				handle_dequeue_reply(m);
				break;
			case MSG_ASYNCH_DQ_REPLY :
				handle_asynch_dq_reply(m);
				break;
			case MSG_DATA_READY :
				handle_data_ready(m);
				break;
			case MSG_DECLARE_BLOCK :
				handle_declare_block(m);
				break;
#ifdef R3000
			case MSG_VBBC_INFO :
				handle_vbbc_info(m);
				break;
#endif
			case MSG_USR_PROC_DEAD :
				handle_usr_proc_dead(m);
				break;
			case MSG_EXPLICIT_MSG :
				handle_explicit_msg(m);
				break;
			case MSG_LIST_MOUNT_REPLY :
				cps_get_list_mount_reply(m,&mount_reply_status,
					&mount_reply_pos,&mount_reply_protection,
					mount_reply_dev,mount_reply_vsn,
					mount_reply_msg);
				acp_treq_complete++;
				break;
			case MSG_SPEC_MOUNT_REPLY :
				cps_get_spec_mount_reply(m,&mount_reply_status,
					mount_reply_dev,mount_reply_msg);
				acp_treq_complete++;
				break;
			default:
				break;
		}

		if (m != NULL) acp_q_push1(0,q);
	}

	restore_error_info();
	inside_handler = 0;
#if defined(MIPS) || defined(R3000)
	in_sig_handler[SIGUSR1] = 0;
#endif
}

acp_handle_sigusr2(sig,code,sc)
int sig,code;
#if defined(R3000)
char *sc;
#else
void *sc;
#endif
{
	int pid;
	int status;

	pid = fork();

	if( pid == 0 )
	{
	}
	else
	{
		kill( pid, SIGTRAP );
		wait(&status);
	}
}
