/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991,1992 Universities Research Association, Inc.	    
	   All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

	Source File:  datacon.c

	Explanation: 

	Exported Routines:


*/

#define DIAG2

#include <errno.h>

#ifdef DIAG2
static char diagmsg[256];
#endif

#include "all.h"
#include <fcntl.h>
#include "../inc/types.h"
#include "../inc/time.h"

#ifdef AIX
#include <sys/select.h>
#endif

#define SELECT_TIMEOUT 50000
/* #define SELECT_TIMEOUT 200000 */

static int double_count = 0;

/* Maximum number of permanent data connections allowed */

/* 102794
#define MAX_PERM 16
*/
#define MAX_PERM 32

static int nperm = 0;	/* Current number of permanent connections */

static int cperm[MAX_PROCESSES+1];	 /* 0 = temporary, -1 = permanent */

static int nready;		/* Number of sockets ready for reading */

static int nldone[MAX_PROCESSES+1]; /* Number of transfers finished for which
					  direct_xfer msgs are still bouncing
					  around */
static int nrdone[MAX_PROCESSES+1];

/* Set of file descriptors associated with open data connections */

static fd_set fdset;
static struct timeval poll_t;


/************************************************************************/
/*

	Function: acp_init_dxfer

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int acp_init_dxfer()
{
	char *x;
	int nbytes;

	x = (char *) &acp_r_dxfer[0];
	nbytes = (MAX_PROCESSES+1)*sizeof(struct direct_xfer);
	bzero(x,nbytes);

	x = (char *) &acp_l_dxfer[0][0];
	nbytes = MAX_L_XFERS*(MAX_PROCESSES+1)*sizeof(struct direct_xfer);
	bzero(x,nbytes);

	x = (char *) &acp_dxfer_ptr[0];
	nbytes = (MAX_PROCESSES+1)*sizeof(int);
	bzero(x,nbytes);

	x = (char *) &acp_dxfer_nleft[0];
	nbytes = (MAX_PROCESSES+1)*sizeof(int);
	bzero(x,nbytes);

	x = (char *) &nldone[0];
	nbytes = (MAX_PROCESSES+1)*sizeof(int);
	bzero(x,nbytes);

	x = (char *) &nrdone[0];
	nbytes = (MAX_PROCESSES+1)*sizeof(int);
	bzero(x,nbytes);

	FD_ZERO(&fdset);
	poll_t.tv_sec = 0;
	poll_t.tv_usec = SELECT_TIMEOUT;
}

/************************************************************************/
/*

	Function:  accept_data_conn

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int accept_data_conn(lpn)
int lpn;
{
	int lpn1;
	int sock;
	char msg[20];
	int permreq;

	/* Check to see if we are already connected */

	if (acp_data_sock[lpn] != -1) return 0;

start:

	/* Accept the connection request on our passive socket */

/* Wait a maximum of 30 seconds for the connection (v2.9) */

	if (socket_ready(acp_dsock,30) != 1) {
		return -1;
	}

	sock = accept_connection(acp_dsock);
	if (sock < 0) return -1;

	/* Find out who this is */

	if (read_counted(sock,msg,20) < 0) {
		socket_close(sock);
		return -1;
	}
	lpn1 = atoi(msg);

	/* Get the request for either a permanent or temporary socket */

	if (read_counted(sock,msg,20) < 0) {
		socket_close(sock);
		return -1;
	}
	permreq = 0;
	if (msg[0] == 'p') permreq = -1;

	/* Check to see if we can keep another permanent connection */

	cperm[lpn1] = 0;	/* Temporary */
	if (permreq && (nperm < MAX_PERM)) {
		nperm++;
		cperm[lpn1] = -1;
	}

	/* Only respond if a permanent connection request was made */

	if (permreq) {
		if (!cperm[lpn1]) msg[0] = 't';
		if (send_counted(sock,msg) < 0) {
			socket_close(sock);
			return -1;
		}
	}

	/* Mark the socket non-blocking */
	/*
	if (fcntl(sock,F_SETFL,FNDELAY) < 0) return -1;
*/
	/* Store the socket's file descriptor */

	acp_data_sock[lpn1] = sock;
	FD_SET(sock,&fdset);

#ifdef DIAG2
	if (cperm[lpn1]) {
		sprintf(diagmsg,"Accepted permanent connection from process %d\n",lpn1);
	}
	else {
		sprintf(diagmsg,"Accepted temporary connection from process %d\n",lpn1);
	}
	dbgbuf_insert(diagmsg);
#endif

	/* Check lpn1 against the lpn expected */

	if (lpn1 != lpn) goto start;

	return 0;
}

/************************************************************************/
/*

	Function:  initiate_data_conn

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int initiate_data_conn(lpn)
int lpn;
{
	char *hostname;
	int port,sock;
	int i;

	char msg[20];
	int permreq;

	/* Check to see if we are already connected */

	if (acp_data_sock[lpn] != -1) return 0;

	/* Get the hostname and port */

	i = acp_pmap[lpn-FIRST_USER_LPN].shmid;
	hostname = acp_smap[i].hostname;
	port = acp_port[lpn];
	if (port == -1) return -1;

	/* Create an active socket */

	sock = create_active_socket();
	if (sock == -1) return -1;

	/* Initiate a connection with the remote user process */

	if (initiate_connection(sock,hostname,port) == -1) {
		socket_close(sock);
		return -1;
	}

	/* Send our lpn */

	sprintf(msg,"%d",acp_local_lpn);
	if (send_counted(sock,msg) < 0) {
		socket_close(sock);
		return -1;
	}

	/* Check to see if we can keep another permanent connection */

	permreq = 0;
	cperm[lpn] = 0;	/* Temporary */
	if (nperm < MAX_PERM) {
		nperm++;
		cperm[lpn] = -1;
		permreq = -1;
	}

	/* Send a 'p' for permanent or a 't' for temporary */

	msg[0] = 't';
	msg[1] = '\0';
	if (permreq) msg[0] = 'p';
	if (send_counted(sock,msg) < 0) {
		socket_close(sock);
		return -1;
	}

	/* If we sent a 'p', look for a response */

	if (permreq) {
		if (read_counted(sock,msg,20) < 0) {
			socket_close(sock);
			return -1;
		}
		if (msg[0] == 't') cperm[lpn] = 0;
	}
#ifdef DIAG2
	if (cperm[lpn]) {
		sprintf(diagmsg,"Made permanent connection to process %d\n",lpn);
	}
	else {
		sprintf(diagmsg,"Made temporary connection to process %d\n",lpn);
	}
	dbgbuf_insert(diagmsg);
#endif

	/* Mark the socket non-blocking */
	/*
	if (fcntl(sock,F_SETFL,FNDELAY) < 0) return -1;
*/

	acp_data_sock[lpn] = sock;
	FD_SET(sock,&fdset);
	return 0;
}

/************************************************************************/
/*

	Function:  open_data_conn

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int open_data_conn(lpn)
int lpn;
{

	/* Check to see if we are already connected */

	if (acp_data_sock[lpn] != -1) return 0;

	/* If our lpn is lower than the remote lpn, initiate the connection */

	if (acp_local_lpn < lpn) {
		return initiate_data_conn(lpn);
	}
	else {
		return accept_data_conn(lpn);
	}
}

/************************************************************************/
/*

	Function:  close_data_conn

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int close_data_conn(lpn,force)
int lpn,force;
{
	if (force || !cperm[lpn]) {
#ifdef DIAG2
		sprintf(diagmsg,"Closing connection to %d\n",lpn);
		dbgbuf_insert(diagmsg);
#endif
		socket_close(acp_data_sock[lpn]);
		FD_CLR(acp_data_sock[lpn],&fdset);
		acp_data_sock[lpn] = -1;
	}
	return 0;
}

/************************************************************************/
/*

	Function:  send_go_ahead

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int send_go_ahead(lpn,x,origin)
int lpn,origin;
struct direct_xfer *x;
{
	char msg[20];

	set_suppress(-1);

	sprintf(msg,"%d",DX_GO_MSG);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

	/* Will need to reverse origin when read at the other end */

	sprintf(msg,"%d",origin);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Sent go msg, lpn=%d, origin=%d\n",lpn,origin);
	dbgbuf_insert(diagmsg);
#endif

	restore_suppress();

	x->state = XFER_NEED_DATA;
	return 0;
}

/************************************************************************/
/*

	Function:  send_repeat_go

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int send_repeat_go(lpn,x,origin)
int lpn,origin;
struct direct_xfer *x;
{
	char msg[20];

	set_suppress(-1);

	sprintf(msg,"%d",DX_REPEAT_GO_MSG);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

	sprintf(msg,"%d",origin);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Sent repeat go, lpn=%d, origin=%d\n",lpn,origin);
	dbgbuf_insert(diagmsg);
#endif

	restore_suppress();
	x->state = XFER_NEED_GO;
	return 0;
}

/************************************************************************/
/*

	Function:  send_data

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int send_data(lpn,x,origin,done)
int lpn,origin,*done;
struct direct_xfer *x;
{
	char msg[20];

	static int count = 0;

#ifdef DIAG2
	sprintf(diagmsg,"Sending data, lpn=%d, origin=%d, nbytes=%d\n"
	    ,lpn,origin,x->nbytes);
	dbgbuf_insert(diagmsg);
#endif

	set_suppress(-1);				/* Suppress error messages */

/* AUTO-RECONNECT: Un-comment this code to test auto-reconnect mechanism. */
/*
	count++;
	if (count == 4) {
		socket_close(acp_data_sock[lpn]);
	}
*/
	sprintf(msg,"%d",DX_DATA_MSG);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

	sprintf(msg,"%d",origin);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

	if (socket_n_send(acp_data_sock[lpn],x->buf,x->nbytes) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

	restore_suppress();

	*done = -1;
	x->state = XFER_DONE;

/* A value of 0 in x->op indicates no-operation. */

	if (x->op == 1) {
		(*(x->done))++;
	}
	else if (x->op == 2) {
		*(x->done) = -1;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Sent data, lpn=%d, origin=%d, nbytes=%d\n"
	    ,lpn,origin,x->nbytes);
	dbgbuf_insert(diagmsg);
#endif
	return 0;
}

/************************************************************************/
/*

	Function:  handle_data

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int handle_data(lpn,lx,rx,ldone,rdone)
int lpn,*ldone,*rdone;
struct direct_xfer *lx,*rx;
{
	char msg[20];
	int origin;
	struct direct_xfer *x;
	int *done;

	set_suppress(-1);

	if (read_counted(acp_data_sock[lpn],msg,20) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 */
		restore_suppress();
		return -1;
	}
	if (sscanf(msg,"%d",&origin) != 1) abort();

	/* Reverse the origin */

	if (origin == XFER_LOCAL_ORIGIN) {
		origin = XFER_REMOTE_ORIGIN;
		x = rx;
		done = rdone;
	}
	else {
		origin = XFER_LOCAL_ORIGIN;
		x = lx;
		done = ldone;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Reading data, lpn=%d, nbytes=%d, origin=%d\n",lpn,x->nbytes,
	    origin);
	dbgbuf_insert(diagmsg);
#endif

	/* Read the data */

	if (x->need_to_clear) {
		if (socket_n_sum(acp_data_sock[lpn],x->buf,x->nbytes) < 0) {
			acp_data_sock[lpn] = -1;
			restore_suppress();
			return -1;
		}
	}
	else {
		if (socket_n_read(acp_data_sock[lpn],x->buf,x->nbytes) < 0) {
			acp_data_sock[lpn] = -1;
			restore_suppress();
			return -1;
		}
	}


	restore_suppress();


#ifdef DIAG2
	sprintf(diagmsg,"Got data, lpn=%d, nbytes=%d, origin=%d\n",lpn,x->nbytes,
	    origin);
	dbgbuf_insert(diagmsg);
#endif

	/* If this was a synchronous or asynchronous acp_send, indicate that the
   block has arrived */

	if (origin == XFER_REMOTE_ORIGIN) {
#ifdef DIAG2
		sprintf(diagmsg,"Local block arrived: blockno=%d\n",x->blockno);
		dbgbuf_insert(diagmsg);
#endif
		acp_local_block_arrived1(x->blockno,x->nbytes);
	}

	/* Decrement the number of pending asynchronous gets */
	/* Only if this was an asynchronous get */

	if ((origin == XFER_LOCAL_ORIGIN) && (x->aflag)) {
#ifdef DIAG2
		sprintf(diagmsg,"Remote block arrived: lpn=%d, blockno=%d\n",lpn,x->blockno);
		dbgbuf_insert(diagmsg);
#endif
		/*
	acp_dec_pget(x->blockno);
	*/
		remote_block_arrived1(lpn,x->blockno,x->nbytes);
	}

	/* Indicate that the transfer has finished */

	*done = -1;
	x->state = XFER_DONE;

/* A value of 0 in x->op indicates no-operation. */

	if (x->op == 1) {
		(*(x->done))++;
	}
	else if (x->op == 2) {
		*(x->done) = -1;
	}
	return 0;
}

/************************************************************************/
/*

	Function:  handle_info

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int handle_info(lpn,rx)
int lpn;
struct direct_xfer *rx;
{
	char msg[80];
	int t,blockno,offset,nbytes;
	struct acp_data_block *block;
	struct direct_xfer *x;

#ifdef DIAG2
	sprintf(diagmsg,"Getting xfer info, lpn=%d\n",lpn);
	dbgbuf_insert(diagmsg);
#endif

	x = rx;
	if (rx == NULL) {
		rx = &acp_r_dxfer[lpn];
	}

	/* Get the data transfer information */

	set_suppress(-1);

	if (read_counted(acp_data_sock[lpn],msg,80) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 */
		restore_suppress();
		return -1;
	}

	restore_suppress();

	if (sscanf(msg,"%d%d%d%d",&t,&blockno,&offset,&nbytes) != 4) abort();
	rx->blockno = blockno;
	rx->offset = offset;
	rx->nbytes = nbytes;

#ifdef DIAG2
	sprintf(diagmsg,"Got xfer info: blockno=%d, offset=%d, nbytes=%d\n",blockno,
	    offset,nbytes);
	dbgbuf_insert(diagmsg);
#endif

	/* Reverse the transfer type */

	if (t == XFER_GET) {
#ifdef DIAG2
		sprintf(diagmsg,"Got xfer info: this process will send");
		dbgbuf_insert(diagmsg);
#endif
		t = XFER_SEND;
	}
	else {
#ifdef DIAG2
		sprintf(diagmsg,"Got xfer info: this process will get");
		dbgbuf_insert(diagmsg);
#endif
		t = XFER_GET;
	}
	rx->type = t;

	/* Find the address of the data block and add the offset */

	block = acp_find_block(acp_blocks,blockno);
	if (block == NULL) abort();
	rx->buf = block->buf + offset;

	/* If we are to receive data, send the go message, otherwise wait for go */

	if (rx->type == XFER_GET) {
		if (x != NULL) {
			return send_go_ahead(lpn,rx,XFER_REMOTE_ORIGIN);
		}
		else {
			rx->state = XFER_SEND_GO;
			return 0;
		}
	}
	else {
		rx->state = XFER_NEED_GO;
		return 0;
	}
}

/************************************************************************/
/*

	Function:  handle_repeat_go

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int handle_repeat_go(lpn,lx,rx)
int lpn;
struct direct_xfer *lx,*rx;
{
	char msg[20];
	int origin;

	/* Get the origin */

	set_suppress(-1);

	if (read_counted(acp_data_sock[lpn],msg,20) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 */
		restore_suppress();
		return -1;
	}
	if (sscanf(msg,"%d",&origin) != 1) abort();

	restore_suppress();

	/* Reverse the origin */

	if (origin == XFER_LOCAL_ORIGIN) {
		origin = XFER_REMOTE_ORIGIN;
	}
	else {
		origin = XFER_LOCAL_ORIGIN;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Received repeat-go, lpn=%d, origin=%d\n",lpn,origin);
	dbgbuf_insert(diagmsg);
#endif

	/* If we are to receive data, send the go message */

	if (origin == XFER_LOCAL_ORIGIN) {
		return send_go_ahead(lpn,lx,origin);
	}
	else {
		return send_go_ahead(lpn,rx,origin);
	}
}

/************************************************************************/
/*

	Function:  handle_go

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int handle_go(lpn,lx,rx,ldone,rdone)
int lpn,*ldone,*rdone;
struct direct_xfer *lx,*rx;
{
	char msg[20];
	int origin;
	struct direct_xfer *x;
	int status;

	/* Get the information */

	set_suppress(-1);

	if (read_counted(acp_data_sock[lpn],msg,20) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 */
		restore_suppress();
		return -1;
	}
	if (sscanf(msg,"%d",&origin) != 1) abort();
	if ((origin != XFER_LOCAL_ORIGIN) && (origin != XFER_REMOTE_ORIGIN))
		abort();

	restore_suppress();

	/* Reverse the origin */

	if (origin == XFER_LOCAL_ORIGIN) {
		origin = XFER_REMOTE_ORIGIN;
	}
	else {
		origin = XFER_LOCAL_ORIGIN;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Received go-ahead, lpn=%d, origin=%d\n",lpn,origin);
	dbgbuf_insert(diagmsg);
#endif

	/* If another transfer is simultaneously active, send a repeat-go if we
   are the larger lpn.  Otherwise, send the data. */

	if (origin == XFER_REMOTE_ORIGIN) {
		if ((lx != NULL) && (lx->state != XFER_DONE)) {
			x =  rx;
			if (rx == NULL) {

				/* In this case, the DX_INFO_MSG and DX_GO_MSG have both arrived before
   the message from the shared memory manager. */

				rx = &acp_r_dxfer[lpn];
#ifdef DIAG2
				sprintf(diagmsg,"handle_go: checkpoint1");
				dbgbuf_insert(diagmsg);
#endif

			}

			if (lpn < acp_local_lpn) {
				return send_repeat_go(lpn,rx,origin);
			}
			else {
#ifdef DIAG2
				sprintf(diagmsg,"handle_go: checkpoint2");
				dbgbuf_insert(diagmsg);
#endif
				status = send_data(lpn,rx,origin,rdone);
				if (x == NULL) {
					rx->state = XFER_START;
					nrdone[lpn]++;
#ifdef DIAG2
					sprintf(diagmsg,"handle_go: nrdone=%d, lpn=%d",nrdone[lpn],lpn);
					dbgbuf_insert(diagmsg);
#endif
				}
				return status;
			}
		}
		else {
#ifdef DIAG2
			sprintf(diagmsg,"handle_go: checkpoint3");
			dbgbuf_insert(diagmsg);
#endif
			return send_data(lpn,rx,origin,rdone);
		}
	}
	else {
		if ((rx != NULL) && (rx->state != XFER_DONE)) {
			if (lpn < acp_local_lpn) {
				return send_repeat_go(lpn,lx,origin);
			}
			else {
#ifdef DIAG2
				sprintf(diagmsg,"handle_go: checkpoint4");
				dbgbuf_insert(diagmsg);
#endif
				return send_data(lpn,lx,origin,ldone);
			}
		}
		else {
#ifdef DIAG2
			sprintf(diagmsg,"handle_go: checkpoint5");
			dbgbuf_insert(diagmsg);
#endif
			return send_data(lpn,lx,origin,ldone);
		}
	}

}

/************************************************************************/
/*

	Function:  read_data_channel

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int read_data_channel(lpn,lx,rx,ldone,rdone)
int lpn,*ldone,*rdone;
struct direct_xfer *lx,*rx;
{
	char msg[20];
	int msgtyp;

	/* Get the type of message */

	set_suppress(-1);
	if (read_counted(acp_data_sock[lpn],msg,20) < 0) {
		restore_suppress();
		goto lostconn;
	}
	restore_suppress();

	msgtyp = atoi(msg);

#ifdef DIAG2
	sprintf(diagmsg,"Message type = %d, LPN = %d\n",msgtyp,lpn);
	dbgbuf_insert(diagmsg);
#endif

	if (msgtyp == DX_GO_MSG) {
		if (handle_go(lpn,lx,rx,ldone,rdone) < 0) goto lostconn;
	}
	else if (msgtyp == DX_INFO_MSG) {
		if ((rx != NULL) &&
		    (rx->state == XFER_DONE)) {
			nrdone[lpn]++;
#ifdef DIAG2
			sprintf(diagmsg,"read_data_channel: nrdone=%d, lpn=%d",nrdone[lpn],
			    lpn);
			dbgbuf_insert(diagmsg);
#endif
		}
		if (handle_info(lpn,rx) < 0) goto lostconn;
	}
	else if (msgtyp == DX_DATA_MSG) {
		if (handle_data(lpn,lx,rx,ldone,rdone) < 0) goto lostconn;
	}
	else if (msgtyp == DX_REPEAT_GO_MSG) {
		if (handle_repeat_go(lpn,lx,rx) < 0) goto lostconn;
	}
	else {
		abort();	/* Bug in code */
	}

	return;

lostconn:

	acp_data_sock[lpn] = -1;

#ifdef DIAG2
	sprintf(diagmsg,"Lost connection!  LPN = %d\n",lpn);
	dbgbuf_insert(diagmsg);
#endif

	if (lx != NULL) lx->state = XFER_NEED_CONNECT;
	if (rx != NULL) rx->state = XFER_NEED_CONNECT;
	return;
}

/************************************************************************/
/*

	Function:  send_xfer_info

	Description:

	Arguments:

	Function Value:  -1 if connection was lost, otherwise 0.

*/
/************************************************************************/

static int send_xfer_info(lpn,x)
int lpn;
struct direct_xfer *x;
{
	char msg[80];

	if (acp_dxfer_nleft[lpn] <= 0) abort();	/* Bug in program */

	set_suppress(-1);

	sprintf(msg,"%d",DX_INFO_MSG);
	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}


	/* Will need to reverse type at the other end */

	sprintf(msg,"%d %d %d %d",x->type,x->blockno,x->offset,x->nbytes);

	if (send_counted(acp_data_sock[lpn],msg) < 0) {
		acp_data_sock[lpn] = -1;		/* 2.8.1 OK */
		x->state = XFER_NEED_CONNECT;
		restore_suppress();
		return -1;
	}

#ifdef DIAG2
	sprintf(diagmsg,"Sent xfer info, type=%d, block=%d, nbytes=%d"
	    ,x->type,x->blockno,x->nbytes);
	dbgbuf_insert(diagmsg);
#endif

	if (x->type == XFER_GET) {
		x->state = XFER_SEND_GO;
	}
	else if (x->type == XFER_SEND) {
		x->state = XFER_NEED_GO;
	}
	else {
		abort();
	}

	restore_suppress();

	return 0;
}

/************************************************************************/
/*

	Function:  data_ready

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int data_ready(lpn)
int lpn;
{
	fd_set fds;

	/* This could be optimized so that select is not called every time. */

	fds = fdset;
	retry_select:
	nready = select(FD_SETSIZE,&fds,(fd_set *)0,(fd_set *)0,&poll_t);
	if (nready == -1) {
        	if (errno == EINTR)
        		goto retry_select;
        	else
                	abort();
    	}
	if (nready == 0) return 0;
#ifdef DIAG2
	sprintf(diagmsg,"fds=%d",fds.fds_bits[0]);
	dbgbuf_insert(diagmsg);
#endif

	if (lpn != -1) {
		if (acp_data_sock[lpn] == -1) abort();
		if (FD_ISSET(acp_data_sock[lpn],&fds)) {
#ifdef DIAG2
			sprintf(diagmsg,"data ready from lpn %d",lpn);
			dbgbuf_insert(diagmsg);
#endif
			return -1;
		}
		else {
			return 0;
		}
	}
	else {
		return 0;
	}

}

/************************************************************************/
/*

	Function:  do_direct_xfer

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int reading_state(s)
int s;
{
	if (s == XFER_NEED_INFO) return -1;
	if (s == XFER_NEED_GO) return -1;
	if (s == XFER_NEED_DATA) return -1;
	return 0;
}

static int non_sending_state(s)
int s;
{
	if (s == XFER_SEND_INFO) return 0;
	if (s == XFER_SEND_GO) return 0;
	if (s == XFER_SEND_DATA) return 0;
	return -1;
}

static int check_state(s)
int s;
{
	if ((s <= XFER_NEED_CONNECT) || (s >= XFER_DONE)) abort();
}

static int do_direct_xfer(lpn,lx,rx,ldone,rdone)
int lpn;
struct direct_xfer *lx,*rx;
int *ldone,*rdone;
{
	int lstate,rstate;		/* State of data transfer, -1 means no state */

	rstate = lstate = -1;
	if (lx != NULL) {
		lstate = lx->state;
		*rdone = 0;
		check_state(lstate);
	}
	if (rx != NULL) {
		rstate = rx->state;
		*ldone = 0;
		check_state(rstate);
	}

	while (1) {

		if (lstate == XFER_DONE) lstate = -1;
		if (rstate == XFER_DONE) rstate = -1;

		if ((lstate == -1) && (rstate == -1)) return 0;

#ifdef DIAG2
		sprintf(diagmsg,"(1) lstate=%d, rstate=%d\n",lstate,rstate);
		dbgbuf_insert(diagmsg);
#endif

		if (lstate == XFER_SEND_INFO) {
			send_xfer_info(lpn,lx);
			lstate = lx->state;
		}
		else if (lstate == XFER_SEND_GO) {
			send_go_ahead(lpn,lx,XFER_LOCAL_ORIGIN);
			lstate = lx->state;
		}

		if (lstate == XFER_NEED_CONNECT) return 0;

		if (rstate == XFER_SEND_GO) {
			send_go_ahead(lpn,rx,XFER_REMOTE_ORIGIN);
			rstate = rx->state;
		}

		if (rstate == XFER_NEED_CONNECT) return 0;

		if (reading_state(lstate) || reading_state(rstate)) {
			if (data_ready(lpn)) {
				read_data_channel(lpn,lx,rx,ldone,rdone);
			}
			else {
				/*
		if (double_count < 4) {
			poll_t.tv_usec *= 2;
			while (poll_t.tv_usec >= 1000000) {
			poll_t.tv_usec -= 1000000;
			poll_t.tv_sec += 1;
			}
			double_count++;
		}
*/
#ifdef DIAG2
				sprintf(diagmsg,"(2) lstate=%d, rstate=%d\n",lstate,rstate);
				dbgbuf_insert(diagmsg);
#endif
				if (non_sending_state(lstate) && 
				    non_sending_state(rstate)) {
					return 0;
				}
				else {
					continue;
				}
			}
			if (rx != NULL) rstate = rx->state;
			if (lx != NULL) lstate = lx->state;
		}

#ifdef DIAG2
		sprintf(diagmsg,"(3) lstate=%d, rstate=%d\n",lstate,rstate);
		dbgbuf_insert(diagmsg);
#endif

		if ((rstate == XFER_NEED_CONNECT) ||
		    (lstate == XFER_NEED_CONNECT)) {
			return 0;
		}

		if (lstate == XFER_SEND_DATA) {
			send_data(lpn,lx,XFER_LOCAL_ORIGIN,ldone);
			lstate = lx->state;
		}

		if (lstate == XFER_NEED_CONNECT) return 0;

		if (rstate == XFER_SEND_DATA) {
			send_data(lpn,rx,XFER_REMOTE_ORIGIN,rdone);
			rstate = rx->state;
		}

		if (rstate == XFER_NEED_CONNECT) return 0;

	}

}

/************************************************************************/
/*

	Function:  check_connection

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int check_connection(lpn,x,origin)
int lpn,origin;
struct direct_xfer *x;
{

	/* Check remotely initiated transfers */

	if (origin == XFER_REMOTE_ORIGIN) {
		if (x->state == XFER_NEED_CONNECT) {
			if (acp_data_sock[lpn] == -1) {
				if (open_data_conn(lpn) == 0) {
					x->state = XFER_NEED_INFO;
				}
			}
			else {
				x->state = XFER_NEED_INFO;
			}
		}
	}
	else {

		/* Check locally initiated transfers */

		if (x->state == XFER_NEED_CONNECT) {
			if (acp_data_sock[lpn] == -1) {
				if (open_data_conn(lpn) == 0) {
					x->state = XFER_SEND_INFO;
				}
			}
			else {
				x->state = XFER_SEND_INFO;
			}
		}

	}
}

/************************************************************************/
/*

	Function: abort_direct_xfer

	Description: This routine is called when a connection attempt with a
		 remote process fails.

	Arguments:

	Function Value:

*/
/************************************************************************/

static int abort_direct_xfer(x)
struct direct_xfer *x;
{

	static int count = 0;

	count++;
	if (count < 10) return -1;

/* This routine is new in v2.9 */

/* If x->op equals zero, then abort & done do not point to anything */

	if (x->op != 0) (*(x->abort))++;
	if (x->op == 1) {
		(*(x->done))++;
	}
	else if (x->op == 2) {
		*(x->done) = -1;
	}

	return 0;
}

/************************************************************************/
/*

	Function:  acp_handle_direct_xfer

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int acp_handle_direct_xfer(m,q)
cps_message *m;
struct qelem *q;
{
	int lpn;
	int i,j;

	int origin;

	int src_lpn;

	cps_message *m1;
	struct qelem *q1;

	struct direct_xfer *lx,*rx;
	int ldone,rdone;			/* 0 = transfer not done, -1 = done,
					   if lx = NULL, ldone means nothing,
						   same for rdone. */


	src_lpn = m->src_lpn;
	cps_get_direct_xfer(m,&lpn);

#ifdef DIAG2
	sprintf(diagmsg,"handle_direct_xfer: lpn=%d, src_lpn=%d",lpn,src_lpn);
	dbgbuf_insert(diagmsg);
#endif

	if (src_lpn != acp_local_lpn) {	/* Remotely initiated transfer */

		if (nrdone[src_lpn] > 0) {
#ifdef DIAG2
			sprintf(diagmsg,"handle_direct_xfer: nrdone=%d, return1",
			    nrdone[src_lpn]);
			dbgbuf_insert(diagmsg);
#endif
			nrdone[src_lpn]--;
			return 0;
		}

		origin = XFER_REMOTE_ORIGIN;

		/* Check to see if the transfer needs to be initiated */

		rx = &acp_r_dxfer[src_lpn];

		if (rx->state == XFER_START) {

			rx->state = XFER_NEED_CONNECT;
			rx->op = 0;
			rx->need_to_clear = 0;

		}
		else if (rx->state < XFER_START) {	/* Bug in code */
			abort();
		}

	}
else {	/* This transfer was initiated locally and should already
			   contain state information */

		if (nldone[lpn] > 0) {
#ifdef DIAG2
			sprintf(diagmsg,"handle_direct_xfer: nldone=%d, return2",nldone[lpn]);
			dbgbuf_insert(diagmsg);
#endif
			nldone[lpn]--;
			return 0;
		}

		origin = XFER_LOCAL_ORIGIN;

		if (acp_dxfer_nleft[lpn] <= 0) {	/* Bug in code */
			abort();
		}

		i = acp_dxfer_ptr[lpn];
		lx = &acp_l_dxfer[lpn][i];

		if (lx->state <= XFER_START) {		/* Bug in code! */
			abort();
		}

	}

	/* Establish connections (changes state from XFER_NEED_CONNECT to 
   either XFER_NEED_INFO or XFER_SEND_INFO) */

	for (i=1; i<=num_user_proc; i++) {
		check_connection(i,&acp_r_dxfer[i],XFER_REMOTE_ORIGIN);
		if (acp_r_dxfer[i].state == XFER_NEED_CONNECT) {
			if (abort_direct_xfer(&acp_r_dxfer[i]) == 0) {
				nrdone[i]++;	/* v2.9 */
			}
		}
		if (acp_dxfer_nleft[i] > 0) {
			j = acp_dxfer_ptr[i];
			check_connection(i,&acp_l_dxfer[i][j],XFER_LOCAL_ORIGIN);
			if (acp_l_dxfer[i][j].state == XFER_NEED_CONNECT) {
				if (abort_direct_xfer(&acp_l_dxfer[i][j]) == 0) {
					nldone[i]++; /* v2.9 */
					acp_dxfer_nleft[i] = 0;
				}
			}
		}
	}

	/* Attempt to complete each possible transfer */

	for (i=1; i<=num_user_proc; i++) {

		/* If there is no connection, go to the next process */

		if (acp_data_sock[i] == -1) continue;

		/* Determine if a locally initiated or remote initiated transfer needs
   to be completed. */

		lx = NULL;
		rx = NULL;
		ldone = 0;
		rdone = 0;

		j = acp_dxfer_ptr[i];
		if ((acp_dxfer_nleft[i] > 0) && 
		    (acp_l_dxfer[i][j].state > XFER_NEED_CONNECT)) {
			lx = &acp_l_dxfer[i][j];
		}
		if (acp_r_dxfer[i].state > XFER_NEED_CONNECT) {
			rx = &acp_r_dxfer[i];
		}

		/* If nothing needs to be done, continue with the next process. */

		if ((lx == NULL) && (rx == NULL)) continue;

		/* Attempt to complete either the locally initiated transfer, the
   remotely initiated transfer, or both. */

		do_direct_xfer(i,lx,rx,&ldone,&rdone);

		/* Check to see if the remotely initiated transfer has completed */

		if ((rx != NULL) && rdone) {

			/* Reset the transfer state */

			rx->state = XFER_START;

			/* Close the connection if its temporary and a locally initiated transfer is
   not in progress. */

			if ((lx == NULL) || 
			    ((lx != NULL) && ldone)) close_data_conn(i,0);

#ifdef DIAG2
			sprintf(diagmsg,"handle_direct_xfer: incrementing nrdone=%d, i=%d"
			    ,nrdone[i],i);
			dbgbuf_insert(diagmsg);
#endif

			nrdone[i]++;
		}

		/* Check to see if the locally initiated transfer has completed */

		if ((lx != NULL) && ldone) {

			/* Reset the transfer state */

			lx->state = XFER_START;

			/* Eliminate the bouncing message if the "last" flag is set. */

			if (lx->lflag) {
				nldone[i]++;
			}

#ifdef DIAG2
			sprintf(diagmsg,"handle_direct_xfer: incrementing nldone=%d, i=%d",
			    nldone[i],i);
			dbgbuf_insert(diagmsg);
#endif

			/* Check to see if there are more locally initiated transfers in the list */

			acp_dxfer_nleft[i]--;
			if (acp_dxfer_nleft[i] != 0) {

				/* Point to the next locally initiated data transfer */

				acp_dxfer_ptr[i]++;
				if (acp_dxfer_ptr[i] == MAX_L_XFERS) acp_dxfer_ptr[i] = 0;

				/* Start the next transfer in the list */

				j = acp_dxfer_ptr[i];
				check_connection(i,&acp_l_dxfer[i][j],XFER_LOCAL_ORIGIN);

				/* Tell the remote process to look for another transfer */

				q1 = acp_q_pop1(S_FREE_Q);
				m1 = (cps_message *) SMALL_MSG(q1);
				cps_bld_direct_xfer(m1,acp_local_lpn,i,acp_local_lpn);
				acp_q_push1(acp_shm_qnum,q1);
				acp_signal_shm();

				i--;	/* Re-do loop with same i-value */

#ifdef DIAG2
				sprintf(diagmsg,"handle_direct_xfer: re-do loop, i=%d",i);
				dbgbuf_insert(diagmsg);
#endif

			}
			else {

				acp_dxfer_ptr[i] = 0;

				/* Close the connection if its temporary and a remotely initiated transfer is
   not in progress. */

				if ((rx == NULL) || 
				    ((rx != NULL) && rdone)) close_data_conn(i,0);

			}

		}

	}

	/* Determine if the MSG_DIRECT_XFER message should come back or not. */

	if ((origin == XFER_LOCAL_ORIGIN) && (nldone[lpn] > 0)) {

		nldone[lpn]--;

#ifdef DIAG2
		sprintf(diagmsg,"handle_direct_xfer: dec(1) nldone=%d, lpn=%d",
		    nldone[lpn],lpn);
		dbgbuf_insert(diagmsg);
#endif
		return 0;
	}
	else if ((origin == XFER_REMOTE_ORIGIN) && (nrdone[src_lpn] > 0)) {
		nrdone[src_lpn]--;
#ifdef DIAG2
		sprintf(diagmsg,"handle_direct_xfer: dec(2) nrdone=%d, lpn=%d",
		    nldone[lpn],lpn);
		dbgbuf_insert(diagmsg);
#endif
		return 0;
	}
	else {
		acp_q_push1(acp_shm_qnum,q);	/* Message will come back to us */
		acp_signal_shm();
#ifdef DIAG2
		sprintf(diagmsg,"handle_direct_xfer: bounce message");
		dbgbuf_insert(diagmsg);
#endif
		return -1;
	}
}
