/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				INIT.C													*/
/*																		*/
/*																		*/
/************************************************************************/

#define DIAG2

/*****************/
/* Include Files */
/*****************/

#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include "../inc/types.h"
#include "../inc/socket.h"
#include "../inc/perr.h"
#include "../inc/param.h"
#include "../inc/shm.h"
#include "../inc/xfunc.h"
#include "../inc/cps_q.h"
#include "../inc/subs.h"
#include "../inc/version.h"
#include "user_param.h"
#include "user_struct.h"
#include "user_globals.h"

#include <stdarg.h>

#define STARTFD 2

static int shared_mem_id;

extern int assigned_port_number;

extern void acp_handle_sigusr1();
extern void acp_handle_sigusr2();

int acp_get_shmid() { return(shared_mem_id); }

display_error(msg)
char *msg;
{
	again1:
	if (write(STARTFD,msg,strlen(msg)) == -1) {
		if (errno == EINTR) goto again1;
		abort();
	}
	again2:
	if (write(STARTFD,"\n",1) == -1) {
		if (errno == EINTR) goto again2;
		abort();
	}
}

void acp_handle_sigill_(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	exit(0);
}
 
void acp_handle_sigill(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	exit(0);
}

void acp_handle_sigterm(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	cps_sighold();	/*	Since this process is exiting, ignore any messages
					  	not yet received.  The printf below somehow lets
					  	the SIGUSR1 signal through, which can cause the
					  	process to get SEMLIB/E6 if the semaphores
						have already been removed. */

	printf("Stopped by the Shared Memory Manager\n");
	exit(0);
}

void acp_handle_sigterm_(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	acp_handle_sigterm(sig,code,sc);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_init

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  None. If anything fails in this, the program will exit.

								  */
/******************************************************************/

cps_init_() { cps_init(); }

cps_init()
{
	char *base_addr,*addr;
	int shmid;
	int i,j,p;
	char msg[80];
	FILE *fopen();
	char fname[32];
	int s;
	struct hostent *hp;

#ifdef DIAG2
	dbgbuf_alloc(100,80);
	dbgbuf_setsig(SIGINT);
#endif

/* Catch the SIGUSR1 signal */

	sig_establish_handler(SIGUSR1,acp_handle_sigusr1);

/* Ignore SIGPIPE */

	sig_establish_handler(SIGPIPE,SIG_IGN);

/* Catch the SIGTERM signal */

	sig_establish_handler(SIGTERM,acp_handle_sigterm);

/* Catch illegal instructions (temporary CDF fix) */

	sig_establish_handler(SIGILL,acp_handle_sigill);

/* Catch the SIGUSR2 signal */

	sig_establish_handler(SIGUSR2,acp_handle_sigusr2);

/* Hold SIGUSR1 */

	cps_sighold();

	gethostname(acp_local_host,64);

	if( ( hp = gethostbyname( acp_local_host ) ) != NULL )
		strcpy( acp_local_host, hp->h_name );

	/* Write out NULL character for neatness of user output */
	sprintf(msg,"%c",'\0');
	write(STARTFD,msg,strlen(msg)+1);

	sprintf(msg,"Process starting on %s %s",acp_local_host,libacp_ident);
	write(STARTFD,msg,strlen(msg)+1);

/* Write this message to synchronize with the Job Manager */

	sprintf(msg,"Cooperative Processes Software %-8s %2s\n",
		cps_version,cps_relcode);
	write(STARTFD,msg,strlen(msg)+1);

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(STARTFD,msg,80) < 0)
		exit(0);
	if (strcmp(msg,"suicide") == 0) exit();

/* Get our lpn */

	if (read_counted(STARTFD,msg,80) < 0)
		exit(0);
	acp_local_lpn = atoi(msg);

/* Get system hooks flag */
	if (read_counted(STARTFD,msg,80) < 0)
		exit(0);
	proc_hooks = atoi(msg);

/* Open a log file for debugging */

#ifdef DIAG
	sprintf(fname,"proc_%d.log",acp_local_lpn);
	acp_fp = fopen(fname,"w");
#endif

/* Send our pid */

	sprintf(msg,"%d",getpid());
	if (send_counted(STARTFD,msg) < 0)
		exit(0);

/* Send our working directory */

	sprintf(msg,"%s", getcwd((char *)NULL, 64) );
	if (send_counted(STARTFD,msg) < 0)
		exit(0);

/* Create a passive socket for accepting connections */

	if ((acp_dsock = create_passive_socket(0)) == -1)
									/* This process will not	 */
		assigned_port_number = -1;	/* participate in direct	 */
									/* data transfers.		 */

/* Send our port number */

	sprintf(msg,"%d",assigned_port_number);
	if (send_counted(STARTFD,msg) < 0)
		exit(0);

/* Initialize user ports & sockets */

	for (i=1; i<=MAX_PROCESSES; i++)
	{
		acp_port[i] = -1;
		acp_data_sock[i] = -1;
	}

/* Get the port numbers of other user processes */

	while (1)
	{
		if (read_counted(STARTFD,msg,80) < 0)
			exit(0);
		sscanf(msg,"%d %d",&i,&p);
		if (i == 0) break;
		acp_port[i] = p;
	}

/* Check to see if we should kill ourselves (may be necessary if JM fails
   to start enough other processes) */

	if (read_counted(STARTFD,msg,80) < 0)
		exit(0);
	if (strcmp(msg,"suicide") == 0)
		exit();

/* Get the shared memory id */

	if (read_counted(STARTFD,msg,80) < 0)
		exit(0);
	shmid = atoi(msg);
	shared_mem_id = shmid;

/* Get class to process map */

	for (i=1; i<=MAX_CLASSES; i++) {
		acp_class_proc[i] = NULL;
		acp_static_cp[i] = NULL;	/* v2.9 */
	}

	getmap:
	if (read_counted(STARTFD,msg,80) < 0)
		exit(0);
	i = atoi(msg);	/* Get class */
	if (i != 0)
	{
		acp_class_proc[i] = create_empty_set( MAX_PROCESSES );
		acp_static_cp[i] = create_empty_set( MAX_PROCESSES );	/* v2.9 */
		getproc:
		if (read_counted(STARTFD,msg,80) < 0)
		{
			exit(0);
		}
		j = atoi(msg);
		if (j != 0)
		{
			if (acp_local_lpn == j) acp_local_class = i;
			add_to_set(j,acp_class_proc[i]);
			add_to_set(j,acp_static_cp[i]);	/* v2.9 */
			goto getproc;
		}
		goto getmap;
	}

/* Attach to the shared memory */

	if ((int)(base_addr = shm_attach(shmid)) == -1)
	{
		printerr(FACILITY,ERROR_MSG,"Failed to attach to shared memory",NULL,1);
		exit(0);
	}

/* Initialize pointers into shared memory */

	addr = base_addr;
	acp_nprocs = (int *) addr;
	addr += 3 * sizeof(int);
	num_user_proc = acp_nprocs[0];
	num_shm_proc = acp_nprocs[1];
	num_local_proc = acp_nprocs[2];
	num_semaphores = num_local_proc + 3;

	acp_semids = (int *) addr;
	addr += 2 * sizeof(int);
	q_semid = acp_semids[0];

	acp_pmap = (struct procmap *) addr;
	addr += num_user_proc * sizeof(struct procmap);
	acp_smap = (struct shmmap *) addr;
	addr += (num_shm_proc + 1) * sizeof(struct shmmap);  /* One for the JM */

	acp_shmem = (struct shmem *) addr;
	addr += SHM_FUNC(num_semaphores) * sizeof(cps_message);
	acp_sshmem = (struct shmem_s *) addr;
	addr += S_SHM_FUNC(num_semaphores) * sizeof(cps_msg);

	qelem = (struct qelem *) addr;
	addr += (SHM_FUNC(num_semaphores) + S_SHM_FUNC(num_semaphores))
		* sizeof(struct qelem);
	q_head = (int *) addr;
	addr += num_semaphores * sizeof(int *);
	q_tail = (int *) addr;
	addr += num_semaphores * sizeof(int *);
	q_block = (int *) addr;

/* Set the pid & local fields in our entry to acp_pmap */

	i = acp_local_lpn - 1;
	acp_pmap[i].local = -1;
	acp_pmap[i].pid = getpid();

	acp_local_qnum = acp_pmap[i].semnum;

/* Record our shm's semaphore number for easy access */

	acp_local_shm = acp_pmap[acp_local_lpn-FIRST_USER_LPN].shmid;
	acp_shm_lpn = acp_local_shm + FIRST_SHM_LPN - 1;
	acp_shm_qnum = acp_smap[acp_local_shm].semnum;

/* Initialize queue information */

	for (i=0; i<=MAX_QUEUE_NUMBER; i++)
	{
		acp_queue_declared[i] = -1;
		acp_queue_bsum[i] = 0;
	}

/* Initialize remote block lists */

	for (i=0; i<num_user_proc; i++)
		acp_remote_blocks[i] = NULL;

/* Tell the semaphore software how many semaphores there are */

	set_nsems(num_semaphores);

/* Initialize explicit message queues */

	acp_init_msg_queue();

/* Initialize array which counts number of pending asynchronous gets */

	acp_init_pget();

/* Initialize direct transfer structures */

	acp_init_dxfer();

/* Tell JM we are done initializing */

	if (send_counted(STARTFD,"done") < 0)
		exit(0);

/* Indicate that this routine has been called */

	acp_init_called = -1;

/* Release SIGUSR1 */

	cps_sigrelease();

	CHECK_CPS_HOOK( FUNC_INIT )
}
