/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnfaus.fnal.gov

			Copyright (c) 1991 Universities Research Association, Inc.	    
			All Rights Reserved


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or	     
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

	Source File: qxfer.c

	Explanation: Asynchronous data transfers

	Exported Routines:


*/

#include "all.h"

static int eoq[33] = { 0,0,0,0,0,0,0,0,0,
			   0,0,0,0,0,0,0,0,
			   0,0,0,0,0,0,0,0,
			   0,0,0,0,0,0,0,0 };

#define NPGETS 25

static struct pending_gets {

	int blockno;
	int count;

} pgets[NPGETS];


int acp_init_pget()
{
	int i;
	for (i=0; i<NPGETS; i++) {
	    pgets[i].blockno = 0;
	    pgets[i].count = 0;
	}
}

int acp_get_npget(blockno)
int blockno;
{
	int i;
	i = blockno % NPGETS;
	while (pgets[i].blockno != blockno) {
	    if (pgets[i].blockno == 0) {
	        pgets[i].blockno = blockno;
	    }
	    else {
	        i++;
	        if (i == NPGETS) i = 0;
	    }
	}
	return pgets[i].count;
}

acp_inc_pget(blockno)
int blockno;
{
	int i;
	i = blockno % NPGETS;
	while (pgets[i].blockno != blockno) {
	    if (pgets[i].blockno == 0) {
	        pgets[i].blockno = blockno;
	    }
	    else {
	        i++;
	        if (i == NPGETS) i = 0;
	    }
	}
	pgets[i].count++;
}

acp_dec_pget(blockno)
int blockno;
{
	int i;
	i = blockno % NPGETS;
	while (pgets[i].blockno != blockno) {
	    if (pgets[i].blockno == 0) {
	        pgets[i].blockno = blockno;
	    }
	    else {
	        i++;
	        if (i == NPGETS) i = 0;
	    }
	}
	pgets[i].count--;
}

acp_zero_pget(blockno)
int blockno;
{
	int i;
	i = blockno % NPGETS;
	while (pgets[i].blockno != blockno) {
	    if (pgets[i].blockno == 0) {
	        pgets[i].blockno = blockno;
	    }
	    else {
	        i++;
	        if (i == NPGETS) i = 0;
	    }
	}
	pgets[i].count = 0;
}

/************************************************************************/
/*

	Function: cps_queue_xfer

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

cps_queue_xfer(addr,nbytes,blockid,flag,bcount)
char *addr;
int *nbytes,*blockid,*flag,*bcount;
{
	struct queued_xfer *qx;

	acp_clear_error_info();
	acp_func = "queue_xfer";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_QUEUE_XFER )

/* Create a data xfer record */

	qx = (struct queued_xfer *) hmalloc(sizeof(struct queued_xfer));
	if (qx == NULL) abort();
	qx->addr = addr;
	qx->nbytes = *nbytes;
	qx->blockid = *blockid;
	qx->flag = flag;
	qx->abort = &acp_num_aborted_xfers;		/* v2.9 */
	qx->bcount = bcount;
	qx->next = NULL;
	if (acp_qxfer_h == NULL) {
		acp_qxfer_h = qx;
	}
	else {
		acp_qxfer_t->next = qx;
	}
	acp_qxfer_t = qx;

}

cps_queue_xfer_(addr,nbytes,blockid,flag,bcount)
char *addr;
int *nbytes,*blockid,*flag,*bcount;
{
	cps_queue_xfer(addr,nbytes,blockid,flag,bcount);
}

/************************************************************************/
/*

	Function:  cps_dq_send

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int cpssta_dq_send(qid,lpn)
int *qid;
int *lpn;
{
	struct active_qxfer *aqx;
	int s;
	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_func = "dq_send";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DQ_SEND )

/* Check the queue number */

	s = acp_check_qnum(*qid);
	if (s != CPS_OK) return(s);

/* Hold SIGUSR1 signals */

	cps_sighold();

/* Create a record to hold the operation (and put it on the list) */

	aqx = (struct active_qxfer *) hmalloc(sizeof(struct active_qxfer));
	if (aqx == NULL) abort();
	aqx->id = acp_next_qx_id++;
	aqx->lpn = -1;
	aqx->lpn_addr = lpn;
	aqx->head = acp_qxfer_h;
	aqx->tail = acp_qxfer_t;
	aqx->next = acp_aqx;
	acp_aqx = aqx;

/* Re-initialize the queue of background data xfers */

	acp_qxfer_h = NULL;
	acp_qxfer_t = NULL;

/* Release SIGUSR1 signals */

	cps_sigrelease();

/* Initiate the operation by sending asynch dequeue message to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_asynch_dq(m,acp_local_lpn,JM_LPN,*qid,aqx->id);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	return CPS_OK;
}

int cpssta_dq_send_(qid,lpn)
int *qid;
int *lpn;
{
	return(cpssta_dq_send(qid,lpn));
}

cps_dq_send(qid,lpn)
int *qid;
int *lpn;
{
	acp_fntype = "cps_";
	cpssta_dq_send(qid,lpn);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_dq_send_(qid,lpn)
int *qid;
int *lpn;
{
	cps_dq_send(qid,lpn);
}

/************************************************************************/
/*

	Function:  cps_dq_get

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int cpssta_dq_get(qid,lpn)
int *qid;
int *lpn;
{
	struct queued_xfer *qx;
	struct active_qxfer *aqx;
	int s;
	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_func = "dq_get";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DQ_GET )

/* Check the queue number */

	s = acp_check_qnum(*qid);
	if (s != CPS_OK) return(s);

/* Hold SIGUSR1 signals */

	cps_sighold();

/* Create a record to hold the operation (and put it on the list) */

	aqx = (struct active_qxfer *) hmalloc(sizeof(struct active_qxfer));
	if (aqx == NULL) abort();
	aqx->id = acp_next_qx_id++;
	aqx->id *= -1;		/* Indicates a get rather than a send */
	aqx->lpn = -1;
	aqx->lpn_addr = lpn;
	aqx->head = acp_qxfer_h;
	aqx->tail = acp_qxfer_t;
	aqx->next = acp_aqx;
	acp_aqx = aqx;

/* Increment counts for pending gets */

	qx = acp_qxfer_h;
	while (qx != NULL) {
		acp_inc_pget(qx->blockid);
		qx = qx->next;
	}

/* Re-initialize the queue of background data xfers */

	acp_qxfer_h = NULL;
	acp_qxfer_t = NULL;

/* Release SIGUSR1 signals */

	cps_sigrelease();

/* Initiate the operation by sending asynch dequeue message to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_asynch_dq(m,acp_local_lpn,JM_LPN,*qid,aqx->id);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	return CPS_OK;
}

int cpssta_dq_get_(qid,lpn)
int *qid;
int *lpn;
{
	return(cpssta_dq_get(qid,lpn));
}

cps_dq_get(qid,lpn)
int *qid;
int *lpn;
{
	acp_fntype = "cps_";
	cpssta_dq_get(qid,lpn);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_dq_get_(qid,lpn)
int *qid;
int *lpn;
{
	cps_dq_get(qid,lpn);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_data_ready

								  */
/******************************************************************/

cps_data_ready(blockno,bcount)
int *blockno,*bcount;
{
	acp_fntype = "cps_";
	cpssta_data_ready(blockno,bcount);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_data_ready_(blockno,bcount)
int *bcount,*blockno;
{
	cps_data_ready(blockno,bcount);
}

cpssta_data_ready_(blockno,bcount)
int *blockno,*bcount;
{
	return(cpssta_data_ready(blockno,bcount));
}

cpssta_data_ready(blockno,bcount)
int *bcount,*blockno;
{
	struct qelem *q;
	cps_msg *m;
	struct acp_data_block *block;

	acp_clear_error_info();
	acp_func = "data_ready";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DATA_READY )

/* Check to see if the block is declared for this process */

/*
	if( ( block = acp_find_block( acp_blocks, *blockno ) ) == NULL )
	{
		sprintf(acp_error_info,
			"Data block %d not declared in process %d",*blockno,acp_local_lpn);
		acp_return_status = CPS_BLOCK_UNDECLARED;
		return(CPS_BLOCK_UNDECLARED);
	}
*/

/* Check to see if we request too many bytes */

/*
	if( *bcount > block->bufsiz )
	{
		sprintf(acp_error_info,
			"Too many bytes specified for block %d (count=%d) (size=%d)",
				*blockno,*bcount,block->bufsiz);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}
*/

/* Send a message to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_data_ready(m,acp_local_lpn,JM_LPN,acp_local_lpn,*blockno,*bcount);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	return(CPS_OK);
}

/************************************************************************/
/*

	Function:   find_bufset
		find_buffer
*/
/************************************************************************/

static struct buffer_set *find_bufset(id)
int id;
{
	struct buffer_set *bs;

	bs = acp_bufset;
	while (bs != NULL) {
		if (bs->id == id) break;
		bs = bs->next;
	}
	return bs;
}

static struct buffer_list *find_buf_status(b,status)
struct buffer_list *b;
int status;
{
	while (b != NULL) {
		if (b->status == status) break;
		b = b->next;
	}
	return b;
}

static struct buffer_list *buffer_wait(bs,status)
struct buffer_set *bs;
int status;
{
	struct buffer_list *b;

	while (1) {
		b = bs->b;
		while (b != NULL) {
			if (b->status == status) return b;
			b = b->next;
		}
		sleep(1);;
	}
}

static struct buffer_list *find_buf_addr(buf)
char *buf;
{
	struct buffer_set *bs;
	struct buffer_list *b;

	bs = acp_bufset;
	while (bs != NULL) {
		b = bs->b;
		while (b != NULL) {
			if (b->buf == buf) return b;
			b = b->next;
		}
		bs = bs->next;
	}
	return NULL;
}

/************************************************************************/
/*

	Function:  cps_buffer_index

*/
/************************************************************************/

int cps_buffer_index(buf,nbytes,maxbufs)
int buf;
int *nbytes,*maxbufs;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_buffer_index(buf,nbytes,maxbufs);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_buffer_index_(buf,nbytes,maxbufs)
int buf;
int *nbytes,*maxbufs;
{

	return cps_buffer_index(buf,nbytes,maxbufs);
}

int cpssta_buffer_index_(buf,nbytes,maxbufs)
int buf;
int *nbytes,*maxbufs;
{
	return(cpssta_buffer_index(buf,nbytes,maxbufs));
}

int cpssta_buffer_index(buf,nbytes,maxbufs)
int buf;
int *nbytes,*maxbufs;
{
	struct buffer_set *bs;
	struct buffer_list *b;

	int i;

	acp_clear_error_info();
	acp_func = "buffer_index";

/* Find the buffer set */

	bs = find_bufset(buf);
	if (bs == NULL) {		/* Create a new set */
		bs = (struct buffer_set *) hmalloc(sizeof(struct buffer_set));
		if (bs == NULL) abort();
		bs->id = buf;
		bs->bufsiz = *nbytes;
		bs->maxbufs = *maxbufs;
		bs->nbufs = 0;
		bs->b = NULL;
		bs->next = acp_bufset;
		acp_bufset = bs;
	}
	else {
		bs->maxbufs = bs->maxbufs > *maxbufs ? bs->maxbufs : *maxbufs;
	}

/* Check for an empty buffer */

	b = find_buf_status(bs->b,BUFFER_FREE);
	if (b != NULL) {
		b->status = BUFFER_ALLOC;
		i = (int) b->buf - buf;
		i /= 4;
		i += 1;
		return i;
	}

/* If we haven't used the maximum number of buffers, create a new one */

	if (bs->nbufs < bs->maxbufs) {
		b = (struct buffer_list *) hmalloc(sizeof(struct buffer_list));
		if (b == NULL) abort();
		b->buf = (char *) hmalloc(bs->bufsiz);
		if (b->buf == NULL) abort();
		b->status = BUFFER_ALLOC;
		b->next = bs->b;
		bs->b = b;
		bs->nbufs++;

		i = (int) b->buf - buf;
		i /= 4;
		i += 1;
		return i;
	}

/* Wait for a buffer to become free */

	b = buffer_wait(bs,BUFFER_FREE);
	b->status = BUFFER_ALLOC;
	i = (int) b->buf - buf;
	i /= 4;
	i += 1;
	return i;

}

/************************************************************************/
/*

	Function:  cps_start_send

*/
/************************************************************************/

cps_start_send(buf,nbytes,blockno,qnum)
char *buf;
int *blockno,*nbytes,*qnum;
{
	acp_fntype = "cps_";
	cpssta_start_send(buf,nbytes,blockno,qnum);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_start_send_(buf,nbytes,blockno,qnum)
char *buf;
int *blockno,*nbytes,*qnum;
{
	cps_start_send(buf,nbytes,blockno,qnum);
}

int cpssta_start_send_(buf,nbytes,blockno,qnum)
char *buf;
int *blockno,*nbytes,*qnum;
{
	return(cpssta_start_send(buf,nbytes,blockno,qnum));
}

int cpssta_start_send(buf,nbytes,blockno,qnum)
char *buf;
int *blockno,*nbytes,*qnum;
{
	struct buffer_set *bs;
	struct buffer_list *b;

	int i;

	acp_clear_error_info();
	acp_func = "start_send";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_START_SEND )

/* Find the buffer */

	b = find_buf_addr(buf);
	if (b == NULL) {		/* User error */
		sprintf(acp_error_info,
			"Buffer not found, need to get index with cps_buffer_index");
		acp_return_status = CPS_BUFFER_NOT_FOUND;
		return CPS_BUFFER_NOT_FOUND;
	}

	b->status = BUFFER_EMPTYING;
	cps_queue_xfer(b->buf,nbytes,blockno,&(b->status),&(b->bcount));
	cps_dq_send(qnum,&(b->lpn));
	return CPS_OK;
}

/************************************************************************/
/*

	Function:  cps_wait_for_get

	Description:  Waits for an asynchronous get to complete.

*/
/************************************************************************/

int cps_wait_for_get(n,b)
int *n,*b;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_wait_for_get(n,b);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_wait_for_get_(n,b)
int *n,*b;
{
	return cps_wait_for_get(n,b);
}

int cpssta_wait_for_get_(n,b)
int *n,*b;
{
	return(cpssta_wait_for_get(n,b));
}

int cpssta_wait_for_get(nblocks,b)
int *nblocks,*b;
{
	int n,i,j;

	struct acp_data_block *block;

	acp_clear_error_info();
	acp_func = "wait_for_get";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_WAIT_FOR_GET )

	n = *nblocks;

	while(1)
	{
/* Return if no gets are pending or end of queue */

		j = 0;
		for (i=0; i<n; i++)
			if (acp_get_npget(b[i]) != 0) j++;

		if (j == 0) return 0;

/* Check remote blocks one by one */

		for (i=0; i<n; i++)
		{
			for (j=0; j<num_user_proc; j++)
			{
				block = acp_find_block(acp_remote_blocks[j],b[i]);
				if (block != NULL)
				{
					if (block->arrived)
					{
						block->arrived = 0;
						acp_dec_pget(b[i]);
						return b[i];
					}
				}
			}
		}

/* Sleep until there is background activity or a maximum of 1 second */
		sleep(1);
	}
}

/************************************************************************/
/*

	Function:  cps_wait_for_data

	Description:  Waits for data to arrive in a local block.  Data is
		  sent from another process by synchronous or asynchronous
		  sends.

*/
/************************************************************************/

/* Added 7/29/94 */
int cps_wait_data_size = 0;

int cps_wait_for_data(n,b)
int *n,*b;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_wait_for_data(n,b);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_wait_for_data_(n,b)
int *n,*b;
{
	return cps_wait_for_data(n,b);
}

int cpssta_wait_for_data_(n,b)
int *n,*b;
{
	return(cpssta_wait_for_data(n,b));
}

int cpssta_wait_for_data(nblocks,b)
int *nblocks,*b;
{
	int n,i,j;

	struct acp_data_block *block;

	acp_clear_error_info();
	acp_func = "wait_for_data";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_WAIT_FOR_DATA )

	n = *nblocks;
	cps_wait_data_size = 0;

	while(1) {

/* Check remote blocks one by one */

		for (i=0; i<n; i++) {
			block = acp_find_block(acp_blocks,b[i]);
			if (block != NULL) {
				if (block->arrived) {
					block->arrived = 0;

					if (block->bcount == -1) {
						cps_wait_data_size = 
							block->bufsiz;
					}
					else {
						cps_wait_data_size =
							block->bcount;
					}
					return b[i];
				}
			}
		}

/* Sleep until there is background activity or a maximum of 1 second */

		sleep(1);
	}
}

/************************************************************************/
/*

	Function:  cps_start_gets

*/
/************************************************************************/

int cps_start_gets(buf,maxbytes,maxbufs,blockno,q1)
int buf;
int *maxbytes,*maxbufs,*blockno,*q1;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_start_gets(buf,maxbytes,maxbufs,blockno,q1);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_start_gets_(buf,maxbytes,maxbufs,blockno,q1)
int buf;
int *maxbytes,*maxbufs,*blockno,*q1;
{

	return cps_start_gets(buf,maxbytes,maxbufs,blockno,q1);
}

int cpssta_start_gets_(buf,maxbytes,maxbufs,blockno,q1)
int buf;
int *maxbytes,*maxbufs,*blockno,*q1;
{
	return(cpssta_start_gets(buf,maxbytes,maxbufs,blockno,q1));
}

int cpssta_start_gets(buf,maxbytes,maxbufs,blockno,qu1)
int buf;
int *maxbytes,*maxbufs,*blockno,*qu1;
{
	struct buffer_set *bs;
	struct buffer_list *b;

	int j,mbufs,mbytes;

	mbufs = *maxbufs;
	mbytes = *maxbytes;

	acp_clear_error_info();
	acp_func = "start_gets";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_START_GETS )

/* Make sure the queue hasn't ended */

	if (eoq[*qu1]) return CPS_QUEUE_ENDED;

/* Find the buffer set */

	bs = find_bufset(buf);
	if (bs == NULL) {

/* Create a new buffer set */

		bs = (struct buffer_set *) hmalloc(sizeof(struct buffer_set));
		if (bs == NULL) abort();
		bs->id = buf;
		bs->bufsiz = mbytes;
		bs->maxbufs = mbufs;
		bs->nbufs = 0;
		bs->b = NULL;
		bs->next = acp_bufset;
		acp_bufset = bs;

/* Allocate all buffers */

		for (j=0; j<mbufs; j++) {
			b = (struct buffer_list *) hmalloc(sizeof(struct buffer_list));
			if (b == NULL) abort();
			b->buf = (char *) hmalloc(mbytes);
			if (b->buf == NULL) abort();
			b->status = BUFFER_EMPTY;
			b->next = bs->b;
			bs->b = b;
		}

	}

/* Start any possible transfers */

	b = bs->b;
	while (b != NULL) {		/* Loop over buffers */
		if (b->status == BUFFER_EMPTY) {
			b->status = BUFFER_FILLING;
			b->srcq = *qu1;
			cps_queue_xfer(b->buf,maxbytes,blockno,&(b->status),&(b->bcount));
			cps_dq_get(qu1,&(b->lpn));
		}
		b = b->next;
	}

	return CPS_OK;
}

/************************************************************************/
/*

	Function:  cps_handle_data

*/
/************************************************************************/

int cps_handle_data(buf,ufun,qnum)
int buf;
int (*ufun)();
int *qnum;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_handle_data(buf,ufun,qnum);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_handle_data_(buf,ufun,qnum)
int buf;
int (*ufun)();
int *qnum;
{

	return cps_handle_data(buf,ufun,qnum);
}

int cpssta_handle_data_(buf,ufun,qnum)
int buf;
int (*ufun)();
int *qnum;
{
	return(cpssta_handle_data(buf,ufun,qnum));
}

int cpssta_handle_data(buf,ufun,qnum)
int buf;
int (*ufun)();
int *qnum;
{
	struct buffer_set *bs;
	struct buffer_list *b;

	acp_clear_error_info();
	acp_func = "handle_data";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_HANDLE_DATA )

/* Find the buffer set */

	bs = find_bufset(buf);
	if (bs == NULL) {
		sprintf(acp_error_info,
			"Buffers not allocated, need to call cps_start_gets first");
		acp_return_status = CPS_BUFFER_NOT_FOUND;
		return CPS_BUFFER_NOT_FOUND;
	}

/* Call the user's function for any full buffers, then re-queue */

	b = bs->b;
	while (b != NULL) {		/* Loop over buffers */
		if (b->status == BUFFER_FULL) {
			if (b->lpn != CPS_END_OF_QUEUE) {
				(*ufun)(b->buf,&(b->bcount));
				cps_queue_process(&(b->lpn),qnum);
			}
			else {
				eoq[b->srcq] = -1;
			}
			b->status = BUFFER_EMPTY;
		}
		b = b->next;
	}

	return CPS_OK;
}

/************************************************************************/
/*

	Function:  cps_end_of_data

*/
/************************************************************************/

int cps_end_of_data(buf,qnum)
int buf;
int *qnum;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_end_of_data(buf,qnum);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_end_of_data_(buf,qnum)
int buf;
int *qnum;
{

	return cps_end_of_data(buf,qnum);
}

int cpssta_end_of_data_(buf,qnum)
int buf;
int *qnum;
{
	return(cpssta_end_of_data(buf,qnum));
}

int cpssta_end_of_data(buf,qnum)
int buf;
int *qnum;
{
	struct buffer_set *bs;
	struct buffer_list *b;

	acp_clear_error_info();
	acp_func = "end_of_data";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_END_OF_DATA )

/* If queue has not ended, return false */

	if (eoq[*qnum] == 0) return 0;

/* Find the buffer set */

	bs = find_bufset(buf);
	if (bs == NULL) {
		sprintf(acp_error_info,
			"Buffers not allocated, need to call cps_start_gets first");
		acp_return_status = CPS_BUFFER_NOT_FOUND;
		return CPS_BUFFER_NOT_FOUND;
	}

/* If any buffers are non-empty, return false */

	b = bs->b;
	while (b != NULL) {
		if (b->status != BUFFER_EMPTY) return 0;
		b = b->next;
	}

/* All data has arrived and been processed */

	return -1;
}

/************************************************************************/
/*

	Function:  cps_buffer_set_wait

*/
/************************************************************************/

int cps_buffer_set_wait(buf,q1)
int buf,*q1;
{
	int x;
	acp_fntype = "cps_";
	x = cpssta_buffer_set_wait(buf,q1);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
	return x;
}

int cps_buffer_set_wait_(buf,q1)
int buf,*q1;
{

	return cps_buffer_set_wait(buf,q1);
}

int cpssta_buffer_set_wait_(buf,q1)
int buf,*q1;
{
	return(cpssta_buffer_set_wait(buf,q1));
}

int cpssta_buffer_set_wait(buf,q1)
int buf,*q1;
{
	struct buffer_set *bs;
	struct buffer_list *b;
	int i,j;

	acp_clear_error_info();
	acp_func = "buffer_set_wait";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_BUFFER_SET_WAIT )

/* Find the buffer set */

	bs = find_bufset(buf);
	if (bs == NULL) {
		sprintf(acp_error_info,
			"Buffers not allocated, need to call cps_start_gets first");
		acp_return_status = CPS_BUFFER_NOT_FOUND;
		return CPS_BUFFER_NOT_FOUND;
	}

/* If any buffers are empty or full return */

	while (1) {
		b = bs->b;
		i = j = 0;
		while (b != NULL) {
			i++;
			if ((eoq[*q1] == 0) && (b->status == BUFFER_EMPTY)) return CPS_OK;
			if ((eoq[*q1] == -1) && (b->status == BUFFER_EMPTY)) j++;
			if (b->status == BUFFER_FULL) return CPS_OK;
			b = b->next;
		}
		if (i == j) return CPS_OK;

		sleep(1);
	}

}
