/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				SYNC.C													*/
/*																		*/
/*																		*/
/************************************************************************/


/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include <signal.h>


/******************************************************************/
/*
	FUNCTION NAME:

	  cps_wait_queue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_wait_queue(qnum,state)
int *qnum,*state;
{
	char x[20];		/* Bug in AIX compiler requires us to do
				   this stupid string copy instead of
				   assigning acp_fntype directly in the
				   last line of the routine. */
	strcpy(x,"cpssta_");

	acp_fntype = "cps_";
	cpssta_wait_queue(qnum,state);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = x;
}

cps_wait_queue_(qnum,state)
int *qnum,*state;
{
	cps_wait_queue(qnum,state);
}

cpssta_wait_queue_(qnum,state)
int *qnum,*state;
{
	return(cpssta_wait_queue(qnum,state));
}

cpssta_wait_queue(qnum,state)
int *qnum,*state;
{
	int s;

	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_func = "wait_queue";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_WAIT_QUEUE )

/* Check the queue number */

	s = acp_check_qnum(*qnum);
	if (s != CPS_OK) return(s);


	if (acp_queue_declared[*qnum] == -1) {
		s = acp_request_queue_info(*qnum);
		if (s != CPS_OK) return(s);
	}

	else if (acp_queue_declared[*qnum] == -2) {
		if (*state == CPS_FULL) return(CPS_OK);
	}

	acp_wait_complete = 0;

/* Send a wait-queue message to the Job Manager */
	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_wait_queue(m,acp_local_lpn,JM_LPN,*qnum,*state);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for a response from the JM */

	acp_wait_condition(&acp_wait_complete,1);

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_sync

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_sync(sop,sync_num)
int *sop,*sync_num;
{
	acp_fntype = "cps_";
	cpssta_sync(sop,sync_num);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_sync_(sop,sync_num)
int *sop,*sync_num;
{
	cps_sync(sop,sync_num);
}

cpssta_sync_(sop1,sync_num)
int *sop1,*sync_num;
{
	return(cpssta_sync(sop1,sync_num));
}

cpssta_sync(sop1,sync_num)
int *sop1,*sync_num;
{
	int sop;

	int s;

	int nclasses,c[MAX_CLASSES],mask;

	int i;

	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_func = "sync";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_SYNC )

	sop = *sop1;

/* Check set of processes */

	if ((sop & 0x80000000) == 0) {
		sprintf(acp_error_info,"Set of processes required");
		acp_return_status = CPS_SET_OF_PROC_REQUIRED;
		return(CPS_SET_OF_PROC_REQUIRED);
	}
	else if ((*sync_num < 0) || (*sync_num > MAX_SYNC_NUMBER)) {
		sprintf(acp_error_info,
			"Illegal sync number (sync# = %d); must be between 0 and %d"
			,*sync_num,MAX_SYNC_NUMBER);
		acp_return_status = CPS_ILLEGAL_SYNC_NUMBER;
		return(CPS_ILLEGAL_SYNC_NUMBER);
	}

	if (sop == CPS_ALL_PROCESSES) {
		nclasses = 0;			/* Indicates all classes */
	}

/* Build a list of classes from sop */

	else {
		sop &= 0x7FFFFFFF;
		mask = 2;
		nclasses = 0;
		for (i=0; i<MAX_CLASSES; i++) {
			if ((mask & sop) != 0) {
				c[nclasses] = i+1;
				nclasses++;
			}
			mask <<= 1;;
		}
	}

/* Send a wait-sync message to the Job Manager */

	acp_sync_complete = 0;

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_wait_sync(m,acp_local_lpn,JM_LPN,*sync_num,nclasses,c);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for a response from the JM */

	acp_wait_condition(&acp_sync_complete,1);

	return(CPS_OK);
}
