/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				TAPE.C													*/
/*																		*/
/*																		*/
/************************************************************************/


/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include "../inc/types.h"
#include <sys/ioctl.h>
#ifdef SGI
#include <sys/mtio.h>
#endif
#include <errno.h>

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_send_mreq

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acp_send_mreq(devnam,vsn,write_enabled,comment)
char *devnam,*comment,*vsn;
int write_enabled;
{

	struct qelem *q;
	cps_msg *m;

	acp_treq_complete = 0;

/* Send a mount request to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_mount_request(m,acp_local_lpn,JM_LPN,devnam,vsn,
	write_enabled,comment);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for the reply */

	acp_wait_condition(&acp_treq_complete,1);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_verify_tape

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acp_verify_tape(devnam,vsn,write_enabled)
char *devnam,*vsn;
int write_enabled;
{
#if defined(SGI)

	char errmsg[256];
	char fname[80];
	FILE *fp;
	int fildes;

	int st;
	struct mtget devstat;

	char *prot;

	int retry_count = 0;

#endif



/* Sleep for 90 seconds */

	sleep_1(90);

#if defined(SGI)

/* Open the tape drive */

	if (write_enabled) {
		prot = "a";
	}
	else {
		prot = "r";
	}

	start:
	sprintf(fname,"/usr/cpsdev/%s_nsv",devnam);
	fp = fopen(fname,prot);
	if (fp == NULL) {
		sprintf(errmsg,"Failed to verify tape mount: %s",fname);
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,19);
		return;
	}
	fildes = fileno(fp);

	retry:
	st = ioctl(fildes,MTIOCGET,&devstat);
	if (st < 0) {
		if (errno == EINTR) goto retry;
		if ((errno == EAGAIN) && (retry_count < 5)) {
			sleep_1(2);
			retry_count++;
			goto retry;
		}
		fclose(fp);
		printerr(FACILITY,ERROR_MSG,NULL,NULL,20);
		return;
	}
	fclose(fp);

/* Check to see if anything is in the tape drive */

	if (!devstat.mt_dposn) {
		sprintf(errmsg,
			"There is no tape in drive %s",devnam);
		acp_send_mreq(devnam,vsn,write_enabled,errmsg);
		if (mount_reply_status != CPS_OK) return;
		goto start;
	}

/* Check to see if the write protection is correct */

	if (write_enabled) {
		if (devstat.mt_dposn & MT_WPROT) {
			sprintf(errmsg,
				"Tape in drive %s needs to be write ENABLED",
				devnam);
			acp_send_mreq(devnam,vsn,write_enabled,errmsg);
			if (mount_reply_status != CPS_OK) return;
			goto start;
		}
	}
	else {
		if (!(devstat.mt_dposn & MT_WPROT)) {
			sprintf(errmsg,
				"Tape in drive %s needs to be write PROTECTED",
				devnam);
			acp_send_mreq(devnam,vsn,write_enabled,errmsg);
			if (mount_reply_status != CPS_OK) return;
			goto start;
		}
	}
	return;	

#endif
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_get_next_tape

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_get_next_tape(listnum,device,vsn,pos,status,errmsg,devlen,vsnlen,errlen)
int *listnum,*pos,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	acp_fntype = "cps_";
	cpssta_get_next_tape(listnum,device,vsn,pos,status,errmsg,
	devlen,vsnlen,errlen);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_get_next_tape_(listnum,device,vsn,pos,status,errmsg,devlen,vsnlen,errlen)
int *listnum,*pos,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	cps_get_next_tape(listnum,device,vsn,pos,status,errmsg,
	devlen,vsnlen,errlen);
}

int
cpssta_get_next_tape_(listnum,device,vsn,pos,status,errmsg,devlen,vsnlen,errlen)
int *listnum,*pos,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	return(cpssta_get_next_tape(listnum,device,vsn,pos,status,errmsg,
	devlen,vsnlen,errlen));
}

int
cpssta_get_next_tape(listnum,device,vsn,pos,status,errmsg,devlen,vsnlen,errlen)
int *listnum,*pos,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	int i;
	char dev1[80],vsn1[80],err1[80];

	struct qelem *q;
	cps_msg *m;

	int s;

	acp_clear_error_info();
	acp_func = "get_next_tape";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_GET_NEXT_TAPE )

	acp_treq_complete = 0;

/* Send a mount request to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_get_next_tape(m,acp_local_lpn,JM_LPN,*listnum);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for the reply */

	acp_wait_condition(&acp_treq_complete,1);
	
/* Copy the mount reply information to the arguments */

	*status = mount_reply_status;
	*pos = mount_reply_pos;
	strncpy(device,mount_reply_dev,devlen);
	strncpy(vsn,mount_reply_vsn,vsnlen);
	strncpy(errmsg,mount_reply_msg,errlen);
	i = devlen-1;
	while ((i >= 0) && (device[i] == '\0')) device[i--] = ' ';
	i = vsnlen-1;
	while ((i >= 0) && (vsn[i] == '\0')) vsn[i--] = ' ';
	i = errlen-1;
	while ((i >= 0) && (errmsg[i] == '\0')) errmsg[i--] = ' ';

/* Verify that the tape is mounted correctly

	mount_reply_protection:	  0 = write protected
								 1 = unprotected
*/
	if (mount_reply_status != CPS_OK) return(CPS_OK);

	acp_verify_tape(mount_reply_dev,mount_reply_vsn,
	mount_reply_protection);

/* Copy the mount reply information to the arguments */

	*status = mount_reply_status;
	*pos = mount_reply_pos;
	strncpy(device,mount_reply_dev,devlen);
	strncpy(vsn,mount_reply_vsn,vsnlen);
	strncpy(errmsg,mount_reply_msg,errlen);
	i = devlen-1;
	while ((i >= 0) && (device[i] == '\0')) device[i--] = ' ';
	i = vsnlen-1;
	while ((i >= 0) && (vsn[i] == '\0')) vsn[i--] = ' ';
	i = errlen-1;
	while ((i >= 0) && (errmsg[i] == '\0')) errmsg[i--] = ' ';

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_done_with_tape

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_done_with_tape(device,devlen)
char *device;
int devlen;
{
	acp_fntype = "cps_";
	cpssta_done_with_tape(device,devlen);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_done_with_tape_(device,devlen)
char *device;
int devlen;
{
	cps_done_with_tape(device,devlen);
}

int
cpssta_done_with_tape_(device,devlen)
char *device;
int devlen;
{
	return(cpssta_done_with_tape(device,devlen));
}

int
cpssta_done_with_tape(device,devlen)
char *device;
int devlen;
{
	int i;
	char dev1[80],fname[132];

	struct qelem *q;
	cps_msg *m;

	FILE *fp;
	int fildes;

	char errmsg[256];

	int retry_count = 0;

	int st;
#ifdef SGI
	struct mtget devstat;
#endif

	acp_clear_error_info();
	acp_func = "done_with_tape";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DONE_WITH_TAPE )

	str_f_to_c(device,devlen,dev1,80);
	str_switch(dev1,' ','\0');

/* Send the message to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_done_with_tape(m,acp_local_lpn,JM_LPN,dev1);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

#ifdef SGI

/* Build the device filename */

	sprintf(fname,"/usr/cpsdev/%s_nsv",dev1);

/* Sleep for 90 seconds */

	sleep_1(90);

/* Rewind and eject the tape */

	fp = fopen(fname,"r");
	if (fp == NULL) {
		sprintf(errmsg,"cps_done_with_tape: cannot open %s",fname);
		printerr(FACILITY,WARNING_MSG,errmsg,
			"Close file before calling cps_done_with_tape",22);
		return(CPS_OK);
	}
	fildes = fileno(fp);

	retry:
	st = ioctl(fildes,MTIOCGET,&devstat);
	if (st < 0) {
		if (errno == EINTR) goto retry;
		if ((errno == EAGAIN) && (retry_count < 5)) {
			sleep_1(2);
			retry_count++;
			goto retry;
		}
		fclose(fp);
		printerr(FACILITY,ERROR_MSG,NULL,NULL,20);
		return(CPS_OK);
	}

/* Check to see if anything is in the tape drive */

	if (!devstat.mt_dposn) {	/* No tape is in the drive */
		fclose(fp);
		return(CPS_OK);
	}

	acp_rewind(fildes);
	acp_eject_tape(fildes);	/* Should eject tape only for drives which
				   support this operation */
	fclose(fp);

#endif

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_get_spec_tape

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_get_spec_tape(device,devtyp,vsn,ring,status,errmsg,devlen,vsnlen,errlen)
int *devtyp,*ring,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	acp_fntype = "cps_";
	cpssta_get_spec_tape(device,devtyp,vsn,ring,status,errmsg,
	devlen,vsnlen,errlen);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_get_spec_tape_(device,devtyp,vsn,ring,status,errmsg,devlen,vsnlen,errlen)
int *devtyp,*ring,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	cps_get_spec_tape(device,devtyp,vsn,ring,status,errmsg,
	devlen,vsnlen,errlen);
}

int
cpssta_get_spec_tape_(device,devtyp,vsn,ring,status,errmsg,devlen,vsnlen,errlen)
int *devtyp,*ring,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	return(cpssta_get_spec_tape(device,devtyp,vsn,ring,status,errmsg,
	devlen,vsnlen,errlen));
}

int
cpssta_get_spec_tape(device,devtyp,vsn,ring,status,errmsg,devlen,vsnlen,errlen)
int *devtyp,*ring,*status,devlen,vsnlen,errlen;
char *device,*vsn,*errmsg;
{
	int i;
	char vsn1[80],dev1[80],err1[80];

	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_func = "get_spec_tape";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_GET_SPEC_TAPE )

	str_f_to_c(vsn,vsnlen,vsn1,80);
	str_switch(vsn1,' ','\0');

	acp_treq_complete = 0;

/* Send a mount request to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_get_spec_tape(m,acp_local_lpn,JM_LPN,*devtyp,*ring,vsn1);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for the reply */

	acp_wait_condition(&acp_treq_complete,1);

/* Copy the mount reply information to the arguments */

	*status = mount_reply_status;
	strncpy(device,mount_reply_dev,devlen);
	strncpy(errmsg,mount_reply_msg,errlen);
	i = devlen-1;
	while ((i >= 0) && (device[i] == '\0')) device[i--] = ' ';
	i = errlen-1;
	while ((i >= 0) && (errmsg[i] == '\0')) errmsg[i--] = ' ';
	mount_reply_protection = *ring;

/* Verify that the tape is mounted correctly

	mount_reply_protection:	  0 = write protected
								 1 = unprotected
*/
	if (mount_reply_status != CPS_OK) return(CPS_OK);

	acp_verify_tape(mount_reply_dev,mount_reply_vsn,
	mount_reply_protection);

/* Copy the mount reply information to the arguments */

	*status = mount_reply_status;
	strncpy(device,mount_reply_dev,devlen);
	strncpy(errmsg,mount_reply_msg,errlen);
	i = devlen-1;
	while ((i >= 0) && (device[i] == '\0')) device[i--] = ' ';
	i = errlen-1;
	while ((i >= 0) && (errmsg[i] == '\0')) errmsg[i--] = ' ';

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_mount_request

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_mount_request_(device,vsn,protection,status,errmsg,devlen,vsnlen,errlen)
char *device,*vsn,*errmsg;
int *status,*protection,devlen,vsnlen,errlen;
{
	cps_mount_request(device,vsn,protection,status,errmsg,devlen,vsnlen,errlen);
}

cps_mount_request(device,vsn,protection,status,errmsg,devlen,vsnlen,errlen)
char *device,*vsn,*errmsg;
int *status,*protection,devlen,vsnlen,errlen;
{

	acp_fntype = "cps_";
	cpssta_mount_request(device,vsn,protection,status,errmsg,devlen,vsnlen,
		errlen);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";

}

int cpssta_mount_request_(device,vsn,protection,status,errmsg,devlen,vsnlen,errlen)
char *device,*vsn,*errmsg;
int *status,*protection,devlen,vsnlen,errlen;
{
	return(cpssta_mount_request(device,vsn,protection,status,errmsg,devlen,vsnlen,errlen));
}

int cpssta_mount_request(device,vsn,protection,status,errmsg,devlen,vsnlen,errlen)
char *device,*vsn,*errmsg;
int *status,*protection,devlen,vsnlen,errlen;
{
	char dev1[80],vsn1[80];
	int i;

	acp_clear_error_info();
	acp_func = "mount_request";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_MOUNT_REQUEST )

	str_f_to_c(device,devlen,dev1,80);
	str_switch(dev1,' ','\0');

	str_f_to_c(vsn,vsnlen,vsn1,80);
	str_switch(vsn1,' ','\0');

	acp_send_mreq(dev1,vsn1,*protection,"no_comment");
	if (mount_reply_status == CPS_OK) acp_verify_tape(dev1,vsn1,*protection);

/* Copy the mount reply information to the arguments */

	*status = mount_reply_status;
	strncpy(errmsg,mount_reply_msg,errlen);
	i = errlen-1;
	while ((i >= 0) && (errmsg[i] == '\0')) errmsg[i--] = ' ';

	return(CPS_OK);
}
