/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

struct acp_vbbc_info {

    int crate;				/* Crate number */
    int vme_addr;			/* VME address */
    int npaddr;				/* Number of addresses in the list */
    unsigned long *paddr;		/* Array of physical addresses */
    int offset;				/* Offset in bytes of first
					   physical block */
};

struct acp_data_block {

    int blockno;			/* Block number */
    char *buf;				/* Pointer to memory (virtual addr) */
    int bufsiz;				/* #bytes in block */
    struct acp_vbbc_info *vbbc;		/* For R3000 only */
    struct acp_data_block *next;	/* Next declared block */
    int bcount;				/* Current number of valid bytes in
					   the data block (used for asynch.
					   acp_dq_get) A value of -1 means
					   to get the entire block. */
    int arrived;			/* Non-zero if data has arrived */
};

struct acp_sub_instance {

    int lpn;		/* Server process number */
    int active;		/* 0 = inactive, -1 = active */
    int action;		/* ACP_WAIT or ACP_NO_WAIT */
    char **baddr;	/* Addresses to copy back arguments */
    int complete;	/* -1 if completed, otherwise 0 */

};

struct acp_remote_sub {

    int id;
    int (*func)();	/* Pointer to the function */
    int nargs;
    int bcount[32];
    int bsum;		/* Sum of byte counts */
    int argtyp[32];	/* Argument types (read-only, write-only, read-write) */

    int num_active;	/* Number of active instances */
    struct acp_sub_instance **inst;
    
    struct acp_remote_sub *next;
};

struct acp_msg_queue {

    struct acp_msg_block *head;
    struct acp_msg_block *tail;

};

struct acp_msg_block {

    char *buf;
    int index;
    struct acp_msg_block *next;

};

struct queued_xfer {

    char *addr;			/* Non-volatile source address */
    int nbytes;			/* Number of bytes */
    int blockid;		/* Destination block number */
    int *flag;			/* Flag to set when xfer is done */
    int *abort;			/* Integer to increment for aborted xfers */
    int *bcount;		/* Number of bytes received or sent */
    struct queued_xfer *next;

};

struct active_qxfer {

    int id;			/* >0 = asynch send, <0 = asynch get */
    int lpn;			/* -1 means still waiting for dequeue */
    int *lpn_addr;		/* Address to put lpn of process dequeued */
    struct queued_xfer *head;	/* Queued data xfers */
    struct queued_xfer *tail;
    struct active_qxfer *next;

};

/* Definitions for asynchronous sends */

#define BUFFER_FREE -1
#define BUFFER_ALLOC 1
#define BUFFER_EMPTYING 0

/* Definitions for asynchronous gets */

#define BUFFER_EMPTY 0
#define BUFFER_FULL -1
#define BUFFER_FILLING 1

struct buffer_list {

    char *buf;			/* Actual buffer */
    int status;			/* Empty, full, or filling */
    int bcount;
    int lpn;
    int srcq;			/* Queue number from which lpn was obtained */
    struct buffer_list *next;

};

/* Global var is acp_bufset */

struct buffer_set {

    int id;			/* Address of user's reference point */
    int bufsiz;			/* Number of bytes in a buffer */
    int maxbufs;		/* Maximum number of buffers */
    int nbufs;			/* Number of buffers in the list */
    struct buffer_list *b;	/* List of buffers */
    struct buffer_set *next;	/* Next buffer set in the list */

};

/* Direct data transfers */

/* Transfer directions */

#define XFER_GET 0
#define XFER_SEND 1

/* Transfer origins */

#define XFER_LOCAL_ORIGIN 0
#define XFER_REMOTE_ORIGIN 1

/* Transfer states */

#define XFER_START 0
#define XFER_NEED_CONNECT 1
#define XFER_NEED_INFO 2
#define XFER_SEND_INFO 3
#define XFER_NEED_GO 4
#define XFER_SEND_GO 5
#define XFER_NEED_DATA 6
#define XFER_SEND_DATA 7
#define XFER_DONE 8

/* Data connection message types */

#define DX_INFO_MSG 1
#define DX_GO_MSG 2
#define DX_REPEAT_GO_MSG 3
#define DX_DATA_MSG 4

struct direct_xfer {

    int type;		/* XFER_GET or XFER_SEND */
    int need_to_clear;  /* For processes initiating acp_gets */
    int lflag;		/* "Last" flag.  For locally init. transfers, if this is
			   set, then when the transfer completes, the
			   bouncing message will be killed. */
    int aflag;		/* -1 if this is an asynchronous xfer, else 0 */

    char *buf;		/* Local buffer */
    int blockno;
    int offset;
    int nbytes;

    int state;		/* XFER_NEED_CONNECT, XFER_NEED_INFO, ... */

    int op;		/* 0 = no op, 1 = increment done, 2 = set done to -1 */
    int *done;		/* Increments this memory location when xfer is done,
			   sets it to -1, or does nothing. */

    int *abort;		/* Increments this memory location for each xfer that
			   was aborted.  (The done field is still updated
			   for aborted transfers) (New in v2.9) */
};


/*

Initiator:

 0) Connection established.
 1) Send xfer info over data connection.
*2) Get the go-ahead.
 3) If transfer possible, send data, else send repeat-go-ahead.

Remote:

 0) Connection established.
*1) Receive xfer infor over data connection.
 2) Send the go-ahead.
*3) Get reply, if data, store it and complete xfer, else go to other xfers,
    then bounce back.

*/
