/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnfaus.fnal.gov

			Copyright (c) 1991 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

	Source File: xfer.c

	Explanation: 

	Exported Routines:


*/

#include "../inc/types.h"
#include <sys/times.h>
#include <sys/param.h>
#ifndef HZ
#define HZ 60
#endif
#include "all.h"
#include <signal.h>

struct acp_data_block *acp_find_block(list,blockno)
struct acp_data_block *list;
int blockno;
{
	struct acp_data_block *dblock;

	dblock = list;
	while (dblock != NULL)
	{
		if (dblock->blockno == blockno) return(dblock);
		dblock = dblock->next;
	}
	return(NULL);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  r3000_send_virtual

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	   msgtyp is ignored.

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

#ifdef R3000

static int r3000_send_virtual(blockno,offset,sop,buf,bufsiz)
int blockno,offset;
char *buf;
int bufsiz;
unsigned long *sop;
{
	int *plist;

	struct qelem *q;
	cps_message *m;

	int n = 0;

	int i,j;

	int nsends = 0;
	clock_t start,stop;
	struct tms tbuf;
	int elapsed,bps;
	float felapsed,fbcount,bytes_per_sec;

	struct acp_data_block *dblock,*blist;

/* Start the timer */

	start = times(&tbuf);

/* Reset acp_send_complete to zero */

	acp_send_complete = 0;

/* Send transfer requests to the JM, or do the transfer over the VBBC */

	plist = set_query(sop,MAX_PROCESSES);

	i = 0;
	while (plist[i] != 0) {

		nsends++;

/* Determine what type of send to do */

		j = plist[i] - FIRST_USER_LPN;
		blist = acp_remote_blocks[j];
		dblock = acp_find_block(blist,blockno);
		if ((dblock->vbbc != NULL) &&
			(acp_pmap[j].local == 0)) {		/* Process is non-local */
			acp_vbbc_send(buf,bufsiz,dblock->vbbc,offset);
		}
		else {		/* Do non-VBBC transfer */
			n++;
			q = acp_q_pop2(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
			cps_bld_xfer_request(m,acp_local_lpn,acp_shm_lpn,0,acp_local_lpn,
			plist[i],buf,blockno,offset,bufsiz);
			acp_q_push2(acp_shm_qnum,q);
			acp_signal_shm();
		}
		i++;
	}

/* Tell the JM we are waiting for a transfer to complete */

	if (n > 0) {
		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_WAIT_SEND,blockno);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

/* Wait until all MSG_XFER_COMPLETE messages have arrived */

		acp_wait_condition(&acp_send_complete,n);

/* Tell the JM the transfer has completed */

		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_SEND_DONE,bufsiz*n);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
	}

/* Increment number of sends */

	q = acp_q_pop2(S_FREE_Q);
	m = (cps_message *) SMALL_MSG(q);
	stop = times(&tbuf);
	elapsed = (stop - start);
	if (elapsed == 0) elapsed = 1;
	felapsed = (float) elapsed;
	felapsed /= (float) HZ;
	fbcount = (float) bufsiz;
	fbcount *= (float) nsends;
	bytes_per_sec = fbcount/felapsed;
	bps = (int) bytes_per_sec;
	cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_INC_CALL,acp_local_lpn,
		INC_SEND,bps);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	return(CPS_OK);
}

#endif

/******************************************************************/
/*
	FUNCTION NAME:

	  send_virtual

								  */
/******************************************************************/

static int send_virtual(blockno,offset,sop,buf,bufsiz)
int blockno,offset;
char *buf;
int bufsiz;
unsigned long *sop;
{
	int *plist;

	struct qelem *q;
	cps_message *m;

	int n = 0;

	int i,j,k;

	char go_msg[20];

	int nsends = 0;
	clock_t start,stop;
	struct tms tbuf;
	int elapsed,bps;
	float felapsed,fbcount,bytes_per_sec;

	start = times(&tbuf);

/* Reset acp_send_complete to zero */

	acp_send_complete = 0;
	acp_send_aborts = 0;	/* v2.9 */

/* For each process determine how to accomplish each transfer */

	plist = set_query(sop,MAX_PROCESSES);
	i = 0;
	while (plist[i] != 0) {

		nsends++;
		n++;

/* If the destination process is local, send it the old way */

		j = plist[i] - FIRST_USER_LPN;
		if (acp_pmap[j].local) {

/* Send a transfer request to the shm */

			q = acp_q_pop2(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
			cps_bld_xfer_request(m,acp_local_lpn,acp_shm_lpn,0,acp_local_lpn,
				plist[i],buf,blockno,offset,bufsiz);
			acp_q_push2(acp_shm_qnum,q);
			acp_signal_shm();

		}

/* Otherwise, do a direct data transfer */

		else {

/* Hold signals. */

			cps_sighold();

/* Fill in the information in the direct transfer record */

			j = plist[i];
			if (acp_dxfer_nleft[j] >= MAX_L_XFERS) {

			}
			k = acp_dxfer_nleft[j] + acp_dxfer_ptr[j];

			acp_l_dxfer[j][k].type = XFER_SEND;
			acp_l_dxfer[j][k].need_to_clear = 0;
			acp_l_dxfer[j][k].state = XFER_NEED_CONNECT;
			acp_l_dxfer[j][k].op = 1;
			acp_l_dxfer[j][k].done = &acp_send_complete;
			acp_l_dxfer[j][k].abort = &acp_send_aborts; /* v2.9 */
			acp_l_dxfer[j][k].buf = buf;
			acp_l_dxfer[j][k].blockno = blockno;
			acp_l_dxfer[j][k].offset = offset;
			acp_l_dxfer[j][k].nbytes = bufsiz;
			acp_l_dxfer[j][k].lflag = -1;
			acp_l_dxfer[j][k].aflag = 0;

			acp_dxfer_nleft[j]++;

/* Tell the remote process to look for a transfer */

			q = acp_q_pop2(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
			cps_bld_direct_xfer(m,acp_local_lpn,j,acp_local_lpn);
			acp_q_push2(acp_shm_qnum,q);
			acp_signal_shm();

/* Tell ourselves to do the transfer */

			q = acp_q_pop2(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
			cps_bld_direct_xfer(m,acp_local_lpn,acp_local_lpn,j);
			acp_q_push2(acp_shm_qnum,q);
			acp_signal_shm();

/* Release signals */

			cps_sigrelease();

		}

		i++;
	}

	if (n > 0) {

/* Tell the JM we are waiting for a transfer to complete */
/*
		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_WAIT_SEND,blockno);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
*/
/* Wait until all MSG_XFER_COMPLETE messages have arrived */

		acp_wait_condition(&acp_send_complete,n);

/* Tell the JM the transfer has completed */
/*
		q = acp_q_pop2(S_FREE_Q);
		m = (cps_message *) SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_SEND_DONE,bufsiz*n);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
*/
	}

/* Increment number of sends */

	q = acp_q_pop2(S_FREE_Q);
	m = (cps_message *) SMALL_MSG(q);
	stop = times(&tbuf);
	elapsed = (stop - start);
	if (elapsed == 0) elapsed = 1;
	felapsed = (float) elapsed;
	felapsed /= (float) HZ;
	fbcount = (float) bufsiz;
	fbcount *= (float) nsends;
	bytes_per_sec = fbcount/felapsed;
	bps = (int) bytes_per_sec;
	cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_INC_CALL,acp_local_lpn,
		INC_SEND,bps);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	if (acp_send_aborts == 0) {
		return(CPS_OK);
	}
	else {
		sprintf(acp_error_info,
"Send failed because of lost connection or dead process");
		acp_return_status = CPS_XFER_FAILED;
		return(CPS_XFER_FAILED); /* One or more transfers failed */
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_declare_block

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_declare_block(buf,bufsiz,blockno)
char *buf;
int *bufsiz,*blockno;
{
	acp_fntype = "cps_";
	cpssta_declare_block(buf,bufsiz,blockno);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_declare_block_(buf,bufsiz,blockno)
char *buf;
int *bufsiz,*blockno;
{
	cps_declare_block(buf,bufsiz,blockno);
}

int
cpssta_declare_block_(buf,bufsiz,blockno)
char *buf;
int *bufsiz,*blockno;
{
	return(cpssta_declare_block(buf,bufsiz,blockno));
}

int
cpssta_declare_block(buf,bufsiz,blockno)
char *buf;
int *bufsiz,*blockno;
{
	struct acp_data_block *dblock;

	char warning[80];

	struct qelem *q;
	cps_msg *m;

	int msgsiz;
	struct acp_vbbc_info *vbbc;

	int newblock;

	acp_clear_error_info();
	acp_func = "declare_block";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DECLARE_BLOCK )

/* Check to see if bufsiz is a multiple of 4 */

	if (*bufsiz % 4) {
		sprintf(acp_error_info,
			"Byte count must be a multiple of four (count=%d)",*bufsiz);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}

/* If the block was already declared, send a warning to the JM and
   substitute new information */

	if ((dblock = acp_find_block(acp_blocks,*blockno)) != NULL) {
		sprintf(warning,"%s%s: Data block %d declared more than once",
			acp_fntype,acp_func,*blockno);
		acp_log_msg(warning);
		newblock = 0;
	}
	else {
		dblock = (struct acp_data_block *) 
			hmalloc(sizeof(struct acp_data_block));
		newblock = -1;
	}

	dblock->blockno = *blockno;
	dblock->buf = buf;
	dblock->bufsiz = *bufsiz;
	dblock->vbbc = NULL;
	dblock->bcount = -1;
	dblock->arrived = 0;
	if (newblock) {
		dblock->next = acp_blocks;
		acp_blocks = dblock;
	}

/* Send the declaration to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_declare_block(m,acp_local_lpn,JM_LPN,acp_local_lpn,*blockno);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* If we are on an R3000 node, find the physical addresses of the block,
   and send them to the JM. */

#ifdef R3000
	vbbc = (struct acp_vbbc_info *) hmalloc(sizeof(struct acp_vbbc_info));
	dblock->vbbc = vbbc;
	getcratevmeaddress(dblock->vbbc);
	acp_set_vbbc_info(dblock->vbbc,buf,*bufsiz);
	msgsiz = 16+16+vbbc->npaddr*4;
	if (msgsiz > MAX_MESSAGE_SIZE) {		/* Error, block too large */
		abort();
	}
	else if (msgsiz <= SMALL_MSG_SIZE) {
		q = acp_q_pop2(S_FREE_Q);
		m = SMALL_MSG(q);
	}
	else {
		q = acp_q_pop2(L_FREE_Q);
		m = (cps_msg *) LARGE_MSG(q);
	}
	cps_bld_vbbc_info(m,acp_local_lpn,JM_LPN,acp_local_lpn,*blockno,
	vbbc->crate,vbbc->vme_addr,vbbc->offset,vbbc->npaddr,vbbc->paddr);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();
#endif

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_send

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_send(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	acp_fntype = "cps_";
	cpssta_send(sop,buf,bufsiz,blockno,offset);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_send_(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	cps_send(sop,buf,bufsiz,blockno,offset);
}

int
cpssta_send_(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	return(cpssta_send(sop,buf,bufsiz,blockno,offset));
}

int
cpssta_send(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	static unsigned long *ssop = NULL;
	int *plist,i,j;

	int this_process_included = 0;

	struct acp_data_block *block,*blist;
	char *c;

	acp_clear_error_info();
	acp_func = "send";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_SEND )

/* Translate CPS set of processes into standard set of processes */

	if (ssop == NULL) ssop = create_empty_set(MAX_PROCESSES);

	acp_translate_sop(*sop,ssop);

/* Check to see if the block is declared for each process in the set */

	plist = set_query(ssop,MAX_PROCESSES);
	i = 0;
	while (plist[i] != 0) {
		if (plist[i] == acp_local_lpn) {
			blist = acp_blocks;
		}
		else {
			j = plist[i]-FIRST_USER_LPN;
			blist = acp_remote_blocks[j];
		}
		if (acp_find_block(blist,*blockno) == NULL) {
			sprintf(acp_error_info,
				"Data block %d not declared in process %d",*blockno,plist[i]);
			acp_return_status = CPS_BLOCK_UNDECLARED;
			return(CPS_BLOCK_UNDECLARED);
		}
		i++;
	}

	if (set_member(acp_local_lpn,ssop)) {
		remove_from_set(acp_local_lpn,ssop);
		this_process_included = -1;
	}

/* Check to see if bufsiz is a multiple of 4 */

	if (*bufsiz % 4) {
		sprintf(acp_error_info,
			"Byte count must be a multiple of four (count=%d)",*bufsiz);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}

/* Check to see if offset is a multiple of 4 */

	if (*offset % 4) {
		sprintf(acp_error_info,
			"Offset must be a multiple of four (offset=%d)",*offset);
		acp_return_status = CPS_ILLEGAL_OFFSET;
		return(CPS_ILLEGAL_OFFSET);
	}

/* If this process is included, do the send right here */

	if (this_process_included) {
		block = acp_find_block(acp_blocks,*blockno);
		c = block->buf;
		c += *offset;
		bcopy(buf,c,*bufsiz);
	}

#ifdef R3000
	return(r3000_send_virtual(*blockno,*offset,ssop,buf,*bufsiz));
#else
	return(send_virtual(*blockno,*offset,ssop,buf,*bufsiz));
#endif
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_get

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_get(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	acp_fntype = "cps_";
	cpssta_get(sop,buf,bufsiz,blockno,offset);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_get_(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	cps_get(sop,buf,bufsiz,blockno,offset);
}

int
cpssta_get_(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	return(cpssta_get(sop,buf,bufsiz,blockno,offset));
}

int
cpssta_get(sop,buf,bufsiz,blockno,offset)
int *sop;
char *buf;
int *bufsiz,*blockno,*offset;
{
	static unsigned long *ssop = NULL;
	int *plist;
	int i,j,k,n;
	struct qelem *q;
	cps_msg *m;

	int ngets;

	int s,status;

	char errmsg[80];

	int this_process_included = 0;

	char *c;
	struct acp_data_block *block,*blist;

	int *isrc,*idst;

	int remote_shm_lpn;

	clock_t start,stop;
	struct tms tbuf;
	int elapsed,bps;
	float felapsed,fbcount,bytes_per_sec;

	int need_to_clear = 0;

	acp_clear_error_info();
	acp_func = "get";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_GET )

	start = times(&tbuf);

/* Translate CPS set of processes into standard set of processes */

	if (ssop == NULL) ssop = create_empty_set(MAX_PROCESSES);

	acp_translate_sop(*sop,ssop);

/* Check to see if the block is declared for each process in the set */
/* Count the number of processes we are getting data from and store this
   in ngets. */
/* Determine if we need to clear the destination buffer */

	plist = set_query(ssop,MAX_PROCESSES);
	i = 0;
	ngets = 0;			/* Number of processes to get data from */
	while (plist[i] != 0)
	{
		ngets++;
		if (plist[i] == acp_local_lpn)
		{
			blist = acp_blocks;
		}
		else
		{
			j = plist[i]-FIRST_USER_LPN;
			blist = acp_remote_blocks[j];
			if (acp_pmap[j].local) need_to_clear = -1;
		}
		if (( block = acp_find_block(blist,*blockno)) == NULL )
		{
			sprintf(acp_error_info,
				"Data block %d not declared in process %d",*blockno,plist[i]);
			acp_return_status = CPS_BLOCK_UNDECLARED;
			return(CPS_BLOCK_UNDECLARED);
		}
		i++;
	}
	if (ngets > 1) need_to_clear = -1;

/* Check to see if we are requesting data from ourself */

	if (set_member(acp_local_lpn,ssop))
	{
		remove_from_set(acp_local_lpn,ssop);
		this_process_included = -1;
	}

/* Check to see if bufsiz is a multiple of 4 */

	if (*bufsiz % 4)
	{
		sprintf(acp_error_info,
			"Byte count must be a multiple of four (count=%d)",*bufsiz);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}

/* Check to see if we request too many bytes */

/*
	if (*bufsiz > block->bufsiz)
	{
		sprintf(acp_error_info,
			"Too many bytes requested for block %d (count=%d) (size=%d)",
				*blockno,*bufsiz,block->bufsiz);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}
*/

/* Check to see if offset is a multiple of 4 */

	if (*offset % 4)
	{
		sprintf(acp_error_info,
			"Offset must be a multiple of four (offset=%d)",*offset);
		acp_return_status = CPS_ILLEGAL_OFFSET;
		return(CPS_ILLEGAL_OFFSET);
	}

/* Clear the buffer for summation */

	if (this_process_included)
	{
		block = acp_find_block(acp_blocks,*blockno);
		c = block->buf;
		c += *offset;
		isrc = (int *) c;
		idst = (int *) buf;
		n = *bufsiz>>2;
		for (i=0; i<n; i++) idst[i] = isrc[i];
	}
	else if (need_to_clear)
	{
		bzero(buf,*bufsiz);
	}

/* Reset the get_complete counter to zero */

	acp_get_complete = 0;
	acp_get_aborts = 0;	/* v2.9 */

/* Send a data request out to each process, the data request is sent to
   the shm that is local to each process from which we are getting data */

	plist = set_query(ssop,MAX_PROCESSES);

	n = 0;
	i = 0;
	while (plist[i] != 0) {

#ifdef R3000

/* Get data over VME if possible */

		j = plist[i]-FIRST_USER_LPN;
		if (acp_pmap[j].local == 0) {
			blist = acp_remote_blocks[j];
			block = acp_find_block(blist,*blockno);
			if (block->vbbc != NULL) {
				acp_vbbc_get(buf,*bufsiz,block->vbbc,*offset);
				i++;
				continue;
			}
		}
#endif

/* If destination process is local, get data the old way */

		j = plist[i]-FIRST_USER_LPN;
		if (acp_pmap[j].local) {
			n++;
			q = acp_q_pop2(S_FREE_Q);
			m = SMALL_MSG(q);
			j = acp_pmap[plist[i]-FIRST_USER_LPN].shmid;
			remote_shm_lpn = j + FIRST_SHM_LPN - 1;
			cps_bld_xfer_request(m,acp_local_lpn,remote_shm_lpn,1,plist[i],
			acp_local_lpn,buf,*blockno,*offset,*bufsiz);
			acp_q_push2(acp_shm_qnum,q);
			acp_signal_shm();
		}

/* Otherwise do a direct transfer */

		else {

			n++;

/* Hold signals */

			cps_sighold();

/* Fill in the information in the direct transfer record */

			j = plist[i];

			if (acp_dxfer_nleft[j] >= MAX_L_XFERS) {

			}
			k = acp_dxfer_nleft[j] + acp_dxfer_ptr[j];

			acp_l_dxfer[j][k].type = XFER_GET;
			acp_l_dxfer[j][k].need_to_clear = need_to_clear;
			acp_l_dxfer[j][k].state = XFER_NEED_CONNECT;
			acp_l_dxfer[j][k].op = 1;
			acp_l_dxfer[j][k].done = &acp_get_complete;
			acp_l_dxfer[j][k].abort = &acp_get_aborts; /* v2.9 */
			acp_l_dxfer[j][k].buf = buf;
			acp_l_dxfer[j][k].blockno = *blockno;
			acp_l_dxfer[j][k].offset = *offset;
			acp_l_dxfer[j][k].nbytes = *bufsiz;
			acp_l_dxfer[j][k].lflag = -1;
			acp_l_dxfer[j][k].aflag = 0;

			acp_dxfer_nleft[j]++;

/* Release the signals */

			cps_sigrelease();

/* Tell the remote process to look for a transfer */

			q = acp_q_pop2(S_FREE_Q);
			m = SMALL_MSG(q);
			cps_bld_direct_xfer(m,acp_local_lpn,j,acp_local_lpn);
			acp_q_push2(acp_shm_qnum,q);

/* Tell ourselves to do the transfer */

			q = acp_q_pop2(S_FREE_Q);
			m = SMALL_MSG(q);
			cps_bld_direct_xfer(m,acp_local_lpn,acp_local_lpn,j);
			acp_q_push2(acp_shm_qnum,q);
			acp_signal_shm();

		}

		i++;
	}

	if (n > 0) {

/* Tell the JM we are waiting for data */
/*
		q = acp_q_pop2(S_FREE_Q);
		m = SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_WAIT_GET,*blockno);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
*/
/* Wait until all MSG_XFER_COMPLETE messages have arrived */

		acp_wait_condition(&acp_get_complete,n);

/* Tell the JM we got the data */
/*
		q = acp_q_pop2(S_FREE_Q);
		m = SMALL_MSG(q);
		cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
			PSTATE_GET_DONE,*bufsiz);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();
*/
	}

/* Increment number of gets */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	stop = times(&tbuf);
	elapsed = (stop - start);
	if (elapsed == 0) elapsed = 1;
	felapsed = (float) elapsed;
	felapsed /= (float) HZ;
	fbcount = (float) *bufsiz;
	fbcount *= (float) ngets;
	bytes_per_sec = fbcount/felapsed;
	bps = (int) bytes_per_sec;
	cps_bld_mon_update(m,acp_local_lpn,JM_LPN,MOP_INC_CALL,acp_local_lpn,
		INC_GET,bps);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	if (acp_get_aborts == 0) {	/* v2.9 */
		return(CPS_OK);
	}
	else {
		sprintf(acp_error_info,
		"Get failed because of lost connection or dead process");
		acp_return_status = CPS_XFER_FAILED;
		return CPS_XFER_FAILED;
	}
}
