/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: dbgbuf.c

    Explanation: Keeps a circular buffer of messages that will be
		 printed when a signal arrives.

    Exported Routines:


*/

#include <stdio.h>
#include "../inc/perr.h"

static int bufsiz = -1;
static int nbufs = -1;
static int sig = -1;

static char *buf = NULL;
static int bp = 0;
static int sbp = 0;
static int inc_sbp = 0;

extern char logmsg_fac[80];

void dbgbuf_print(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	int i,j;

	i = j = sbp;

	while (1) {
		if (buf[i*bufsiz] == '\0') break;
		logmsg(1,INFO_MSG,&buf[i*bufsiz]);
		i++;
		if (i == nbufs) i = 0;
		if (i == j) break;
	}
}

int dbgbuf_alloc(nbuf,size)
int nbuf,size;
{
	int i;
	if (buf != NULL) return 0;
	buf = (char *) hmalloc(nbuf*size);
	if (buf == NULL) {
		logmsg(2,ERROR_MSG,"Failed to allocate circular buffer");
		return -1;
	}
	nbufs = nbuf;
	bufsiz = size;
	for (i=0; i<nbufs; i++) buf[i*bufsiz] = '\0';
	return 0;
}

void dbgbuf_setsig(signum)
int signum;
{
	sig = signum;
	sig_establish_handler(sig,dbgbuf_print);
	strcpy(logmsg_fac,"CPS_UTIL");
}

void dbgbuf_insert(msg)
char *msg;
{
	if (sig == -1) return;
	if (buf == NULL) return;

	if (inc_sbp) {
		inc_sbp = 0;
		sbp++;
		if (sbp == nbufs) sbp = 0;
	}
	strncpy(&buf[bp*bufsiz],msg,bufsiz-1);
	buf[(bp+1)*bufsiz - 1] = '\0';
	bp++;
	if (bp == nbufs) bp = 0;
	if (bp == sbp) {
		inc_sbp = -1;
	}
}

/*
main(argc,argv)
int argc;
char *argv[];
{
    char msg[80];
    int n,i;
    dbgbuf_setsig(16);
    n = atoi(argv[1]);
    printf("n = %d\n",n);
    fflush(stdout);
    dbgbuf_alloc(12,80);
    for (i=0; i<n; i++) {
	sprintf(msg,"%d",i);
	dbgbuf_insert(msg);
    }
    while (1) sleep(1);
}
*/
