/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: internet.c

    Explanation: Socket level routines.

    Exported Routines:


*/

#include <stdio.h>
#include <errno.h>
#include "../inc/types.h"
#include "../inc/time.h"

#ifdef VMS
#include "multinet_root:[multinet.include.sys]types.h"
#include "multinet_root:[multinet.include.sys]socket.h"
#include "multinet_root:[multinet.include.netinet]in.h"
#include "multinet_root:[multinet.include]netdb.h"
#include "multinet_root:[multinet.include]errno.h"
#include "multinet_root:[multinet.include.vms]inetiodef.h"
#endif

#ifdef SGI
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif

#ifdef HPUX
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif

#ifdef AIX
#include <sys/select.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif

#ifdef MIPS
#include <sys/socket.h>
#include <netdb.h>
#include <bsd/netinet/in.h>
#include <bsd/netinet/tcp.h>
#endif

#ifdef R3000
#include <bsd/sys/socket.h>
#include <bsd/netdb.h>
#include <bsd/netinet/in.h>
#include <bsd/netinet/tcp.h>
#endif

#ifdef SUNOS
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif

#ifdef ULTRIX
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#endif

#include "../inc/perr.h"

#define FACILITY "MH"
#define PACKET_SIZE 4096
#define MESSAGE_BUFFER_SIZE 512

/* The following global variables are set by create_passive_socket */
    
int assigned_port_number = 0;
char local_hostname[64];
char local_hostaddr[24]; /* Internet address in d.d.d.d representation */
struct sockaddr_in local_socket_name;

int mh_nbsocks = 0;	/* Non-blocking sockets */

static int local_name_unknown = -1;
static int local_addr_unknown = -1;

static char errmsg[132];

static struct msgbuff {
    int more;		/* 0 = last block in message, -1 = more to come */
    int nbyte;		/* number of bytes in this block */
    char msg[4096];	/* binary message */
};

static int inside_read_bin = 0;
static int inside_send_bin = 0;

extern char *strchr();

extern int h_errno;

/************************************************************************/
/*

    Private Functions:	get_local_hostname

*/
/************************************************************************/

int get_local_hostname() {

/* Find out what host we are on */

    if (local_name_unknown) {
        retry1:
        if (gethostname(local_hostname,64) < 0) {
            if (errno == EINTR) goto retry1;
            printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
            *local_hostname = '\0';
            return(-1);
        }
        local_name_unknown = 0;
    }
    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      gethostid (for HPUX)

  DESCRIPTION:

  FORMAL PARAMETERS:

      none

  FUNCTION VALUE:

								  */
/******************************************************************/

#if defined (HPUX) || defined(VMS)
unsigned long gethostid()
{
	struct hostent *hp;
	struct in_addr hostaddr;

	if (local_name_unknown) {
	    if (get_local_hostname() == -1) return(0xFFFFFFFF);
	}

	if ((hp = gethostbyname(local_hostname)) == NULL) {
            printerr(FACILITY,ERROR_MSG,
		"Failed to get local host by name",NULL,2);
	    return(0xFFFFFFFF);
	}

	bcopy(hp->h_addr,&hostaddr,hp->h_length);
	return(hostaddr.s_addr);
}
#endif

/******************************************************************/
/*
    FUNCTION NAME:

      anchor_socket

  DESCRIPTION:

  FORMAL PARAMETERS:

      none

  FUNCTION VALUE:

      int - 0 = success, -1 = failure

								  */
/******************************************************************/

int anchor_socket(s,portnum)
int s,portnum;
{
    unsigned short portno;

    static struct sockaddr_in sin;

    static struct hostent *hp = NULL;
    struct hostent *gethostbyname();

    unsigned long gethostid();
    struct in_addr addr;
    char *inet_ntoa();

    portno = (unsigned short) portnum;

/* Find out what host we are on */

    if (local_name_unknown) {
	if (get_local_hostname() == -1) return(0xFFFFFFFF);
    }

/* Get our internet address */

    if (local_addr_unknown) {
    	addr.s_addr = gethostid();
    	strcpy(local_hostaddr,inet_ntoa(addr));
	local_addr_unknown = 0;
    }

    if (hp == NULL) {
    	hp = gethostbyname(local_hostname);
    	if (hp == NULL) {
	    printerr(FACILITY,ERROR_MSG,"Failed to get local host by name",
		NULL,3);
	    return(-1);
	}
    	sin.sin_family = AF_INET;
    	bcopy(hp->h_addr,(char *)&(sin.sin_addr),hp->h_length);
    }
    sin.sin_port = htons(portno);

/* Bind the socket to our address */

    retry2:
    if (bind(s,&sin,sizeof(struct sockaddr_in))) {
	if (errno == EINTR) goto retry2;
	printerr(FACILITY,ERROR_MSG,NULL,NULL,4);
	return(-1);
    }

    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      create_active_socket

  DESCRIPTION:

      Creates a socket which will be used to make a connection with a server
      socket located at a well-known address.

  FORMAL PARAMETERS:

      none

  FUNCTION VALUE:

      int - socket descriptor if the socket creation succeeded,
	    -1 otherwise.

								  */
/******************************************************************/

int create_active_socket()
{
    int s;
    int on = 1;

/* Create the socket */

    retry:
    s = socket(AF_INET,SOCK_STREAM,0);
    if (s < 0) {
	if (errno == EINTR) goto retry;
	printerr(FACILITY,ERROR_MSG,NULL,NULL,5);
	return(-1);
    }
    setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on));

/* Anchor the socket */

    anchor_socket(s,0);

    return(s);
}

/******************************************************************/
/*
    FUNCTION NAME:

      establish_connection

  DESCRIPTION:

  FORMAL PARAMETERS:

      int s - socket descriptor.
     
  FUNCTION VALUE:

      int - Zero if connection was successful, -1 if it failed.

								  */
/******************************************************************/

int establish_connection(s,rs)
int s;
struct sockaddr_in *rs;
{
    int on = 1;
    int level;
    struct protoent *proto;

/* Connect to the remote socket */

    retry:
    if (connect(s,rs,sizeof(struct sockaddr_in)) < 0) {
	if (errno == EINTR) goto retry;
	if (errno != ECONNREFUSED) printerr(FACILITY,ERROR_MSG,NULL,NULL,6);
	return(-1);
    }
    else {   /* Connection is complete */

/* Set the TCP_NODELAY socket option */

    	proto = getprotobyname("tcp");
	if (proto == NULL) proto = getprotobyname("TCP");
	if (proto == NULL) {
	    printerr(FACILITY,ERROR_MSG,"Failed to getprotobyname for TCP",
		NULL,76);
	    return -1;
	}

    	level = proto->p_proto;
    	setsockopt(s,level,TCP_NODELAY,&on,sizeof(on));

	return(0);
    }
}

/************************************************************************/
/*

    Function:  get_sockaddr

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

struct sockaddr_in *get_sockaddr(hostname,port)
char *hostname;
int port;
{
    static struct sockaddr_in sa;

    unsigned short portno;

    struct hostent *hp,*gethostbyname();

    char alias[64],*c;

    portno = (unsigned short) port;

/* Get hostent structure (network host entry) */

    again:

    hp = gethostbyname(hostname);
    if (hp != NULL) goto got_it;

/* Try to get the information by a common alias if possible */

    strcpy(alias,hostname);
    if ((c = strchr(alias,'.')) != NULL) {
	*c = 0;
	hp = gethostbyname(alias);
	if (hp != NULL) goto got_it;
    }

/* Could not get host entry */

    if (h_errno == TRY_AGAIN) {
    	sprintf(errmsg,"Failed to get host by name: %s, trying again",hostname);
    	printerr(FACILITY,ERROR_MSG,errmsg,NULL,7);
	sleep(1);
	sleep(1);
	sleep(1);
	sleep(1);
	goto again;
    }
    sprintf(errmsg,"Failed to get host by name: %s",hostname);
    printerr(FACILITY,ERROR_MSG,errmsg,NULL,7);
    return(NULL);

/* Got the host entry */

    got_it:
    bcopy(hp->h_addr,&sa.sin_addr,hp->h_length);
    sa.sin_port = htons(portno);
    sa.sin_family = AF_INET;

    return(&sa);
}

/************************************************************************/
/*

    Function:  initiate_connection

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int initiate_connection(s,hostname,port)
int s,port;
char *hostname;
{
    return(establish_connection(s,get_sockaddr(hostname,port)));
}

/******************************************************************/
/*
    FUNCTION NAME:

      create_passive_socket

  DESCRIPTION:

      Creates a socket which will accept connect requests by client processes.
      The name of the local host and the port number used are returned in
      the global variables local_hostname and assigned_port_number.

  FORMAL PARAMETERS:

      int portno1 -	Port number to be used.  Set this parameter
	to zero if you want the system to choose one for you.

  IMPLICIT OUTPUTS:

    The following global variables are set:

      assigned_port_number - The port number actually used.

      local_hostname - The name of the host machine the server is running
		       on.

      local_hostaddr - The internet address of the host machine in base
		     256 d.d.d.d representation.

      local_socket_name - Structure of type sockaddr_in of socket created.

  FUNCTION VALUE:

      int - The socket descriptor if a socket was successfully created,
	    -1 if the socket could not be created.

								  */
/******************************************************************/

int create_passive_socket(portno)
int portno;
{
    int s;
    int on = 1;
    int length;

/* Create a passive socket */

    retry1:
    s = socket(AF_INET,SOCK_STREAM,0);
    if (s < 0) {
	if (errno == EINTR) goto retry1;
	printerr(FACILITY,ERROR_MSG,NULL,NULL,8);
	return(-1);
    }
    setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on));

/* Anchor the socket */

    anchor_socket(s,portno);

/* Find out the port number assigned */

    retry2:
    length = sizeof(struct sockaddr_in);
    if (getsockname(s,&local_socket_name,&length)) {
	if (errno == EINTR) goto retry2;
	printerr(FACILITY,ERROR_MSG,NULL,NULL,9);
	return(-1);
    }
    assigned_port_number = (int) ntohs(local_socket_name.sin_port);

/* Apply the listen function so we can start accepting connections */

    listen(s,5);

    return(s);
}

/******************************************************************/
/*
    FUNCTION NAME:

      accept_connection

  DESCRIPTION:

      Accepts a connection and returns the socket descriptor of the socket
      created to communicate with the client.

  FORMAL PARAMETERS:

      int s - socket descriptor used solely for accepting connections.

  FUNCTION VALUE:

      int - socket descriptor of the socket created to talk to the newly
	    connected client.  If the connection fails, then a -1 will
	    be returned.  A -2 is returned if the connection already exists.

								  */
/******************************************************************/

int accept_connection(s)
int s;
{
    int ns; /* New socket descriptor */
    int on = 1;
    int level;
    struct protoent *proto;

    reaccept:
    ns = accept(s,0,0);
    if (ns == -1) {
	if (errno == EINTR) goto reaccept;
	printerr(FACILITY,ERROR_MSG,NULL,NULL,10);
    }

/* Set the TCP_NODELAY socket option */

    	proto = getprotobyname("tcp");
	if (proto == NULL) proto = getprotobyname("TCP");
	if (proto == NULL) {
	    printerr(FACILITY,ERROR_MSG,"Failed to getprotobyname for TCP",
		NULL,76);
	    return -1;
	}

    level = proto->p_proto;
    setsockopt(ns,level,TCP_NODELAY,&on,sizeof(on));

    return(ns);
}

/******************************************************************/
/*
    FUNCTION NAME:

      socket_close

  DESCRIPTION:

      Closes a socket.  Include here because the SRI_MULTINET package 
      supplies this subroutine to do a close on a socket.

  FORMAL PARAMETERS:

      int s - Socket to close.

								  */
/******************************************************************/

#ifndef VMS
socket_close(s)
int s;
{
    close(s);
}
#endif

/************************************************************************/
/*

    Function: socket_connected

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int socket_connected(sock)
int sock;
{
    struct sockaddr_in name;
    int namelen;

    int status;

    namelen = sizeof(struct sockaddr_in);
    status = getpeername(sock,&name,&namelen);

    return !status;
}

/************************************************************************/
/*

    Function: peer_hostname

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

char *peer_hostname(sock)
int sock;
{
    struct sockaddr_in name;
    int namelen;

    char *inet_ntoa();

    int status;

    namelen = sizeof(struct sockaddr_in);
    status = getpeername(sock,&name,&namelen);
    if (status == 0) {
	return (char *) inet_ntoa(name.sin_addr);
    }
    else {
	return NULL;
    }
}

/******************************************************************/
/*
    FUNCTION NAME:

      socket_wait

  DESCRIPTION:

      Waits until the sockets connected to the PM are ready for reading
      or writing.

  FORMAL PARAMETERS:

      int insock - Socket descriptor for reading messages from the PM.
		   Set to -1 if we are not interested in this socket.

      int outsock - Socket descriptor for writing messages to the PM.
		    Set to -1 if we are not interested in this socket.

  FUNCTION VALUE:

      (int) - 0 if the call to select was interrupted by a signal.
	      1 if insock is ready for reading.
	      2 if outsock is ready for writing.
	      3 if both insock and outsock are ready.
	     -1 if there was an error.
								  */
/******************************************************************/

#ifndef VMS

int socket_wait(insock,outsock)
int insock,outsock;
{
    int nb,rv;

    fd_set readmask,writemask;

    FD_ZERO(&readmask);
    FD_ZERO(&writemask);

    if (insock != -1) FD_SET(insock,&readmask);
    if (outsock != -1) FD_SET(outsock,&writemask);

    nb = select(FD_SETSIZE,&readmask,&writemask,(fd_set *) 0, NULL);
    if (nb <= 0)
	if (errno == EINTR)
	    return(0);
	else {
	    printerr(FACILITY,ERROR_MSG,NULL,NULL,11);
	    return(-1);
	}
 
    rv = 0;
    if (insock != -1) {
	if (FD_ISSET(insock,&readmask)) rv++;
    }
    if (outsock != -1) {
	if (FD_ISSET(outsock,&writemask)) rv += 2;
    }

    return(rv);
}

#endif

/******************************************************************/
/*
    FUNCTION NAME:

      socket_poll

  DESCRIPTION:

      Checks a socket to see if its ready for reading.

  FORMAL PARAMETERS:

      int insock - Socket descriptor for reading messages from the PM.

  FUNCTION VALUE:

      (int) - 0 if the socket is not ready yet.
	      1 if insock is ready for reading.
	     -1 if the call was interrupted by a signal.
	     -2 if there was another error.

								  */
/******************************************************************/
#ifndef VMS

int socket_poll(insock)
int insock;
{
    int nb;

    struct timeval t;

    fd_set readmask;

    t.tv_sec = 0;
    t.tv_usec = 0;

    FD_ZERO(&readmask);
    FD_SET(insock,&readmask);
    nb = select(FD_SETSIZE,&readmask,(fd_set *) 0,(fd_set *) 0, &t);
    if (nb == 0) {
	return(0);
    }
    if (nb < 0) {
	if (errno == EINTR) {
	    return(-1);
	}
	else {
	    printerr(FACILITY,ERROR_MSG,NULL,NULL,12);
	    return(-2);
	}
    }
    if (FD_ISSET(insock,&readmask)) {
	return(1);
    }
    return(0);
}

#endif

/******************************************************************/
/*
    FUNCTION NAME:

      send_block

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

static int send_block(s,buf,nbyte)
int s;
char *buf;
int nbyte;
{
    int n;

    sndblk:
    n = send(s,buf,nbyte,0);
    if (n <  0) {
	if (errno == ENOBUFS) {
	    printerr(FACILITY,WARNING_MSG,NULL,NULL,13);
	    goto sndblk;
	}
	else if (errno == EINTR) {
	    goto sndblk;
	}
#ifdef HPUX
	else if (errno == EAGAIN) {
	    goto sndblk;
	}
#else
	else if (errno == EWOULDBLOCK) {
	    goto sndblk;
	}
#endif
	else if ((errno == EHOSTUNREACH) || (errno == ENETUNREACH)) {
	    sleep(1);
	    goto sndblk;
	}
	printerr(FACILITY,ERROR_MSG,NULL,NULL,14);
	return(-1);
    }
    else if (n == 0) {
	printerr(FACILITY,ERROR_MSG,"connection lost",NULL,15);
        return(-1);
    }
    else if (n != nbyte) {
	nbyte -= n;
	buf += n;
	goto sndblk;
    }
    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      send_binary

  DESCRIPTION:

      Sends a block of binary data over a TCP socket.

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

int send_binary(s,buf,nbyte)
int s;
char *buf;
int nbyte;
{
    int bleft;
    char *ptr;

    struct msgbuff sb;

    if (inside_send_bin) abort();
    inside_send_bin = -1;

	bleft = nbyte;
	ptr = buf;
	do {
	    if (bleft <= 4096) {
		sb.more = 0;
		sb.nbyte = bleft;
		memcpy(sb.msg,ptr,bleft);
	    }
	    else {
		sb.more = -1;
		sb.nbyte = 4096;
		memcpy(sb.msg,ptr,4096);
	    }
	    if (send_block(s,&sb,sb.nbyte+2*sizeof(int)) < 0) {
    		inside_send_bin = 0;
		return(-1);  
	    }
	    bleft -= sb.nbyte;
	    ptr += sb.nbyte;

	} while (bleft > 0);

    inside_send_bin = 0;
    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      get_bcount

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

static int get_bcount(s,more,nbyte)
int s;
int *more;
int *nbyte;
{
    int n;
    struct msgbuff b;

    char *buf;
    int bytes_left = 2*sizeof(int);
    int index = 0;

    buf = (char *) &b;

    rdmsg:
    n = recv(s,&buf[index],bytes_left,0);
    if (n == 0) {
	return(-1);
    }
    else if (n < 0) {
	if (errno == EINTR) goto rdmsg;
#ifdef HPUX
	if (errno == EAGAIN) goto rdmsg;
#else
	if (errno == EWOULDBLOCK) goto rdmsg;
#endif
	printerr(FACILITY,ERROR_MSG,NULL,NULL,16);
	return(-1);
    }
    if (n < bytes_left) {
	bytes_left -= n;
	index += n;
	goto rdmsg;
    }
    *more = b.more;
    *nbyte = b.nbyte;
    return(n);
}

/******************************************************************/
/*
    FUNCTION NAME:

      read_binary

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

int read_binary(s,buf,nbyte)
int s;
char *buf;
int nbyte;
{
    int n;

    int more;
    int nb;

    char *pos;
    int bleft;
    int discard,d;
    int hole[128];

    char text[80];

    if (inside_read_bin) abort();
    inside_read_bin = -1;

    pos = buf;
    bleft = nbyte;
    discard = 0;
    do {
	if (get_bcount(s,&more,&nb) < 0) {
    	    inside_read_bin = 0;
	    return(-1);
	}
	if (nb > bleft) {
	    discard = nb - bleft;
	    if (discard > 0) {
		sprintf(text,
    "read_binary: buffer too small (%x), bcount=(%x,%x), sock = %d",
		    bleft,more,nb,s);
		printerr(FACILITY,ERROR_MSG,text,NULL,17);
	    }
	    nb = bleft;
	}

	while (nb > 0) {
	    rdmsg1:
	    n = recv(s,pos,nb,0);
	    if (n == 0) {
    		inside_read_bin = 0;
		return(-1);
	    }
	    else if (n < 0) {
		if (errno == EINTR) goto rdmsg1;
#ifdef HPUX
		if (errno == EAGAIN) goto rdmsg1;
#else
		if (errno == EWOULDBLOCK) goto rdmsg1;
#endif
		printerr(FACILITY,ERROR_MSG,NULL,NULL,18);
    		inside_read_bin = 0;
		return(-1);
	    }
	    pos += n;
	    bleft -= n;
	    nb -= n;
	}
	while (discard > 0) {
	    d = (discard < 512) ? discard : 512;    /* d = min(discard,512) */
	    rdmsg2:
	    n = recv(s,hole,d,0);
	    if (n == 0) {
		inside_read_bin = 0;
		return(-1);
	    }
	    else if (n < 0) {
		if (errno == EINTR) goto rdmsg2;
#ifdef HPUX
		if (errno == EAGAIN) goto rdmsg2;
#else
		if (errno == EWOULDBLOCK) goto rdmsg2;
#endif
		printerr(FACILITY,ERROR_MSG,NULL,NULL,19);
    		inside_read_bin = 0;
		return(-1);
	    }
	    discard -= n;
	}

    } while (more);

    inside_read_bin = 0;
    return(nbyte-bleft);
}

/************************************************************************/
/*

    Function: read_counted

    Description:

    Arguments:

    Function Value: 0 = messages successfully read
		   -1 = error, lost connection
		    1 = we want non-blocking sockets and the read would
			have blocked.

*/
/************************************************************************/

int read_counted(s,buf,nbyte)
int s;
char *buf;
int nbyte;
{
    char count;
    int n;

    char *pos;
    int nb,bleft;

    char text[256];

    pos = text;

    get_count:
    n = recv(s,&count,1,0);		/* Get the count (0-255) */
    if (n < 0) {
	if (errno == EINTR) goto get_count;
#ifdef HPUX
	if (errno == EAGAIN) {
#else
	if (errno == EWOULDBLOCK) {
#endif
	    if (mh_nbsocks) {		/* Return for non-blocking sockets */
		return 1;
	    }
	    else {
		goto get_count;
	    }
	}
	printerr(FACILITY,ERROR_MSG,NULL,NULL,20);
	return(-1);
    }
    else if (n == 0) {
	return(-1);
    }
    bleft = (int) count;
    nb = bleft + 1;
    text[bleft] = '\0';

    while (bleft > 0) {
	rdmsg1:
	n = recv(s,pos,bleft,0);
	if (n == 0) {
	    return(-1);
	}
	else if (n < 0) {
	    if (errno == EINTR) goto rdmsg1;
#ifdef HPUX
	    if (errno == EAGAIN) {
#else
	    if (errno == EWOULDBLOCK) {
#endif
	    	if (mh_nbsocks) {	/* Return for non-blocking sockets */
		    return 1;
	    	}
	    	else {
		    goto rdmsg1;
	    	}
	    }
	    printerr(FACILITY,ERROR_MSG,NULL,NULL,21);
	    return(-1);
	}
	pos += n;
	bleft -= n;
    }
    n = nb < nbyte ? nb : nbyte;
    strncpy(buf,text,n);
    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      send_counted

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

                                                                  */
/******************************************************************/

int send_counted(s,buf)
int s;
char *buf;
{
    int n,buflen;
    char nb;
    char text[80];

    buflen = strlen(buf);
    nb = (char) buflen;

    send_count:
    n = send(s,&nb,1,0);
    if (n <  0) {
	if (errno == ENOBUFS) {
	    printerr(FACILITY,WARNING_MSG,NULL,NULL,22);
	    goto send_count;
	}
	else if (errno == EINTR) {
	    goto send_count;
	}
#ifdef HPUX
	else if (errno == EAGAIN) {
#else
	else if (errno == EWOULDBLOCK) {
#endif
	    if (mh_nbsocks) {		/* Return for non-blocking sockets */
		return 1;
	    }
	    else {
		goto send_count;
	    }
	}
	else if ((errno == EHOSTUNREACH) || (errno == ENETUNREACH)) {
	    sleep(1);
	    goto send_count;
	}
	printerr(FACILITY,ERROR_MSG,NULL,NULL,23);
	return(-1);
    }
    if (n == 0) return(-1);

    sndblk:
    n = send(s,buf,buflen,0);
    if (n <  0) {
	if (errno == ENOBUFS) {
	    printerr(FACILITY,WARNING_MSG,NULL,NULL,24);
	    goto sndblk;
	}
	else if (errno == EINTR) {
	    goto sndblk;
	}
#ifdef HPUX
	else if (errno == EAGAIN) {
#else
	else if (errno == EWOULDBLOCK) {
#endif
	    if (mh_nbsocks) {		/* Return for non-blocking sockets */
		return 1;
	    }
	    else {
		goto sndblk;
	    }
	}
	else if ((errno == EHOSTUNREACH) || (errno == ENETUNREACH)) {
	    sleep(1);
	    goto sndblk;
	}
	printerr(FACILITY,ERROR_MSG,NULL,NULL,25);
	return(-1);
    }
    if (n != buflen) {
	sprintf(text,"Not all bytes sent: %d/%d",n,buflen);
	printerr(FACILITY,ERROR_MSG,text,NULL,26);
        return(-1);
    }
    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      read_message

  DESCRIPTION:

      Receives a message from a connected socket.  The read will block
      until data is available on the socket.  If the connection is
      destroyed while the read is pending, a -1 will be returned.
      Any other error will also cause a -1 to be returned.

  FORMAL PARAMETERS:

      int s - socket descriptor.

      char *buf - buffer to receive the message.

      int buflen - number of bytes in the buffer.

  FUNCTION VALUE:

      int - number of messages remaining to be read out of the buffer.
	    -1 if there was an error.
								  */
/******************************************************************/

int read_message( sock, return_message )
int sock;
char **return_message;
{
	char buffer[ MESSAGE_BUFFER_SIZE ];
	int nchars;

	static char *streambuffer = NULL;
	static int totalchars = 0;
	static int start = 0;
	static int number_of_messages = 0;

	int index;

/*	loop forever, returning when a complete message is found,
	warning, message must end in a NULL character */
	for(;;)
	{
		/*	If there are no complete messages in streambuffer,
			read more data from socket */
		if( number_of_messages == 0 )
		{
			nchars = recv( sock, buffer, MESSAGE_BUFFER_SIZE, 0 );
			if( nchars == 0 )
			{
				if (totalchars > 0) {
					buffer[0] = '\0';
					nchars = 1;
				}
				else {
					start = 0;
					number_of_messages = 0;
					return( -1 );
				}
			}
			else if( nchars < 0 )
			{
				if( errno == EINTR )
				{
					nchars = 0;
				}
				else
				{
					totalchars = 0;
					start = 0;
					number_of_messages = 0;
					printerr(FACILITY,ERROR_MSG,NULL,NULL,28);
					return( -1 );
				}
			}
			if (streambuffer == NULL)
			{
				streambuffer = (char *)malloc( totalchars + nchars );
			}
			else
			{
				streambuffer = (char *)realloc( streambuffer,
					totalchars + nchars );
			}

			if( streambuffer == NULL )
			{
				printerr(FACILITY,FATAL_MSG,
					"Malloc failed - no available memory or memory corrupted",
					NULL,10);
				exit(0);
			}

			/*	cannot do a strcpy because there may be several NULL
				characters in buffer */
			for( index = 0; index < nchars; index ++ )
				streambuffer[ totalchars + index ] = buffer[ index ];
			totalchars += nchars;

			/*	count number of messages in streambuffer */
			for( index = start; index < totalchars; index++ )
				if( streambuffer[ index ] == 0 )
					number_of_messages++;
		}

		for( index = start; index < totalchars; index++ )
		{
			/*	Search for complete message, must have NULL character */
			if( streambuffer[ index ] == 0 )
			{
				/* test for a solitary NULL character message */
				if( index == start )
				{
					start++;
					number_of_messages--;
					break;
				}
				/*	copy streambuffer to return_message, return_message
					will have a NULL character at the end */
					*return_message = (char *)hmalloc( index - start + 1 );
					strcpy( *return_message, &streambuffer[ start ] );

				/*	if the end of the message is also the end of the
					streambuffer, we can discard the streambuffer
					and reinitialize it */
				if( index == totalchars - 1 )
				{
					start = 0;
					totalchars = 0;
					free( streambuffer );
					streambuffer = NULL;
				}

				/*	set start to beginning of next message */
				else
					start = index + 1;

				/*	decrement number of messages to process
					since we are returning one */
				number_of_messages--;
				return( 0 );
			}
		}
	}
}

/************************************************************************/
/*

    Function:  socket_n_read

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int socket_n_read(s,buf,bleft)
int s;
char *buf;
int bleft;
{
    int n;

    while (bleft > 0) {
	rdmsg1:
	n = recv(s,buf,bleft,0);
	if (n == 0) {
	    return(-1);
	}
	else if (n < 0) {
	    if (errno == EINTR) goto rdmsg1;
#ifdef HPUX
	    if (errno == EAGAIN) goto rdmsg1;
#else
	    if (errno == EWOULDBLOCK) goto rdmsg1;
#endif
	    printerr(FACILITY,ERROR_MSG,NULL,NULL,27);
	    return(-1);
	}
	buf += n;
	bleft -= n;
    }

    return(0);
}

/************************************************************************/
/*

    Function:  socket_n_send

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int socket_n_send(s,buf,bleft)
int s,bleft;
char *buf;
{
    int n;
    int len;

    sndblk:
    len = bleft > PACKET_SIZE ? PACKET_SIZE : bleft;
    n = send(s,buf,len,0);
    if (n <  0) {
	if (errno == ENOBUFS) {
	    printerr(FACILITY,WARNING_MSG,NULL,NULL,28);
	    sleep(1);
	    goto sndblk;
	}
/*
	else if (errno == EPIPE) {
	    printerr(FACILITY,WARNING_MSG,NULL,NULL,28);
	    sleep(1);
	    goto sndblk;
	}
*/
	else if (errno == EINTR) {
	    goto sndblk;
	}
#ifdef HPUX
	else if (errno == EAGAIN) {
	    goto sndblk;
	}
#else
	else if (errno == EWOULDBLOCK) {
	    goto sndblk;
	}
#endif
	else if ((errno == EHOSTUNREACH) || (errno == ENETUNREACH)) {
	    sleep(1);
	    goto sndblk;
	}

	printerr(FACILITY,ERROR_MSG,NULL,NULL,29);
	return(-1);
    }
    buf += n;
    bleft -= n;
    if (bleft != 0) goto sndblk;

    return(0);
}

/************************************************************************/
/*

    Function:  socket_n_sum

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int socket_n_sum(s,buf,bleft)
int s;
int *buf;
int bleft;
{
    int n_recv,n,m,i;
    int len;
    int tmpbuf[(PACKET_SIZE/sizeof(int))+1];
    char *tmpcbuf;
    int leftover;

    tmpcbuf = (char *) tmpbuf;
    leftover = 0;
    while (bleft > 0) {
	rdmsg1:
    	len = bleft > PACKET_SIZE ? PACKET_SIZE : bleft;
	n_recv = n = recv(s,tmpcbuf,len,0);
	if (n == 0) {
	    return(-1);
	}
	else if (n < 0) {
	    if (errno == EINTR) goto rdmsg1;
#ifdef HPUX
	    if (errno == EAGAIN) goto rdmsg1;
#else
	    if (errno == EWOULDBLOCK) goto rdmsg1;
#endif
	    printerr(FACILITY,ERROR_MSG,NULL,NULL,27);
	    return(-1);
	}
	n += leftover;
	m = n/sizeof(int);
	leftover = (n % sizeof(int));
	for (i=0; i<m; i++) {
	    *buf += tmpbuf[i];
	    buf++;
	}
	tmpbuf[0] = tmpbuf[m];		/* Copy any possible leftover bytes */
	tmpcbuf = (char *) tmpbuf;
	tmpcbuf += leftover;
	bleft -= n_recv;
    }

    return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      socket_ready

  DESCRIPTION:

      Waits for a socket to become ready for reading, or until a timeout.

  FORMAL PARAMETERS:


  FUNCTION VALUE:

      (int) - 0 if the socket is not ready yet.
	      1 ready for reading.
	     -1 error in select.
								  */
/******************************************************************/

int socket_ready(s,maxt)
int s,maxt;
{
    int n;

    struct timeval t;

    fd_set readmask;

    t.tv_sec = maxt;
    t.tv_usec = 0;

retry:
    FD_ZERO(&readmask);
    FD_SET(s,&readmask);
    n = select(FD_SETSIZE,&readmask,(fd_set *) 0,(fd_set *) 0, &t);
    if (n < 0) {
	if (errno == EINTR) {
	    goto retry;
	}
	else {
	    printerr(FACILITY,ERROR_MSG,NULL,NULL,75);
	}
    }
    return n;
}
