/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1992 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: logmsg.c

    Explanation: 

    Exported Routines:


*/

#include <stdio.h>
#include <varargs.h>
#include "../inc/perr.h"

char logmsg_fac[80];

/************************************************************************/
/*

    Function:	logmsg

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

void logmsg(va_alist)
va_dcl
{
	va_list ap;

	char *p;

	int ival;
	double dval;
	char *sval;

	int n,sev;
	char *fmt;

	char msg[1024];
	int i;

	va_start(ap);
	n = va_arg(ap,int);
	sev = va_arg(ap,int);
	fmt = va_arg(ap,char *);
	if (fmt == NULL) {
		printerr(logmsg_fac,sev,NULL,NULL,n);
		if (sev == FATAL_MSG) exit(0);
		return;
	}

	msg[0] = '\0';
	i = 0;

	for (p = fmt; *p; p++) {

		if (*p != '%') {
			msg[i] = *p;
			msg[i+1] = '\0';
			i++;
			continue;
		}

		switch (*++p) {
		case 'd':
			ival = va_arg(ap,int);
			sprintf(&msg[i],"%d",ival);
			i = strlen(msg);
			break;
		case 'f':
			dval = va_arg(ap,double);
			sprintf(&msg[i],"%f",dval);
			i = strlen(msg);
			break;
		case 's':
			sval = va_arg(ap,char *);
			sprintf(&msg[i],"%s",sval);
			i = strlen(msg);
			break;
		default:
			msg[i] = *p;
			msg[i+1] = '\0';
			i++;
			break;
		}
	}

	va_end(ap);

	printerr(logmsg_fac,sev,msg,NULL,n);
	if (sev == FATAL_MSG) exit(0);
}
