/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/****************************************************************************/
/*									    */
/*   Message building & parsing routines:				    */
/*									    */
/*	msg_build - Builds a message given a command and arguments.	    */
/*	msg_parse - Retrieves a command and arguments from a message.	    */
/*									    */
/****************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include <stdio.h>

/********************/
/* Global Variables */
/********************/

char *msg_cmd;	    /* Message command */
char **msg_args;    /* Argument list terminated by a NULL string */


/******************************************************************/
/*
    FUNCTION NAME:

      msg_build

  DESCRIPTION:

      Builds a message from a command string and a set of argument strings.

  FORMAL PARAMETERS:

      char *msgcmd - Command string.

      char **msgargs - Argument strings terminated by a NULL string.

  FUNCTION VALUE:

      *char - Pointer to a string containing the message.
								  */
/******************************************************************/

char *msg_build(msgcmd,msgargs)
char *msgcmd;
char **msgargs;
{

	static char msg[1024];

	strcpy(msg,msgcmd);
	while(*msgargs != NULL) {
		strcat(msg,"|");
		strcat(msg,*msgargs);
		msgargs++;
	}

	return(msg);
}


/******************************************************************/
/*
    FUNCTION NAME:

      msg_parse

  DESCRIPTION:

      Parses a message into a command string and argument strings.

  FORMAL PARAMETERS:

      char *msg1 - Message to be parsed.

  IMPLICIT OUTPUTS:

      char *msg_cmd (global variable) - Command string.

      char **msg_args (global variable) - Argument strings.

  FUNCTION VALUE:

      none
								  */
/******************************************************************/

msg_parse(msg1)
char *msg1;
{
	static char *argptr[33];
	static char msg[1024];

	int i;
	char *s,*c,*strchr();

	char delim = '|';

	strcpy(msg,msg1);

	msg_cmd = msg;

	/* For each argument, replace the '|' character with a '\0' and set an
   element of argptr to point to the argument */

	i = 0;
	s = msg;
	while((c = strchr(s,delim)) != NULL) {
		*c = '\0';
		c++;
		s = c;
		argptr[i] = c;
		i++;
	}
	argptr[i] = NULL;
	msg_args = argptr;

}
