/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: multi_pm.c

    Explanation: 

    Exported Routines:


*/

#include <stdio.h>
#include <errno.h>
#include <search.h>
#include "../inc/types.h"
#include "../inc/time.h"
#include "../inc/perr.h"
#include "../inc/database.h"

#ifdef AIX
#include <sys/select.h>
#endif

#define FACILITY "MULTI_PMCOMM"

/* #define DIAG 1 */

static struct pm_info {

	int sock;			/* TCP socket */
	int portno;			/* Port number */
	char hostname[64];		/* IP Address */
	int connected;		/* -1 = connected, 0 = unconnected */
	int once;			/* True if we've once been connected */
	char domain[64];		/* NIS domain name */

};

static int hash_created = 0;
static int clear_db = -1;

static fd_set fdset;

static int firstsock = -1;
static int nextsock[FD_SETSIZE];
static struct pm_info *pmptr[FD_SETSIZE];

#define MAX_DOMAINS 32

static int ndomains = 0;
static char *domains[MAX_DOMAINS];

static struct pm_info *connecting_pm = NULL;

#define MPM_TAPE 1
#define MPM_PS 4
#define MPM_DOMAIN 5

extern struct database db;
extern int pm_sock;
extern int pm_port;
extern int connected_to_pm;
extern int once_connected;
extern char pm_hostname[64];
extern char pm_domain[80];
extern int reconnect_flag;
extern int asynch_flag;
extern int get_batch_domain;
extern int multi_pm_interface;

extern int db_clear_flag;

extern char *msg_cmd,**msg_args;


/************************************************************************/
/*

    Function: remove_sock

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int remove_sock(s)
int s;
{
	int s1;

start:
	/*
    printf("sockets: ");
    s1 = firstsock;
    while (s1 != -1) {
	printf("%d ",s1);
	s1 = nextsock[s1];
    }
    printf("\n");
    fflush(stdout);
*/
	if (firstsock == s) {
		firstsock = nextsock[firstsock];	/* Remove head */
		goto start;
	}

	s1 = firstsock;				/* Remove all occurances */
	while (s1 != -1) {
		if (nextsock[s1] == s) nextsock[s1] = nextsock[s];
		s1 = nextsock[s1];
	}

	return 0;
}

/************************************************************************/
/*

    Function: read_domains

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int read_domains()
{
	FILE *fp,*fopen();
	char fname[132],d[64];
	int n,i;

#ifdef DIAG
	printf("reading domains\n");
#endif

	sprintf(fname,"%s/.domains",getenv("HOME"));
	fp = fopen(fname,"r");
	if (fp == NULL) {
		printerr(FACILITY,FATAL_MSG,".domains file does not exist",NULL,3);
		exit(0);
	}
	i = 0;
	while ((n = fscanf(fp,"%s",d)) == 1) {
		domains[i] = (char *) hmalloc(strlen(d)+1);
		strcpy(domains[i],d);
#ifdef DIAG
		printf("domain=%s\n",domains[i]);
#endif
		i++;
	}
	fclose(fp);
#ifdef DIAG
	fflush(stdout);
#endif
	ndomains = i;
	return 0;
}

/************************************************************************/
/*

    Function: 	get_pm

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static struct pm_info *get_pm(t,str)
int t;
char *str;
{
	ENTRY item,*found_item,*hsearch();

	char key[256];

	struct pm_info *pm;

	sprintf(key,"%d%s",t,str);
	item.key = key;
#ifdef DIAG
	printf("get_pm: key=%s\n",key);
	fflush(stdout);
#endif

	found_item = (ENTRY *) hsearch(item,FIND);
	if (found_item == NULL) return NULL;
	pm = (struct pm_info *) found_item->data;
#ifdef DIAG
	printf("get_pm: pm=%s\n",pm->hostname);
	fflush(stdout);
#endif
	return (struct pm_info *)found_item->data;
}

/************************************************************************/
/*

    Function: new_pm

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static struct pm_info *new_pm(domain)
char *domain;
{
	struct pm_info *pm;
	char *d;

	ENTRY item,*hsearch();

#ifdef DIAG
	printf("new_pm: domain=%s\n",domain);
	fflush(stdout);
#endif

	/* Create a structure for the PM info */

	pm = (struct pm_info *) hmalloc(sizeof(struct pm_info));
	pm->sock = -1;
	pm->portno = -1;
	pm->connected = 0;
	pm->once = 0;
	strcpy(pm->domain,domain);

	/* Allocate some space for the domain */

	d = (char *) hmalloc(strlen(domain)+5);
	sprintf(d,"%d%s",MPM_DOMAIN,domain);

	/* Enter the PM into the hash table */

	item.key = d;
	item.data = (char *) pm;
	hsearch(item,ENTER);
#ifdef DIAG
	printf("new_pm: %s\n",item.key);
	fflush(stdout);
#endif

	return pm;
}

/************************************************************************/
/*

    Function: new_entry

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int new_entry(t,str,pm)
int t;
char *str;
struct pm_info *pm;
{
	char *d;

	ENTRY item,*hsearch(),*result;

#ifdef DIAG
	printf("new_entry: t=%d, str=%s\n",t,str);
	fflush(stdout);
#endif

	/* Allocate some space for the key */

	d = (char *) hmalloc(strlen(str)+5);
	sprintf(d,"%d%s",t,str);

	/* Enter the key into the hash table */

	item.key = d;
	item.data = (char *) pm;
	result = hsearch(item,FIND);
	if (result != NULL) {
#ifdef DIAG
		printf("Key already in table: %s\n",d);
		fflush(stdout);
#endif
		hfree(d);
	}
	else {
		result = hsearch(item,ENTER);
		if (result == NULL) {
			printerr(FACILITY,ERROR_MSG,"Hash table is full",NULL,4);
		}
	}
	return 0;
}

/************************************************************************/
/*

    Function: set_state

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int set_state(pm)
struct pm_info *pm;
{
#ifdef DIAG
	printf("set_state:\n");
	fflush(stdout);
#endif
	pm_sock = pm->sock;
	connected_to_pm = pm->connected;
	once_connected = pm->once;
	strcpy(pm_hostname,pm->hostname);
}

/************************************************************************/
/*

    Function: copyback_state

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int copyback_state(pm)
struct pm_info *pm;
{
#ifdef DIAG
	printf("copyback_state:\n");
	fflush(stdout);
#endif
	pm->connected = connected_to_pm;
	pm->sock = pm_sock;
	strcpy(pm->hostname,pm_hostname);
	pm->portno = pm_port;
	pm->once = once_connected;
}

/************************************************************************/
/*

    Function: hash_msg

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int hash_msg(pm)
struct pm_info *pm;
{

	int msgtyp;
	char *simple_hostname();
	char temp[132];

	if (strcmp(msg_cmd,"db_update") == 0) {

		msgtyp = atoi(msg_args[0]);

		switch(msgtyp) {
		case DB_ADD_TAPE :
			sprintf(temp,"%s:%s",simple_hostname(msg_args[3]),msg_args[1]);
#ifdef DIAG
			printf("Adding tape: %s\n",temp);
			fflush(stdout);
#endif
			new_entry(MPM_TAPE,temp,pm);
			break;
		case DB_ADD_PRODSYS :
			new_entry(MPM_PS,msg_args[1],pm);
#ifdef DIAG
			printf("Adding prodsys: %s\n",msg_args[1]);
			fflush(stdout);
#endif
			break;
		default :
			break;
		}

	}

}

/************************************************************************/
/*

    Function: getdomainbytape
	      getdomainbyps

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

char *getdomainbytape(hostname,devname)
char *hostname,*devname;
{
	char temp[132],*simple_hostname();

	struct pm_info *pm;

	sprintf(temp,"%s:%s",simple_hostname(hostname),devname);
	pm = get_pm(MPM_TAPE,temp);
	if (pm == NULL) return NULL;
	return pm->domain;
}

char *getdomainbyps(prodname)
char *prodname;
{
	struct pm_info *pm;

	pm = get_pm(MPM_PS,prodname);
	if (pm == NULL) return NULL;
	return pm->domain;
}

/************************************************************************/
/*

    Function: m_pm_startup_msg

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int m_pm_startup_msg()
{
	if (connecting_pm != NULL) hash_msg(connecting_pm);
}

/************************************************************************/
/*

    Function: select_pm

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static struct pm_info *select_pm(msg)
char *msg;
{
	int msgtyp;

	char temp[132],*simple_hostname();

	struct pm_info *pm;

	msg_parse(msg);

	if (strcmp(msg_cmd,"db_update") == 0) {

		msgtyp = atoi(msg_args[0]);

		switch(msgtyp) {
		case DB_ADD_CPU :
		case DB_MODIFY_CPU :
		case DB_DELETE_CPU :
			printerr(FACILITY,ERROR_MSG,
			    "Cannot select PM for adding/modifying/deleting CPUs"
			    ,NULL,6);
			pm = NULL;
			break;
		case DB_ADD_TAPE :
			printerr(FACILITY,ERROR_MSG,
			    "Cannot select PM for adding tape",NULL,7);
			pm = NULL;
			break;
		case DB_ADD_PRODSYS :
			printerr(FACILITY,ERROR_MSG,
			    "Cannot select PM for adding production system",NULL,8);
			pm = NULL;
			break;
		case DB_ADD_FILESYSTEM :
		case DB_REMOVE_FILESYSTEM :
			printerr(FACILITY,ERROR_MSG,
			    "Cannot select PM for adding/deleting filesystem"
			    ,NULL,9);
			pm = NULL;
			break;
		case DB_MODIFY_TAPE :
			sprintf(temp,"%s:%s",simple_hostname(msg_args[3]),msg_args[1]);
			pm = get_pm(MPM_TAPE,temp);
			break;
		case DB_DELETE_TAPE :
		case DB_MOUNT_INFO :
			sprintf(temp,"%s:%s",simple_hostname(msg_args[2]),msg_args[1]);
			pm = get_pm(MPM_TAPE,temp);
			break;
		case DB_MODIFY_PRODSYS :
		case DB_DELETE_PRODSYS :
		case DB_ADD_CPU_TO_PROD :
		case DB_MODIFY_CPU_IN_PROD :
		case DB_DELETE_CPU_FROM_PROD :
		case DB_ADD_TAPE_TO_PROD :
		case DB_DELETE_TAPE_FROM_PROD :
		case DB_ADD_JOB :
		case DB_FILL_SUBMIT_INFO :
		case DB_ADVANCE_JOB :
		case DB_REMOVE_JOB :
		case DB_SET_JOB_AREA :
		case DB_UPDATE_JOB_STATUS :
		case DB_ADD_JOB_CLASS :
		case DB_ADD_TAPE_LIST :
		case DB_ADD_TAPE_REEL :
		case DB_INSERT_TAPE_REEL :
		case DB_REMOVE_TAPE_REEL :
		case DB_TAPE_REEL_STATUS :
		case DB_MOVE_JOB_TO_QUEUE :
		case DB_MOVE_JOB_TO_ON_DECK :
		case DB_MOVE_JOB_TO_CURRENT :
		case DB_MOVE_JOB_TO_RUNDOWN :
		case DB_ALLOCATE_AREA :
		case DB_FREE_AREA :
		case DB_JCSS_PID :
		case DB_SET_ABORT_FLAG :
		case DB_CLEAR_ABORT_FLAG :
		case DB_SET_PHASE_TIMES :
		case DB_TAPE_TIME_REQ :
		case DB_TAPE_TIME_MNT :
		case DB_TAPE_TIME_DONE :
		case DB_ADD_USER :
		case DB_REMOVE_USER :
		case DB_SET_NEXT_ENTRY :
			pm = get_pm(MPM_PS,msg_args[1]);
			break;
		default :
			sprintf(temp,"Cannot route message, type=%d",msgtyp);
			printerr(FACILITY,ERROR_MSG,temp,NULL,10);
			pm = NULL;
			break;
		}

	}
	else if (strcmp(msg_cmd,"m_request") == 0) {
		sprintf(temp,"%s:%s",simple_hostname(msg_args[12]),msg_args[11]);
		pm = get_pm(MPM_TAPE,temp);
#ifdef DIAG
		printf("Looking for tape: %s\n",temp);
		if (pm != NULL)
			printf("Found!\n");
		else
			printf("Not found!\n");
		fflush(stdout);
#endif
	}
	else if (strcmp(msg_cmd,"m_reply") == 0) {
		sprintf(temp,"%s:%s",simple_hostname(msg_args[11]),msg_args[10]);
		pm = get_pm(MPM_TAPE,temp);
#ifdef DIAG
		printf("Looking for tape: %s\n",temp);
		if (pm != NULL)
			printf("Found!\n");
		else
			printf("Not found!\n");
		fflush(stdout);
#endif
	}
	else {
		sprintf(temp,"Cannot route message: %s",msg);
		printerr(FACILITY,ERROR_MSG,temp,NULL,11);
		pm = NULL;
	}

	return pm;
}

/************************************************************************/
/*

    Function: m_pm_connect

    Description:

    Arguments:	ndom - Number of domains.
		domlst - List of domain names.
		stat - Status of each PM connection, 
			 0 = connected,
			-1 = disconnected

    Function Value:

*/
/************************************************************************/

int m_pm_connect(ndom,dom,stat)
int ndom;
char *dom[];
int stat[];
{
	int i;

	struct pm_info *pm;

	int stat1[MAX_DOMAINS];

	int ndom1;

	get_batch_domain = 0;
	ndom1 = ndom;

	if (ndom == -1) {
		stat = stat1;
		dom = domains;
	}

	/* Create a hash table if one hasn't already been created */

	if (!hash_created) {
		hcreate(2500);
		hash_created = -1;
	}

	/* Check to see if the database has been initialized */

	if (clear_db) {
		clear_db = -1;
		FD_ZERO(&fdset);
		db_clear_flag = -1;

		/* Read .domains file */

		if (ndom == -1) {
			read_domains();
		}
	}
	else {
		db_clear_flag = 0;
	}

	if (ndom == -1) {
		ndom = ndomains;
	}

	/* Make sure we are connected to each PM in the list */

	for (i=0; i<ndom; i++) {

		/* Find the PM info record for this domain */

		pm = get_pm(MPM_DOMAIN,dom[i]);
		if (pm == NULL) {
			pm = new_pm(dom[i]);
		}

		/* If we are already connected, go to the next PM */

		if (pm->connected) {
#ifdef DIAG
			printf("Already connected\n");
			fflush(stdout);
#endif
			stat[i] = 0;
			continue;
		}

		/* Connect to this PM (and get its database) */

		set_state(pm);
		strcpy(pm_domain,dom[i]);

		multi_pm_interface = -1;
		connecting_pm = pm;
		if (pm_connect() == 0) {	/* Connection succeeded */
			db_clear_flag = 0;
			copyback_state(pm);
			stat[i] = 0;
			FD_SET(pm_sock,&fdset);
			nextsock[pm_sock] = firstsock;
			firstsock = pm_sock;
			pmptr[pm_sock] = pm;
#ifdef DIAG
			printf("Connected to PM: %s,%d\n",pm->hostname,pm->portno);
			fflush(stdout);
#endif
		}
		else {
			stat[i] = -1;
		}

	}

	/* If domains were explicitly passed, copy domain list to global variables */

	if (ndom1 != -1) {
		for (i=0; i<ndom1; i++) {
			pm = get_pm(MPM_DOMAIN,dom[i]);
			domains[i] = pm->domain;
		}
		ndomains = ndom1;
	}

	return 0;
}

/************************************************************************/
/*

    Function: m_pm_reconnect

    Description: Attempt to reconnect to any PM's with which we've lost
		 connections.

    Arguments:

    Function Value:

*/
/************************************************************************/

int m_pm_reconnect()
{
	int i;

	int connection_lost[MAX_DOMAINS];

	int n_reconnect = 0;

	struct pm_info *pm;

	/* Check to see which PMs are alive */

	m_pm_check(-1,0,connection_lost);

	for (i=0; i<ndomains; i++) {
		if (connection_lost[i]) {
			pm = get_pm(MPM_DOMAIN,domains[i]);
			if (pm != NULL) {
				set_state(pm);
				strcpy(pm_domain,domains[i]);
				multi_pm_interface = -1;
				connecting_pm = pm;
				db_clear_flag = 0;
				if (pm_connect() == 0) {
					copyback_state(pm);
					FD_SET(pm_sock,&fdset);
					nextsock[pm_sock] = firstsock;
					firstsock = pm_sock;
					pmptr[pm_sock] = pm;
					n_reconnect++;
				}
			}
		}
	}

	return n_reconnect;
}

/************************************************************************/
/*

    Function: m_pm_disconnect

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int m_pm_disconnect(ndom,dom)
int ndom;
char *dom[];
{

	int i,j;

	struct pm_info *pm;

	if (ndom == -1) {
		ndom = ndomains;
		dom = domains;
	}

	for (i=0; i<ndom; i++) {

		pm = get_pm(MPM_DOMAIN,dom[i]);
		if (pm == NULL) {
			pm = new_pm(dom[i]);
		}

		if (pm->connected) {
			FD_CLR(pm->sock,&fdset);
			socket_close(pm->sock);
			pm->connected = 0;
		}
	}

	return 0;
}

/************************************************************************/
/*

    Function: partial_delete

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

static int partial_delete(pm)
struct pm_info *pm;
{
	struct pm_info *pm1;

	struct tape_drive_list *t;

	struct prod_sys *p;

	char temp[256],temp1[64];

	/* Delete production systems belonging to this PM */

	p = db.ps;
	while (p != NULL) {

		pm1 = get_pm(MPM_PS,p->ps_name);
		if (pm1 == pm) {
#ifdef DIAG
			printf("partial_delete: deleting production system %s\n",
			    p->ps_name);
			fflush(stdout);
#endif
			strcpy(temp,p->ps_name);
			delete_prodsys(&(db.ps),temp);
			p = db.ps;
		}
		else {
			p = p->next;
		}

	}

	/* Delete tape drives belonging to this PM */

	t = db.tape;
	while (t != NULL) {

		pm1 = get_pm(MPM_TAPE,temp);

		if (pm == pm1) {
			sprintf(temp,"%s:%s",simple_hostname(t->cpu->cpu_name),t->dev_name);
			strcpy(temp,t->cpu->cpu_name);
			strcpy(temp1,t->dev_name);
			delete_tape(&(db.tape),temp,temp1);
			t = db.tape;
		}
		else {
			t = t->next;
		}

	}

	return 0;
}

/************************************************************************/
/*

    Function: m_pm_check

    Description:

    Arguments:	ndom - Number of domains.
		domlst - List of domain names.
		stat - Status of each PM connection, 
			 0 = connected,
			-1 = disconnected

    Function Value:

*/
/************************************************************************/

int m_pm_check(ndom,dom,stat)
int ndom;
char *dom[];
int stat[];
{
	int i;

	struct pm_info *pm;

	int retstat = 0;

	if (ndom == -1) {
		dom = domains;
		ndom = ndomains;
	}

	for (i=0; i<ndom; i++) {

		pm = get_pm(MPM_DOMAIN,dom[i]);
		if (pm != NULL) {

			if (pm->sock != -1) {
				if (!socket_connected(pm->sock)) {
					FD_CLR(pm->sock,&fdset);
					socket_close(pm->sock);
					remove_sock(pm->sock);
					pm->sock = -1;
					pm->connected = 0;
				}
			}

			stat[i] = !(pm->connected);
			if (stat[i] != 0) {	    /* Connection was lost */
				partial_delete(pm);	/* Delete production systems & 
									   tape drives belonging to this PM */
				retstat = -1;
			}
		}
		else {
			stat[i] = -1;
		}
	}

	return retstat;
}

/************************************************************************/
/*

    Function: m_send_to_pm

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int m_send_to_pm(msg)
char *msg;
{
	struct pm_info *pm;

	int status;

	char tmpmsg[160];

	/* Determine which PM is to get the message */

	pm = select_pm(msg);

	if (pm == NULL) {		/* Error */
		sprintf(tmpmsg,"PM not found for msg: %s",msg);
		printerr(FACILITY,ERROR_MSG,tmpmsg,NULL,5);
		return -1;
	}

	set_state(pm);
	status = send_to_pm(msg);
	copyback_state(pm);
	return status;
}

/************************************************************************/
/*

    Function: m_pm_poll

    Description:

    Arguments:

    Function Value:  	 0 = Non-db_update message received.
			 1 = db_update message received.
			 2 = No messages, may have lost a connection.
			-1 = Not connected to any PMs.
			-2 = Error in call to select.

*/
/************************************************************************/

int m_pm_poll()
{
	fd_set fds;
	int n;
	int sock;
	int aflag;
	int status[FD_SETSIZE];

	struct timeval poll_t;

	poll_t.tv_sec = 0;
	poll_t.tv_usec = 0;

start:

	/* Check to see if we are connected to any PMs */

	if (firstsock == -1) return -1;	/* Not connected to any PMs */

	/* Poll the sockets */

select_msg:
	fds = fdset;
	n = select(FD_SETSIZE,&fds,(fd_set *)0,(fd_set *)0,&poll_t);
	if (n < 0) {
		if (errno == EINTR) goto select_msg;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
		return -2;
	}
	if (n == 0) {
		return 2;
	}
	aflag = asynch_flag;
	asynch_flag = -1;

	/* Check each socket */

	sock = firstsock;
	while (sock != -1) {
		if (FD_ISSET(sock,&fds)) {
			set_state(pmptr[sock]);
			status[sock] = read_from_pm();
			if (status[sock] == -1) {
				copyback_state(pmptr[sock]);
				FD_CLR(sock,&fdset);
				socket_close(sock);
				pmptr[sock]->sock = -1;
				pmptr[sock]->connected = 0;
				remove_sock(sock);
				goto start;
			}
			if ((status[sock] == 0) ||
			    (status[sock] == 1)) {

				/* Possibly enter something into the hash table */

				hash_msg(pmptr[sock]);

			}
			copyback_state(pmptr[sock]);
			if (status[sock] == 0) {
				asynch_flag = aflag;
				return 0;
			}
			if (aflag && (status[sock] == 1)) {
				return 1;
			}
		}
		sock = nextsock[sock];
	}

	goto start;
}

/************************************************************************/
/*

    Function: m_read_from_pm

    Description:

    Arguments:

    Function Value:  	 0 = Non-db_update message received.
			 1 = db_update message received.
			 2 = Never returns this status.
			-1 = Not connected to any PMs.
			-2 = Error in call to select.

*/
/************************************************************************/

int m_read_from_pm()
{
	fd_set fds;
	int n;
	int sock;
	int aflag;
	int status[FD_SETSIZE];

start:

	/* Check to see if we are connected to any PMs */

	if (firstsock == -1) return -1;

	/* Wait for data to arrive on one of the sockets */

select_msg:
	fds = fdset;
	n = select(FD_SETSIZE,&fds,(fd_set *)0,(fd_set *)0,0);
	if (n < 0) {
		if (errno == EINTR) goto select_msg;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
		return -2;
	}
	if (n == 0) {
		printerr(FACILITY,WARNING_MSG,"Zero status return from select",NULL,12);
		goto select_msg;
	}
	aflag = asynch_flag;
	asynch_flag = -1;

	/* Check each socket */

	sock = firstsock;
	while (sock != -1) {
		if (FD_ISSET(sock,&fds)) {
			set_state(pmptr[sock]);
			status[sock] = read_from_pm();
			if (status[sock] == -1) {
				copyback_state(pmptr[sock]);
				FD_CLR(sock,&fdset);
				socket_close(sock);
				pmptr[sock]->sock = -1;
				pmptr[sock]->connected = 0;
				remove_sock(sock);
				goto start;
			}
			if ((status[sock] == 0) ||
			    (status[sock] == 1)) {

				/* Possibly enter something into the hash table */

				hash_msg(pmptr[sock]);

			}
			copyback_state(pmptr[sock]);
			if (status[sock] == 0) {
				asynch_flag = aflag;
				return 0;
			}
			if (aflag && (status[sock] == 1)) {
				return 1;
			}
		}
		sock = nextsock[sock];
	}

	goto start;
}

/************************************************************************/
/*

    Function:  m_pm_msg_rdy

    Description:

    Arguments:

    Function Value:  -1 = error
		     otherwise returns the number of sockets with data ready.

*/
/************************************************************************/

int m_pm_msg_rdy() {

	fd_set fds;
	int n;
	struct timeval poll_t;

	poll_t.tv_sec = 0;
	poll_t.tv_usec = 0;

	/* Check to see if we are connected to any PMs */

	if (firstsock == -1) {
		printerr(FACILITY,ERROR_MSG,"Not connected to any PMs",NULL,14);
		return -1;
	}

	/* Poll sockets to see how many have data */

	fds = fdset;
	n = select(FD_SETSIZE,&fds,(fd_set *)0,(fd_set *)0,&poll_t);
	if (n < 0) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,13);
		return n;
	}

	return n;
}
