/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*                                                                      */
/*                                                                      */
/*                       Message Queue Utility Routines                 */
/*                                                                      */
/*                                                                      */
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include <stdio.h>
#include "../inc/param.h"
#include "../inc/cps_q.h"


/******************************************************************/
/*
    FUNCTION NAME:

      cps_create_queue

  DESCRIPTION:

      Creates a message queue.

								  */
/******************************************************************/

struct cps_msg_queue *cps_create_queue()
{
	struct cps_msg_queue *q;

	q = (struct cps_msg_queue *) hmalloc(sizeof(struct cps_msg_queue));
	q->front = NULL;
	q->back = NULL;
	q->len = 0;
	q->lock = 0;
	return(q);
}

/******************************************************************/
/*
    FUNCTION NAME:

      cps_enter_q

  DESCRIPTION:

      Places an element onto a queue.

  FORMAL PARAMETERS:

      struct cps_msg_queue *q - Queue to enter onto.

      struct cps_queue_element *e - Element to enter.

								  */
/******************************************************************/


cps_enter_q(q,e)
struct cps_msg_queue *q;
struct cps_queue_element *e;
{

	if (q->lock == 1) {
		printf("Recursive queue operation (cps_enter_q)\n");
		exit(1);
	}

	/* Place element in the back of the queue */

	q->lock = 1;	    /* Operation in progress */
	if (q->back == NULL) {
		q->back = e;
		q->front = e;
		q->len = 1;
		e->next = NULL;
	}
	else {
		q->back->next = e;
		q->back = e;
		q->len++;
		e->next = NULL;
	}
	q->lock = 0;

}

/******************************************************************/
/*
    FUNCTION NAME:

      cps_remove_q

  DESCRIPTION:

      Removes an element from a queue.

  FORMAL PARAMETERS:

      struct cps_msg_queue *q - Queue to remove element from.

  FUNCTION VALUE:

      (struct cps_queue_element *) - Element removed from the queue.

								  */
/******************************************************************/

struct cps_queue_element *cps_remove_q(q)
struct cps_msg_queue *q;
{

	struct cps_queue_element *e,*e1;

	if (q->lock == 1) {
		printf("Recursive queue operation (cps_remove_q)\n");
		exit(1);
	}

	q->lock = 1;
	e = q->front;
	if (e == NULL) {
		q->lock = 0;
		return(e);
	}
	e1 = e->next;
	e->next = NULL;
	q->front = e1;
	q->len--;
	if (e1 == NULL) q->back = NULL;
	q->lock = 0;
	return(e);

}

/******************************************************************/
/*
    FUNCTION NAME:

      cps_alloc_qelem

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

struct cps_queue_element *cps_alloc_qelem(free_q,size)
struct cps_msg_queue *free_q;
int size;
{
	struct cps_queue_element *e;

	if (free_q == NULL) free_q = cps_create_queue();
	e = cps_remove_q(free_q);
	if (e == NULL) {
		e = (struct cps_queue_element *) hmalloc(sizeof(struct cps_queue_element));
		if (size > 0)
			e->msg = (char *) hmalloc(size+1);
		else
			e->msg = NULL;
	}
	return(e);
}

/******************************************************************/
/*
    FUNCTION NAME:

      cps_dealloc_qelem

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

								  */
/******************************************************************/

cps_dealloc_qelem(free_q,e)
struct cps_msg_queue *free_q;
struct cps_queue_element *e;
{
	cps_enter_q(free_q,e);
}

/******************************************************************/
/*
    FUNCTION NAME:

      cps_extract_from_queue

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

cps_extract_from_queue(lpn,q,fq,zq)
int lpn;
struct cps_msg_queue *q,*fq,*zq;
{
	struct cps_queue_element *e;

	e = cps_alloc_qelem(zq,0);
	e->lpn = -1;
	cps_enter_q(q,e);
	do {
		e = cps_remove_q(q);
		if (e->lpn == lpn)
			cps_enter_q(fq,e);
		else if (e->lpn == -1)
			cps_enter_q(zq,e);
		else
			cps_enter_q(q,e);
	} while (e->lpn != -1);
}

cps_extract1_from_queue(lpn,q,fq,zq)
int lpn;
struct cps_msg_queue *q,*fq,*zq;
{
	struct cps_queue_element *e;
	int flag = 1;

	e = cps_alloc_qelem(zq,0);
	e->lpn = -1;
	cps_enter_q(q,e);
	do {
		e = cps_remove_q(q);
		if (flag && (e->lpn == lpn)) {
			cps_enter_q(fq,e);
			flag = 0;
		}
		else if (e->lpn == -1)
			cps_enter_q(zq,e);
		else
			cps_enter_q(q,e);
	} while (e->lpn != -1);
}

/******************************************************************/
/*
    FUNCTION NAME:

      cps_find_qelem

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

struct cps_queue_element *cps_find_qelem(q,lpn)
int lpn;
struct cps_msg_queue *q;
{
	struct cps_queue_element *e;

	e = q->front;
	while (e != NULL) {
		if (e->lpn == lpn) break;
		e = e->next;
	}
	return(e);
}

cps_printq(q)
struct cps_msg_queue *q;
{
	struct cps_queue_element *e;
	printf("queue: ");
	e = q->front;
	while (e != NULL) {
		printf("%d ",e->lpn);
		e = e->next;
	}
	printf("\n");
	fflush(stdout);
}
