/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/



#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <errno.h>

#include "../inc/perr.h"
#include "../inc/param.h"
#include "../inc/shm.h"

#define FACILITY "SEMLIB"

int q_semid;

static int nsems = 0;	/* Number of semaphores */

struct qelem *qelem;
int *q_head,*q_tail;
int *q_block;

static int inside_q_pop = 0;
static int inside_q_push = 0;

int set_nsems(n)
int n;
{ 
	nsems = n; 
}

print_qinfo(str)
char *str;
{
	/*
    struct qelem *q;

    int i,nl,ns;

    printf("%s ",str);

    for (i=0; i<nsems; i++) {
	if (q_head[i] == -1)
	    q = NULL;
	else
	    q = &qelem[q_head[i]];
	nl = ns = 0;
	while (q != NULL) {
	    if (q->size == 0)
		ns++;
	    else
		nl++;
	    if (q->next == -1)
		q = NULL;
	    else
		q = &(qelem[q->next]);
	}
	printf("(%d:%d,%d) ",i,ns,nl);
    }
    printf("\n");
    fflush(stdout);
*/
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_init

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

q_init(n)
int n;		/* Number of semaphores */
{
	int i;

	int nq1,nq2;

	nsems = n;

	q_block[0] = 0;	/* Large free queue */
	q_block[n-1] = 0;	/* Small free queue */

	/* User & shm queues are empty */

	for (i=1; i<n-1; i++) {
		q_block[i] = 0;		/* Unblocked */
		q_head[i] = -1;
		q_tail[i] = -1;
	}

	/* Put all large queue elements on the large free queue */

	nq1 = SHM_FUNC(n);
	q_head[0] = 0;
	q_tail[0] = nq1-1;
	for (i=0; i<nq1; i++) {
		qelem[i].index = i;
		qelem[i].size = 1;
		qelem[i].buf = i;
		if (i == (nq1 - 1))
			qelem[i].next = -1;
		else
			qelem[i].next = i + 1;
	}

	/* Put all small queue elements on the small free queue */

	nq2 = S_SHM_FUNC(n);
	q_head[n-1] = nq1;
	q_tail[n-1] = nq1+nq2-1;
	for (i=nq1; i<nq1+nq2; i++) {
		qelem[i].index = i;
		qelem[i].size = 0;
		qelem[i].buf = i-nq1;
		if (i == (nq1+nq2-1))
			qelem[i].next = -1;
		else
			qelem[i].next = i + 1;
	}

}

/******************************************************************/
/*
    FUNCTION NAME:

      sem_init

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

int sem_init(n)
int n;			/* Number of semaphores */
{

	union semun {
		int val;
		struct semid_ds *buf;
		ushort *array;
	} arg;

	int semflg;
	int i;

	semflg = 0666 | IPC_CREAT | IPC_EXCL;

	arg.val = 1;
	if ((q_semid = semget(IPC_PRIVATE,2*n,semflg)) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,1);
		return(-1);
	}
	else if (semctl(q_semid,0,SETVAL,arg) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,2);
		return(-1);
	}
	else if (semctl(q_semid,n-1,SETVAL,arg) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,7);
		return(-1);
	}
	else {
		arg.val = 0;
		for (i=1; i<n-1; i++) {
			if (semctl(q_semid,i,SETVAL,arg) == -1) {
				printerr(FACILITY,ERROR_MSG,NULL,NULL,3);
				return(-1);
			}
		}
		arg.val = 1;
		for (i=n; i<2*n; i++) {
			if (semctl(q_semid,i,SETVAL,arg) == -1) {
				printerr(FACILITY,ERROR_MSG,NULL,NULL,5);
				return(-1);
			}
		}
	}

	return(0);

}

/******************************************************************/
/*
    FUNCTION NAME:

      q_pop

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

struct qelem *q_pop(i)
int i;
{
	struct qelem *q;

	struct sembuf sops[2];

	int nsops;

	if (inside_q_pop) {
		printerr(FACILITY,ERROR_MSG,"Re-entered q_pop",NULL,13);
		abort();
	}
	inside_q_pop = -1;

	/* Wait for semaphores allowing q-op and non-empty queue */

	sops[0].sem_num = (short) i;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) nsems+i;
	sops[1].sem_op = (short) -1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,2) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,6);
		inside_q_pop = 0;
		return NULL;
	}

	/* Pop */

	if (q_head[i] == -1)
		q = NULL;
	else
		q = &qelem[q_head[i]];

	if (q == NULL) abort();		/* Queue was empty, this is an error */

	if (qelem[q_head[i]].next == -1) {	/* List is now empty */
		q_head[i] = -1;
		q_tail[i] = -1;
	}
	else {
		q_head[i] = q->next;
	}

	/* Indicate that the operation is over, and that the queue is empty */

	if (q_head[i] == -1)
		nsops = 1;
	else
		nsops = 2;

	sops[0].sem_num = (short) nsems+i;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) i;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,8);
		abort();
	}

	inside_q_pop = 0;
	return(q);
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_pop_retry

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

struct qelem *q_pop_retry(i)
int i;
{
	struct qelem *q;

	struct sembuf sops[2];

	int nsops;

	if (inside_q_pop) {
		printerr(FACILITY,ERROR_MSG,"Re-entered q_pop",NULL,20);
		abort();
	}
	inside_q_pop = -1;

	/* Wait for semaphores allowing q-op and non-empty queue */

retry1:
	sops[0].sem_num = (short) i;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) nsems+i;
	sops[1].sem_op = (short) -1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,2) == -1) {
		if (errno == EINTR) goto retry1;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,25);
		abort();
	}

	/* Pop */

	if (q_head[i] == -1)
		q = NULL;
	else
		q = &qelem[q_head[i]];

	if (q == NULL) abort();

	if (qelem[q_head[i]].next == -1) {	/* List is now empty */
		q_head[i] = -1;
		q_tail[i] = -1;
	}
	else {
		q_head[i] = q->next;
	}

	/* Indicate that the operation is over, and that the queue is empty */

	if (q_head[i] == -1)
		nsops = 1;
	else
		nsops = 2;

retry2:
	sops[0].sem_num = (short) nsems+i;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) i;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		if (errno == EINTR) goto retry2;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,26);
		abort();
	}

	inside_q_pop = 0;
	return(q);
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_pop_no_wait

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

struct qelem *q_pop_no_wait(i)
int i;
{
	struct qelem *q;

	struct sembuf sops[2];

	int nsops;

	char msg[80];

	if (inside_q_pop) {
		printerr(FACILITY,ERROR_MSG,"Re-entered q_pop",NULL,16);
		abort();
	}
	inside_q_pop = -1;

	/* Check semaphores allowing q-op and non-empty queue */

	sops[0].sem_num = (short) i;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) IPC_NOWAIT;
	sops[1].sem_num = (short) nsems+i;
	sops[1].sem_op = (short) -1;
	sops[1].sem_flg = (short) IPC_NOWAIT;

	if (semop(q_semid,sops,2) == -1) {
		if (errno == EAGAIN) {
			inside_q_pop = 0;
			return(NULL);
		}
		sprintf(msg,"    q_semid=%d",q_semid);
		printerr(FACILITY,ERROR_MSG,NULL,msg,17);
		abort();
	}

	/* Pop */

	if (q_head[i] == -1)
		q = NULL;
	else
		q = &qelem[q_head[i]];

	if (q == NULL) abort();

	if (qelem[q_head[i]].next == -1) {	/* List is now empty */
		q_head[i] = -1;
		q_tail[i] = -1;
	}
	else {
		q_head[i] = q->next;
	}

	/* Indicate that the operation is over, and that the queue is empty */

	if (q_head[i] == -1)
		nsops = 1;
	else
		nsops = 2;

	sops[0].sem_num = (short) nsems+i;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) i;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,18);
		abort();
	}

	inside_q_pop = 0;
	return(q);
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_push

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

int q_push(j,q)
int j;
struct qelem *q;
{
	int was_empty;

	struct sembuf sops[2];

	int nsops;

	if (inside_q_push) abort();

	inside_q_push = -1;

	/* If this queue is blocked, put the message on the free queue */

	if (q_block[j]) j = 0;

	/* Check to see if it is the correct free queue */

	if (j == 0) {
		if (q->size != 1) j = nsems-1;	/* Small free queue */
	}
	else if (j == nsems-1) {
		if (q->size != 0) j = 0;	/* Large free queue */
	}

	if (q_block[j]) abort();

	/* Wait till no queue operations are in progress for this particular queue */

	sops[0].sem_num = (short) nsems+j;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) 0;

	if (semop(q_semid,sops,1) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,10);
		abort();
	}

	/* Push */

	if (q_tail[j] == -1) {
		was_empty = -1;
		q->next = -1;
		q_tail[j] = q->index;
		q_head[j] = q->index;
	}
	else {
		was_empty = 0;
		qelem[q_tail[j]].next = q->index;
		q->next = -1;
		q_tail[j] = q->index;
	}

	/* Indicate that the operation is over, and that the queue is non-empty */

	if (was_empty)
		nsops = 2;
	else
		nsops = 1;

	sops[0].sem_num = (short) nsems+j;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) j;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,11);
		abort();
	}

	inside_q_push = 0;
	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_push_urgent

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

int q_push_urgent(j,q)
int j;
struct qelem *q;
{
	int was_empty;

	struct sembuf sops[2];

	int nsops;

	if (inside_q_push) abort();

	inside_q_push = -1;

	/* If this queue is blocked, put the message on the free queue */

	if (q_block[j]) j = 0;

	/* Check to see if it is the correct free queue */

	if (j == 0) {
		if (q->size != 1) j = nsems-1;	/* Small free queue */
	}
	else if (j == nsems-1) {
		if (q->size != 0) j = 0;	/* Large free queue */
	}

	if (q_block[j]) abort();

	/* Wait till no queue operations are in progress for this particular queue */

	sops[0].sem_num = (short) nsems+j;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) 0;

	if (semop(q_semid,sops,1) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,10);
		abort();
	}

	/* Push onto front of queue (instead of tail) */

	if (q_tail[j] == -1) {
		was_empty = -1;
		q->next = -1;
		q_tail[j] = q->index;
		q_head[j] = q->index;
	}
	else {
		was_empty = 0;
		q->next = q_head[j];
		q_head[j] = q->index;
	}

	/* Indicate that the operation is over, and that the queue is non-empty */

	if (was_empty)
		nsops = 2;
	else
		nsops = 1;

	sops[0].sem_num = (short) nsems+j;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) j;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,11);
		abort();
	}

	inside_q_push = 0;
	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_push_retry

  DESCRIPTION:

      [@tbs@]

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

int q_push_retry(j,q)
int j;
struct qelem *q;
{
	int was_empty;

	struct sembuf sops[2];

	int nsops;

	if (inside_q_push) abort();

	inside_q_push = -1;

	/* If this queue is blocked, put the message on the free queue */

	if (q_block[j]) j = 0;

	/* Check to see if it is the correct free queue */

	if (j == 0) {
		if (q->size != 1) j = nsems-1;	/* Small free queue */
	}
	else if (j == nsems-1) {
		if (q->size != 0) j = 0;	/* Large free queue */
	}

	if (q_block[j]) abort();

	/* Wait till no queue operations are in progress for this particular queue */

retry1:
	sops[0].sem_num = (short) nsems+j;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) 0;

	if (semop(q_semid,sops,1) == -1) {
		if (errno == EINTR) goto retry1;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,21);
		abort();
	}

	/* Push */

	if (q_tail[j] == -1) {
		was_empty = -1;
		q->next = -1;
		q_tail[j] = q->index;
		q_head[j] = q->index;
	}
	else {
		was_empty = 0;
		qelem[q_tail[j]].next = q->index;
		q->next = -1;
		q_tail[j] = q->index;
	}

	/* Indicate that the operation is over, and that the queue is non-empty */

	if (was_empty)
		nsops = 2;
	else
		nsops = 1;

retry2:
	sops[0].sem_num = (short) nsems+j;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) j;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		if (errno == EINTR) goto retry2;
		printerr(FACILITY,ERROR_MSG,NULL,NULL,22);
		abort();
	}

	inside_q_push = 0;
	return(0);
}

/******************************************************************/
/*
    FUNCTION NAME:

      q_pop_no_wait_n

  DESCRIPTION:

      Pop one or more messages off a queue.

  FORMAL PARAMETERS:

      [@description or none@]

  FUNCTION VALUE:

      [@description or none@]

								  */
/******************************************************************/

int q_pop_no_wait_n(i,qptrs,max_qptrs)
int i,max_qptrs;
struct qelem *qptrs[];
{
	struct qelem *q;

	struct sembuf sops[2];

	int nsops;

	char msg[80];

	int num_popped = 0;
	int j;

	if (inside_q_pop) {
		printerr(FACILITY,ERROR_MSG,"Re-entered q_pop",NULL,25);
		abort();
	}
	inside_q_pop = -1;

	/* Check semaphores allowing q-op and non-empty queue */

	sops[0].sem_num = (short) i;
	sops[0].sem_op = (short) -1;
	sops[0].sem_flg = (short) IPC_NOWAIT;
	sops[1].sem_num = (short) nsems+i;
	sops[1].sem_op = (short) -1;
	sops[1].sem_flg = (short) IPC_NOWAIT;

	if (semop(q_semid,sops,2) == -1) {
		if (errno == EAGAIN) {
			inside_q_pop = 0;
			return(0);
		}
		sprintf(msg,"    q_semid=%d",q_semid);
		printerr(FACILITY,ERROR_MSG,NULL,msg,26);
		abort();
	}

	/* Pop up to max_qptrs messages */

	j = 0;
	while ((q_head[i] != -1) && (j < max_qptrs)) {
		qptrs[j] = &qelem[q_head[i]];
		j++;
		if (qelem[q_head[i]].next == -1) {	/* List is now empty */
			q_head[i] = -1;
			q_tail[i] = -1;
		}
		else {
			q = &qelem[q_head[i]];
			q_head[i] = q->next;
		}
	}
	if (j == 0) abort();	/* Error, messages should have been there */
	num_popped = j;

	/* Indicate that the operation is over, and that the queue is empty */

	if (q_head[i] == -1)
		nsops = 1;
	else
		nsops = 2;

	sops[0].sem_num = (short) nsems+i;
	sops[0].sem_op = (short) 1;
	sops[0].sem_flg = (short) 0;
	sops[1].sem_num = (short) i;
	sops[1].sem_op = (short) 1;
	sops[1].sem_flg = (short) 0;

	if (semop(q_semid,sops,nsops) == -1) {
		printerr(FACILITY,ERROR_MSG,NULL,NULL,27);
		abort();
	}

	inside_q_pop = 0;
	return(num_popped);
}
