/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: ypmaps.h

    Explanation: 

    Exported Routines:  map_lookup
			mapfile_lookup
			get_yp_domain

*/

#include <stdio.h>
#include "../inc/perr.h"

#define FACILITY "YPMAPS"


/************************************************************************/
/*

    Function:  get_yp_domain

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int get_yp_domain(domain)
char *domain;
{

	int status;
	char *d,*getenv();
	char errmsg[256];

	/* First try the BATCH_DOMAIN environment variable */

	d = getenv("BATCH_DOMAIN");
	if (d != NULL) {
		strcpy(domain,d);
		return(0);
	}

	/* Get the default YP domain */

	status = yp_get_default_domain(&d);
	if (status != 0) {
		sprintf(errmsg,"%s",yperr_string(status));
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,1);
		return(-1);
	}
	strcpy(domain,d);
	return(0);
}

/************************************************************************/
/*

    Function:  mapfile_lookup

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int mapfile_lookup(fp,key,val,vallen)
FILE *fp;
char *key,*val;
int vallen;
{
	char line[256];
	char k[256],*v;
	int l;

	while (fgets(line,256,fp) != NULL) {
		k[0] = '\0';
		sscanf(line,"%s",k);
		line[strlen(k)] = '\0';
		if (strcmp(k,key) == 0) {
			v = &line[strlen(k)+1];
			while ((*v == ' ') || (*v == '\t')) v++;
			l = strlen(v) < vallen ? strlen(v) : vallen-1;
			while (v[l-1] == '\n') l--;
			strncpy(val,v,l);
			val[l] = '\0';
			return(0);		/* Entry found */
		}
	}
	return(-1);	/* Entry not found */
}

/************************************************************************/
/*

    Function:  map_lookup

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int map_lookup(mapfile,domain,mapname,key,val,vallen)
char *mapfile,*domain,*mapname,*key,*val;
int vallen;
{
	char *v;
	int vlen;
	FILE *fp,*fopen();
	int status;
	char errmsg[132];

	/* First, see if the entry can be found in the file */

	if (mapfile == NULL) goto check_yp;

	fp = fopen(mapfile,"r");
	if (fp != NULL) {
		if (mapfile_lookup(fp,key,val,vallen) == 0) {
			fclose(fp);
			return 0;	/* Found the entry */
		}
		fclose(fp);
	}

	/* Otherwise, check the YP map */

check_yp:
	status = yp_match(domain,mapname,key,strlen(key),&v,&vlen);
	if (status != 0) {
		sprintf(errmsg,"%s (domain=%s)",yperr_string(status),domain);
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,2);
		return(-1);
	}
	vlen = vlen < vallen ? vlen : vallen-1;
	strncpy(val,v,vlen);
	val[vlen] = '\0';

	return 0;
}
