/*
**
** Demo program:
** 
** Uses the exponential map method to 
** integrate a non-Hamltonian, 
** nonlinear vector
** field that spirals inward.
** 
** --- Leo Michelotti
** --- August 10, 1993
**
*/

#include <stdlib.h>
#include "mxyzptlk.rsc"

main( int argc, char** argv ) {

 if( argc != 3 ) {
  printf( "\n" );
  printf( "Usage: %s  d  t                           \n", argv[0] );
  printf( "where  d  = highest degree derivative     \n" );
  printf( "       t  = upper limit of integration.   \n" );
  printf( "\n" );
  exit(0);
 }

 int    deg   = atoi( argv[1] );
 double limit = atof( argv[2] );

 Jet::setup( 3, deg, 2 );
 coord x( 0.0 );
 coord y( 0.0 );
 coord t( limit );
 LieOperator v;
 Jet f, g;
 
 f = 1.0 - exp( - x*x - y*y );
 v.setVariable(          f*y, 0 );
 v.setVariable( -f*( x + y ), 1 );

 printf( "\n-*-*-*-  x  coefficients  -*-*-*-*-*\n" );
 ( f = v.expMap( t, x ) ).printCoeffs();
 printf( "\n-*-*-*-  y  coefficients  -*-*-*-*-*\n" );
 ( g = v.expMap( t, y ) ).printCoeffs();

 // Enter a point and evaluate the map.
 double s[3];
 double tlo, thi, delta;
 printf( "Enter x and y:  ");
 scanf( "%lf  %lf", s, s+1);
 getchar();
 printf( "Enter range and increment  ");
 scanf( "%lf  %lf  %lf", &tlo, &thi, &delta );
 getchar();
 s[2] = limit;
 printf( "( %lf, %lf ) maps to ( %lf, %lf )\n\n",
          s[0], s[1] , f(s), g(s) );
 for( limit = tlo; limit < thi + delta / 2.0 ; limit += delta ) {
  s[2] = limit;
  printf( "%lf: ( %lf, %lf )\n",
          limit, f(s), g(s) );
 }
}
