/*
 * Implementation of class JetVector
 * 
 * Version 1.0       April, 1995
 * 
 * Leo Michelotti
 * michelot@calvin.fnal.gov
 * 
 */


#include <stdlib.h>
#include "JetVector.h"

// Constructors and the destructor ...

#define CHECKOUT(test,fcn,message)          \
  if( test ) {                              \
    cout << "\n\n"                          \
            "*** ERROR ***\n"               \
            "*** ERROR ***" fcn "\n"        \
            "*** ERROR ***" message "\n"    \
            "*** ERROR ***\n"               \
            "*** ERROR ***"                 \
         << endl;                           \
         exit(1);                           \
  }

JetVector::JetVector( const int n, const Jet* x )
{
  int i;

  CHECKOUT(n <= 0, "JetVector::JetVector", "Dimension must be positive.")

  dim = n;
  comp = new Jet [ dim ];
  if( x ) for ( i = 0; i < dim; i++ ) comp[i] = x[i];
  else    for ( i = 0; i < dim; i++ ) comp[i] = 0.0;
}

JetVector::JetVector( const JetVector& x )
{
  dim = x.dim;
  comp = new Jet [ dim ];
  for ( int i = 0; i < dim; i++ ) comp[i] = x.comp[i];
}

JetVector::~JetVector()
{
  delete [] comp;
}



// Assignment ...

void JetVector::Set( const Jet* x )
{
  for ( int i = 0; i < 3; i++ ) comp[i] = x[i];
}

Jet& JetVector::operator() ( int i ) const
{
  static Jet dummy ( - 1. / 137. );
  if ( ( 0 <= i ) && ( i < dim ) ) return comp[i];
  else                             return dummy;
}


// Algebraic functions ...

JetVector& JetVector::operator= ( const JetVector& x )
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator=", "Incompatible dimensions.")
#endif

  for ( int i = 0; i < dim; i++ ) comp[i] = x.comp[i];
  return *this;
}


JetVector JetVector::operator+ ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator+", "Incompatible dimensions.")
#endif

  JetVector z( dim );
  for ( int i = 0; i < dim; i++ ) 
    z.comp[i] = comp[i] + x.comp[i];
  return z;
}


JetVector JetVector::operator+= ( const JetVector& x )
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator+=", "Incompatible dimensions.")
#endif

  for ( int i = 0; i < dim; i++ ) comp[i] += x.comp[i];
  return *this;
}


JetVector JetVector::operator- ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator-", "Incompatible dimensions.")
#endif

  JetVector z( dim );
  for ( int i = 0; i < dim; i++ ) 
    z.comp[i] = comp[i] - x.comp[i];
  return z;
}


JetVector operator- ( const JetVector& x )
{
  JetVector z( x.dim );
  for ( int i = 0; i < x.dim; i++ ) 
    z.comp[i] = - x.comp[i];
  return z;
}


JetVector JetVector::operator-= ( const JetVector& x )
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator-=", "Incompatible dimensions.")
#endif

  for ( int i = 0; i < dim; i++ ) comp[i] -= x.comp[i];
  return *this;
}


JetVector JetVector::operator* ( const Jet& c ) const
{
  JetVector z( dim );
  for ( int i = 0; i < dim; i++ ) z.comp[i] = c * comp[i];
  return z;
}

JetVector JetVector::operator* ( double c ) const
{
  JetVector z( dim );
  for ( int i = 0; i < dim; i++ ) z.comp[i] = c * comp[i];
  return z;
}

JetVector operator* ( const Jet& c, const JetVector& x )
{
  JetVector z( x.dim );
  for ( int i = 0; i < x.dim; i++ ) z.comp[i] = c * x.comp[i];
  return z;
}

JetVector operator* ( double c, const JetVector& x )
{
  JetVector z( x.dim );
  for ( int i = 0; i < x.dim; i++ ) z.comp[i] = c * x.comp[i];
  return z;
}

JetVector JetVector::operator*=( const Jet& c ) {
 for( int i = 0; i < dim; i++ ) comp[i] *= c;
 return *this;
}


JetVector JetVector::operator*=( double c ) {
 for( int i = 0; i < dim; i++ ) comp[i] *= c;
 return *this;
}


JetVector JetVector::operator/ ( const Jet& c ) const
{
 JetVector z( dim );
 for( int i = 0; i < dim; i++ ) z.comp[i] = comp[i] / c;
 return z;
}

JetVector JetVector::operator/ ( double c ) const
{
 JetVector z( dim );
 for( int i = 0; i < dim; i++ ) z.comp[i] = comp[i] / c;
 return z;
}

JetVector JetVector::operator/=( const Jet& c ) {
 for( int i = 0; i < dim; i++ ) comp[i] /= c;
 return *this;
}

JetVector JetVector::operator/=( double c ) {
 for( int i = 0; i < dim; i++ ) comp[i] /= c;
 return *this;
}


Jet JetVector::operator* ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator*", "Dimensions incompatible.")
#endif

  Jet u( 0.0 );
  int i;
  for ( i = 0; i < dim; i++ ) u += comp[i] * x.comp[i];
  return u;
}


JetVector JetVector::operator^ ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT((dim != 3) || ( x.dim != 3 ),
           "JetVector::operator^",
           "Dimension must be 3." )
#endif

  static JetVector z( 3 );
  z.comp[ 0 ] = comp[ 1 ] * x.comp[ 2 ] - comp[ 2 ] * x.comp[ 1 ];
  z.comp[ 1 ] = comp[ 2 ] * x.comp[ 0 ] - comp[ 0 ] * x.comp[ 2 ];
  z.comp[ 2 ] = comp[ 0 ] * x.comp[ 1 ] - comp[ 1 ] * x.comp[ 0 ];
  return z;
}


JetVector JetVector::operator^ ( Vector x ) const
{
#ifndef NOCHECKS
  CHECKOUT((dim != 3) || ( x.Dim() != 3 ),
           "JetVector::operator^",
           "Dimension must be 3." )
#endif

  static JetVector z( 3 );
  z.comp[ 0 ] = comp[ 1 ] * x( 2 ) - comp[ 2 ] * x( 1 );
  z.comp[ 1 ] = comp[ 2 ] * x( 0 ) - comp[ 0 ] * x( 2 );
  z.comp[ 2 ] = comp[ 0 ] * x( 1 ) - comp[ 1 ] * x( 0 );
  return z;
}

JetVector operator^ ( Vector y, const JetVector& x )
{
#ifndef NOCHECKS
  CHECKOUT(( y.Dim() != 3 ) || ( x.dim != 3 ),
           "JetVector::operator^",
           "Dimension must be 3." )
#endif

  static JetVector z( 3 );
  z.comp[ 0 ] = y( 1 ) * x.comp[ 2 ] - y( 2 ) * x.comp[ 1 ];
  z.comp[ 1 ] = y( 2 ) * x.comp[ 0 ] - y( 0 ) * x.comp[ 2 ];
  z.comp[ 2 ] = y( 0 ) * x.comp[ 1 ] - y( 1 ) * x.comp[ 0 ];
  return z;
}



// Boolean functions ...

char JetVector::operator== ( const JetVector& x ) const
{
  if( dim != x.dim ) return 0;
  char z = 1;
  int  i = -1;
  while ( ( ++i < dim ) && z ) z &= comp[i] == x.comp[i];
  return z;
}

char JetVector::operator!= ( const JetVector& x ) const
{
  return !( operator==( x ) );
}

char JetVector::operator< ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator<", "Dimensions incompatible.")
#endif

  char z = 1;
  int  i = -1;
  while ( ( ++i < dim ) && z ) 
    z &= comp[i].standardPart() < x.comp[i].standardPart();
  return z;
}

char JetVector::operator<= ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator<=", "Dimensions incompatible.")
#endif

  char z = 1;
  int  i = -1;
  while ( ( ++i < dim ) && z ) 
    z &= comp[i].standardPart() <= x.comp[i].standardPart();
  return z;
}

char JetVector::operator> ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator>", "Dimensions incompatible.")
#endif

  char z = 1;
  int  i = -1;
  while ( ( ++i < dim ) && z ) 
    z &= comp[i].standardPart() > x.comp[i].standardPart();
  return z;
}

char JetVector::operator>= ( const JetVector& x ) const
{
#ifndef NOCHECKS
  CHECKOUT(dim != x.dim, "JetVector::operator>=", "Dimensions incompatible.")
#endif

  char z = 1;
  int  i = -1;
  while ( ( ++i < dim ) && z ) 
    z &= comp[i].standardPart() >= x.comp[i].standardPart();
  return z;
}

char JetVector::IsNull() const
{
  int i = -1;
  while ( ++i < dim ) if( comp[i] != 0.0 ) return 0;
  return 1;
}

char JetVector::IsUnit() const
{
  double x = 0.0;
  for( int i = 0; i < dim; i++ ) 
    x += comp[i].standardPart()*comp[i].standardPart();
  return x == 1.0;
}



// Utilities ..

void JetVector::peekAt() const
{
  cout << "\n\nBegin JetVector::peekAt() ......\n";
  for ( int i = 0; i < dim; i++ ) {
    cout << "JetVector::peekAt(): Component " << i << endl;
    comp[i].peekAt();
  }
  cout << "End JetVector::peekAt() ......\n" << endl;
}


void JetVector::printCoeffs() const
{
  cout << "\n\nBegin JetVector::printCoeffs() ......\n";
  for ( int i = 0; i < dim; i++ ) {
    cout << "JetVector::printCoeffs(): Component " << i << endl;
    comp[i].printCoeffs();
  }
  cout << "End JetVector::printCoeffs() ......\n" << endl;
}


Jet JetVector::Norm () const
{
  Jet x;
  x = 0.0;
  for ( int i = 0; i < dim; i++ ) x += comp[i]*comp[i];
  return sqrt(x);
}

JetVector JetVector::Unit () const
{
  Jet x;
  x = Norm();
  JetVector z( dim );

  for ( int i = 0; i < dim; i++ ) z.comp[i] = comp[i] / x;
  return z;
}


void JetVector::Rotate ( JetVector& v, double theta ) const
{
#ifndef NOCHECKS
  CHECKOUT((dim != 3) || ( v.dim != 3 ),
           "JetVector::Rotate",
           "Dimension must be 3." )
#endif

  double c, s;
  JetVector e, u;

  e = Unit();
  c = cos( theta );
  s = sin( theta );
  u = ( c*v ) +
      ( s*( e^v) ) +
      ( ( ( 1.0 - c )*(e*v) )*e );
  for ( int i = 0; i < 3; i++ ) v.comp[i] = u.comp[i];
}


void JetVector::Rotate ( JetVector& v, const Jet& theta ) const
{
#ifndef NOCHECKS
  CHECKOUT((dim != 3) || ( v.dim != 3 ),
           "JetVector::Rotate",
           "Dimension must be 3." )
#endif

  Jet c, s;
  JetVector e, u;

  e = Unit();
  c = cos( theta );
  s = sin( theta );
  u = ( c*v ) +
      ( s*( e^v) ) +
      ( ( ( 1.0 - c )*(e*v) )*e );
  for ( int i = 0; i < 3; i++ ) v.comp[i] = u.comp[i];
}


ostream& operator<<( ostream& os, const JetVector& v )
{
  os << "( ";
  for( int i = 0; i < v.dim - 1; i++ )
    os << v.comp[i] << ", ";
  os << v.comp[ v.dim - 1 ] << " )";
  return os;
}

