/*************************************************************************
**************************************************************************
**************************************************************************
******                                                                ****
******  MXYZPTLK:  A C++ implementation of differential algebra.      ****
******                                                                ****
******  Copyright (c) 1990  Universities Research Association, Inc.   ****
******                All Rights Reserved                             ****
******                                                                ****
******  Author:    Leo Michelotti                                     ****
******                                                                ****
******             Fermilab                                           ****
******             P.O.Box 500                                        ****
******             Mail Stop 345                                      ****
******             Batavia, IL   60510                                ****
******                                                                ****
******             Phone: (708) 840 4956                              ****
******             Email: michelot@hazel.fnal.gov                     ****
******                    michelotti@adcalc.fnal.gov                  ****
******                    almond::michelotti                          ****
******                                                                ****
******  Release    Version 1.0 : January   31, 1990                   ****
******  Dates:              .1 : October   11, 1990                   ****
******                      .2 : November   6, 1990                   ****
******                      .3 : October    2, 1992                   ****
******                      .4 : July      20, 1993                   ****
******                     2.0   August    12, 1993                   ****
******                     3.0   May       19, 1994                   ****
******                      .1   September 14, 1994                   ****
******                                                                ****
**************************************************************************
**************************************************************************
******                                                                ****
******  This material resulted from work developed under a            ****
******  Government Contract and is subject to the following           ****
******  license: The Government retains a paid-up, nonexclusive,      ****
******  irrevocable worldwide license to reproduce, prepare           ****
******  derivative works, perform publicly and display publicly       ****
******  by or for the Government, including the right to distribute   ****
******  to other Government contractors.  Neither the United          ****
******  States nor the United States Department of Energy, nor        ****
******  any of their employees, makes any warranty, express or        ****
******  implied, or assumes any legal liability or responsibility     ****
******  for the accuracy, completeness or usefulness of any           ****
******  information, apparatus, product, or process disclosed, or     ****
******  represents that its use would not infringe privately owned    ****
******  rights.                                                       ****
******                                                                ****
**************************************************************************
**************************************************************************
******                                                                ****
******  Notes:                                                        ****
******                                                                ****
******  11/6/90   .. Leo Michelotti                                   ****
******  JL functions  pow( JL&, double )  and  sqrt( JL& ) have       ****
******  been included in the lexicon.                                 ****
******                                                                ****
******                                                                ****
******  10/11/90  .. Leo Michelotti                                   ****
******  JLSetup routines have been changed so that the variable       ****
******  oneEnv.allSet and the array oneEnv.refPoint **
******  are set earlier.  This fixes a bug which first arose in       ****
******  an application involving concatenation.                       ****
******                                                                ****
******                                                                ****
******  10/2/92  .. Leo Michelotti                                    ****
******  In   double JL::weightedDerivative( int* ind ), the cast      ****
******  appearing in declaring a dlist_iterator was changed           ****
******  from (dlist) to (dlist&) in order to bypass a call to         ****
******  the dlist copy constructor.  I suspect g++ does not           ****
******  handle casting correctly, but I cannot yet prove it.          ****
******  It may also be true that the copy constructor is not          ****
******  working properly.                                             ****
******                                                                ****
******                                                                ****
******  7/20/93 .. Leo Michelotti                                     ****
******  In an attempt to improve efficiency, functions that           ****
******  previously returned JL have been changed to return            ****
******  JL&.  In conjunction with this, a disposable base             ****
******  class has been defined in order to provide for some           ****
******  primitive garbate collection.                                 ****
******                                                                ****
******  The code sandwiched between #ifdef POOLED and #endif          ****
******  POOLED statements is still under development.  It             ****
******  should not be activated by the Makefile.                      ****
******                                                                ****
******                                                                ****
******  8/12/93 .. Leo Michelotti                                     ****
******  The LieOperator methods .expMap() and .operator^              ****
******  have now been implemented.  The former                        ****
******  integrates equations of motion by developing                  ****
******  exponential maps, and the latter simply                       ****
******  evaluates a Lie derivative.  The one restriction              ****
******  at the moment is the assumption made by                       ****
******  .expMap() that the point of evaluation is a                   ****
******  fixed point of the vector field.  There are ways              ****
******  of getting around this -- but, another day.                   ****
******  Demos are contained in files Lie_X_Test.cc,                   ****
******  where X = A ... E.                                            ****
******                                                                ****
******     __ This upgrade is significant enough to                   ****
******  warrant a version change, from 1.X to 2.0.                    ****
******  Huzzah!                                                       ****
******                                                                ****
******     __ Addition has been modified so that small                ****
******  terms indistinguishable from zero are not                     ****
******  carried.  Further, arithmetic on LieOperators                 ****
******  has been enabled.                                             ****
******                                                                ****
******                                                                ****
******  3/21/94 .. Leo Michelotti                                     ****
******  The line                                                      ****
******                                                                ****
******  for( w = 1; w <= ( this->weight ); w++ )                      ****
******                                                                ****
******  that appears in ::operator() functions has been changed to    ****
******                                                                ****
******  for( w = 1; w <= this; w++ )                                  ****
******                                                                ****
******  A switch was also put into the concat() routine so as to      ****
******  break the calculational loop if the weight of a JLterm        ****
******  is greater than the accurate weight of the polynomial.        ****
******                                                                ****
******  viz: if( p->weight > x.accuWgt ) break;                       ****
******                                                                ****
******  A more correct fix for all of this is to never store     ???  ****
******  nonaccurate coefficients.  The ::D() operator should be  ???  ****
******  modified to insure this.                                 ???  ****
******                                                                ****
******                                                                ****
**************************************************************************
*************************************************************************/

#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}


// ================================================================

#ifndef MX_SMALL
#define MX_SMALL    1.0e-12 // Used by Jet::addTerm to decide 
                            //   removal of a Jetterm.
#endif  // MX_SMALL

#ifndef MX_MAXITER
#define MX_MAXITER  100     // Maximum number of iterations allowed
                            //   in iterative routines
#endif  // MX_MAXITER




//***************************************************************
//***************************************************************
//***************************************************************
//
//    Implementation of class LieOperator
//

//    Constructors and destructors    |||||||||||||||||||||||||||

LieOperator::LieOperator() {
 int i;
 
 weight = -1;
 accuWgt = Jet::oneEnv.MaxWeight;
 
 if( Jet::oneEnv.allSet == 137 ) {
   if( Jet::oneEnv.SpaceDim == 0 ) {
     printf( "\n*** ERROR: LieOperator::LieOperator called " );
     printf( "after Jet::Setup declared no vectors.\n" );
     exit(0);
   }
   v = new Jet[ Jet::oneEnv.NumVar ];
   refPoint = new double[ Jet::oneEnv.NumVar ];
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = Jet::oneEnv.refPoint[i];
 }
 else {
   v = 0;
   refPoint = 0;
   Jet::oneEnv.predefLieOp.insert( this );
 }
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator::LieOperator( LieOperator& x ) {
 int i;
 
 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** LieOperator::LieOperator( LieOperator& ) ERROR *** " );
   printf( "Prior call to Jet::setup required.\n" );
   exit(0);
   }
 
 v = new Jet[ Jet::oneEnv.NumVar ];

 for ( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   delete v[i].jl;
   v[i].jl = x.v[i].jl;    // Perform shallow copy.
   ( v[i]->rc )++;         // Increase reference counter.
 }

 weight = x.weight;
 accuWgt = x.accuWgt;
 
 refPoint = new double[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator::LieOperator( Jet* x ) {    // A dangerous routine.
 int i;                                // It assumes the correct number
                                       // of components in the argument.

 static char firstCall = 1;

 if( firstCall ) {
  firstCall = 0;
  printf( "\n" );
  printf( "*** WARNING ***					\n" );
  printf( "*** WARNING *** LieOperator::LieOperator( Jet* x )	\n" );
  printf( "*** WARNING *** A dangerous routine.			\n" );
  printf( "*** WARNING *** It assumes the correct number	\n" );
  printf( "*** WARNING *** of components in the argument.	\n" );
  printf( "*** WARNING *** I highly recommend against its use!	\n" );
  printf( "*** WARNING ***					\n" );
 }

 if( Jet::oneEnv.SpaceDim == 0 ) {
  printf( "\n*** ERROR: LieOperator::LieOperator called " );
  printf( "after Jet::Setup declared no vectors.\n" );
  exit(0);
 }
 
 if( Jet::oneEnv.allSet != 137 ) {
  printf( "\n*** LieOperator::LieOperator( LieOperator& ) ERROR *** " );
  printf( "Prior call to Jet::Setup required.\n" );
  exit(0);
 }
 
 weight = -1;
 accuWgt = Jet::oneEnv.MaxWeight;

 refPoint = new double[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x[0]->refPoint[i];

 v = new Jet[ Jet::oneEnv.NumVar ];
 for ( i = 0; i < Jet::oneEnv.NumVar; i++ ) setVariable( x[i], i );

 reset();
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator::LieOperator( const Jet& x ) { 
 int i;
 int s = Jet::oneEnv.SpaceDim;
 int n = Jet::oneEnv.NumVar;
 int* ndx = new int [n];

 if( 2*(s/2) != s ) {
  cout << "*** ERROR ***                                           \n"
          "*** ERROR *** LieOperator::LieOperator( const Jet& )    \n"
          "*** ERROR *** Phase space has odd dimension = " << s
                                                           << "    \n"
          "*** ERROR ***                                           \n"
       << endl;
  exit(0);
 }

 if( s == 0 ) {
  printf( "\n*** ERROR: LieOperator::LieOperator called " );
  printf( "after Jet::Setup declared no vectors.\n" );
  exit(0);
 }
 
 if( Jet::oneEnv.allSet != 137 ) {
  printf( "\n*** ERROR *** LieOperator::LieOperator " );
  printf( "*** ERROR *** Prior call to Jet::Setup required.\n" );
  exit(0);
 }
 
 weight = -1;
 accuWgt = Jet::oneEnv.MaxWeight;

 refPoint = new double[n];
 for( i = 0; i < n; i++ ) refPoint[i] = x->refPoint[i];

 v = new Jet[n];
 for( i = 0; i < n; i++ ) ndx[i] = 0;

 for( i = 0; i < s/2; i++ ) {
  ndx[i + s/2] = 1;
  v[i] = x.D( ndx );
  ndx[i + s/2] = 0;
 }

 for( i = s/2; i < s; i++ ) {
  ndx[i - s/2] = 1;
  v[i] = - x.D( ndx );
  ndx[i - s/2] = 0;
 }

 for( i = s; i < n; i++ ) v[i] = 0.0;

 reset();
 delete [] ndx;
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator::LieOperator( char*  ) { 
 int i;
 int nv;
 
 if( Jet::oneEnv.SpaceDim == 0 ) {
   printf( "\n*** ERROR: LieOperator::LieOperator called " );
   printf( "after Jet::Setup declared no vectors.\n" );
   exit(0);
   }
 
 if( Jet::oneEnv.allSet != 137 ) {
   printf( "\n*** LieOperator::LieOperator( char* ) ERROR *** " );
   printf( "Prior call to Jet::Setup required.\n" );
   exit(0);
   }
 
 nv = Jet::oneEnv.NumVar;
 v = new Jet[ nv ];

 refPoint = new double[ nv ];
 for( i = 0; i < nv; i++ ) {
  refPoint[i] = Jet::oneEnv.refPoint[i];
  v[i]->setVariable( i );
 }

 weight = 1;
 accuWgt = Jet::oneEnv.MaxWeight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator::~LieOperator() {
 if( v ) { delete [] v;
           delete [] refPoint; }
}

void LieOperator::reset() {
 int i, j;
 weight = -1;
 accuWgt = Jet::oneEnv.MaxWeight;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
  if( weight < v[i]->weight ) weight = v[i]->weight;
  if( accuWgt > v[i]->accuWgt ) accuWgt = v[i]->accuWgt;
  refPoint[i] = v[0]->refPoint[i];
 }
 i = 1;
 while( i < Jet::oneEnv.NumVar ) {
  for( j = 0; j < Jet::oneEnv.NumVar; j++ ) 
   if( refPoint[j] != v[i]->refPoint[j] ) {
    printf( "\n" );
    printf( "*** ERROR ***                                      \n" );
    printf( "*** ERROR *** void LieOperator::reset()            \n" );
    printf( "*** ERROR *** refPoint error for i = %d            \n", i );
    printf( "*** ERROR ***                                      \n" );
    exit(0);
   }
  i++;
 }
}


//     Member functions(public)  |||||||||||||||||||||||||||||||

void LieOperator::fixReference( double* r ) {
 int i, j;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = r[i];
 if( v ) {
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) v[j]->refPoint[i] = r[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::fixReference() {
 int i, j;

 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  refPoint[i] = Jet::oneEnv.refPoint[i];

 if( v ) {
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = Jet::oneEnv.refPoint[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::fixReference( Jet& x ) {
 int i, j;

 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  refPoint[i] = x->refPoint[i];

 if( v ) {
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = x->refPoint[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::fixReferenceAtEnd( const LieOperator& x ) {
 int i, j;
 double* u = new double [ Jet::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = u[i];

 if( v ) {
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = u[i];
 }

 delete [] u;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::fixReferenceAtStart( const LieOperator& x ) {
 int i, j;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];

 if( v ) {
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
    v[j]->refPoint[i] = x.refPoint[i];
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::setVariable( const Jet& x, const int j ) {
 int i;
 
 if( j < 0 || Jet::oneEnv.NumVar <= j ) {
  printf( "\n***ERROR ***                                              \n" );
  printf(   "***ERROR *** LieOperator::setVariable( Jet&, int )         \n" );
  printf(   "***ERROR *** It is certainly not true                     \n" );
  printf(   "***ERROR *** that 0 <= %d < %d                            \n",
            j, Jet::oneEnv.NumVar );
  printf(   "***ERROR ***                                              \n" );
  printf(   "***ERROR *** Sorry.  Good bye.                            \n" );
  printf(   "***ERROR ***                                              \n" );
  exit(0);
 }
 
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) if( refPoint[i] != x->refPoint[i] ) {
  printf( "\n*** LieOperator::setVariable ERROR: Incompatible references.\n" );
  exit(0);
 }
 
 v[j] = x;
 if ( weight  < x->weight  ) weight  = x->weight;
 if ( accuWgt > x->accuWgt ) accuWgt = x->accuWgt;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::setVariable( const double x, const int j ) {
 int* ndx;
 int i;
 
 if( j < 0 || Jet::oneEnv.NumVar <= j ) {
  printf( "\n***ERROR ***                                              \n" );
  printf(   "***ERROR *** LieOperator::setVariable( double, int )         \n" );
  printf(   "***ERROR *** It is certainly not true                     \n" );
  printf(   "***ERROR *** that 0 <= %d < %d                            \n",
            j, Jet::oneEnv.NumVar );
  printf(   "***ERROR ***                                              \n" );
  printf(   "***ERROR *** Sorry.  Good bye.                            \n" );
  printf(   "***ERROR ***                                              \n" );
  exit(0);
 }
 
 v[j].clear();
 // if( ( j < 0 ) || ( j >= Jet::oneEnv.NumVar ) ) return;  ??? Old form.
 
 refPoint[j] = x;
 Jet::oneEnv.refPoint[j] = x;
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 ndx[j] = 1;
 v[j]->insert( new JLterm( ndx, 1.0 ) );
 delete [] ndx;
 
 ndx = new int[ Jet::oneEnv.NumVar ];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) ndx[i] = 0;
 v[j]->insert( new JLterm( ndx, x ) );
 delete [] ndx;

 if ( weight < 1 ) weight = 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::getReference( double* r ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) r[i] = refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char LieOperator::isNilpotent() {
 for( int i = 0; i <  Jet::oneEnv.SpaceDim; i++ ) 
  if( !(v[i]->isNilpotent()) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::peekAt() {
 int i;
 printf( "\n******************\nLieOperator::peekAt ********\n" );
 printf( "Weight: %d   Address: %d\n", weight, this );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   printf( "\n******************\n**** Component index = %d:\n", i );
   v[i]->peekAt();
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

dlist* LieOperator::image() {
 dlist* ret = new dlist[ Jet::oneEnv.NumVar ];  // ??? Change this!!!
 for( int i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  ret[i] = v[i]->image();
 return ret;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::printCoeffs() const {
 int i;
 cout << "\n************ Begin LieOperator::printCoeffs ********" << endl;
 cout << "Weight: " << weight << endl;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   cout << "\n******************\n**** Component index = " << i << endl;
   v[i]->printCoeffs();
   }
 cout << "\n************ End   LieOperator::printCoeffs ********" << endl;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::peekAt( char* fileName ) {
 int i;
 Jet::scratchFile = fopen( fileName, "w" );
 fprintf( Jet::scratchFile,
          "\n******************\nLieOperator::peekAt ********\n" );
 fprintf( Jet::scratchFile, "Weight: %d   Address: %d\n", weight, this );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   fprintf( Jet::scratchFile,
            "\n******************\n**** Component index = %d:\n", i );
   v[i]->peekAt( 1 );
   }
 fclose( Jet::scratchFile );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::printCoeffs( char* fileName ) const {
 int i;
 Jet::scratchFile = fopen( fileName, "w" );
 fprintf( Jet::scratchFile,
          "\n************ Begin LieOperator::printCoeffs ********\n" );
 fprintf( Jet::scratchFile, "Weight: %d \n", weight );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   fprintf( Jet::scratchFile, "\n******************\n**** Component index = %d:\n", i );
   v[i]->printCoeffs( 1 );
   }
 fprintf( Jet::scratchFile, 
          "\n************ End   LieOperator::printCoeffs ********\n" );
 fclose( Jet::scratchFile );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::standardPart( double* x ) const {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )  x[i] = v[i]->standardPart();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::weightedDerivative( int* m, double* x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) x[i] = v[i]->weightedDerivative( m );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void LieOperator::derivative( int* m, double* x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) x[i] = v[i]->derivative( m );
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::filter( MX_R_FUNCPTR f [] ) {
 LieOperator z;
 int i;        

 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
  z.v[i] = v[i].filter( f[i] );

 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::filter( int lower, int upper ) {
 LieOperator z;
 int i;        

 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
  z.v[i] = v[i].filter( lower, upper );

 z.reset();

 return z;
}


//     Operators   |||||||||||||||||||||||||||||||||||||||||||||

istream& operator>>(istream& is,  LieOperator& x) {
 int i;
 int ival;
  char buf[100];
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> x.weight;
  for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
    is >> buf;
    is >> buf;
    is >> buf;
    is >> buf;
    is >> buf;
    is >> ival;
    is >> x.v[i];
 }
 return  is;
}

ostream& operator<<(ostream& os,  LieOperator& x) {
 int i;
 os << "\n************ Begin LieOperator  printCoeffs ********\n";
 os << "Weight: " << x.weight << endl;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   os << "\n******************\n**** Component index = " << i << endl;
   os << x.v[i];
   }
 return  os << "\n************ End   LieOperator  printCoeffs ********\n";
}

char operator==( const LieOperator& x, const LieOperator& y ) {
 int i;
 if( x.weight  != y.weight  ) return 0;
 if( x.accuWgt != y.accuWgt ) return 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) if( x.refPoint[i] != y.refPoint[i] ) return 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) if( !( x.v[i] == y.v[i] ) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const LieOperator& x, const LieOperator& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char LieOperator::operator==( const double x ) {
 char answer = 1;
 for( int i = 0; i < Jet::oneEnv.SpaceDim; i++ ) 
  answer = answer && ( v[i] == x );
 return answer;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char LieOperator::operator!=( const double x ) {
 return !( this->operator==( x ) );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

MatrixD LieOperator::Jacobian() const {
 int            i, j;
 int            nsd = Jet::oneEnv.SpaceDim;
 int            nv  = Jet::oneEnv.NumVar;
 int*           d = new int[ nv ];
 MatrixD        M( nsd, nsd, 0.0 );

 for( i = 0; i < nv; i++  ) d[i] = 0;
 for( j = 0; j < nsd; j++ ) {
  d[j] = 1;
  for( i = 0; i < nsd; i++ ) M( i, j ) = v[i]->derivative( d );
  d[j] = 0;
 }

 delete [] d;
 return M;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::inverse() {           // ??? REMOVE: Obsolete

 static char firstTime = 1;
 if( firstTime ) {
  firstTime = 0;
  cout << "\n*** WARNING ***                                  \n"
          "*** WARNING *** You are using the obsolete         \n"
          "*** WARNING *** LieOperator::inverse routine,      \n"
          "*** WARNING *** soon to vanish from the face of    \n"
          "*** WARNING *** the Earth.  Use the new and        \n"
          "*** WARNING *** improved LieOperator::Inverse.     \n"
          "*** WARNING ***                                    \n"
       << endl;
 }

 LieOperator z;
 LieOperator  id( "ident" );
 LieOperator  v;
 int          nsd  = Jet::oneEnv.SpaceDim;
 int          i = 0;
 MatrixD M( nsd, nsd, 0.0 );

 M = ( this->Jacobian() ).inverse();

 z = M*id;
 v = ( operator()(z) - id );
 // Old form                                           ??? REMOVE
 // while ( ( i++ < MX_MAXITER ) && ( v != 0.0 ) ) {   ??? REMOVE
 // New form                                           ??? REMOVE
 while ( ( i++ < MX_MAXITER ) && ( ( v - v.filter(0,1)) != 0.0 ) ) {
               // This assumes linear is handled well enough
               // by the MatrixD routine.  
 //                                                    ??? REMOVE
  z = z - M*v;
  v = ( operator()(z) - id );
 }

 if( i >= MX_MAXITER ) {
  printf( "*** WARNING ***                                        \n" );
  printf( "*** WARNING *** LieOperator& LieOperator::inverse()    \n" );
  printf( "*** WARNING *** Over %d iterations used;               \n", MX_MAXITER );
  printf( "*** WARNING *** result may be incorrect.               \n" );
  printf( "*** WARNING ***                                        \n" );
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::Inverse() const { 

 if( Jet::oneEnv.SpaceDim <= 0 ) {
  cout << "*** ERROR ***                                     \n"
          "*** ERROR *** LieOperator::Inverse                \n"
          "*** ERROR *** Prior call to Jet::Setup routine    \n"
          "*** ERROR *** required.                           \n"
          "*** ERROR ***                                     \n"
          "*** ERROR *** oneEnv.SpaceDim = " << Jet::oneEnv.SpaceDim << "\n"
          "*** ERROR ***                                     \n"
       << endl;
  exit(1);
 }


 int         i, j;
 char*       t1  = new char    [Jet::oneEnv.SpaceDim];
 char*       t2  = new char    [Jet::oneEnv.SpaceDim];
 char        t3;
 double*     ref = new double  [Jet::oneEnv.NumVar];
 double*    dref = new double  [Jet::oneEnv.NumVar];
 JLterm**    p   = new JLterm* [Jet::oneEnv.SpaceDim];
 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) p[j] = 0;

 // Store the default reference point

 for( j = 0; j < Jet::oneEnv.NumVar; j++ ) 
  dref[j] = Jet::oneEnv.refPoint[j];
 Jet::FixReference( Jet::oneEnv.AllZeroes ); 


 // If zero maps to zero, return result

 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) {        // Actually, all
  t1[j] = 1;					      // componenets should
  for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )         // have the same
   t1[j] &= ( v[j]->refPoint[i] == 0.0 );             // reference point!
 }                                                    // ??? CHANGE THIS

 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ )
  t2[j] = ( v[j].standardPart() == 0.0 );

 t3 = 1;
 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) t3 &= ( t1[j] & t2[j] );
 if( t3 ) {
   Jet::FixReference( dref ); 
   delete [] t1;
   delete [] t2;
   delete [] ref;
   delete [] dref;
   delete [] p;

   return EpsInverse();
 }


 // Otherwise, construct an idempotent and compute its inverse.

 LieOperator z;
 z = *this;   // Deep copy needed here.

 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) {
  if( !t2[j] ) { 
   p[j] = z.v[j].get();
   z.v[j]->addTerm( new JLterm( Jet::oneEnv.AllZeroes, 0.0 ) );
  }
 }

 if( !t1[0] ) 
  for( i = 0; i < Jet::oneEnv.SpaceDim; i++ ) {
   z.refPoint[i] = 0.0;
   for( j = 0; j < Jet::oneEnv.SpaceDim; j++ )    // A little paranoia
    z.v[j]->refPoint[i] = 0.0;                    // never hurt.
  }

 z = z.EpsInverse();


 // Make final adjustments before returning.

 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) {
  if( t2[j] ) ref[j] = 0.0;
  else        ref[j] = p[j]->value;
 }
 while( j < Jet::oneEnv.NumVar ) {
  ref[j] = refPoint[j];
  j++;
 }

 z.fixReference( ref );

 if( !t1[0] ) {   // ??? This should be sufficient.
  // A little check
  z.standardPart( ref );
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   if( ref[j] != 0.0 ) {
    cout << "*** WARNING ***                                              \n"
            "*** WARNING *** Jet::Inverse()                               \n"
            "*** WARNING *** ref[" << j << "] = " << ref[j] << "          \n"
            "*** WARNING ***                                              \n"
         << endl;
   }
  for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) 
   z.v[j]->addTerm( new JLterm( Jet::oneEnv.AllZeroes, refPoint[j] ) );
 }


 // Clean up and leave

 Jet::FixReference( dref ); 
 delete [] t1;
 delete [] t2;
 delete [] ref;
 delete [] dref;
 for( j = 0; j < Jet::oneEnv.SpaceDim; j++ ) delete p[j];
 delete [] p;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::EpsInverse() const {
 LieOperator  z;
 LieOperator  id( "ident" );
 LieOperator  v;
 int          nsd  = Jet::oneEnv.SpaceDim;
 int          i = 0;
 MatrixD M( nsd, nsd, 0.0 );

 M = Jacobian().inverse();

 z = M*id;
 v = ( operator()(z) - id );
 while ( ( i++ < MX_MAXITER ) && ( ( v - v.filter(0,1)) != 0.0 ) ) {
               // This assumes linear is handled well enough
               // by the MatrixD routine.  
  z = z - M*v;
  v = ( operator()(z) - id );
 }

 if( i >= MX_MAXITER ) {
  cout << endl;
  cout << "*** WARNING ***                                            \n" 
          "*** WARNING *** LieOperator& LieOperator::EpsInverse()     \n" 
          "*** WARNING *** Over " << MX_MAXITER << " iterations used. \n"
          "*** WARNING *** result may be incorrect.                   \n" 
          "*** WARNING ***                                            \n" 
       << endl;
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::operator()( LieOperator& x ) const {
 int i;
 LieOperator z;
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++) {
  z.v[i] = v[i]( x );
 }
 while( i < Jet::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator& LieOperator::operator=( const LieOperator& x ) {
 int i;
 int w;
 weight = -1;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   v[i] = x.v[i];
   if ( weight < ( w = v[i]->weight ) ) weight = w;
   refPoint[i] = x.refPoint[i];
   }
 accuWgt = x.accuWgt;
 return *this;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet LieOperator::operator[]( int i ) {  // ??? Review the use of this???
 return v[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet LieOperator::operator^( Jet& x ) { 
 Jet answer;
 int  s      = Jet::oneEnv.SpaceDim;
 int* ndx    = Jet::oneEnv.AllZeroes;
 int aw, i, jw;
 char adjustWeight = 1;

 aw = Jet::oneEnv.MaxWeight;
 ndx[0] = 1;
 answer = v[0]*x.D( ndx );
 ndx[0] = 0;

        // Differentiation will normally lower 
        //  the maximum accurate order by one.  This is 
        //  offset if the constant term in v[] vanishes.
        //
        if( v[0].standardPart() == 0.0 ) {
         jw = v[0]->accuWgt;
         if( jw > x->accuWgt ) jw = x->accuWgt;
         if( jw < aw ) aw = jw;
        }
        else adjustWeight = 0;

 if( s > 1 ) for( i = 1; i < s; i++ ) {
  ndx[i] = 1;
  answer += v[i]*x.D( ndx );
  ndx[i] = 0;

        if( v[i]->standardPart() == 0.0 ) {
         jw = v[i]->accuWgt;
         if( jw > x->accuWgt ) jw = x->accuWgt;
         if( jw < aw ) aw = jw;
        }
        else adjustWeight = 0;

 }

 // The accuracy of the answer is adjusted only if
 //  all components of v[] are differentials.
 // 
 if( adjustWeight ) answer->accuWgt = aw;

 answer.stacked = 1;
 return answer;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator^( /* const */ LieOperator& x, /* const */ LieOperator& y ) {
 // ??? const has to be systematically fixed throughout MXYZPTLK
 LieOperator z;
 int i;
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ ) 
  z.SetComponent( i, ( x^(y(i)) ) - ( y^(x(i)) ) );

 while( i < Jet::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet LieOperator::expMap( const double t, const Jet& x ) { 
 Jet answer;
 Jet u;
 double f;
 int count;
 
 answer = x;
 u      = x;
 f      = 1.0;
 count  = 0;

 u = ( t/f++ )*( operator^( u ) );
 while( ( count++ < MX_MAXITER ) && ( u != 0.0 ) ) {
  answer += u;                       // ??? How does this
  u = ( t/f++ )*( operator^( u ) );  // ??? affect efficiency?
 }

 if( count >= MX_MAXITER ) {
  printf( "\n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "*** WARNING *** Jet LieOperator::expMap()               \n" );
  printf( "*** WARNING *** Number of iterations has exceeded %d    \n", MX_MAXITER );
  printf( "*** WARNING *** without achieving convergence.          \n" );
  printf( "*** WARNING *** Results may be incorrect.               \n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "\n" );
 }

 answer.stacked = 1;
 return answer;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet LieOperator::expMap( const Jet& t, const Jet& x ) { 
 Jet answer;
 Jet u;
 double f;
 int count;
 
 answer = x;
 u      = x;
 f      = 1.0;
 count  = 0;

 u = ( t/f++ )*( operator^( u ) );
 while( ( count++ < MX_MAXITER ) && ( u != 0.0 ) ) {
  answer += u;                       // ??? How does this
  u = ( t/f++ )*( operator^( u ) );  // ??? affect efficiency?
 }

 if( count >= MX_MAXITER ) {
  printf( "\n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "*** WARNING *** Jet LieOperator::expMap()               \n" );
  printf( "*** WARNING *** Number of iterations has exceeded %d    \n", MX_MAXITER );
  printf( "*** WARNING *** without achieving convergence.          \n" );
  printf( "*** WARNING *** Results may be incorrect.               \n" );
  printf( "*** WARNING ***                                         \n" );
  printf( "\n" );
 }

 answer.stacked = 1;
 return answer;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::expMap( const double t , const LieOperator& x ) {
 LieOperator z;

 for( int i = 0; i < Jet::oneEnv.SpaceDim; i++) 
  z.v[i] = expMap( t, x.v[i] );
 while( i < Jet::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator LieOperator::expMap( const Jet& t, const LieOperator& x ) {
 LieOperator z;

 for( int i = 0; i < Jet::oneEnv.SpaceDim; i++) 
  z.v[i] = expMap( t, x.v[i] );
 while( i < Jet::oneEnv.NumVar ) z.v[i] = x.v[i++];
 z.reset();

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator+ ( const LieOperator& x, const LieOperator& y ) { 
 LieOperator z;
 int i;

 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( x.refPoint[i] != ( z.refPoint[i] = y.refPoint[i] ) ) {
     printf( "\n*** LieOperator ADDITION ERROR: Inconsistency in reference points.\n" );
     exit(0);
     }
   }
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 z.weight = 0;
 z.accuWgt = Jet::oneEnv.MaxWeight;
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  {
  z.v[i] = x.v[i] + y.v[i];
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
 }
 
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator- ( const LieOperator& x, const LieOperator& y ) {
 LieOperator z;
 int i;

 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   if( x.refPoint[i] != ( z.refPoint[i] = y.refPoint[i] ) ) {
     printf( "\n*** LieOperator ADDITION ERROR: Inconsistency in reference points.\n" );
     exit(0);
     }
   }
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 z.weight = 0;
 z.accuWgt = Jet::oneEnv.MaxWeight;
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  { 
  z.v[i] = x.v[i] - y.v[i];
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator- ( const LieOperator& x ) {
 LieOperator z;
 int i;

 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  z.v[i] = - x.v[i];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];
 z.weight  = x.weight;
 z.accuWgt = x.accuWgt;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator* ( const double& c, const LieOperator& x ) {
 LieOperator z;
 int i;

 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  z.v[i] = c*x.v[i];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];
 z.weight  = x.weight;
 z.accuWgt = x.accuWgt;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator* ( const LieOperator& x, const double& c ) {
 LieOperator z;
 int i;

 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  z.v[i] = c*x.v[i];
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];
 z.weight  = x.weight;
 z.accuWgt = x.accuWgt;

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator* ( const Jet& c, const LieOperator& x ) {
 LieOperator z;
 int i;

 z.weight = 0;
 z.accuWgt = Jet::oneEnv.MaxWeight;
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  {
  z.v[i] = c*x.v[i];
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
  }

 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator* ( const LieOperator& x, const Jet& c ) {
 return c*x;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator*(  MatrixD& M, const LieOperator& x ) {
 LieOperator z;
 int i, j;

 z.weight = x.weight;
 z.accuWgt = x.accuWgt;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = x.refPoint[i];

 int nsd = Jet::oneEnv.SpaceDim;

 if( ( M.rows() != nsd ) || 
     ( M.cols() != nsd )) {
  printf( "\n" );
  printf( "*** ERROR ***                                           \n" );
  printf( "*** ERROR *** LieOperator& operator*(  MatrixD& M, LieOperator& x ) \n" );
  printf( "*** ERROR *** Rows and/or columns of the matrix         \n" );
  printf( "*** ERROR *** are not correct.                          \n" );
  printf( "*** ERROR ***                                           \n" );
  exit(0);
 }

 double temp;    // ??? I don't like this A LOT!!!
 for( i = 0; i < nsd; i++ ) {
  temp = M( i, 0 );
  z.v[i] = temp * x.v[0];
  j = 1;
  while( j < nsd )   { 
    temp = M( i, j );
    z.v[i] += temp * ( x.v[j++] ) ;
  }
 }

 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

LieOperator operator*( const LieOperator& x, const LieOperator& y ) { 
 LieOperator z;
 LieOperator u;
 Jet term;
 int i;
 int j;
 int w;
 JLterm* p;
 
 // Check consistency of reference points.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ )
   if( x.refPoint[i] != ( y.v[i] ).weightedDerivative( Jet::oneEnv.AllZeroes ) ) {
     printf( "\n*** LieOperator CONCATENATION ERROR: Inconsistent reference points.\n" );
     exit(0);
     }
 
 // Subtract reference point prior to concatenation.
 u = y;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) { 
   p = ( u.v[i] ).get();
   if( p->weight != 0 ) {
     printf( "\n*** LieOperator CONCATENATION ERROR: Argument has no nonstandard part!!\n" );
     exit(0);
   }
   delete p;
 }
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 Jet::oneEnv.JLmonomial[0] = 1.0;
 Jet::oneEnv.JLmonomial[0].fixReferenceAtStart( y );
 
 // For all higher weights ...
 for( w = 1; w <= x.weight; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, Jet::oneEnv.NumVar, Jet::oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( Jet::oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( Jet::oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = Jet::oneEnv.JLmonomial[ Jet::monoRank() ];
     ( Jet::oneEnv.exponent[i] )++;
     Jet::monoCode();
     Jet::oneEnv.JLmonomial[ Jet::monoRank() ] = term * u.v[i];   // ??? Is this OK???
 
     } // End while loop.
 
 // Monomials have been stored.
 // Now traverse the Jet variable and evaluate.
 
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  {
   z.v[i] = concat( x.v[i] );
   for( j = 0; j < Jet::oneEnv.NumVar; j++ ) ( z.v[i] )->refPoint[j] = y.refPoint[j];
   }
 
 for( i = Jet::oneEnv.SpaceDim; i < Jet::oneEnv.NumVar; i++ ) {
   if( x.v[i] != y.v[i] )
     printf( "\n*** LieOperator CONCATENATION ERROR!!! Horrible!!!\n" );
   z.v[i] = y.v[i];
   }
 
 // Set the correct reference point and weights.
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) z.refPoint[i] = y.refPoint[i];
 
 z.weight = 0;
 z.accuWgt = Jet::oneEnv.MaxWeight;
 for( i = 0; i < Jet::oneEnv.SpaceDim; i++ )  {
  if( z.weight  < z.v[i]->weight  ) z.weight  = z.v[i]->weight;
  if( z.accuWgt > z.v[i]->accuWgt ) z.accuWgt = z.v[i]->accuWgt;
 }
 
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void operator*=( LieOperator& x, const LieOperator& y ) {
 x = x*y;
}
