#include "ML.hxx"
#define M_SMALL  1.0e-14

const complex complex_0 = complex(0.0,0.0);
const complex complex_1 = complex(1.0,0.0);

MLD::MLD() {
  m = 0;
  r = 0;
  c = 0;
  rc = 1;
}

MLD::MLD(int rows, int columns, double initval) {
  int i,j;
  r = rows;
  c = columns;
  rc = 1;
  m = new double *[r];
  for(i = 0; i < r; i++)
    m[i] = new double[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = initval;

  }
}

MLD::MLD(int rows, int columns, double* values) {
  int i,j;
  r = rows;
  c = columns;
  rc = 1;
  m = new double *[r];
  for(i = 0; i < r; i++)
    m[i] = new double[c];
  int index = 0;
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = values[index++];

  }
}

MLD::MLD(const MLD& X) {
  static char firstCall= 1;

  if(firstCall) {
    firstCall = 0;
    cerr << "*** WARNING *** The MLD copy constructor was called for" << endl;
    cerr   <<   "*** WARNING *** some inexplicable reason." << endl;
    cerr   <<   "*** WARNING *** You have a problem!" << endl;
  }

  int i,j;
  r = X.r;
  c = X.c;
  m = new double *[r];
  for(i = 0; i < r; i++)
    m[i] = new double[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = X.m[i][j];

  }
  rc = 1;
}



MLD::~MLD() {
  clear();
}

void MLD::clear() {
  if((r == 0) && (c == 0)) return;
  for(int i = 0; i < r; i++)
    delete []m[i];
  delete m;
  r = 0;
  c = 0;
}

MLD& MLD::operator=(const MLD& X) {

  if( this == &X ) return *this;
  
  clear();
  int i,j;

  r = X.r;
  c = X.c;
  m = new double *[r];
  for(i = 0; i < r; i++)
    m[i] = new double[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = X.m[i][j];

  }
}

char MLD::operator==(const double& x) {
  if(r != c) return 0;
  for(int i=0; i< r; i++) {
    if(m[i][i] != x) return 0;
    for(int j =0; j < c; j++)
      if((i != j) && (m[i][j] != 0.0)) return 0;
  }
  return 1;
}

MLD& MLD::operator+=( const double& x) {
  for(int i = 0; i < r; i++) {
    m[i][i] += x;
    if(fabs(m[i][i]) < M_SMALL*fabs(x))
      m[i][i] = 0.0;
  }
  return *this;
}

MLD& MLD::operator-=( const double& x) {
  for(int i = 0; i < r; i++) {
    m[i][i] -= x;
    if(fabs(m[i][i]) < M_SMALL*fabs(x))
      m[i][i] = 0.0;
  }
  return *this;
}

char operator==(const MLD& x, const MLD& y) {
  if((x.r != y.r) || (x.c != y.c)) return 0;
  for(int i = 0; i < x.r; i++) {
    for(int j = 0; j < x.c; j++)
      if(x.m[i][j] != y.m[i][j]) return 0;
  }
  return 1;
}


MLC::MLC() {
  m = 0;
  r = 0;
  c = 0;
  rc = 1;
}

MLC::MLC(int rows, int columns, complex initval) {
  int i,j;
  r = rows;
  c = columns;
  rc = 1;
  m = new complex *[r];
  for(i = 0; i < r; i++)
    m[i] = new complex[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = initval;

  }
}

MLC::MLC(int rows, int columns, complex* values) {
  int i,j;
  r = rows;
  c = columns;
  rc = 1;
  m = new complex *[r];
  for(i = 0; i < r; i++)
    m[i] = new complex[c];
  int index = 0;
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = values[index++];

  }
}

MLC::MLC(const MLC& X) {
  static char firstCall= 1;

  if(firstCall) {
    firstCall = 0;
    cerr << "*** WARNING *** The MLC copy constructor was called for" << endl;
    cerr   <<   "*** WARNING *** some inexplicable reason." << endl;
    cerr   <<   "*** WARNING *** You have a problem!" << endl;
  }

  int i,j;
  r = X.r;
  c = X.c;
  m = new complex *[r];
  for(i = 0; i < r; i++)
    m[i] = new complex[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = X.m[i][j];

  }
  rc = 1;
}
MLC::MLC(const MLD& X) {
  int i,j;
  r = X.r;
  c = X.c;
  m = new complex *[r];
  for(i = 0; i < r; i++)
    m[i] = new complex[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = complex(X.m[i][j],0.0);
  }
  rc = 1;
}



MLC::~MLC() {
  clear();
}

void MLC::clear() {
  if((r == 0) && (c == 0)) return;
  for(int i = 0; i < r; i++)
    delete []m[i];
  delete m;
  r = 0;
  c = 0;
}

MLC& MLC::operator=(const MLC& X) {

  if( this == &X ) return *this;
  
  clear();
  int i,j;

  r = X.r;
  c = X.c;
  m = new complex *[r];
  for(i = 0; i < r; i++)
    m[i] = new complex[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = X.m[i][j];

  }
}

char MLC::operator==(const complex& x) {
  if(r != c) return 0;
  for(int i=0; i< r; i++) {
    if(m[i][i] != x) return 0;
    for(int j =0; j < c; j++)
      if((i != j) && (m[i][j] != complex_0)) return 0;
  }
  return 1;
}

MLC& MLC::operator+=( const complex& x) {
  for(int i = 0; i < r; i++) {
    m[i][i] += x;
    if(abs(m[i][i]) < M_SMALL*abs(x))
      m[i][i] = complex_0;
  }
  return *this;
}

MLC& MLC::operator-=( const complex& x) {
  for(int i = 0; i < r; i++) {
    m[i][i] -= x;
    if(abs(m[i][i]) < M_SMALL*abs(x))
      m[i][i] = complex_0;
  }
  return *this;
}

char operator==(const MLC& x, const MLC& y) {
  if((x.r != y.r) || (x.c != y.c)) return 0;
  for(int i = 0; i < x.r; i++) {
    for(int j = 0; j < x.c; j++)
      if(x.m[i][j] != y.m[i][j]) return 0;
  }
  return 1;
}

MLI::MLI() {
  m = 0;
  r = 0;
  c = 0;
  rc = 1;
}

MLI::MLI(int rows, int columns, int initval) {
  int i,j;
  r = rows;
  c = columns;
  rc = 1;
  m = new int *[r];
  for(i = 0; i < r; i++)
    m[i] = new int[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = initval;

  }
}

MLI::MLI(int rows, int columns, int* values) {
  int i,j;
  r = rows;
  c = columns;
  rc = 1;
  m = new int *[r];
  for(i = 0; i < r; i++)
    m[i] = new int[c];
  int index = 0;
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = values[index++];

  }
}

MLI::MLI(const MLI& X) {
  static char firstCall= 1;

  if(firstCall) {
    firstCall = 0;
    cerr << "*** WARNING *** The MLI copy constructor was called for" << endl;
    cerr   <<   "*** WARNING *** some inexplicable reason." << endl;
    cerr   <<   "*** WARNING *** You have a problem!" << endl;
  }

  int i,j;
  r = X.r;
  c = X.c;
  m = new int *[r];
  for(i = 0; i < r; i++)
    m[i] = new int[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = X.m[i][j];

  }
  rc = 1;
}



MLI::~MLI() {
  clear();
}

void MLI::clear() {
  if((r == 0) && (c == 0)) return;
  for(int i = 0; i < r; i++)
    delete []m[i];
  delete m;
  r = 0;
  c = 0;
}

MLI& MLI::operator=(const MLI& X) {

  if( this == &X ) return *this;
  
  clear();
  int i,j;

  r = X.r;
  c = X.c;
  m = new int *[r];
  for(i = 0; i < r; i++)
    m[i] = new int[c];
  for(i = 0; i< r; i++) {
    for(j = 0; j < c; j++)
      m[i][j] = X.m[i][j];

  }
}

char MLI::operator==(const int& x) {
  if(r != c) return 0;
  for(int i=0; i< r; i++) {
    if(m[i][i] != x) return 0;
    for(int j =0; j < c; j++)
      if((i != j) && (m[i][j] != 0)) return 0;
  }
  return 1;
}

MLI& MLI::operator+=( const int& x) {
  for(int i = 0; i < r; i++) {
    m[i][i] += x;
  }
  return *this;
}

MLI& MLI::operator-=( const int& x) {
  for(int i = 0; i < r; i++) {
    m[i][i] -= x;
  }
  return *this;
}

char operator==(const MLI& x, const MLI& y) {
  if((x.r != y.r) || (x.c != y.c)) return 0;
  for(int i = 0; i < x.r; i++) {
    for(int j = 0; j < x.c; j++)
      if(x.m[i][j] != y.m[i][j]) return 0;
  }
  return 1;
}


